import { APIClient, FetchProvider, PrivateKey, Serializer, Action, Transaction, SignedTransaction, KeyType, Bytes, Checksum256, PublicKey, Name, ABI, Weight } from '@wharfkit/antelope';
import fetch$1 from 'cross-fetch';
import Debug from 'debug';
import { encode as encode$4 } from 'universal-base64url';
import rb from '@consento/sync-randombytes';
import 'crypto';
import * as nc from 'node:crypto';
import { getResolver } from '@tonomy/antelope-did-resolver';
import { getDidKeyResolver } from '@veramo/did-provider-key';
import 'ethers';
import Decimal from 'decimal.js';
import { io } from 'socket.io-client';
import '@veramo/core';
import '@veramo/did-manager';
import '@veramo/did-provider-ethr';
import '@veramo/key-manager';
import '@veramo/kms-local';
import '@veramo/credential-w3c';
import '@veramo/did-resolver';
import 'ethr-did-resolver';
import 'web-did-resolver';
import '@veramo/data-store';
import 'typeorm';
import fs from 'fs';
import path, { dirname } from 'path';
import { fileURLToPath } from 'url';
import argon2 from 'argon2';
import os from 'node:os';
import 'node:fs/promises';
import 'node:child_process';
import { parse as parse$1 } from 'csv-parse/sync';

class HttpError extends Error {
  constructor(httpError) {
    super('HTTP Error');
    // Ensure the name of this error is the same as the class name
    this.path = void 0;
    this.response = void 0;
    this.code = void 0;
    this.line = void 0;
    this.column = void 0;
    this.sourceURL = void 0;
    this.name = this.constructor.name;
    // This clips the constructor invocation from the stack trace.
    // It's not absolutely essential, but it does make the stack trace a little nicer.
    //  @see Node.js reference (bottom)
    if (Error.captureStackTrace) {
      Error.captureStackTrace(this, this.constructor);
    }
    this.stack = new Error().stack;
    this.path = httpError.path;
    this.response = httpError.response;
    if (httpError.line) this.line = httpError.line;
    if (httpError.column) this.line = httpError.column;
    if (httpError.sourceURL) this.sourceURL = httpError.sourceURL;
    if (httpError.code) this.code = httpError.code;
  }
}
class CommunicationError extends Error {
  constructor(communicationError) {
    super('Communication Error');
    this.exception = void 0;
    this.name = 'CommunicationError';
    this.message = communicationError.exception.message;
    this.exception = communicationError.exception;
    // This clips the constructor invocation from the stack trace.
    // It's not absolutely essential, but it does make the stack trace a little nicer.
    //  @see Node.js reference (bottom)
    if (Error.captureStackTrace) {
      Error.captureStackTrace(this, this.constructor);
    } else {
      this.stack = new Error().stack;
    }
  }
}
class SdkError extends Error {
  constructor(message) {
    super(message);
    // Ensure the name of this error is the same as the class name
    this.code = void 0;
    this.name = this.constructor.name;
    // This clips the constructor invocation from the stack trace.
    // It's not absolutely essential, but it does make the stack trace a little nicer.
    if (Error.captureStackTrace) {
      Error.captureStackTrace(this, this.constructor);
    } else {
      this.stack = new Error().stack;
    }
  }
}
function createSdkError(message, code) {
  let error = new SdkError(message);
  if (code) {
    error = new SdkError(code + ': ' + message);
    error.code = code;
  }
  return error;
}
// using never to suppress error https://bobbyhadz.com/blog/typescript-function-that-throws-error#:~:text=To%20declare%20a%20function%20that,terminate%20execution%20of%20the%20program.
function throwError(message, code) {
  throw createSdkError(message, code);
}
var SdkErrors;
(function (SdkErrors) {
  SdkErrors["AccountDoesntExist"] = "AccountDoesntExist";
  SdkErrors["AccountNotFound"] = "AccountNotFound";
  SdkErrors["AccountServiceError"] = "AccountServiceError";
  SdkErrors["CommunicationNotConnected"] = "CommunicationNotConnected";
  SdkErrors["CommunicationNotLoggedIn"] = "CommunicationNotLoggedIn";
  SdkErrors["CommunicationTimeout"] = "CommunicationTimeout";
  SdkErrors["CouldntCreateApi"] = "CouldntCreateApi";
  SdkErrors["DataQueryNoRowDataFound"] = "DataQueryNoRowDataFound";
  SdkErrors["InvalidLinkAuthRequest"] = "InvalidLinkAuthRequest";
  SdkErrors["InvalidChallenge"] = "InvalidChallenge";
  SdkErrors["InvalidData"] = "InvalidData";
  SdkErrors["InvalidKey"] = "InvalidKey";
  SdkErrors["InvalidArgumentType"] = "InvalidArgumentType";
  SdkErrors["InvalidKeyLevel"] = "InvalidKeyLevel";
  SdkErrors["InvalidQrCode"] = "InvalidQrCode";
  SdkErrors["InvalidUsername"] = "InvalidUsername";
  SdkErrors["InvalidRequestType"] = "InvalidRequestType";
  SdkErrors["InvalidRequestResponseType"] = "InvalidRequestResponseType";
  SdkErrors["JwtNotValid"] = "JwtNotValid";
  SdkErrors["KeyNotFound"] = "KeyNotFound";
  SdkErrors["LinkAuthFailed"] = "LinkAuthFailed";
  SdkErrors["MessageSendError"] = "MessageSendError";
  SdkErrors["MissingChallenge"] = "MissingChallenge";
  SdkErrors["MissingParams"] = "MissingParams";
  SdkErrors["OriginNotFound"] = "OriginNotFound";
  SdkErrors["OriginMismatch"] = "OriginMismatch";
  SdkErrors["PasswordFormatInvalid"] = "PasswordFormatInvalid";
  SdkErrors["PasswordInvalid"] = "PasswordInvalid";
  SdkErrors["RequestsNotFound"] = "RequestsNotFound";
  SdkErrors["ResponsesNotFound"] = "ResponsesNotFound";
  SdkErrors["PasswordTooCommon"] = "PasswordTooCommon";
  SdkErrors["PinInvalid"] = "PinInvalid";
  SdkErrors["ReferrerEmpty"] = "ReferrerEmpty";
  SdkErrors["SenderNotAuthorized"] = "SenderNotAuthorized";
  SdkErrors["SettingsNotInitialized"] = "SettingsNotInitialized";
  SdkErrors["UpdateKeysTransactionNoKeys"] = "UpdateKeysTransactionNoKeys";
  SdkErrors["UserDeactivated"] = "UserDeactivated";
  SdkErrors["UsernameNotFound"] = "UsernameNotFound";
  SdkErrors["UsernameTaken"] = "UsernameTaken";
  SdkErrors["UsernameNotDefined"] = "UsernameNotDefined";
  SdkErrors["UserNotLoggedIn"] = "UserNotLoggedIn";
  SdkErrors["UserNotLoggedInWithThisApp"] = "UserNotLoggedInWithThisApp";
  SdkErrors["UserRefreshed"] = "UserRefreshed";
  SdkErrors["WrongOrigin"] = "WrongOrigin";
  SdkErrors["UserCancelled"] = "UserCancelled";
  SdkErrors["UserLogout"] = "UserLogout";
})(SdkErrors || (SdkErrors = {}));
// eslint-disable-next-line @typescript-eslint/no-namespace
(function (SdkErrors) {
  /*
   * Returns the index of the enum value
   *
   * @param value The value to get the index of
   */
  function indexFor(value) {
    return Object.keys(SdkErrors).indexOf(value);
  }
  SdkErrors.indexFor = indexFor;
  /*
   * Creates an SdkErrors from a string or index of the level
   *
   * @param value The string or index
   */
  function from(value) {
    let index;
    if (typeof value !== 'number') {
      index = SdkErrors.indexFor(value);
    } else {
      index = value;
    }
    return Object.values(SdkErrors)[index];
  }
  SdkErrors.from = from;
})(SdkErrors || (SdkErrors = {}));

const debug$9 = Debug('tonomy-sdk:settings');
if (typeof Proxy === 'undefined') {
  throw new Error("This environment doesn't support Proxy");
}
let settings$1;
let fetchFunction;
let initialized = false;
function setSettings(newSettings) {
  debug$9('Settings initialized', newSettings);
  settings$1 = newSettings;
  initialized = true;
}
function getFetch() {
  return fetchFunction;
}
function getSettings() {
  if (!initialized) {
    throwError('Settings not yet initialized', SdkErrors.SettingsNotInitialized);
  }
  const proxy = new Proxy(settings$1, {
    get(target, name, receiver) {
      if (Reflect.has(target, name)) {
        return Reflect.get(target, name, receiver);
      }
      throw new Error(`Tonomy SDK settings has not been initialized using setSettings() with property for variable: ${name.toString()}`);
    }
  });
  return proxy;
}
function isProduction() {
  debug$9('Checking if production for environment:', settings$1.environment);
  if (!initialized) {
    throwError('Settings not yet initialized', SdkErrors.SettingsNotInitialized);
  }
  return ['production', 'staging', 'testnet'].includes(settings$1.environment);
}

const debug$8 = Debug('tonomy-sdk:services:blockchain:eosio:eosio');
let api;
async function getApi() {
  if (api) return api;
  const settings = getSettings();
  api = new APIClient({
    url: settings.blockchainUrl,
    provider: new FetchProvider(settings.blockchainUrl, {
      fetch: fetch$1
    })
  });
  if (!api) throwError('Could not create API client', SdkErrors.CouldntCreateApi);
  return api;
}
/**
 * This function serializes one action into hex string
 *
 * @param {string} account - name of the contract account to pull the ABI from
 * @param {string } type - name of the action that will be executed
 * @param {object} data - data of the action that will be executed
 * @returns {string} - hex string of the serialized action
 */
async function serializeActionData(account, type, data) {
  const {
    abi
  } = await (await getApi()).v1.chain.get_abi(account);
  if (!abi) {
    throw new Error(`No ABI for ${account}`);
  }
  const {
    hexString
  } = Serializer.encode({
    object: data,
    abi,
    type
  });
  return hexString;
}
async function getChainInfo() {
  const api = await getApi();
  return await api.v1.chain.get_info();
}
async function getAccount(account) {
  try {
    const api = await getApi();
    return await api.v1.chain.get_account(account);
  } catch (e) {
    if (e.message === 'Account not found at /v1/chain/get_account') {
      throwError('Account "' + account.toString() + '" not found', SdkErrors.AccountDoesntExist);
    } else {
      throw e;
    }
  }
}
async function getProducers() {
  const api = await getApi();
  return await api.v1.chain.get_producer_schedule();
}
function getTonomyOperationsKey() {
  if (isProduction() && !process.env.TONOMY_OPS_PRIVATE_KEY) throw new Error('TONOMY_OPS_PRIVATE_KEY must be set in production');
  if (process.env.TONOMY_OPS_PRIVATE_KEY) {
    debug$8('Using TONOMY_OPS_PRIVATE_KEY from env');
    return PrivateKey.from(process.env.TONOMY_OPS_PRIVATE_KEY);
  }
  return PrivateKey.from('PVT_K1_24kG9VcMk3VkkgY4hh42X262AWV18YcPjBTd2Hox4YWoP8vRTU');
}

const debug$7 = Debug('tonomy-sdk:services:blockchain:eosio:transaction');
function createSigner$1(privateKey) {
  return {
    async sign(digest) {
      return privateKey.signDigest(digest);
    }
  };
}
function createKeyManagerSigner(keyManager, level, challenge) {
  return {
    async sign(digest) {
      return await keyManager.signData({
        level,
        data: digest,
        challenge
      });
    }
  };
}
class AntelopePushTransactionError extends Error {
  constructor(err) {
    var _err$error;
    super('AntelopePushTransactionError');
    this.code = void 0;
    // HTTP error code
    this.message = void 0;
    // HTTP error message
    this.error = void 0;
    this.actions = void 0;
    this.contract = void 0;
    this.code = err.code;
    this.message = err.error.what;
    if (Array.isArray((_err$error = err.error) == null ? void 0 : _err$error.details) && err.error.details.length > 0) {
      this.message += ': ' + err.error.details[0].message;
    }
    this.error = err.error;
    this.contract = err.contract;
    this.actions = err.actions;
    this.stack = new Error().stack;
    // Ensure the name of this error is the same as the class name
    this.name = this.constructor.name;
    // This clips the constructor invocation from the stack trace.
    // It's not absolutely essential, but it does make the stack trace a little nicer.
    //  @see Node.js reference (bottom)
    if (Error.captureStackTrace) {
      Error.captureStackTrace(this, this.constructor);
    }
  }
  hasErrorCode(code) {
    return this.error.code === code;
  }
  hasTonomyErrorCode(code) {
    // TODO: iterate over deatils array instead of only looking at first element
    return this.error.details[0].message.search(code) > 0;
  }
}
async function transact(contract, actions, signer) {
  // Get the ABI
  const api = await getApi();
  const abi = await api.v1.chain.get_abi(contract);
  // Create the action data
  const actionData = [];
  actions.forEach(data => {
    actionData.push(Action.from({
      account: contract,
      ...data
    }, abi.abi));
  });
  // Construct the transaction
  const info = await api.v1.chain.get_info();
  const header = info.getTransactionHeader();
  const transaction = Transaction.from({
    ...header,
    actions: actionData
  });
  // Create signature
  if (!Array.isArray(signer)) signer = [signer];
  const signDigest = transaction.signingDigest(info.chain_id);
  const signatures = await Promise.all(signer.map(s => s.sign(signDigest)));
  const signedTransaction = SignedTransaction.from({
    ...transaction,
    signatures
  });
  // Send to the node
  let res;
  try {
    debug$7('Pushing transaction', JSON.stringify(actions, null, 2));
    res = await api.v1.chain.push_transaction(signedTransaction);
  } catch (e) {
    var _e$response;
    debug$7('Error pushing transaction', e);
    if ((_e$response = e.response) != null && _e$response.headers) {
      var _e$response2;
      if ((_e$response2 = e.response) != null && _e$response2.json) {
        throw new AntelopePushTransactionError({
          ...e.response.json,
          contract,
          actions
        });
      }
      throw new HttpError(e);
    }
    throw e;
  }
  return res;
}

// From utf8 to base64url and visa versa
function objToBase64Url(obj) {
  return encode$4(JSON.stringify(obj));
}

function asUint8Array(buf) {
  if (globalThis.Buffer != null) {
    return new Uint8Array(buf.buffer, buf.byteOffset, buf.byteLength);
  }
  return buf;
}

function allocUnsafe(size = 0) {
  if (globalThis.Buffer != null && globalThis.Buffer.allocUnsafe != null) {
    return asUint8Array(globalThis.Buffer.allocUnsafe(size));
  }
  return new Uint8Array(size);
}

function concat$1(arrays, length) {
  if (!length) {
    length = arrays.reduce((acc, curr) => acc + curr.length, 0);
  }
  const output = allocUnsafe(length);
  let offset = 0;
  for (const arr of arrays) {
    output.set(arr, offset);
    offset += arr.length;
  }
  return asUint8Array(output);
}

function base$2(ALPHABET, name) {
  if (ALPHABET.length >= 255) {
    throw new TypeError('Alphabet too long');
  }
  var BASE_MAP = new Uint8Array(256);
  for (var j = 0; j < BASE_MAP.length; j++) {
    BASE_MAP[j] = 255;
  }
  for (var i = 0; i < ALPHABET.length; i++) {
    var x = ALPHABET.charAt(i);
    var xc = x.charCodeAt(0);
    if (BASE_MAP[xc] !== 255) {
      throw new TypeError(x + ' is ambiguous');
    }
    BASE_MAP[xc] = i;
  }
  var BASE = ALPHABET.length;
  var LEADER = ALPHABET.charAt(0);
  var FACTOR = Math.log(BASE) / Math.log(256);
  var iFACTOR = Math.log(256) / Math.log(BASE);
  function encode(source) {
    if (source instanceof Uint8Array);
    else if (ArrayBuffer.isView(source)) {
      source = new Uint8Array(source.buffer, source.byteOffset, source.byteLength);
    } else if (Array.isArray(source)) {
      source = Uint8Array.from(source);
    }
    if (!(source instanceof Uint8Array)) {
      throw new TypeError('Expected Uint8Array');
    }
    if (source.length === 0) {
      return '';
    }
    var zeroes = 0;
    var length = 0;
    var pbegin = 0;
    var pend = source.length;
    while (pbegin !== pend && source[pbegin] === 0) {
      pbegin++;
      zeroes++;
    }
    var size = (pend - pbegin) * iFACTOR + 1 >>> 0;
    var b58 = new Uint8Array(size);
    while (pbegin !== pend) {
      var carry = source[pbegin];
      var i = 0;
      for (var it1 = size - 1; (carry !== 0 || i < length) && it1 !== -1; it1--, i++) {
        carry += 256 * b58[it1] >>> 0;
        b58[it1] = carry % BASE >>> 0;
        carry = carry / BASE >>> 0;
      }
      if (carry !== 0) {
        throw new Error('Non-zero carry');
      }
      length = i;
      pbegin++;
    }
    var it2 = size - length;
    while (it2 !== size && b58[it2] === 0) {
      it2++;
    }
    var str = LEADER.repeat(zeroes);
    for (; it2 < size; ++it2) {
      str += ALPHABET.charAt(b58[it2]);
    }
    return str;
  }
  function decodeUnsafe(source) {
    if (typeof source !== 'string') {
      throw new TypeError('Expected String');
    }
    if (source.length === 0) {
      return new Uint8Array();
    }
    var psz = 0;
    if (source[psz] === ' ') {
      return;
    }
    var zeroes = 0;
    var length = 0;
    while (source[psz] === LEADER) {
      zeroes++;
      psz++;
    }
    var size = (source.length - psz) * FACTOR + 1 >>> 0;
    var b256 = new Uint8Array(size);
    while (source[psz]) {
      var carry = BASE_MAP[source.charCodeAt(psz)];
      if (carry === 255) {
        return;
      }
      var i = 0;
      for (var it3 = size - 1; (carry !== 0 || i < length) && it3 !== -1; it3--, i++) {
        carry += BASE * b256[it3] >>> 0;
        b256[it3] = carry % 256 >>> 0;
        carry = carry / 256 >>> 0;
      }
      if (carry !== 0) {
        throw new Error('Non-zero carry');
      }
      length = i;
      psz++;
    }
    if (source[psz] === ' ') {
      return;
    }
    var it4 = size - length;
    while (it4 !== size && b256[it4] === 0) {
      it4++;
    }
    var vch = new Uint8Array(zeroes + (size - it4));
    var j = zeroes;
    while (it4 !== size) {
      vch[j++] = b256[it4++];
    }
    return vch;
  }
  function decode(string) {
    var buffer = decodeUnsafe(string);
    if (buffer) {
      return buffer;
    }
    throw new Error(`Non-${ name } character`);
  }
  return {
    encode: encode,
    decodeUnsafe: decodeUnsafe,
    decode: decode
  };
}
var src$2 = base$2;
var _brrp__multiformats_scope_baseX = src$2;

const coerce = o => {
  if (o instanceof Uint8Array && o.constructor.name === 'Uint8Array')
    return o;
  if (o instanceof ArrayBuffer)
    return new Uint8Array(o);
  if (ArrayBuffer.isView(o)) {
    return new Uint8Array(o.buffer, o.byteOffset, o.byteLength);
  }
  throw new Error('Unknown type, must be binary type');
};
const fromString$1 = str => new TextEncoder().encode(str);
const toString$1 = b => new TextDecoder().decode(b);

class Encoder {
  constructor(name, prefix, baseEncode) {
    this.name = name;
    this.prefix = prefix;
    this.baseEncode = baseEncode;
  }
  encode(bytes) {
    if (bytes instanceof Uint8Array) {
      return `${ this.prefix }${ this.baseEncode(bytes) }`;
    } else {
      throw Error('Unknown type, must be binary type');
    }
  }
}
class Decoder {
  constructor(name, prefix, baseDecode) {
    this.name = name;
    this.prefix = prefix;
    if (prefix.codePointAt(0) === undefined) {
      throw new Error('Invalid prefix character');
    }
    this.prefixCodePoint = prefix.codePointAt(0);
    this.baseDecode = baseDecode;
  }
  decode(text) {
    if (typeof text === 'string') {
      if (text.codePointAt(0) !== this.prefixCodePoint) {
        throw Error(`Unable to decode multibase string ${ JSON.stringify(text) }, ${ this.name } decoder only supports inputs prefixed with ${ this.prefix }`);
      }
      return this.baseDecode(text.slice(this.prefix.length));
    } else {
      throw Error('Can only multibase decode strings');
    }
  }
  or(decoder) {
    return or(this, decoder);
  }
}
class ComposedDecoder {
  constructor(decoders) {
    this.decoders = decoders;
  }
  or(decoder) {
    return or(this, decoder);
  }
  decode(input) {
    const prefix = input[0];
    const decoder = this.decoders[prefix];
    if (decoder) {
      return decoder.decode(input);
    } else {
      throw RangeError(`Unable to decode multibase string ${ JSON.stringify(input) }, only inputs prefixed with ${ Object.keys(this.decoders) } are supported`);
    }
  }
}
const or = (left, right) => new ComposedDecoder({
  ...left.decoders || { [left.prefix]: left },
  ...right.decoders || { [right.prefix]: right }
});
class Codec {
  constructor(name, prefix, baseEncode, baseDecode) {
    this.name = name;
    this.prefix = prefix;
    this.baseEncode = baseEncode;
    this.baseDecode = baseDecode;
    this.encoder = new Encoder(name, prefix, baseEncode);
    this.decoder = new Decoder(name, prefix, baseDecode);
  }
  encode(input) {
    return this.encoder.encode(input);
  }
  decode(input) {
    return this.decoder.decode(input);
  }
}
const from = ({name, prefix, encode, decode}) => new Codec(name, prefix, encode, decode);
const baseX = ({prefix, name, alphabet}) => {
  const {encode, decode} = _brrp__multiformats_scope_baseX(alphabet, name);
  return from({
    prefix,
    name,
    encode,
    decode: text => coerce(decode(text))
  });
};
const decode$4 = (string, alphabet, bitsPerChar, name) => {
  const codes = {};
  for (let i = 0; i < alphabet.length; ++i) {
    codes[alphabet[i]] = i;
  }
  let end = string.length;
  while (string[end - 1] === '=') {
    --end;
  }
  const out = new Uint8Array(end * bitsPerChar / 8 | 0);
  let bits = 0;
  let buffer = 0;
  let written = 0;
  for (let i = 0; i < end; ++i) {
    const value = codes[string[i]];
    if (value === undefined) {
      throw new SyntaxError(`Non-${ name } character`);
    }
    buffer = buffer << bitsPerChar | value;
    bits += bitsPerChar;
    if (bits >= 8) {
      bits -= 8;
      out[written++] = 255 & buffer >> bits;
    }
  }
  if (bits >= bitsPerChar || 255 & buffer << 8 - bits) {
    throw new SyntaxError('Unexpected end of data');
  }
  return out;
};
const encode$3 = (data, alphabet, bitsPerChar) => {
  const pad = alphabet[alphabet.length - 1] === '=';
  const mask = (1 << bitsPerChar) - 1;
  let out = '';
  let bits = 0;
  let buffer = 0;
  for (let i = 0; i < data.length; ++i) {
    buffer = buffer << 8 | data[i];
    bits += 8;
    while (bits > bitsPerChar) {
      bits -= bitsPerChar;
      out += alphabet[mask & buffer >> bits];
    }
  }
  if (bits) {
    out += alphabet[mask & buffer << bitsPerChar - bits];
  }
  if (pad) {
    while (out.length * bitsPerChar & 7) {
      out += '=';
    }
  }
  return out;
};
const rfc4648$2 = ({name, prefix, bitsPerChar, alphabet}) => {
  return from({
    prefix,
    name,
    encode(input) {
      return encode$3(input, alphabet, bitsPerChar);
    },
    decode(input) {
      return decode$4(input, alphabet, bitsPerChar, name);
    }
  });
};

const identity$1 = from({
  prefix: '\0',
  name: 'identity',
  encode: buf => toString$1(buf),
  decode: str => fromString$1(str)
});

var identityBase = {
    __proto__: null,
    identity: identity$1
};

const base2 = rfc4648$2({
  prefix: '0',
  name: 'base2',
  alphabet: '01',
  bitsPerChar: 1
});

var base2$1 = {
    __proto__: null,
    base2: base2
};

const base8 = rfc4648$2({
  prefix: '7',
  name: 'base8',
  alphabet: '01234567',
  bitsPerChar: 3
});

var base8$1 = {
    __proto__: null,
    base8: base8
};

const base10 = baseX({
  prefix: '9',
  name: 'base10',
  alphabet: '0123456789'
});

var base10$1 = {
    __proto__: null,
    base10: base10
};

const base16 = rfc4648$2({
  prefix: 'f',
  name: 'base16',
  alphabet: '0123456789abcdef',
  bitsPerChar: 4
});
const base16upper = rfc4648$2({
  prefix: 'F',
  name: 'base16upper',
  alphabet: '0123456789ABCDEF',
  bitsPerChar: 4
});

var base16$1 = {
    __proto__: null,
    base16: base16,
    base16upper: base16upper
};

const base32 = rfc4648$2({
  prefix: 'b',
  name: 'base32',
  alphabet: 'abcdefghijklmnopqrstuvwxyz234567',
  bitsPerChar: 5
});
const base32upper = rfc4648$2({
  prefix: 'B',
  name: 'base32upper',
  alphabet: 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567',
  bitsPerChar: 5
});
const base32pad = rfc4648$2({
  prefix: 'c',
  name: 'base32pad',
  alphabet: 'abcdefghijklmnopqrstuvwxyz234567=',
  bitsPerChar: 5
});
const base32padupper = rfc4648$2({
  prefix: 'C',
  name: 'base32padupper',
  alphabet: 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567=',
  bitsPerChar: 5
});
const base32hex = rfc4648$2({
  prefix: 'v',
  name: 'base32hex',
  alphabet: '0123456789abcdefghijklmnopqrstuv',
  bitsPerChar: 5
});
const base32hexupper = rfc4648$2({
  prefix: 'V',
  name: 'base32hexupper',
  alphabet: '0123456789ABCDEFGHIJKLMNOPQRSTUV',
  bitsPerChar: 5
});
const base32hexpad = rfc4648$2({
  prefix: 't',
  name: 'base32hexpad',
  alphabet: '0123456789abcdefghijklmnopqrstuv=',
  bitsPerChar: 5
});
const base32hexpadupper = rfc4648$2({
  prefix: 'T',
  name: 'base32hexpadupper',
  alphabet: '0123456789ABCDEFGHIJKLMNOPQRSTUV=',
  bitsPerChar: 5
});
const base32z = rfc4648$2({
  prefix: 'h',
  name: 'base32z',
  alphabet: 'ybndrfg8ejkmcpqxot1uwisza345h769',
  bitsPerChar: 5
});

var base32$1 = {
    __proto__: null,
    base32: base32,
    base32upper: base32upper,
    base32pad: base32pad,
    base32padupper: base32padupper,
    base32hex: base32hex,
    base32hexupper: base32hexupper,
    base32hexpad: base32hexpad,
    base32hexpadupper: base32hexpadupper,
    base32z: base32z
};

const base36 = baseX({
  prefix: 'k',
  name: 'base36',
  alphabet: '0123456789abcdefghijklmnopqrstuvwxyz'
});
const base36upper = baseX({
  prefix: 'K',
  name: 'base36upper',
  alphabet: '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
});

var base36$1 = {
    __proto__: null,
    base36: base36,
    base36upper: base36upper
};

const base58btc = baseX({
  name: 'base58btc',
  prefix: 'z',
  alphabet: '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
});
const base58flickr = baseX({
  name: 'base58flickr',
  prefix: 'Z',
  alphabet: '123456789abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ'
});

var base58 = {
    __proto__: null,
    base58btc: base58btc,
    base58flickr: base58flickr
};

const base64 = rfc4648$2({
  prefix: 'm',
  name: 'base64',
  alphabet: 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/',
  bitsPerChar: 6
});
const base64pad = rfc4648$2({
  prefix: 'M',
  name: 'base64pad',
  alphabet: 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=',
  bitsPerChar: 6
});
const base64url = rfc4648$2({
  prefix: 'u',
  name: 'base64url',
  alphabet: 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_',
  bitsPerChar: 6
});
const base64urlpad = rfc4648$2({
  prefix: 'U',
  name: 'base64urlpad',
  alphabet: 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_=',
  bitsPerChar: 6
});

var base64$1 = {
    __proto__: null,
    base64: base64,
    base64pad: base64pad,
    base64url: base64url,
    base64urlpad: base64urlpad
};

const alphabet$1 = Array.from('\uD83D\uDE80\uD83E\uDE90\u2604\uD83D\uDEF0\uD83C\uDF0C\uD83C\uDF11\uD83C\uDF12\uD83C\uDF13\uD83C\uDF14\uD83C\uDF15\uD83C\uDF16\uD83C\uDF17\uD83C\uDF18\uD83C\uDF0D\uD83C\uDF0F\uD83C\uDF0E\uD83D\uDC09\u2600\uD83D\uDCBB\uD83D\uDDA5\uD83D\uDCBE\uD83D\uDCBF\uD83D\uDE02\u2764\uD83D\uDE0D\uD83E\uDD23\uD83D\uDE0A\uD83D\uDE4F\uD83D\uDC95\uD83D\uDE2D\uD83D\uDE18\uD83D\uDC4D\uD83D\uDE05\uD83D\uDC4F\uD83D\uDE01\uD83D\uDD25\uD83E\uDD70\uD83D\uDC94\uD83D\uDC96\uD83D\uDC99\uD83D\uDE22\uD83E\uDD14\uD83D\uDE06\uD83D\uDE44\uD83D\uDCAA\uD83D\uDE09\u263A\uD83D\uDC4C\uD83E\uDD17\uD83D\uDC9C\uD83D\uDE14\uD83D\uDE0E\uD83D\uDE07\uD83C\uDF39\uD83E\uDD26\uD83C\uDF89\uD83D\uDC9E\u270C\u2728\uD83E\uDD37\uD83D\uDE31\uD83D\uDE0C\uD83C\uDF38\uD83D\uDE4C\uD83D\uDE0B\uD83D\uDC97\uD83D\uDC9A\uD83D\uDE0F\uD83D\uDC9B\uD83D\uDE42\uD83D\uDC93\uD83E\uDD29\uD83D\uDE04\uD83D\uDE00\uD83D\uDDA4\uD83D\uDE03\uD83D\uDCAF\uD83D\uDE48\uD83D\uDC47\uD83C\uDFB6\uD83D\uDE12\uD83E\uDD2D\u2763\uD83D\uDE1C\uD83D\uDC8B\uD83D\uDC40\uD83D\uDE2A\uD83D\uDE11\uD83D\uDCA5\uD83D\uDE4B\uD83D\uDE1E\uD83D\uDE29\uD83D\uDE21\uD83E\uDD2A\uD83D\uDC4A\uD83E\uDD73\uD83D\uDE25\uD83E\uDD24\uD83D\uDC49\uD83D\uDC83\uD83D\uDE33\u270B\uD83D\uDE1A\uD83D\uDE1D\uD83D\uDE34\uD83C\uDF1F\uD83D\uDE2C\uD83D\uDE43\uD83C\uDF40\uD83C\uDF37\uD83D\uDE3B\uD83D\uDE13\u2B50\u2705\uD83E\uDD7A\uD83C\uDF08\uD83D\uDE08\uD83E\uDD18\uD83D\uDCA6\u2714\uD83D\uDE23\uD83C\uDFC3\uD83D\uDC90\u2639\uD83C\uDF8A\uD83D\uDC98\uD83D\uDE20\u261D\uD83D\uDE15\uD83C\uDF3A\uD83C\uDF82\uD83C\uDF3B\uD83D\uDE10\uD83D\uDD95\uD83D\uDC9D\uD83D\uDE4A\uD83D\uDE39\uD83D\uDDE3\uD83D\uDCAB\uD83D\uDC80\uD83D\uDC51\uD83C\uDFB5\uD83E\uDD1E\uD83D\uDE1B\uD83D\uDD34\uD83D\uDE24\uD83C\uDF3C\uD83D\uDE2B\u26BD\uD83E\uDD19\u2615\uD83C\uDFC6\uD83E\uDD2B\uD83D\uDC48\uD83D\uDE2E\uD83D\uDE46\uD83C\uDF7B\uD83C\uDF43\uD83D\uDC36\uD83D\uDC81\uD83D\uDE32\uD83C\uDF3F\uD83E\uDDE1\uD83C\uDF81\u26A1\uD83C\uDF1E\uD83C\uDF88\u274C\u270A\uD83D\uDC4B\uD83D\uDE30\uD83E\uDD28\uD83D\uDE36\uD83E\uDD1D\uD83D\uDEB6\uD83D\uDCB0\uD83C\uDF53\uD83D\uDCA2\uD83E\uDD1F\uD83D\uDE41\uD83D\uDEA8\uD83D\uDCA8\uD83E\uDD2C\u2708\uD83C\uDF80\uD83C\uDF7A\uD83E\uDD13\uD83D\uDE19\uD83D\uDC9F\uD83C\uDF31\uD83D\uDE16\uD83D\uDC76\uD83E\uDD74\u25B6\u27A1\u2753\uD83D\uDC8E\uD83D\uDCB8\u2B07\uD83D\uDE28\uD83C\uDF1A\uD83E\uDD8B\uD83D\uDE37\uD83D\uDD7A\u26A0\uD83D\uDE45\uD83D\uDE1F\uD83D\uDE35\uD83D\uDC4E\uD83E\uDD32\uD83E\uDD20\uD83E\uDD27\uD83D\uDCCC\uD83D\uDD35\uD83D\uDC85\uD83E\uDDD0\uD83D\uDC3E\uD83C\uDF52\uD83D\uDE17\uD83E\uDD11\uD83C\uDF0A\uD83E\uDD2F\uD83D\uDC37\u260E\uD83D\uDCA7\uD83D\uDE2F\uD83D\uDC86\uD83D\uDC46\uD83C\uDFA4\uD83D\uDE47\uD83C\uDF51\u2744\uD83C\uDF34\uD83D\uDCA3\uD83D\uDC38\uD83D\uDC8C\uD83D\uDCCD\uD83E\uDD40\uD83E\uDD22\uD83D\uDC45\uD83D\uDCA1\uD83D\uDCA9\uD83D\uDC50\uD83D\uDCF8\uD83D\uDC7B\uD83E\uDD10\uD83E\uDD2E\uD83C\uDFBC\uD83E\uDD75\uD83D\uDEA9\uD83C\uDF4E\uD83C\uDF4A\uD83D\uDC7C\uD83D\uDC8D\uD83D\uDCE3\uD83E\uDD42');
const alphabetBytesToChars = alphabet$1.reduce((p, c, i) => {
  p[i] = c;
  return p;
}, []);
const alphabetCharsToBytes = alphabet$1.reduce((p, c, i) => {
  p[c.codePointAt(0)] = i;
  return p;
}, []);
function encode$2(data) {
  return data.reduce((p, c) => {
    p += alphabetBytesToChars[c];
    return p;
  }, '');
}
function decode$3(str) {
  const byts = [];
  for (const char of str) {
    const byt = alphabetCharsToBytes[char.codePointAt(0)];
    if (byt === undefined) {
      throw new Error(`Non-base256emoji character: ${ char }`);
    }
    byts.push(byt);
  }
  return new Uint8Array(byts);
}
const base256emoji = from({
  prefix: '\uD83D\uDE80',
  name: 'base256emoji',
  encode: encode$2,
  decode: decode$3
});

var base256emoji$1 = {
    __proto__: null,
    base256emoji: base256emoji
};

var encode_1 = encode$1;
var MSB = 128, REST = 127, MSBALL = ~REST, INT = Math.pow(2, 31);
function encode$1(num, out, offset) {
  out = out || [];
  offset = offset || 0;
  var oldOffset = offset;
  while (num >= INT) {
    out[offset++] = num & 255 | MSB;
    num /= 128;
  }
  while (num & MSBALL) {
    out[offset++] = num & 255 | MSB;
    num >>>= 7;
  }
  out[offset] = num | 0;
  encode$1.bytes = offset - oldOffset + 1;
  return out;
}
var decode$2 = read;
var MSB$1 = 128, REST$1 = 127;
function read(buf, offset) {
  var res = 0, offset = offset || 0, shift = 0, counter = offset, b, l = buf.length;
  do {
    if (counter >= l) {
      read.bytes = 0;
      throw new RangeError('Could not decode varint');
    }
    b = buf[counter++];
    res += shift < 28 ? (b & REST$1) << shift : (b & REST$1) * Math.pow(2, shift);
    shift += 7;
  } while (b >= MSB$1);
  read.bytes = counter - offset;
  return res;
}
var N1 = Math.pow(2, 7);
var N2 = Math.pow(2, 14);
var N3 = Math.pow(2, 21);
var N4 = Math.pow(2, 28);
var N5 = Math.pow(2, 35);
var N6 = Math.pow(2, 42);
var N7 = Math.pow(2, 49);
var N8 = Math.pow(2, 56);
var N9 = Math.pow(2, 63);
var length = function (value) {
  return value < N1 ? 1 : value < N2 ? 2 : value < N3 ? 3 : value < N4 ? 4 : value < N5 ? 5 : value < N6 ? 6 : value < N7 ? 7 : value < N8 ? 8 : value < N9 ? 9 : 10;
};
var varint = {
  encode: encode_1,
  decode: decode$2,
  encodingLength: length
};
var _brrp_varint = varint;

const decode$1 = (data, offset = 0) => {
  const code = _brrp_varint.decode(data, offset);
  return [
    code,
    _brrp_varint.decode.bytes
  ];
};

new TextEncoder();
new TextDecoder();

const bases = {
  ...identityBase,
  ...base2$1,
  ...base8$1,
  ...base10$1,
  ...base16$1,
  ...base32$1,
  ...base36$1,
  ...base58,
  ...base64$1,
  ...base256emoji$1
};

function createCodec(name, prefix, encode, decode) {
  return {
    name,
    prefix,
    encoder: {
      name,
      prefix,
      encode
    },
    decoder: { decode }
  };
}
const string = createCodec('utf8', 'u', buf => {
  const decoder = new TextDecoder('utf8');
  return 'u' + decoder.decode(buf);
}, str => {
  const encoder = new TextEncoder();
  return encoder.encode(str.substring(1));
});
const ascii = createCodec('ascii', 'a', buf => {
  let string = 'a';
  for (let i = 0; i < buf.length; i++) {
    string += String.fromCharCode(buf[i]);
  }
  return string;
}, str => {
  str = str.substring(1);
  const buf = allocUnsafe(str.length);
  for (let i = 0; i < str.length; i++) {
    buf[i] = str.charCodeAt(i);
  }
  return buf;
});
const BASES = {
  utf8: string,
  'utf-8': string,
  hex: bases.base16,
  latin1: ascii,
  ascii: ascii,
  binary: ascii,
  ...bases
};

function fromString(string, encoding = 'utf8') {
  const base = BASES[encoding];
  if (!base) {
    throw new Error(`Unsupported encoding "${ encoding }"`);
  }
  if ((encoding === 'utf8' || encoding === 'utf-8') && globalThis.Buffer != null && globalThis.Buffer.from != null) {
    return asUint8Array(globalThis.Buffer.from(string, 'utf-8'));
  }
  return base.decoder.decode(`${ base.prefix }${ string }`);
}

function toString(array, encoding = 'utf8') {
  const base = BASES[encoding];
  if (!base) {
    throw new Error(`Unsupported encoding "${ encoding }"`);
  }
  if ((encoding === 'utf8' || encoding === 'utf-8') && globalThis.Buffer != null && globalThis.Buffer.from != null) {
    return globalThis.Buffer.from(array.buffer, array.byteOffset, array.byteLength).toString('utf8');
  }
  return base.encoder.encode(array).substring(1);
}

function number(n) {
    if (!Number.isSafeInteger(n) || n < 0)
        throw new Error(`positive integer expected, not ${n}`);
}
// copied from utils
function isBytes$2(a) {
    return (a instanceof Uint8Array ||
        (a != null && typeof a === 'object' && a.constructor.name === 'Uint8Array'));
}
function bytes(b, ...lengths) {
    if (!isBytes$2(b))
        throw new Error('Uint8Array expected');
    if (lengths.length > 0 && !lengths.includes(b.length))
        throw new Error(`Uint8Array expected of length ${lengths}, not of length=${b.length}`);
}
function hash(h) {
    if (typeof h !== 'function' || typeof h.create !== 'function')
        throw new Error('Hash should be wrapped by utils.wrapConstructor');
    number(h.outputLen);
    number(h.blockLen);
}
function exists(instance, checkFinished = true) {
    if (instance.destroyed)
        throw new Error('Hash instance has been destroyed');
    if (checkFinished && instance.finished)
        throw new Error('Hash#digest() has already been called');
}
function output(out, instance) {
    bytes(out);
    const min = instance.outputLen;
    if (out.length < min) {
        throw new Error(`digestInto() expects output buffer of length at least ${min}`);
    }
}

// We use WebCrypto aka globalThis.crypto, which exists in browsers and node.js 16+.
const crypto = nc && typeof nc === 'object' && 'webcrypto' in nc ? nc.webcrypto : undefined;

/*! noble-hashes - MIT License (c) 2022 Paul Miller (paulmillr.com) */
const u32 = (arr) => new Uint32Array(arr.buffer, arr.byteOffset, Math.floor(arr.byteLength / 4));
// Cast array to view
const createView = (arr) => new DataView(arr.buffer, arr.byteOffset, arr.byteLength);
// The rotate right (circular right shift) operation for uint32
const rotr = (word, shift) => (word << (32 - shift)) | (word >>> shift);
// The rotate left (circular left shift) operation for uint32
const rotl = (word, shift) => (word << shift) | ((word >>> (32 - shift)) >>> 0);
const isLE = new Uint8Array(new Uint32Array([0x11223344]).buffer)[0] === 0x44;
// The byte swap operation for uint32
const byteSwap = (word) => ((word << 24) & 0xff000000) |
    ((word << 8) & 0xff0000) |
    ((word >>> 8) & 0xff00) |
    ((word >>> 24) & 0xff);
// In place byte swap for Uint32Array
function byteSwap32(arr) {
    for (let i = 0; i < arr.length; i++) {
        arr[i] = byteSwap(arr[i]);
    }
}
/**
 * @example utf8ToBytes('abc') // new Uint8Array([97, 98, 99])
 */
function utf8ToBytes$1(str) {
    if (typeof str !== 'string')
        throw new Error(`utf8ToBytes expected string, got ${typeof str}`);
    return new Uint8Array(new TextEncoder().encode(str)); // https://bugzil.la/1681809
}
/**
 * Normalizes (non-hex) string or Uint8Array to Uint8Array.
 * Warning: when Uint8Array is passed, it would NOT get copied.
 * Keep in mind for future mutable operations.
 */
function toBytes(data) {
    if (typeof data === 'string')
        data = utf8ToBytes$1(data);
    bytes(data);
    return data;
}
/**
 * Copies several Uint8Arrays into one.
 */
function concatBytes$1(...arrays) {
    let sum = 0;
    for (let i = 0; i < arrays.length; i++) {
        const a = arrays[i];
        bytes(a);
        sum += a.length;
    }
    const res = new Uint8Array(sum);
    for (let i = 0, pad = 0; i < arrays.length; i++) {
        const a = arrays[i];
        res.set(a, pad);
        pad += a.length;
    }
    return res;
}
// For runtime check if class implements interface
class Hash {
    // Safe version that clones internal state
    clone() {
        return this._cloneInto();
    }
}
function wrapConstructor(hashCons) {
    const hashC = (msg) => hashCons().update(toBytes(msg)).digest();
    const tmp = hashCons();
    hashC.outputLen = tmp.outputLen;
    hashC.blockLen = tmp.blockLen;
    hashC.create = () => hashCons();
    return hashC;
}
/**
 * Secure PRNG. Uses `crypto.getRandomValues`, which defers to OS.
 */
function randomBytes$1(bytesLength = 32) {
    if (crypto && typeof crypto.getRandomValues === 'function') {
        return crypto.getRandomValues(new Uint8Array(bytesLength));
    }
    throw new Error('crypto.getRandomValues must be defined');
}

// Polyfill for Safari 14
function setBigUint64(view, byteOffset, value, isLE) {
    if (typeof view.setBigUint64 === 'function')
        return view.setBigUint64(byteOffset, value, isLE);
    const _32n = BigInt(32);
    const _u32_max = BigInt(0xffffffff);
    const wh = Number((value >> _32n) & _u32_max);
    const wl = Number(value & _u32_max);
    const h = isLE ? 4 : 0;
    const l = isLE ? 0 : 4;
    view.setUint32(byteOffset + h, wh, isLE);
    view.setUint32(byteOffset + l, wl, isLE);
}
// Choice: a ? b : c
const Chi = (a, b, c) => (a & b) ^ (~a & c);
// Majority function, true if any two inpust is true
const Maj = (a, b, c) => (a & b) ^ (a & c) ^ (b & c);
/**
 * Merkle-Damgard hash construction base class.
 * Could be used to create MD5, RIPEMD, SHA1, SHA2.
 */
class HashMD extends Hash {
    constructor(blockLen, outputLen, padOffset, isLE) {
        super();
        this.blockLen = blockLen;
        this.outputLen = outputLen;
        this.padOffset = padOffset;
        this.isLE = isLE;
        this.finished = false;
        this.length = 0;
        this.pos = 0;
        this.destroyed = false;
        this.buffer = new Uint8Array(blockLen);
        this.view = createView(this.buffer);
    }
    update(data) {
        exists(this);
        const { view, buffer, blockLen } = this;
        data = toBytes(data);
        const len = data.length;
        for (let pos = 0; pos < len;) {
            const take = Math.min(blockLen - this.pos, len - pos);
            // Fast path: we have at least one block in input, cast it to view and process
            if (take === blockLen) {
                const dataView = createView(data);
                for (; blockLen <= len - pos; pos += blockLen)
                    this.process(dataView, pos);
                continue;
            }
            buffer.set(data.subarray(pos, pos + take), this.pos);
            this.pos += take;
            pos += take;
            if (this.pos === blockLen) {
                this.process(view, 0);
                this.pos = 0;
            }
        }
        this.length += data.length;
        this.roundClean();
        return this;
    }
    digestInto(out) {
        exists(this);
        output(out, this);
        this.finished = true;
        // Padding
        // We can avoid allocation of buffer for padding completely if it
        // was previously not allocated here. But it won't change performance.
        const { buffer, view, blockLen, isLE } = this;
        let { pos } = this;
        // append the bit '1' to the message
        buffer[pos++] = 0b10000000;
        this.buffer.subarray(pos).fill(0);
        // we have less than padOffset left in buffer, so we cannot put length in
        // current block, need process it and pad again
        if (this.padOffset > blockLen - pos) {
            this.process(view, 0);
            pos = 0;
        }
        // Pad until full block byte with zeros
        for (let i = pos; i < blockLen; i++)
            buffer[i] = 0;
        // Note: sha512 requires length to be 128bit integer, but length in JS will overflow before that
        // You need to write around 2 exabytes (u64_max / 8 / (1024**6)) for this to happen.
        // So we just write lowest 64 bits of that value.
        setBigUint64(view, blockLen - 8, BigInt(this.length * 8), isLE);
        this.process(view, 0);
        const oview = createView(out);
        const len = this.outputLen;
        // NOTE: we do division by 4 later, which should be fused in single op with modulo by JIT
        if (len % 4)
            throw new Error('_sha2: outputLen should be aligned to 32bit');
        const outLen = len / 4;
        const state = this.get();
        if (outLen > state.length)
            throw new Error('_sha2: outputLen bigger than state');
        for (let i = 0; i < outLen; i++)
            oview.setUint32(4 * i, state[i], isLE);
    }
    digest() {
        const { buffer, outputLen } = this;
        this.digestInto(buffer);
        const res = buffer.slice(0, outputLen);
        this.destroy();
        return res;
    }
    _cloneInto(to) {
        to || (to = new this.constructor());
        to.set(...this.get());
        const { blockLen, buffer, length, finished, destroyed, pos } = this;
        to.length = length;
        to.pos = pos;
        to.finished = finished;
        to.destroyed = destroyed;
        if (length % blockLen)
            to.buffer.set(buffer);
        return to;
    }
}

const U32_MASK64 = /* @__PURE__ */ BigInt(2 ** 32 - 1);
const _32n = /* @__PURE__ */ BigInt(32);
// We are not using BigUint64Array, because they are extremely slow as per 2022
function fromBig(n, le = false) {
    if (le)
        return { h: Number(n & U32_MASK64), l: Number((n >> _32n) & U32_MASK64) };
    return { h: Number((n >> _32n) & U32_MASK64) | 0, l: Number(n & U32_MASK64) | 0 };
}
function split(lst, le = false) {
    let Ah = new Uint32Array(lst.length);
    let Al = new Uint32Array(lst.length);
    for (let i = 0; i < lst.length; i++) {
        const { h, l } = fromBig(lst[i], le);
        [Ah[i], Al[i]] = [h, l];
    }
    return [Ah, Al];
}
const toBig = (h, l) => (BigInt(h >>> 0) << _32n) | BigInt(l >>> 0);
// for Shift in [0, 32)
const shrSH = (h, _l, s) => h >>> s;
const shrSL = (h, l, s) => (h << (32 - s)) | (l >>> s);
// Right rotate for Shift in [1, 32)
const rotrSH = (h, l, s) => (h >>> s) | (l << (32 - s));
const rotrSL = (h, l, s) => (h << (32 - s)) | (l >>> s);
// Right rotate for Shift in (32, 64), NOTE: 32 is special case.
const rotrBH = (h, l, s) => (h << (64 - s)) | (l >>> (s - 32));
const rotrBL = (h, l, s) => (h >>> (s - 32)) | (l << (64 - s));
// Right rotate for shift===32 (just swaps l&h)
const rotr32H = (_h, l) => l;
const rotr32L = (h, _l) => h;
// Left rotate for Shift in [1, 32)
const rotlSH = (h, l, s) => (h << s) | (l >>> (32 - s));
const rotlSL = (h, l, s) => (l << s) | (h >>> (32 - s));
// Left rotate for Shift in (32, 64), NOTE: 32 is special case.
const rotlBH = (h, l, s) => (l << (s - 32)) | (h >>> (64 - s));
const rotlBL = (h, l, s) => (h << (s - 32)) | (l >>> (64 - s));
// JS uses 32-bit signed integers for bitwise operations which means we cannot
// simple take carry out of low bit sum by shift, we need to use division.
function add(Ah, Al, Bh, Bl) {
    const l = (Al >>> 0) + (Bl >>> 0);
    return { h: (Ah + Bh + ((l / 2 ** 32) | 0)) | 0, l: l | 0 };
}
// Addition with more than 2 elements
const add3L = (Al, Bl, Cl) => (Al >>> 0) + (Bl >>> 0) + (Cl >>> 0);
const add3H = (low, Ah, Bh, Ch) => (Ah + Bh + Ch + ((low / 2 ** 32) | 0)) | 0;
const add4L = (Al, Bl, Cl, Dl) => (Al >>> 0) + (Bl >>> 0) + (Cl >>> 0) + (Dl >>> 0);
const add4H = (low, Ah, Bh, Ch, Dh) => (Ah + Bh + Ch + Dh + ((low / 2 ** 32) | 0)) | 0;
const add5L = (Al, Bl, Cl, Dl, El) => (Al >>> 0) + (Bl >>> 0) + (Cl >>> 0) + (Dl >>> 0) + (El >>> 0);
const add5H = (low, Ah, Bh, Ch, Dh, Eh) => (Ah + Bh + Ch + Dh + Eh + ((low / 2 ** 32) | 0)) | 0;
// prettier-ignore
const u64 = {
    fromBig, split, toBig,
    shrSH, shrSL,
    rotrSH, rotrSL, rotrBH, rotrBL,
    rotr32H, rotr32L,
    rotlSH, rotlSL, rotlBH, rotlBL,
    add, add3L, add3H, add4L, add4H, add5H, add5L,
};

// Round contants (first 32 bits of the fractional parts of the cube roots of the first 80 primes 2..409):
// prettier-ignore
const [SHA512_Kh, SHA512_Kl] = /* @__PURE__ */ (() => u64.split([
    '0x428a2f98d728ae22', '0x7137449123ef65cd', '0xb5c0fbcfec4d3b2f', '0xe9b5dba58189dbbc',
    '0x3956c25bf348b538', '0x59f111f1b605d019', '0x923f82a4af194f9b', '0xab1c5ed5da6d8118',
    '0xd807aa98a3030242', '0x12835b0145706fbe', '0x243185be4ee4b28c', '0x550c7dc3d5ffb4e2',
    '0x72be5d74f27b896f', '0x80deb1fe3b1696b1', '0x9bdc06a725c71235', '0xc19bf174cf692694',
    '0xe49b69c19ef14ad2', '0xefbe4786384f25e3', '0x0fc19dc68b8cd5b5', '0x240ca1cc77ac9c65',
    '0x2de92c6f592b0275', '0x4a7484aa6ea6e483', '0x5cb0a9dcbd41fbd4', '0x76f988da831153b5',
    '0x983e5152ee66dfab', '0xa831c66d2db43210', '0xb00327c898fb213f', '0xbf597fc7beef0ee4',
    '0xc6e00bf33da88fc2', '0xd5a79147930aa725', '0x06ca6351e003826f', '0x142929670a0e6e70',
    '0x27b70a8546d22ffc', '0x2e1b21385c26c926', '0x4d2c6dfc5ac42aed', '0x53380d139d95b3df',
    '0x650a73548baf63de', '0x766a0abb3c77b2a8', '0x81c2c92e47edaee6', '0x92722c851482353b',
    '0xa2bfe8a14cf10364', '0xa81a664bbc423001', '0xc24b8b70d0f89791', '0xc76c51a30654be30',
    '0xd192e819d6ef5218', '0xd69906245565a910', '0xf40e35855771202a', '0x106aa07032bbd1b8',
    '0x19a4c116b8d2d0c8', '0x1e376c085141ab53', '0x2748774cdf8eeb99', '0x34b0bcb5e19b48a8',
    '0x391c0cb3c5c95a63', '0x4ed8aa4ae3418acb', '0x5b9cca4f7763e373', '0x682e6ff3d6b2b8a3',
    '0x748f82ee5defb2fc', '0x78a5636f43172f60', '0x84c87814a1f0ab72', '0x8cc702081a6439ec',
    '0x90befffa23631e28', '0xa4506cebde82bde9', '0xbef9a3f7b2c67915', '0xc67178f2e372532b',
    '0xca273eceea26619c', '0xd186b8c721c0c207', '0xeada7dd6cde0eb1e', '0xf57d4f7fee6ed178',
    '0x06f067aa72176fba', '0x0a637dc5a2c898a6', '0x113f9804bef90dae', '0x1b710b35131c471b',
    '0x28db77f523047d84', '0x32caab7b40c72493', '0x3c9ebe0a15c9bebc', '0x431d67c49c100d4c',
    '0x4cc5d4becb3e42b6', '0x597f299cfc657e2a', '0x5fcb6fab3ad6faec', '0x6c44198c4a475817'
].map(n => BigInt(n))))();
// Temporary buffer, not used to store anything between runs
const SHA512_W_H = /* @__PURE__ */ new Uint32Array(80);
const SHA512_W_L = /* @__PURE__ */ new Uint32Array(80);
class SHA512 extends HashMD {
    constructor() {
        super(128, 64, 16, false);
        // We cannot use array here since array allows indexing by variable which means optimizer/compiler cannot use registers.
        // Also looks cleaner and easier to verify with spec.
        // Initial state (first 32 bits of the fractional parts of the square roots of the first 8 primes 2..19):
        // h -- high 32 bits, l -- low 32 bits
        this.Ah = 0x6a09e667 | 0;
        this.Al = 0xf3bcc908 | 0;
        this.Bh = 0xbb67ae85 | 0;
        this.Bl = 0x84caa73b | 0;
        this.Ch = 0x3c6ef372 | 0;
        this.Cl = 0xfe94f82b | 0;
        this.Dh = 0xa54ff53a | 0;
        this.Dl = 0x5f1d36f1 | 0;
        this.Eh = 0x510e527f | 0;
        this.El = 0xade682d1 | 0;
        this.Fh = 0x9b05688c | 0;
        this.Fl = 0x2b3e6c1f | 0;
        this.Gh = 0x1f83d9ab | 0;
        this.Gl = 0xfb41bd6b | 0;
        this.Hh = 0x5be0cd19 | 0;
        this.Hl = 0x137e2179 | 0;
    }
    // prettier-ignore
    get() {
        const { Ah, Al, Bh, Bl, Ch, Cl, Dh, Dl, Eh, El, Fh, Fl, Gh, Gl, Hh, Hl } = this;
        return [Ah, Al, Bh, Bl, Ch, Cl, Dh, Dl, Eh, El, Fh, Fl, Gh, Gl, Hh, Hl];
    }
    // prettier-ignore
    set(Ah, Al, Bh, Bl, Ch, Cl, Dh, Dl, Eh, El, Fh, Fl, Gh, Gl, Hh, Hl) {
        this.Ah = Ah | 0;
        this.Al = Al | 0;
        this.Bh = Bh | 0;
        this.Bl = Bl | 0;
        this.Ch = Ch | 0;
        this.Cl = Cl | 0;
        this.Dh = Dh | 0;
        this.Dl = Dl | 0;
        this.Eh = Eh | 0;
        this.El = El | 0;
        this.Fh = Fh | 0;
        this.Fl = Fl | 0;
        this.Gh = Gh | 0;
        this.Gl = Gl | 0;
        this.Hh = Hh | 0;
        this.Hl = Hl | 0;
    }
    process(view, offset) {
        // Extend the first 16 words into the remaining 64 words w[16..79] of the message schedule array
        for (let i = 0; i < 16; i++, offset += 4) {
            SHA512_W_H[i] = view.getUint32(offset);
            SHA512_W_L[i] = view.getUint32((offset += 4));
        }
        for (let i = 16; i < 80; i++) {
            // s0 := (w[i-15] rightrotate 1) xor (w[i-15] rightrotate 8) xor (w[i-15] rightshift 7)
            const W15h = SHA512_W_H[i - 15] | 0;
            const W15l = SHA512_W_L[i - 15] | 0;
            const s0h = u64.rotrSH(W15h, W15l, 1) ^ u64.rotrSH(W15h, W15l, 8) ^ u64.shrSH(W15h, W15l, 7);
            const s0l = u64.rotrSL(W15h, W15l, 1) ^ u64.rotrSL(W15h, W15l, 8) ^ u64.shrSL(W15h, W15l, 7);
            // s1 := (w[i-2] rightrotate 19) xor (w[i-2] rightrotate 61) xor (w[i-2] rightshift 6)
            const W2h = SHA512_W_H[i - 2] | 0;
            const W2l = SHA512_W_L[i - 2] | 0;
            const s1h = u64.rotrSH(W2h, W2l, 19) ^ u64.rotrBH(W2h, W2l, 61) ^ u64.shrSH(W2h, W2l, 6);
            const s1l = u64.rotrSL(W2h, W2l, 19) ^ u64.rotrBL(W2h, W2l, 61) ^ u64.shrSL(W2h, W2l, 6);
            // SHA256_W[i] = s0 + s1 + SHA256_W[i - 7] + SHA256_W[i - 16];
            const SUMl = u64.add4L(s0l, s1l, SHA512_W_L[i - 7], SHA512_W_L[i - 16]);
            const SUMh = u64.add4H(SUMl, s0h, s1h, SHA512_W_H[i - 7], SHA512_W_H[i - 16]);
            SHA512_W_H[i] = SUMh | 0;
            SHA512_W_L[i] = SUMl | 0;
        }
        let { Ah, Al, Bh, Bl, Ch, Cl, Dh, Dl, Eh, El, Fh, Fl, Gh, Gl, Hh, Hl } = this;
        // Compression function main loop, 80 rounds
        for (let i = 0; i < 80; i++) {
            // S1 := (e rightrotate 14) xor (e rightrotate 18) xor (e rightrotate 41)
            const sigma1h = u64.rotrSH(Eh, El, 14) ^ u64.rotrSH(Eh, El, 18) ^ u64.rotrBH(Eh, El, 41);
            const sigma1l = u64.rotrSL(Eh, El, 14) ^ u64.rotrSL(Eh, El, 18) ^ u64.rotrBL(Eh, El, 41);
            //const T1 = (H + sigma1 + Chi(E, F, G) + SHA256_K[i] + SHA256_W[i]) | 0;
            const CHIh = (Eh & Fh) ^ (~Eh & Gh);
            const CHIl = (El & Fl) ^ (~El & Gl);
            // T1 = H + sigma1 + Chi(E, F, G) + SHA512_K[i] + SHA512_W[i]
            // prettier-ignore
            const T1ll = u64.add5L(Hl, sigma1l, CHIl, SHA512_Kl[i], SHA512_W_L[i]);
            const T1h = u64.add5H(T1ll, Hh, sigma1h, CHIh, SHA512_Kh[i], SHA512_W_H[i]);
            const T1l = T1ll | 0;
            // S0 := (a rightrotate 28) xor (a rightrotate 34) xor (a rightrotate 39)
            const sigma0h = u64.rotrSH(Ah, Al, 28) ^ u64.rotrBH(Ah, Al, 34) ^ u64.rotrBH(Ah, Al, 39);
            const sigma0l = u64.rotrSL(Ah, Al, 28) ^ u64.rotrBL(Ah, Al, 34) ^ u64.rotrBL(Ah, Al, 39);
            const MAJh = (Ah & Bh) ^ (Ah & Ch) ^ (Bh & Ch);
            const MAJl = (Al & Bl) ^ (Al & Cl) ^ (Bl & Cl);
            Hh = Gh | 0;
            Hl = Gl | 0;
            Gh = Fh | 0;
            Gl = Fl | 0;
            Fh = Eh | 0;
            Fl = El | 0;
            ({ h: Eh, l: El } = u64.add(Dh | 0, Dl | 0, T1h | 0, T1l | 0));
            Dh = Ch | 0;
            Dl = Cl | 0;
            Ch = Bh | 0;
            Cl = Bl | 0;
            Bh = Ah | 0;
            Bl = Al | 0;
            const All = u64.add3L(T1l, sigma0l, MAJl);
            Ah = u64.add3H(All, T1h, sigma0h, MAJh);
            Al = All | 0;
        }
        // Add the compressed chunk to the current hash value
        ({ h: Ah, l: Al } = u64.add(this.Ah | 0, this.Al | 0, Ah | 0, Al | 0));
        ({ h: Bh, l: Bl } = u64.add(this.Bh | 0, this.Bl | 0, Bh | 0, Bl | 0));
        ({ h: Ch, l: Cl } = u64.add(this.Ch | 0, this.Cl | 0, Ch | 0, Cl | 0));
        ({ h: Dh, l: Dl } = u64.add(this.Dh | 0, this.Dl | 0, Dh | 0, Dl | 0));
        ({ h: Eh, l: El } = u64.add(this.Eh | 0, this.El | 0, Eh | 0, El | 0));
        ({ h: Fh, l: Fl } = u64.add(this.Fh | 0, this.Fl | 0, Fh | 0, Fl | 0));
        ({ h: Gh, l: Gl } = u64.add(this.Gh | 0, this.Gl | 0, Gh | 0, Gl | 0));
        ({ h: Hh, l: Hl } = u64.add(this.Hh | 0, this.Hl | 0, Hh | 0, Hl | 0));
        this.set(Ah, Al, Bh, Bl, Ch, Cl, Dh, Dl, Eh, El, Fh, Fl, Gh, Gl, Hh, Hl);
    }
    roundClean() {
        SHA512_W_H.fill(0);
        SHA512_W_L.fill(0);
    }
    destroy() {
        this.buffer.fill(0);
        this.set(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    }
}
const sha512 = /* @__PURE__ */ wrapConstructor(() => new SHA512());

/*! noble-curves - MIT License (c) 2022 Paul Miller (paulmillr.com) */
// 100 lines of code in the file are duplicated from noble-hashes (utils).
// This is OK: `abstract` directory does not use noble-hashes.
// User may opt-in into using different hashing library. This way, noble-hashes
// won't be included into their bundle.
const _0n$5 = /* @__PURE__ */ BigInt(0);
const _1n$7 = /* @__PURE__ */ BigInt(1);
const _2n$5 = /* @__PURE__ */ BigInt(2);
function isBytes$1(a) {
    return (a instanceof Uint8Array ||
        (a != null && typeof a === 'object' && a.constructor.name === 'Uint8Array'));
}
function abytes(item) {
    if (!isBytes$1(item))
        throw new Error('Uint8Array expected');
}
// Array where index 0xf0 (240) is mapped to string 'f0'
const hexes = /* @__PURE__ */ Array.from({ length: 256 }, (_, i) => i.toString(16).padStart(2, '0'));
/**
 * @example bytesToHex(Uint8Array.from([0xca, 0xfe, 0x01, 0x23])) // 'cafe0123'
 */
function bytesToHex$1(bytes) {
    abytes(bytes);
    // pre-caching improves the speed 6x
    let hex = '';
    for (let i = 0; i < bytes.length; i++) {
        hex += hexes[bytes[i]];
    }
    return hex;
}
function numberToHexUnpadded(num) {
    const hex = num.toString(16);
    return hex.length & 1 ? `0${hex}` : hex;
}
function hexToNumber(hex) {
    if (typeof hex !== 'string')
        throw new Error('hex string expected, got ' + typeof hex);
    // Big Endian
    return BigInt(hex === '' ? '0' : `0x${hex}`);
}
// We use optimized technique to convert hex string to byte array
const asciis = { _0: 48, _9: 57, _A: 65, _F: 70, _a: 97, _f: 102 };
function asciiToBase16(char) {
    if (char >= asciis._0 && char <= asciis._9)
        return char - asciis._0;
    if (char >= asciis._A && char <= asciis._F)
        return char - (asciis._A - 10);
    if (char >= asciis._a && char <= asciis._f)
        return char - (asciis._a - 10);
    return;
}
/**
 * @example hexToBytes('cafe0123') // Uint8Array.from([0xca, 0xfe, 0x01, 0x23])
 */
function hexToBytes$1(hex) {
    if (typeof hex !== 'string')
        throw new Error('hex string expected, got ' + typeof hex);
    const hl = hex.length;
    const al = hl / 2;
    if (hl % 2)
        throw new Error('padded hex string expected, got unpadded hex of length ' + hl);
    const array = new Uint8Array(al);
    for (let ai = 0, hi = 0; ai < al; ai++, hi += 2) {
        const n1 = asciiToBase16(hex.charCodeAt(hi));
        const n2 = asciiToBase16(hex.charCodeAt(hi + 1));
        if (n1 === undefined || n2 === undefined) {
            const char = hex[hi] + hex[hi + 1];
            throw new Error('hex string expected, got non-hex character "' + char + '" at index ' + hi);
        }
        array[ai] = n1 * 16 + n2;
    }
    return array;
}
// BE: Big Endian, LE: Little Endian
function bytesToNumberBE(bytes) {
    return hexToNumber(bytesToHex$1(bytes));
}
function bytesToNumberLE(bytes) {
    abytes(bytes);
    return hexToNumber(bytesToHex$1(Uint8Array.from(bytes).reverse()));
}
function numberToBytesBE(n, len) {
    return hexToBytes$1(n.toString(16).padStart(len * 2, '0'));
}
function numberToBytesLE(n, len) {
    return numberToBytesBE(n, len).reverse();
}
// Unpadded, rarely used
function numberToVarBytesBE(n) {
    return hexToBytes$1(numberToHexUnpadded(n));
}
/**
 * Takes hex string or Uint8Array, converts to Uint8Array.
 * Validates output length.
 * Will throw error for other types.
 * @param title descriptive title for an error e.g. 'private key'
 * @param hex hex string or Uint8Array
 * @param expectedLength optional, will compare to result array's length
 * @returns
 */
function ensureBytes(title, hex, expectedLength) {
    let res;
    if (typeof hex === 'string') {
        try {
            res = hexToBytes$1(hex);
        }
        catch (e) {
            throw new Error(`${title} must be valid hex string, got "${hex}". Cause: ${e}`);
        }
    }
    else if (isBytes$1(hex)) {
        // Uint8Array.from() instead of hash.slice() because node.js Buffer
        // is instance of Uint8Array, and its slice() creates **mutable** copy
        res = Uint8Array.from(hex);
    }
    else {
        throw new Error(`${title} must be hex string or Uint8Array`);
    }
    const len = res.length;
    if (typeof expectedLength === 'number' && len !== expectedLength)
        throw new Error(`${title} expected ${expectedLength} bytes, got ${len}`);
    return res;
}
/**
 * Copies several Uint8Arrays into one.
 */
function concatBytes(...arrays) {
    let sum = 0;
    for (let i = 0; i < arrays.length; i++) {
        const a = arrays[i];
        abytes(a);
        sum += a.length;
    }
    const res = new Uint8Array(sum);
    for (let i = 0, pad = 0; i < arrays.length; i++) {
        const a = arrays[i];
        res.set(a, pad);
        pad += a.length;
    }
    return res;
}
// Compares 2 u8a-s in kinda constant time
function equalBytes(a, b) {
    if (a.length !== b.length)
        return false;
    let diff = 0;
    for (let i = 0; i < a.length; i++)
        diff |= a[i] ^ b[i];
    return diff === 0;
}
/**
 * @example utf8ToBytes('abc') // new Uint8Array([97, 98, 99])
 */
function utf8ToBytes(str) {
    if (typeof str !== 'string')
        throw new Error(`utf8ToBytes expected string, got ${typeof str}`);
    return new Uint8Array(new TextEncoder().encode(str)); // https://bugzil.la/1681809
}
// Bit operations
/**
 * Calculates amount of bits in a bigint.
 * Same as `n.toString(2).length`
 */
function bitLen(n) {
    let len;
    for (len = 0; n > _0n$5; n >>= _1n$7, len += 1)
        ;
    return len;
}
/**
 * Gets single bit at position.
 * NOTE: first bit position is 0 (same as arrays)
 * Same as `!!+Array.from(n.toString(2)).reverse()[pos]`
 */
function bitGet(n, pos) {
    return (n >> BigInt(pos)) & _1n$7;
}
/**
 * Sets single bit at position.
 */
function bitSet(n, pos, value) {
    return n | ((value ? _1n$7 : _0n$5) << BigInt(pos));
}
/**
 * Calculate mask for N bits. Not using ** operator with bigints because of old engines.
 * Same as BigInt(`0b${Array(i).fill('1').join('')}`)
 */
const bitMask = (n) => (_2n$5 << BigInt(n - 1)) - _1n$7;
// DRBG
const u8n = (data) => new Uint8Array(data); // creates Uint8Array
const u8fr = (arr) => Uint8Array.from(arr); // another shortcut
/**
 * Minimal HMAC-DRBG from NIST 800-90 for RFC6979 sigs.
 * @returns function that will call DRBG until 2nd arg returns something meaningful
 * @example
 *   const drbg = createHmacDRBG<Key>(32, 32, hmac);
 *   drbg(seed, bytesToKey); // bytesToKey must return Key or undefined
 */
function createHmacDrbg(hashLen, qByteLen, hmacFn) {
    if (typeof hashLen !== 'number' || hashLen < 2)
        throw new Error('hashLen must be a number');
    if (typeof qByteLen !== 'number' || qByteLen < 2)
        throw new Error('qByteLen must be a number');
    if (typeof hmacFn !== 'function')
        throw new Error('hmacFn must be a function');
    // Step B, Step C: set hashLen to 8*ceil(hlen/8)
    let v = u8n(hashLen); // Minimal non-full-spec HMAC-DRBG from NIST 800-90 for RFC6979 sigs.
    let k = u8n(hashLen); // Steps B and C of RFC6979 3.2: set hashLen, in our case always same
    let i = 0; // Iterations counter, will throw when over 1000
    const reset = () => {
        v.fill(1);
        k.fill(0);
        i = 0;
    };
    const h = (...b) => hmacFn(k, v, ...b); // hmac(k)(v, ...values)
    const reseed = (seed = u8n()) => {
        // HMAC-DRBG reseed() function. Steps D-G
        k = h(u8fr([0x00]), seed); // k = hmac(k || v || 0x00 || seed)
        v = h(); // v = hmac(k || v)
        if (seed.length === 0)
            return;
        k = h(u8fr([0x01]), seed); // k = hmac(k || v || 0x01 || seed)
        v = h(); // v = hmac(k || v)
    };
    const gen = () => {
        // HMAC-DRBG generate() function
        if (i++ >= 1000)
            throw new Error('drbg: tried 1000 values');
        let len = 0;
        const out = [];
        while (len < qByteLen) {
            v = h();
            const sl = v.slice();
            out.push(sl);
            len += v.length;
        }
        return concatBytes(...out);
    };
    const genUntil = (seed, pred) => {
        reset();
        reseed(seed); // Steps D-G
        let res = undefined; // Step H: grind until k is in [1..n-1]
        while (!(res = pred(gen())))
            reseed();
        reset();
        return res;
    };
    return genUntil;
}
// Validating curves and fields
const validatorFns = {
    bigint: (val) => typeof val === 'bigint',
    function: (val) => typeof val === 'function',
    boolean: (val) => typeof val === 'boolean',
    string: (val) => typeof val === 'string',
    stringOrUint8Array: (val) => typeof val === 'string' || isBytes$1(val),
    isSafeInteger: (val) => Number.isSafeInteger(val),
    array: (val) => Array.isArray(val),
    field: (val, object) => object.Fp.isValid(val),
    hash: (val) => typeof val === 'function' && Number.isSafeInteger(val.outputLen),
};
// type Record<K extends string | number | symbol, T> = { [P in K]: T; }
function validateObject(object, validators, optValidators = {}) {
    const checkField = (fieldName, type, isOptional) => {
        const checkVal = validatorFns[type];
        if (typeof checkVal !== 'function')
            throw new Error(`Invalid validator "${type}", expected function`);
        const val = object[fieldName];
        if (isOptional && val === undefined)
            return;
        if (!checkVal(val, object)) {
            throw new Error(`Invalid param ${String(fieldName)}=${val} (${typeof val}), expected ${type}`);
        }
    };
    for (const [fieldName, type] of Object.entries(validators))
        checkField(fieldName, type, false);
    for (const [fieldName, type] of Object.entries(optValidators))
        checkField(fieldName, type, true);
    return object;
}
// validate type tests
// const o: { a: number; b: number; c: number } = { a: 1, b: 5, c: 6 };
// const z0 = validateObject(o, { a: 'isSafeInteger' }, { c: 'bigint' }); // Ok!
// // Should fail type-check
// const z1 = validateObject(o, { a: 'tmp' }, { c: 'zz' });
// const z2 = validateObject(o, { a: 'isSafeInteger' }, { c: 'zz' });
// const z3 = validateObject(o, { test: 'boolean', z: 'bug' });
// const z4 = validateObject(o, { a: 'boolean', z: 'bug' });

var ut = {
    __proto__: null,
    isBytes: isBytes$1,
    abytes: abytes,
    bytesToHex: bytesToHex$1,
    numberToHexUnpadded: numberToHexUnpadded,
    hexToNumber: hexToNumber,
    hexToBytes: hexToBytes$1,
    bytesToNumberBE: bytesToNumberBE,
    bytesToNumberLE: bytesToNumberLE,
    numberToBytesBE: numberToBytesBE,
    numberToBytesLE: numberToBytesLE,
    numberToVarBytesBE: numberToVarBytesBE,
    ensureBytes: ensureBytes,
    concatBytes: concatBytes,
    equalBytes: equalBytes,
    utf8ToBytes: utf8ToBytes,
    bitLen: bitLen,
    bitGet: bitGet,
    bitSet: bitSet,
    bitMask: bitMask,
    createHmacDrbg: createHmacDrbg,
    validateObject: validateObject
};

/*! noble-curves - MIT License (c) 2022 Paul Miller (paulmillr.com) */
// prettier-ignore
const _0n$4 = BigInt(0), _1n$6 = BigInt(1), _2n$4 = BigInt(2), _3n$1 = BigInt(3);
// prettier-ignore
const _4n = BigInt(4), _5n$1 = BigInt(5), _8n$2 = BigInt(8);
// prettier-ignore
BigInt(9); BigInt(16);
// Calculates a modulo b
function mod(a, b) {
    const result = a % b;
    return result >= _0n$4 ? result : b + result;
}
/**
 * Efficiently raise num to power and do modular division.
 * Unsafe in some contexts: uses ladder, so can expose bigint bits.
 * @example
 * pow(2n, 6n, 11n) // 64n % 11n == 9n
 */
// TODO: use field version && remove
function pow(num, power, modulo) {
    if (modulo <= _0n$4 || power < _0n$4)
        throw new Error('Expected power/modulo > 0');
    if (modulo === _1n$6)
        return _0n$4;
    let res = _1n$6;
    while (power > _0n$4) {
        if (power & _1n$6)
            res = (res * num) % modulo;
        num = (num * num) % modulo;
        power >>= _1n$6;
    }
    return res;
}
// Does x ^ (2 ^ power) mod p. pow2(30, 4) == 30 ^ (2 ^ 4)
function pow2(x, power, modulo) {
    let res = x;
    while (power-- > _0n$4) {
        res *= res;
        res %= modulo;
    }
    return res;
}
// Inverses number over modulo
function invert(number, modulo) {
    if (number === _0n$4 || modulo <= _0n$4) {
        throw new Error(`invert: expected positive integers, got n=${number} mod=${modulo}`);
    }
    // Euclidean GCD https://brilliant.org/wiki/extended-euclidean-algorithm/
    // Fermat's little theorem "CT-like" version inv(n) = n^(m-2) mod m is 30x slower.
    let a = mod(number, modulo);
    let b = modulo;
    // prettier-ignore
    let x = _0n$4, u = _1n$6;
    while (a !== _0n$4) {
        // JIT applies optimization if those two lines follow each other
        const q = b / a;
        const r = b % a;
        const m = x - u * q;
        // prettier-ignore
        b = a, a = r, x = u, u = m;
    }
    const gcd = b;
    if (gcd !== _1n$6)
        throw new Error('invert: does not exist');
    return mod(x, modulo);
}
/**
 * Tonelli-Shanks square root search algorithm.
 * 1. https://eprint.iacr.org/2012/685.pdf (page 12)
 * 2. Square Roots from 1; 24, 51, 10 to Dan Shanks
 * Will start an infinite loop if field order P is not prime.
 * @param P field order
 * @returns function that takes field Fp (created from P) and number n
 */
function tonelliShanks(P) {
    // Legendre constant: used to calculate Legendre symbol (a | p),
    // which denotes the value of a^((p-1)/2) (mod p).
    // (a | p) ≡ 1    if a is a square (mod p)
    // (a | p) ≡ -1   if a is not a square (mod p)
    // (a | p) ≡ 0    if a ≡ 0 (mod p)
    const legendreC = (P - _1n$6) / _2n$4;
    let Q, S, Z;
    // Step 1: By factoring out powers of 2 from p - 1,
    // find q and s such that p - 1 = q*(2^s) with q odd
    for (Q = P - _1n$6, S = 0; Q % _2n$4 === _0n$4; Q /= _2n$4, S++)
        ;
    // Step 2: Select a non-square z such that (z | p) ≡ -1 and set c ≡ zq
    for (Z = _2n$4; Z < P && pow(Z, legendreC, P) !== P - _1n$6; Z++)
        ;
    // Fast-path
    if (S === 1) {
        const p1div4 = (P + _1n$6) / _4n;
        return function tonelliFast(Fp, n) {
            const root = Fp.pow(n, p1div4);
            if (!Fp.eql(Fp.sqr(root), n))
                throw new Error('Cannot find square root');
            return root;
        };
    }
    // Slow-path
    const Q1div2 = (Q + _1n$6) / _2n$4;
    return function tonelliSlow(Fp, n) {
        // Step 0: Check that n is indeed a square: (n | p) should not be ≡ -1
        if (Fp.pow(n, legendreC) === Fp.neg(Fp.ONE))
            throw new Error('Cannot find square root');
        let r = S;
        // TODO: will fail at Fp2/etc
        let g = Fp.pow(Fp.mul(Fp.ONE, Z), Q); // will update both x and b
        let x = Fp.pow(n, Q1div2); // first guess at the square root
        let b = Fp.pow(n, Q); // first guess at the fudge factor
        while (!Fp.eql(b, Fp.ONE)) {
            if (Fp.eql(b, Fp.ZERO))
                return Fp.ZERO; // https://en.wikipedia.org/wiki/Tonelli%E2%80%93Shanks_algorithm (4. If t = 0, return r = 0)
            // Find m such b^(2^m)==1
            let m = 1;
            for (let t2 = Fp.sqr(b); m < r; m++) {
                if (Fp.eql(t2, Fp.ONE))
                    break;
                t2 = Fp.sqr(t2); // t2 *= t2
            }
            // NOTE: r-m-1 can be bigger than 32, need to convert to bigint before shift, otherwise there will be overflow
            const ge = Fp.pow(g, _1n$6 << BigInt(r - m - 1)); // ge = 2^(r-m-1)
            g = Fp.sqr(ge); // g = ge * ge
            x = Fp.mul(x, ge); // x *= ge
            b = Fp.mul(b, g); // b *= g
            r = m;
        }
        return x;
    };
}
function FpSqrt(P) {
    // NOTE: different algorithms can give different roots, it is up to user to decide which one they want.
    // For example there is FpSqrtOdd/FpSqrtEven to choice root based on oddness (used for hash-to-curve).
    // P ≡ 3 (mod 4)
    // √n = n^((P+1)/4)
    if (P % _4n === _3n$1) {
        // Not all roots possible!
        // const ORDER =
        //   0x1a0111ea397fe69a4b1ba7b6434bacd764774b84f38512bf6730d2a0f6b0f6241eabfffeb153ffffb9feffffffffaaabn;
        // const NUM = 72057594037927816n;
        const p1div4 = (P + _1n$6) / _4n;
        return function sqrt3mod4(Fp, n) {
            const root = Fp.pow(n, p1div4);
            // Throw if root**2 != n
            if (!Fp.eql(Fp.sqr(root), n))
                throw new Error('Cannot find square root');
            return root;
        };
    }
    // Atkin algorithm for q ≡ 5 (mod 8), https://eprint.iacr.org/2012/685.pdf (page 10)
    if (P % _8n$2 === _5n$1) {
        const c1 = (P - _5n$1) / _8n$2;
        return function sqrt5mod8(Fp, n) {
            const n2 = Fp.mul(n, _2n$4);
            const v = Fp.pow(n2, c1);
            const nv = Fp.mul(n, v);
            const i = Fp.mul(Fp.mul(nv, _2n$4), v);
            const root = Fp.mul(nv, Fp.sub(i, Fp.ONE));
            if (!Fp.eql(Fp.sqr(root), n))
                throw new Error('Cannot find square root');
            return root;
        };
    }
    // Other cases: Tonelli-Shanks algorithm
    return tonelliShanks(P);
}
// Little-endian check for first LE bit (last BE bit);
const isNegativeLE = (num, modulo) => (mod(num, modulo) & _1n$6) === _1n$6;
// prettier-ignore
const FIELD_FIELDS = [
    'create', 'isValid', 'is0', 'neg', 'inv', 'sqrt', 'sqr',
    'eql', 'add', 'sub', 'mul', 'pow', 'div',
    'addN', 'subN', 'mulN', 'sqrN'
];
function validateField(field) {
    const initial = {
        ORDER: 'bigint',
        MASK: 'bigint',
        BYTES: 'isSafeInteger',
        BITS: 'isSafeInteger',
    };
    const opts = FIELD_FIELDS.reduce((map, val) => {
        map[val] = 'function';
        return map;
    }, initial);
    return validateObject(field, opts);
}
// Generic field functions
/**
 * Same as `pow` but for Fp: non-constant-time.
 * Unsafe in some contexts: uses ladder, so can expose bigint bits.
 */
function FpPow(f, num, power) {
    // Should have same speed as pow for bigints
    // TODO: benchmark!
    if (power < _0n$4)
        throw new Error('Expected power > 0');
    if (power === _0n$4)
        return f.ONE;
    if (power === _1n$6)
        return num;
    let p = f.ONE;
    let d = num;
    while (power > _0n$4) {
        if (power & _1n$6)
            p = f.mul(p, d);
        d = f.sqr(d);
        power >>= _1n$6;
    }
    return p;
}
/**
 * Efficiently invert an array of Field elements.
 * `inv(0)` will return `undefined` here: make sure to throw an error.
 */
function FpInvertBatch(f, nums) {
    const tmp = new Array(nums.length);
    // Walk from first to last, multiply them by each other MOD p
    const lastMultiplied = nums.reduce((acc, num, i) => {
        if (f.is0(num))
            return acc;
        tmp[i] = acc;
        return f.mul(acc, num);
    }, f.ONE);
    // Invert last element
    const inverted = f.inv(lastMultiplied);
    // Walk from last to first, multiply them by inverted each other MOD p
    nums.reduceRight((acc, num, i) => {
        if (f.is0(num))
            return acc;
        tmp[i] = f.mul(acc, tmp[i]);
        return f.mul(acc, num);
    }, inverted);
    return tmp;
}
// CURVE.n lengths
function nLength(n, nBitLength) {
    // Bit size, byte size of CURVE.n
    const _nBitLength = nBitLength !== undefined ? nBitLength : n.toString(2).length;
    const nByteLength = Math.ceil(_nBitLength / 8);
    return { nBitLength: _nBitLength, nByteLength };
}
/**
 * Initializes a finite field over prime. **Non-primes are not supported.**
 * Do not init in loop: slow. Very fragile: always run a benchmark on a change.
 * Major performance optimizations:
 * * a) denormalized operations like mulN instead of mul
 * * b) same object shape: never add or remove keys
 * * c) Object.freeze
 * @param ORDER prime positive bigint
 * @param bitLen how many bits the field consumes
 * @param isLE (def: false) if encoding / decoding should be in little-endian
 * @param redef optional faster redefinitions of sqrt and other methods
 */
function Field(ORDER, bitLen, isLE = false, redef = {}) {
    if (ORDER <= _0n$4)
        throw new Error(`Expected Field ORDER > 0, got ${ORDER}`);
    const { nBitLength: BITS, nByteLength: BYTES } = nLength(ORDER, bitLen);
    if (BYTES > 2048)
        throw new Error('Field lengths over 2048 bytes are not supported');
    const sqrtP = FpSqrt(ORDER);
    const f = Object.freeze({
        ORDER,
        BITS,
        BYTES,
        MASK: bitMask(BITS),
        ZERO: _0n$4,
        ONE: _1n$6,
        create: (num) => mod(num, ORDER),
        isValid: (num) => {
            if (typeof num !== 'bigint')
                throw new Error(`Invalid field element: expected bigint, got ${typeof num}`);
            return _0n$4 <= num && num < ORDER; // 0 is valid element, but it's not invertible
        },
        is0: (num) => num === _0n$4,
        isOdd: (num) => (num & _1n$6) === _1n$6,
        neg: (num) => mod(-num, ORDER),
        eql: (lhs, rhs) => lhs === rhs,
        sqr: (num) => mod(num * num, ORDER),
        add: (lhs, rhs) => mod(lhs + rhs, ORDER),
        sub: (lhs, rhs) => mod(lhs - rhs, ORDER),
        mul: (lhs, rhs) => mod(lhs * rhs, ORDER),
        pow: (num, power) => FpPow(f, num, power),
        div: (lhs, rhs) => mod(lhs * invert(rhs, ORDER), ORDER),
        // Same as above, but doesn't normalize
        sqrN: (num) => num * num,
        addN: (lhs, rhs) => lhs + rhs,
        subN: (lhs, rhs) => lhs - rhs,
        mulN: (lhs, rhs) => lhs * rhs,
        inv: (num) => invert(num, ORDER),
        sqrt: redef.sqrt || ((n) => sqrtP(f, n)),
        invertBatch: (lst) => FpInvertBatch(f, lst),
        // TODO: do we really need constant cmov?
        // We don't have const-time bigints anyway, so probably will be not very useful
        cmov: (a, b, c) => (c ? b : a),
        toBytes: (num) => (isLE ? numberToBytesLE(num, BYTES) : numberToBytesBE(num, BYTES)),
        fromBytes: (bytes) => {
            if (bytes.length !== BYTES)
                throw new Error(`Fp.fromBytes: expected ${BYTES}, got ${bytes.length}`);
            return isLE ? bytesToNumberLE(bytes) : bytesToNumberBE(bytes);
        },
    });
    return Object.freeze(f);
}
/**
 * Returns total number of bytes consumed by the field element.
 * For example, 32 bytes for usual 256-bit weierstrass curve.
 * @param fieldOrder number of field elements, usually CURVE.n
 * @returns byte length of field
 */
function getFieldBytesLength(fieldOrder) {
    if (typeof fieldOrder !== 'bigint')
        throw new Error('field order must be bigint');
    const bitLength = fieldOrder.toString(2).length;
    return Math.ceil(bitLength / 8);
}
/**
 * Returns minimal amount of bytes that can be safely reduced
 * by field order.
 * Should be 2^-128 for 128-bit curve such as P256.
 * @param fieldOrder number of field elements, usually CURVE.n
 * @returns byte length of target hash
 */
function getMinHashLength(fieldOrder) {
    const length = getFieldBytesLength(fieldOrder);
    return length + Math.ceil(length / 2);
}
/**
 * "Constant-time" private key generation utility.
 * Can take (n + n/2) or more bytes of uniform input e.g. from CSPRNG or KDF
 * and convert them into private scalar, with the modulo bias being negligible.
 * Needs at least 48 bytes of input for 32-byte private key.
 * https://research.kudelskisecurity.com/2020/07/28/the-definitive-guide-to-modulo-bias-and-how-to-avoid-it/
 * FIPS 186-5, A.2 https://csrc.nist.gov/publications/detail/fips/186/5/final
 * RFC 9380, https://www.rfc-editor.org/rfc/rfc9380#section-5
 * @param hash hash output from SHA3 or a similar function
 * @param groupOrder size of subgroup - (e.g. secp256k1.CURVE.n)
 * @param isLE interpret hash bytes as LE num
 * @returns valid private scalar
 */
function mapHashToField(key, fieldOrder, isLE = false) {
    const len = key.length;
    const fieldLen = getFieldBytesLength(fieldOrder);
    const minLen = getMinHashLength(fieldOrder);
    // No small numbers: need to understand bias story. No huge numbers: easier to detect JS timings.
    if (len < 16 || len < minLen || len > 1024)
        throw new Error(`expected ${minLen}-1024 bytes of input, got ${len}`);
    const num = isLE ? bytesToNumberBE(key) : bytesToNumberLE(key);
    // `mod(x, 11)` can sometimes produce 0. `mod(x, 10) + 1` is the same, but no 0
    const reduced = mod(num, fieldOrder - _1n$6) + _1n$6;
    return isLE ? numberToBytesLE(reduced, fieldLen) : numberToBytesBE(reduced, fieldLen);
}

/*! noble-curves - MIT License (c) 2022 Paul Miller (paulmillr.com) */
const _0n$3 = BigInt(0);
const _1n$5 = BigInt(1);
// Elliptic curve multiplication of Point by scalar. Fragile.
// Scalars should always be less than curve order: this should be checked inside of a curve itself.
// Creates precomputation tables for fast multiplication:
// - private scalar is split by fixed size windows of W bits
// - every window point is collected from window's table & added to accumulator
// - since windows are different, same point inside tables won't be accessed more than once per calc
// - each multiplication is 'Math.ceil(CURVE_ORDER / 𝑊) + 1' point additions (fixed for any scalar)
// - +1 window is neccessary for wNAF
// - wNAF reduces table size: 2x less memory + 2x faster generation, but 10% slower multiplication
// TODO: Research returning 2d JS array of windows, instead of a single window. This would allow
// windows to be in different memory locations
function wNAF(c, bits) {
    const constTimeNegate = (condition, item) => {
        const neg = item.negate();
        return condition ? neg : item;
    };
    const opts = (W) => {
        const windows = Math.ceil(bits / W) + 1; // +1, because
        const windowSize = 2 ** (W - 1); // -1 because we skip zero
        return { windows, windowSize };
    };
    return {
        constTimeNegate,
        // non-const time multiplication ladder
        unsafeLadder(elm, n) {
            let p = c.ZERO;
            let d = elm;
            while (n > _0n$3) {
                if (n & _1n$5)
                    p = p.add(d);
                d = d.double();
                n >>= _1n$5;
            }
            return p;
        },
        /**
         * Creates a wNAF precomputation window. Used for caching.
         * Default window size is set by `utils.precompute()` and is equal to 8.
         * Number of precomputed points depends on the curve size:
         * 2^(𝑊−1) * (Math.ceil(𝑛 / 𝑊) + 1), where:
         * - 𝑊 is the window size
         * - 𝑛 is the bitlength of the curve order.
         * For a 256-bit curve and window size 8, the number of precomputed points is 128 * 33 = 4224.
         * @returns precomputed point tables flattened to a single array
         */
        precomputeWindow(elm, W) {
            const { windows, windowSize } = opts(W);
            const points = [];
            let p = elm;
            let base = p;
            for (let window = 0; window < windows; window++) {
                base = p;
                points.push(base);
                // =1, because we skip zero
                for (let i = 1; i < windowSize; i++) {
                    base = base.add(p);
                    points.push(base);
                }
                p = base.double();
            }
            return points;
        },
        /**
         * Implements ec multiplication using precomputed tables and w-ary non-adjacent form.
         * @param W window size
         * @param precomputes precomputed tables
         * @param n scalar (we don't check here, but should be less than curve order)
         * @returns real and fake (for const-time) points
         */
        wNAF(W, precomputes, n) {
            // TODO: maybe check that scalar is less than group order? wNAF behavious is undefined otherwise
            // But need to carefully remove other checks before wNAF. ORDER == bits here
            const { windows, windowSize } = opts(W);
            let p = c.ZERO;
            let f = c.BASE;
            const mask = BigInt(2 ** W - 1); // Create mask with W ones: 0b1111 for W=4 etc.
            const maxNumber = 2 ** W;
            const shiftBy = BigInt(W);
            for (let window = 0; window < windows; window++) {
                const offset = window * windowSize;
                // Extract W bits.
                let wbits = Number(n & mask);
                // Shift number by W bits.
                n >>= shiftBy;
                // If the bits are bigger than max size, we'll split those.
                // +224 => 256 - 32
                if (wbits > windowSize) {
                    wbits -= maxNumber;
                    n += _1n$5;
                }
                // This code was first written with assumption that 'f' and 'p' will never be infinity point:
                // since each addition is multiplied by 2 ** W, it cannot cancel each other. However,
                // there is negate now: it is possible that negated element from low value
                // would be the same as high element, which will create carry into next window.
                // It's not obvious how this can fail, but still worth investigating later.
                // Check if we're onto Zero point.
                // Add random point inside current window to f.
                const offset1 = offset;
                const offset2 = offset + Math.abs(wbits) - 1; // -1 because we skip zero
                const cond1 = window % 2 !== 0;
                const cond2 = wbits < 0;
                if (wbits === 0) {
                    // The most important part for const-time getPublicKey
                    f = f.add(constTimeNegate(cond1, precomputes[offset1]));
                }
                else {
                    p = p.add(constTimeNegate(cond2, precomputes[offset2]));
                }
            }
            // JIT-compiler should not eliminate f here, since it will later be used in normalizeZ()
            // Even if the variable is still unused, there are some checks which will
            // throw an exception, so compiler needs to prove they won't happen, which is hard.
            // At this point there is a way to F be infinity-point even if p is not,
            // which makes it less const-time: around 1 bigint multiply.
            return { p, f };
        },
        wNAFCached(P, precomputesMap, n, transform) {
            // @ts-ignore
            const W = P._WINDOW_SIZE || 1;
            // Calculate precomputes on a first run, reuse them after
            let comp = precomputesMap.get(P);
            if (!comp) {
                comp = this.precomputeWindow(P, W);
                if (W !== 1) {
                    precomputesMap.set(P, transform(comp));
                }
            }
            return this.wNAF(W, comp, n);
        },
    };
}
function validateBasic(curve) {
    validateField(curve.Fp);
    validateObject(curve, {
        n: 'bigint',
        h: 'bigint',
        Gx: 'field',
        Gy: 'field',
    }, {
        nBitLength: 'isSafeInteger',
        nByteLength: 'isSafeInteger',
    });
    // Set defaults
    return Object.freeze({
        ...nLength(curve.n, curve.nBitLength),
        ...curve,
        ...{ p: curve.Fp.ORDER },
    });
}

/*! noble-curves - MIT License (c) 2022 Paul Miller (paulmillr.com) */
// Be friendly to bad ECMAScript parsers by not using bigint literals
// prettier-ignore
const _0n$2 = BigInt(0), _1n$4 = BigInt(1), _2n$3 = BigInt(2), _8n$1 = BigInt(8);
// verification rule is either zip215 or rfc8032 / nist186-5. Consult fromHex:
const VERIFY_DEFAULT = { zip215: true };
function validateOpts$1(curve) {
    const opts = validateBasic(curve);
    validateObject(curve, {
        hash: 'function',
        a: 'bigint',
        d: 'bigint',
        randomBytes: 'function',
    }, {
        adjustScalarBytes: 'function',
        domain: 'function',
        uvRatio: 'function',
        mapToCurve: 'function',
    });
    // Set defaults
    return Object.freeze({ ...opts });
}
// It is not generic twisted curve for now, but ed25519/ed448 generic implementation
function twistedEdwards(curveDef) {
    const CURVE = validateOpts$1(curveDef);
    const { Fp, n: CURVE_ORDER, prehash: prehash, hash: cHash, randomBytes, nByteLength, h: cofactor, } = CURVE;
    const MASK = _2n$3 << (BigInt(nByteLength * 8) - _1n$4);
    const modP = Fp.create; // Function overrides
    // sqrt(u/v)
    const uvRatio = CURVE.uvRatio ||
        ((u, v) => {
            try {
                return { isValid: true, value: Fp.sqrt(u * Fp.inv(v)) };
            }
            catch (e) {
                return { isValid: false, value: _0n$2 };
            }
        });
    const adjustScalarBytes = CURVE.adjustScalarBytes || ((bytes) => bytes); // NOOP
    const domain = CURVE.domain ||
        ((data, ctx, phflag) => {
            if (ctx.length || phflag)
                throw new Error('Contexts/pre-hash are not supported');
            return data;
        }); // NOOP
    const inBig = (n) => typeof n === 'bigint' && _0n$2 < n; // n in [1..]
    const inRange = (n, max) => inBig(n) && inBig(max) && n < max; // n in [1..max-1]
    const in0MaskRange = (n) => n === _0n$2 || inRange(n, MASK); // n in [0..MASK-1]
    function assertInRange(n, max) {
        // n in [1..max-1]
        if (inRange(n, max))
            return n;
        throw new Error(`Expected valid scalar < ${max}, got ${typeof n} ${n}`);
    }
    function assertGE0(n) {
        // n in [0..CURVE_ORDER-1]
        return n === _0n$2 ? n : assertInRange(n, CURVE_ORDER); // GE = prime subgroup, not full group
    }
    const pointPrecomputes = new Map();
    function isPoint(other) {
        if (!(other instanceof Point))
            throw new Error('ExtendedPoint expected');
    }
    // Extended Point works in extended coordinates: (x, y, z, t) ∋ (x=x/z, y=y/z, t=xy).
    // https://en.wikipedia.org/wiki/Twisted_Edwards_curve#Extended_coordinates
    class Point {
        constructor(ex, ey, ez, et) {
            this.ex = ex;
            this.ey = ey;
            this.ez = ez;
            this.et = et;
            if (!in0MaskRange(ex))
                throw new Error('x required');
            if (!in0MaskRange(ey))
                throw new Error('y required');
            if (!in0MaskRange(ez))
                throw new Error('z required');
            if (!in0MaskRange(et))
                throw new Error('t required');
        }
        get x() {
            return this.toAffine().x;
        }
        get y() {
            return this.toAffine().y;
        }
        static fromAffine(p) {
            if (p instanceof Point)
                throw new Error('extended point not allowed');
            const { x, y } = p || {};
            if (!in0MaskRange(x) || !in0MaskRange(y))
                throw new Error('invalid affine point');
            return new Point(x, y, _1n$4, modP(x * y));
        }
        static normalizeZ(points) {
            const toInv = Fp.invertBatch(points.map((p) => p.ez));
            return points.map((p, i) => p.toAffine(toInv[i])).map(Point.fromAffine);
        }
        // "Private method", don't use it directly
        _setWindowSize(windowSize) {
            this._WINDOW_SIZE = windowSize;
            pointPrecomputes.delete(this);
        }
        // Not required for fromHex(), which always creates valid points.
        // Could be useful for fromAffine().
        assertValidity() {
            const { a, d } = CURVE;
            if (this.is0())
                throw new Error('bad point: ZERO'); // TODO: optimize, with vars below?
            // Equation in affine coordinates: ax² + y² = 1 + dx²y²
            // Equation in projective coordinates (X/Z, Y/Z, Z):  (aX² + Y²)Z² = Z⁴ + dX²Y²
            const { ex: X, ey: Y, ez: Z, et: T } = this;
            const X2 = modP(X * X); // X²
            const Y2 = modP(Y * Y); // Y²
            const Z2 = modP(Z * Z); // Z²
            const Z4 = modP(Z2 * Z2); // Z⁴
            const aX2 = modP(X2 * a); // aX²
            const left = modP(Z2 * modP(aX2 + Y2)); // (aX² + Y²)Z²
            const right = modP(Z4 + modP(d * modP(X2 * Y2))); // Z⁴ + dX²Y²
            if (left !== right)
                throw new Error('bad point: equation left != right (1)');
            // In Extended coordinates we also have T, which is x*y=T/Z: check X*Y == Z*T
            const XY = modP(X * Y);
            const ZT = modP(Z * T);
            if (XY !== ZT)
                throw new Error('bad point: equation left != right (2)');
        }
        // Compare one point to another.
        equals(other) {
            isPoint(other);
            const { ex: X1, ey: Y1, ez: Z1 } = this;
            const { ex: X2, ey: Y2, ez: Z2 } = other;
            const X1Z2 = modP(X1 * Z2);
            const X2Z1 = modP(X2 * Z1);
            const Y1Z2 = modP(Y1 * Z2);
            const Y2Z1 = modP(Y2 * Z1);
            return X1Z2 === X2Z1 && Y1Z2 === Y2Z1;
        }
        is0() {
            return this.equals(Point.ZERO);
        }
        negate() {
            // Flips point sign to a negative one (-x, y in affine coords)
            return new Point(modP(-this.ex), this.ey, this.ez, modP(-this.et));
        }
        // Fast algo for doubling Extended Point.
        // https://hyperelliptic.org/EFD/g1p/auto-twisted-extended.html#doubling-dbl-2008-hwcd
        // Cost: 4M + 4S + 1*a + 6add + 1*2.
        double() {
            const { a } = CURVE;
            const { ex: X1, ey: Y1, ez: Z1 } = this;
            const A = modP(X1 * X1); // A = X12
            const B = modP(Y1 * Y1); // B = Y12
            const C = modP(_2n$3 * modP(Z1 * Z1)); // C = 2*Z12
            const D = modP(a * A); // D = a*A
            const x1y1 = X1 + Y1;
            const E = modP(modP(x1y1 * x1y1) - A - B); // E = (X1+Y1)2-A-B
            const G = D + B; // G = D+B
            const F = G - C; // F = G-C
            const H = D - B; // H = D-B
            const X3 = modP(E * F); // X3 = E*F
            const Y3 = modP(G * H); // Y3 = G*H
            const T3 = modP(E * H); // T3 = E*H
            const Z3 = modP(F * G); // Z3 = F*G
            return new Point(X3, Y3, Z3, T3);
        }
        // Fast algo for adding 2 Extended Points.
        // https://hyperelliptic.org/EFD/g1p/auto-twisted-extended.html#addition-add-2008-hwcd
        // Cost: 9M + 1*a + 1*d + 7add.
        add(other) {
            isPoint(other);
            const { a, d } = CURVE;
            const { ex: X1, ey: Y1, ez: Z1, et: T1 } = this;
            const { ex: X2, ey: Y2, ez: Z2, et: T2 } = other;
            // Faster algo for adding 2 Extended Points when curve's a=-1.
            // http://hyperelliptic.org/EFD/g1p/auto-twisted-extended-1.html#addition-add-2008-hwcd-4
            // Cost: 8M + 8add + 2*2.
            // Note: It does not check whether the `other` point is valid.
            if (a === BigInt(-1)) {
                const A = modP((Y1 - X1) * (Y2 + X2));
                const B = modP((Y1 + X1) * (Y2 - X2));
                const F = modP(B - A);
                if (F === _0n$2)
                    return this.double(); // Same point. Tests say it doesn't affect timing
                const C = modP(Z1 * _2n$3 * T2);
                const D = modP(T1 * _2n$3 * Z2);
                const E = D + C;
                const G = B + A;
                const H = D - C;
                const X3 = modP(E * F);
                const Y3 = modP(G * H);
                const T3 = modP(E * H);
                const Z3 = modP(F * G);
                return new Point(X3, Y3, Z3, T3);
            }
            const A = modP(X1 * X2); // A = X1*X2
            const B = modP(Y1 * Y2); // B = Y1*Y2
            const C = modP(T1 * d * T2); // C = T1*d*T2
            const D = modP(Z1 * Z2); // D = Z1*Z2
            const E = modP((X1 + Y1) * (X2 + Y2) - A - B); // E = (X1+Y1)*(X2+Y2)-A-B
            const F = D - C; // F = D-C
            const G = D + C; // G = D+C
            const H = modP(B - a * A); // H = B-a*A
            const X3 = modP(E * F); // X3 = E*F
            const Y3 = modP(G * H); // Y3 = G*H
            const T3 = modP(E * H); // T3 = E*H
            const Z3 = modP(F * G); // Z3 = F*G
            return new Point(X3, Y3, Z3, T3);
        }
        subtract(other) {
            return this.add(other.negate());
        }
        wNAF(n) {
            return wnaf.wNAFCached(this, pointPrecomputes, n, Point.normalizeZ);
        }
        // Constant-time multiplication.
        multiply(scalar) {
            const { p, f } = this.wNAF(assertInRange(scalar, CURVE_ORDER));
            return Point.normalizeZ([p, f])[0];
        }
        // Non-constant-time multiplication. Uses double-and-add algorithm.
        // It's faster, but should only be used when you don't care about
        // an exposed private key e.g. sig verification.
        // Does NOT allow scalars higher than CURVE.n.
        multiplyUnsafe(scalar) {
            let n = assertGE0(scalar); // 0 <= scalar < CURVE.n
            if (n === _0n$2)
                return I;
            if (this.equals(I) || n === _1n$4)
                return this;
            if (this.equals(G))
                return this.wNAF(n).p;
            return wnaf.unsafeLadder(this, n);
        }
        // Checks if point is of small order.
        // If you add something to small order point, you will have "dirty"
        // point with torsion component.
        // Multiplies point by cofactor and checks if the result is 0.
        isSmallOrder() {
            return this.multiplyUnsafe(cofactor).is0();
        }
        // Multiplies point by curve order and checks if the result is 0.
        // Returns `false` is the point is dirty.
        isTorsionFree() {
            return wnaf.unsafeLadder(this, CURVE_ORDER).is0();
        }
        // Converts Extended point to default (x, y) coordinates.
        // Can accept precomputed Z^-1 - for example, from invertBatch.
        toAffine(iz) {
            const { ex: x, ey: y, ez: z } = this;
            const is0 = this.is0();
            if (iz == null)
                iz = is0 ? _8n$1 : Fp.inv(z); // 8 was chosen arbitrarily
            const ax = modP(x * iz);
            const ay = modP(y * iz);
            const zz = modP(z * iz);
            if (is0)
                return { x: _0n$2, y: _1n$4 };
            if (zz !== _1n$4)
                throw new Error('invZ was invalid');
            return { x: ax, y: ay };
        }
        clearCofactor() {
            const { h: cofactor } = CURVE;
            if (cofactor === _1n$4)
                return this;
            return this.multiplyUnsafe(cofactor);
        }
        // Converts hash string or Uint8Array to Point.
        // Uses algo from RFC8032 5.1.3.
        static fromHex(hex, zip215 = false) {
            const { d, a } = CURVE;
            const len = Fp.BYTES;
            hex = ensureBytes('pointHex', hex, len); // copy hex to a new array
            const normed = hex.slice(); // copy again, we'll manipulate it
            const lastByte = hex[len - 1]; // select last byte
            normed[len - 1] = lastByte & ~0x80; // clear last bit
            const y = bytesToNumberLE(normed);
            if (y === _0n$2) ;
            else {
                // RFC8032 prohibits >= p, but ZIP215 doesn't
                if (zip215)
                    assertInRange(y, MASK); // zip215=true [1..P-1] (2^255-19-1 for ed25519)
                else
                    assertInRange(y, Fp.ORDER); // zip215=false [1..MASK-1] (2^256-1 for ed25519)
            }
            // Ed25519: x² = (y²-1)/(dy²+1) mod p. Ed448: x² = (y²-1)/(dy²-1) mod p. Generic case:
            // ax²+y²=1+dx²y² => y²-1=dx²y²-ax² => y²-1=x²(dy²-a) => x²=(y²-1)/(dy²-a)
            const y2 = modP(y * y); // denominator is always non-0 mod p.
            const u = modP(y2 - _1n$4); // u = y² - 1
            const v = modP(d * y2 - a); // v = d y² + 1.
            let { isValid, value: x } = uvRatio(u, v); // √(u/v)
            if (!isValid)
                throw new Error('Point.fromHex: invalid y coordinate');
            const isXOdd = (x & _1n$4) === _1n$4; // There are 2 square roots. Use x_0 bit to select proper
            const isLastByteOdd = (lastByte & 0x80) !== 0; // x_0, last bit
            if (!zip215 && x === _0n$2 && isLastByteOdd)
                // if x=0 and x_0 = 1, fail
                throw new Error('Point.fromHex: x=0 and x_0=1');
            if (isLastByteOdd !== isXOdd)
                x = modP(-x); // if x_0 != x mod 2, set x = p-x
            return Point.fromAffine({ x, y });
        }
        static fromPrivateKey(privKey) {
            return getExtendedPublicKey(privKey).point;
        }
        toRawBytes() {
            const { x, y } = this.toAffine();
            const bytes = numberToBytesLE(y, Fp.BYTES); // each y has 2 x values (x, -y)
            bytes[bytes.length - 1] |= x & _1n$4 ? 0x80 : 0; // when compressing, it's enough to store y
            return bytes; // and use the last byte to encode sign of x
        }
        toHex() {
            return bytesToHex$1(this.toRawBytes()); // Same as toRawBytes, but returns string.
        }
    }
    Point.BASE = new Point(CURVE.Gx, CURVE.Gy, _1n$4, modP(CURVE.Gx * CURVE.Gy));
    Point.ZERO = new Point(_0n$2, _1n$4, _1n$4, _0n$2); // 0, 1, 1, 0
    const { BASE: G, ZERO: I } = Point;
    const wnaf = wNAF(Point, nByteLength * 8);
    function modN(a) {
        return mod(a, CURVE_ORDER);
    }
    // Little-endian SHA512 with modulo n
    function modN_LE(hash) {
        return modN(bytesToNumberLE(hash));
    }
    /** Convenience method that creates public key and other stuff. RFC8032 5.1.5 */
    function getExtendedPublicKey(key) {
        const len = nByteLength;
        key = ensureBytes('private key', key, len);
        // Hash private key with curve's hash function to produce uniformingly random input
        // Check byte lengths: ensure(64, h(ensure(32, key)))
        const hashed = ensureBytes('hashed private key', cHash(key), 2 * len);
        const head = adjustScalarBytes(hashed.slice(0, len)); // clear first half bits, produce FE
        const prefix = hashed.slice(len, 2 * len); // second half is called key prefix (5.1.6)
        const scalar = modN_LE(head); // The actual private scalar
        const point = G.multiply(scalar); // Point on Edwards curve aka public key
        const pointBytes = point.toRawBytes(); // Uint8Array representation
        return { head, prefix, scalar, point, pointBytes };
    }
    // Calculates EdDSA pub key. RFC8032 5.1.5. Privkey is hashed. Use first half with 3 bits cleared
    function getPublicKey(privKey) {
        return getExtendedPublicKey(privKey).pointBytes;
    }
    // int('LE', SHA512(dom2(F, C) || msgs)) mod N
    function hashDomainToScalar(context = new Uint8Array(), ...msgs) {
        const msg = concatBytes(...msgs);
        return modN_LE(cHash(domain(msg, ensureBytes('context', context), !!prehash)));
    }
    /** Signs message with privateKey. RFC8032 5.1.6 */
    function sign(msg, privKey, options = {}) {
        msg = ensureBytes('message', msg);
        if (prehash)
            msg = prehash(msg); // for ed25519ph etc.
        const { prefix, scalar, pointBytes } = getExtendedPublicKey(privKey);
        const r = hashDomainToScalar(options.context, prefix, msg); // r = dom2(F, C) || prefix || PH(M)
        const R = G.multiply(r).toRawBytes(); // R = rG
        const k = hashDomainToScalar(options.context, R, pointBytes, msg); // R || A || PH(M)
        const s = modN(r + k * scalar); // S = (r + k * s) mod L
        assertGE0(s); // 0 <= s < l
        const res = concatBytes(R, numberToBytesLE(s, Fp.BYTES));
        return ensureBytes('result', res, nByteLength * 2); // 64-byte signature
    }
    const verifyOpts = VERIFY_DEFAULT;
    function verify(sig, msg, publicKey, options = verifyOpts) {
        const { context, zip215 } = options;
        const len = Fp.BYTES; // Verifies EdDSA signature against message and public key. RFC8032 5.1.7.
        sig = ensureBytes('signature', sig, 2 * len); // An extended group equation is checked.
        msg = ensureBytes('message', msg);
        if (prehash)
            msg = prehash(msg); // for ed25519ph, etc
        const s = bytesToNumberLE(sig.slice(len, 2 * len));
        // zip215: true is good for consensus-critical apps and allows points < 2^256
        // zip215: false follows RFC8032 / NIST186-5 and restricts points to CURVE.p
        let A, R, SB;
        try {
            A = Point.fromHex(publicKey, zip215);
            R = Point.fromHex(sig.slice(0, len), zip215);
            SB = G.multiplyUnsafe(s); // 0 <= s < l is done inside
        }
        catch (error) {
            return false;
        }
        if (!zip215 && A.isSmallOrder())
            return false;
        const k = hashDomainToScalar(context, R.toRawBytes(), A.toRawBytes(), msg);
        const RkA = R.add(A.multiplyUnsafe(k));
        // [8][S]B = [8]R + [8][k]A'
        return RkA.subtract(SB).clearCofactor().equals(Point.ZERO);
    }
    G._setWindowSize(8); // Enable precomputes. Slows down first publicKey computation by 20ms.
    const utils = {
        getExtendedPublicKey,
        // ed25519 private keys are uniform 32b. No need to check for modulo bias, like in secp256k1.
        randomPrivateKey: () => randomBytes(Fp.BYTES),
        /**
         * We're doing scalar multiplication (used in getPublicKey etc) with precomputed BASE_POINT
         * values. This slows down first getPublicKey() by milliseconds (see Speed section),
         * but allows to speed-up subsequent getPublicKey() calls up to 20x.
         * @param windowSize 2, 4, 8, 16
         */
        precompute(windowSize = 8, point = Point.BASE) {
            point._setWindowSize(windowSize);
            point.multiply(BigInt(3));
            return point;
        },
    };
    return {
        CURVE,
        getPublicKey,
        sign,
        verify,
        ExtendedPoint: Point,
        utils,
    };
}

/*! noble-curves - MIT License (c) 2022 Paul Miller (paulmillr.com) */
/**
 * ed25519 Twisted Edwards curve with following addons:
 * - X25519 ECDH
 * - Ristretto cofactor elimination
 * - Elligator hash-to-group / point indistinguishability
 */
const ED25519_P = BigInt('57896044618658097711785492504343953926634992332820282019728792003956564819949');
// √(-1) aka √(a) aka 2^((p-1)/4)
const ED25519_SQRT_M1 = /* @__PURE__ */ BigInt('19681161376707505956807079304988542015446066515923890162744021073123829784752');
// prettier-ignore
BigInt(0); const _1n$3 = BigInt(1), _2n$2 = BigInt(2); BigInt(3);
// prettier-ignore
const _5n = BigInt(5), _8n = BigInt(8);
function ed25519_pow_2_252_3(x) {
    // prettier-ignore
    const _10n = BigInt(10), _20n = BigInt(20), _40n = BigInt(40), _80n = BigInt(80);
    const P = ED25519_P;
    const x2 = (x * x) % P;
    const b2 = (x2 * x) % P; // x^3, 11
    const b4 = (pow2(b2, _2n$2, P) * b2) % P; // x^15, 1111
    const b5 = (pow2(b4, _1n$3, P) * x) % P; // x^31
    const b10 = (pow2(b5, _5n, P) * b5) % P;
    const b20 = (pow2(b10, _10n, P) * b10) % P;
    const b40 = (pow2(b20, _20n, P) * b20) % P;
    const b80 = (pow2(b40, _40n, P) * b40) % P;
    const b160 = (pow2(b80, _80n, P) * b80) % P;
    const b240 = (pow2(b160, _80n, P) * b80) % P;
    const b250 = (pow2(b240, _10n, P) * b10) % P;
    const pow_p_5_8 = (pow2(b250, _2n$2, P) * x) % P;
    // ^ To pow to (p+3)/8, multiply it by x.
    return { pow_p_5_8, b2 };
}
function adjustScalarBytes(bytes) {
    // Section 5: For X25519, in order to decode 32 random bytes as an integer scalar,
    // set the three least significant bits of the first byte
    bytes[0] &= 248; // 0b1111_1000
    // and the most significant bit of the last to zero,
    bytes[31] &= 127; // 0b0111_1111
    // set the second most significant bit of the last byte to 1
    bytes[31] |= 64; // 0b0100_0000
    return bytes;
}
// sqrt(u/v)
function uvRatio(u, v) {
    const P = ED25519_P;
    const v3 = mod(v * v * v, P); // v³
    const v7 = mod(v3 * v3 * v, P); // v⁷
    // (p+3)/8 and (p-5)/8
    const pow = ed25519_pow_2_252_3(u * v7).pow_p_5_8;
    let x = mod(u * v3 * pow, P); // (uv³)(uv⁷)^(p-5)/8
    const vx2 = mod(v * x * x, P); // vx²
    const root1 = x; // First root candidate
    const root2 = mod(x * ED25519_SQRT_M1, P); // Second root candidate
    const useRoot1 = vx2 === u; // If vx² = u (mod p), x is a square root
    const useRoot2 = vx2 === mod(-u, P); // If vx² = -u, set x <-- x * 2^((p-1)/4)
    const noRoot = vx2 === mod(-u * ED25519_SQRT_M1, P); // There is no valid root, vx² = -u√(-1)
    if (useRoot1)
        x = root1;
    if (useRoot2 || noRoot)
        x = root2; // We return root2 anyway, for const-time
    if (isNegativeLE(x, P))
        x = mod(-x, P);
    return { isValid: useRoot1 || useRoot2, value: x };
}
const Fp$2 = /* @__PURE__ */ (() => Field(ED25519_P, undefined, true))();
const ed25519Defaults = /* @__PURE__ */ (() => ({
    // Param: a
    a: BigInt(-1), // Fp.create(-1) is proper; our way still works and is faster
    // d is equal to -121665/121666 over finite field.
    // Negative number is P - number, and division is invert(number, P)
    d: BigInt('37095705934669439343138083508754565189542113879843219016388785533085940283555'),
    // Finite field 𝔽p over which we'll do calculations; 2n**255n - 19n
    Fp: Fp$2,
    // Subgroup order: how many points curve has
    // 2n**252n + 27742317777372353535851937790883648493n;
    n: BigInt('7237005577332262213973186563042994240857116359379907606001950938285454250989'),
    // Cofactor
    h: _8n,
    // Base point (x, y) aka generator point
    Gx: BigInt('15112221349535400772501151409588531511454012693041857206046113283949847762202'),
    Gy: BigInt('46316835694926478169428394003475163141307993866256225615783033603165251855960'),
    hash: sha512,
    randomBytes: randomBytes$1,
    adjustScalarBytes,
    // dom2
    // Ratio of u to v. Allows us to combine inversion and square root. Uses algo from RFC8032 5.1.3.
    // Constant-time, u/√v
    uvRatio,
}))();
const ed25519 = /* @__PURE__ */ (() => twistedEdwards(ed25519Defaults))();

function createCommonjsModule(fn) {
  var module = { exports: {} };
	return fn(module, module.exports), module.exports;
}

// base-x encoding / decoding
// Copyright (c) 2018 base-x contributors
// Copyright (c) 2014-2018 The Bitcoin Core developers (base58.cpp)
// Distributed under the MIT software license, see the accompanying
// file LICENSE or http://www.opensource.org/licenses/mit-license.php.
function base$1 (ALPHABET) {
  if (ALPHABET.length >= 255) { throw new TypeError('Alphabet too long') }
  var BASE_MAP = new Uint8Array(256);
  for (var j = 0; j < BASE_MAP.length; j++) {
    BASE_MAP[j] = 255;
  }
  for (var i = 0; i < ALPHABET.length; i++) {
    var x = ALPHABET.charAt(i);
    var xc = x.charCodeAt(0);
    if (BASE_MAP[xc] !== 255) { throw new TypeError(x + ' is ambiguous') }
    BASE_MAP[xc] = i;
  }
  var BASE = ALPHABET.length;
  var LEADER = ALPHABET.charAt(0);
  var FACTOR = Math.log(BASE) / Math.log(256); // log(BASE) / log(256), rounded up
  var iFACTOR = Math.log(256) / Math.log(BASE); // log(256) / log(BASE), rounded up
  function encode (source) {
    if (source instanceof Uint8Array) ; else if (ArrayBuffer.isView(source)) {
      source = new Uint8Array(source.buffer, source.byteOffset, source.byteLength);
    } else if (Array.isArray(source)) {
      source = Uint8Array.from(source);
    }
    if (!(source instanceof Uint8Array)) { throw new TypeError('Expected Uint8Array') }
    if (source.length === 0) { return '' }
        // Skip & count leading zeroes.
    var zeroes = 0;
    var length = 0;
    var pbegin = 0;
    var pend = source.length;
    while (pbegin !== pend && source[pbegin] === 0) {
      pbegin++;
      zeroes++;
    }
        // Allocate enough space in big-endian base58 representation.
    var size = ((pend - pbegin) * iFACTOR + 1) >>> 0;
    var b58 = new Uint8Array(size);
        // Process the bytes.
    while (pbegin !== pend) {
      var carry = source[pbegin];
            // Apply "b58 = b58 * 256 + ch".
      var i = 0;
      for (var it1 = size - 1; (carry !== 0 || i < length) && (it1 !== -1); it1--, i++) {
        carry += (256 * b58[it1]) >>> 0;
        b58[it1] = (carry % BASE) >>> 0;
        carry = (carry / BASE) >>> 0;
      }
      if (carry !== 0) { throw new Error('Non-zero carry') }
      length = i;
      pbegin++;
    }
        // Skip leading zeroes in base58 result.
    var it2 = size - length;
    while (it2 !== size && b58[it2] === 0) {
      it2++;
    }
        // Translate the result into a string.
    var str = LEADER.repeat(zeroes);
    for (; it2 < size; ++it2) { str += ALPHABET.charAt(b58[it2]); }
    return str
  }
  function decodeUnsafe (source) {
    if (typeof source !== 'string') { throw new TypeError('Expected String') }
    if (source.length === 0) { return new Uint8Array() }
    var psz = 0;
        // Skip leading spaces.
    if (source[psz] === ' ') { return }
        // Skip and count leading '1's.
    var zeroes = 0;
    var length = 0;
    while (source[psz] === LEADER) {
      zeroes++;
      psz++;
    }
        // Allocate enough space in big-endian base256 representation.
    var size = (((source.length - psz) * FACTOR) + 1) >>> 0; // log(58) / log(256), rounded up.
    var b256 = new Uint8Array(size);
        // Process the characters.
    while (source[psz]) {
            // Decode character
      var carry = BASE_MAP[source.charCodeAt(psz)];
            // Invalid character
      if (carry === 255) { return }
      var i = 0;
      for (var it3 = size - 1; (carry !== 0 || i < length) && (it3 !== -1); it3--, i++) {
        carry += (BASE * b256[it3]) >>> 0;
        b256[it3] = (carry % 256) >>> 0;
        carry = (carry / 256) >>> 0;
      }
      if (carry !== 0) { throw new Error('Non-zero carry') }
      length = i;
      psz++;
    }
        // Skip trailing spaces.
    if (source[psz] === ' ') { return }
        // Skip leading zeroes in b256.
    var it4 = size - length;
    while (it4 !== size && b256[it4] === 0) {
      it4++;
    }
    var vch = new Uint8Array(zeroes + (size - it4));
    var j = zeroes;
    while (it4 !== size) {
      vch[j++] = b256[it4++];
    }
    return vch
  }
  function decode (string) {
    var buffer = decodeUnsafe(string);
    if (buffer) { return buffer }
    throw new Error('Non-base' + BASE + ' character')
  }
  return {
    encode: encode,
    decodeUnsafe: decodeUnsafe,
    decode: decode
  }
}
var src$1 = base$1;

const textDecoder = new TextDecoder();
/**
 * @param {ArrayBufferView|ArrayBuffer} bytes
 * @returns {string}
 */
const decodeText$1 = (bytes) => textDecoder.decode(bytes);

const textEncoder = new TextEncoder();
/**
 * @param {string} text
 * @returns {Uint8Array}
 */
const encodeText$2 = (text) => textEncoder.encode(text);

/**
 * Returns a new Uint8Array created by concatenating the passed Arrays
 *
 * @param {Array<ArrayLike<number>>} arrs
 * @param {number} length
 * @returns {Uint8Array}
 */
function concat (arrs, length) {
  const output = new Uint8Array(length);
  let offset = 0;

  for (const arr of arrs) {
    output.set(arr, offset);
    offset += arr.length;
  }

  return output
}

var util = { decodeText: decodeText$1, encodeText: encodeText$2, concat };

const { encodeText: encodeText$1 } = util;

/** @typedef {import('./types').CodecFactory} CodecFactory */
/** @typedef {import("./types").BaseName} BaseName */
/** @typedef {import("./types").BaseCode} BaseCode */

/**
 * Class to encode/decode in the supported Bases
 *
 */
class Base {
  /**
   * @param {BaseName} name
   * @param {BaseCode} code
   * @param {CodecFactory} factory
   * @param {string} alphabet
   */
  constructor (name, code, factory, alphabet) {
    this.name = name;
    this.code = code;
    this.codeBuf = encodeText$1(this.code);
    this.alphabet = alphabet;
    this.codec = factory(alphabet);
  }

  /**
   * @param {Uint8Array} buf
   * @returns {string}
   */
  encode (buf) {
    return this.codec.encode(buf)
  }

  /**
   * @param {string} string
   * @returns {Uint8Array}
   */
  decode (string) {
    for (const char of string) {
      if (this.alphabet && this.alphabet.indexOf(char) < 0) {
        throw new Error(`invalid character '${char}' in '${string}'`)
      }
    }
    return this.codec.decode(string)
  }
}

var base = Base;

/** @typedef {import('./types').CodecFactory} CodecFactory */

/**
 * @param {string} string
 * @param {string} alphabet
 * @param {number} bitsPerChar
 * @returns {Uint8Array}
 */
const decode = (string, alphabet, bitsPerChar) => {
  // Build the character lookup table:
  /** @type {Record<string, number>} */
  const codes = {};
  for (let i = 0; i < alphabet.length; ++i) {
    codes[alphabet[i]] = i;
  }

  // Count the padding bytes:
  let end = string.length;
  while (string[end - 1] === '=') {
    --end;
  }

  // Allocate the output:
  const out = new Uint8Array((end * bitsPerChar / 8) | 0);

  // Parse the data:
  let bits = 0; // Number of bits currently in the buffer
  let buffer = 0; // Bits waiting to be written out, MSB first
  let written = 0; // Next byte to write
  for (let i = 0; i < end; ++i) {
    // Read one character from the string:
    const value = codes[string[i]];
    if (value === undefined) {
      throw new SyntaxError('Invalid character ' + string[i])
    }

    // Append the bits to the buffer:
    buffer = (buffer << bitsPerChar) | value;
    bits += bitsPerChar;

    // Write out some bits if the buffer has a byte's worth:
    if (bits >= 8) {
      bits -= 8;
      out[written++] = 0xff & (buffer >> bits);
    }
  }

  // Verify that we have received just enough bits:
  if (bits >= bitsPerChar || 0xff & (buffer << (8 - bits))) {
    throw new SyntaxError('Unexpected end of data')
  }

  return out
};

/**
 * @param {Uint8Array} data
 * @param {string} alphabet
 * @param {number} bitsPerChar
 * @returns {string}
 */
const encode = (data, alphabet, bitsPerChar) => {
  const pad = alphabet[alphabet.length - 1] === '=';
  const mask = (1 << bitsPerChar) - 1;
  let out = '';

  let bits = 0; // Number of bits currently in the buffer
  let buffer = 0; // Bits waiting to be written out, MSB first
  for (let i = 0; i < data.length; ++i) {
    // Slurp data into the buffer:
    buffer = (buffer << 8) | data[i];
    bits += 8;

    // Write out as much as we can:
    while (bits > bitsPerChar) {
      bits -= bitsPerChar;
      out += alphabet[mask & (buffer >> bits)];
    }
  }

  // Partial character:
  if (bits) {
    out += alphabet[mask & (buffer << (bitsPerChar - bits))];
  }

  // Add padding characters until we hit a byte boundary:
  if (pad) {
    while ((out.length * bitsPerChar) & 7) {
      out += '=';
    }
  }

  return out
};

/**
 * RFC4648 Factory
 *
 * @param {number} bitsPerChar
 * @returns {CodecFactory}
 */
const rfc4648$1 = (bitsPerChar) => (alphabet) => {
  return {
    /**
     * @param {Uint8Array} input
     * @returns {string}
     */
    encode (input) {
      return encode(input, alphabet, bitsPerChar)
    },
    /**
     * @param {string} input
     * @returns {Uint8Array}
     */
    decode (input) {
      return decode(input, alphabet, bitsPerChar)
    }
  }
};

var rfc4648_1 = { rfc4648: rfc4648$1 };

const { rfc4648 } = rfc4648_1;
const { decodeText, encodeText } = util;

/** @typedef {import('./types').CodecFactory} CodecFactory */
/** @typedef {import('./types').Codec} Codec */
/** @typedef {import('./types').BaseName} BaseName */
/** @typedef {import('./types').BaseCode} BaseCode */

/** @type {CodecFactory} */
const identity = () => {
  return {
    encode: decodeText,
    decode: encodeText
  }
};

/**
 *
 * name, code, implementation, alphabet
 *
 * @type {Array<[BaseName, BaseCode, CodecFactory, string]>}
 */
const constants = [
  ['identity', '\x00', identity, ''],
  ['base2', '0', rfc4648(1), '01'],
  ['base8', '7', rfc4648(3), '01234567'],
  ['base10', '9', src$1, '0123456789'],
  ['base16', 'f', rfc4648(4), '0123456789abcdef'],
  ['base16upper', 'F', rfc4648(4), '0123456789ABCDEF'],
  ['base32hex', 'v', rfc4648(5), '0123456789abcdefghijklmnopqrstuv'],
  ['base32hexupper', 'V', rfc4648(5), '0123456789ABCDEFGHIJKLMNOPQRSTUV'],
  ['base32hexpad', 't', rfc4648(5), '0123456789abcdefghijklmnopqrstuv='],
  ['base32hexpadupper', 'T', rfc4648(5), '0123456789ABCDEFGHIJKLMNOPQRSTUV='],
  ['base32', 'b', rfc4648(5), 'abcdefghijklmnopqrstuvwxyz234567'],
  ['base32upper', 'B', rfc4648(5), 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'],
  ['base32pad', 'c', rfc4648(5), 'abcdefghijklmnopqrstuvwxyz234567='],
  ['base32padupper', 'C', rfc4648(5), 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567='],
  ['base32z', 'h', rfc4648(5), 'ybndrfg8ejkmcpqxot1uwisza345h769'],
  ['base36', 'k', src$1, '0123456789abcdefghijklmnopqrstuvwxyz'],
  ['base36upper', 'K', src$1, '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'],
  ['base58btc', 'z', src$1, '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'],
  ['base58flickr', 'Z', src$1, '123456789abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ'],
  ['base64', 'm', rfc4648(6), 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'],
  ['base64pad', 'M', rfc4648(6), 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/='],
  ['base64url', 'u', rfc4648(6), 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_'],
  ['base64urlpad', 'U', rfc4648(6), 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_=']
];

/** @type {Record<BaseName,Base>} */
const names = constants.reduce((prev, tupple) => {
  prev[tupple[0]] = new base(tupple[0], tupple[1], tupple[2], tupple[3]);
  return prev
}, /** @type {Record<BaseName,Base>} */({}));

/** @type {Record<BaseCode,Base>} */
const codes = constants.reduce((prev, tupple) => {
  prev[tupple[1]] = names[tupple[0]];
  return prev
}, /** @type {Record<BaseCode,Base>} */({}));

var constants_1 = {
  names,
  codes
};

/**
 * Implementation of the [multibase](https://github.com/multiformats/multibase) specification.
 *
 */

var src = createCommonjsModule(function (module, exports) {


const { encodeText, decodeText, concat } = util;

/** @typedef {import('./base')} Base */
/** @typedef {import("./types").BaseNameOrCode} BaseNameOrCode */
/** @typedef {import("./types").BaseCode} BaseCode */
/** @typedef {import("./types").BaseName} BaseName */

/**
 * Create a new Uint8Array with the multibase varint+code.
 *
 * @param {BaseNameOrCode} nameOrCode - The multibase name or code number.
 * @param {Uint8Array} buf - The data to be prefixed with multibase.
 * @returns {Uint8Array}
 * @throws {Error} Will throw if the encoding is not supported
 */
function multibase (nameOrCode, buf) {
  if (!buf) {
    throw new Error('requires an encoded Uint8Array')
  }
  const { name, codeBuf } = encoding(nameOrCode);
  validEncode(name, buf);

  return concat([codeBuf, buf], codeBuf.length + buf.length)
}

/**
 * Encode data with the specified base and add the multibase prefix.
 *
 * @param {BaseNameOrCode} nameOrCode - The multibase name or code number.
 * @param {Uint8Array} buf - The data to be encoded.
 * @returns {Uint8Array}
 * @throws {Error} Will throw if the encoding is not supported
 *
 */
function encode (nameOrCode, buf) {
  const enc = encoding(nameOrCode);
  const data = encodeText(enc.encode(buf));

  return concat([enc.codeBuf, data], enc.codeBuf.length + data.length)
}

/**
 * Takes a Uint8Array or string encoded with multibase header, decodes it and
 * returns the decoded buffer
 *
 * @param {Uint8Array|string} data
 * @returns {Uint8Array}
 * @throws {Error} Will throw if the encoding is not supported
 *
 */
function decode (data) {
  if (data instanceof Uint8Array) {
    data = decodeText(data);
  }
  const prefix = data[0];

  // Make all encodings case-insensitive except the ones that include upper and lower chars in the alphabet
  if (['f', 'F', 'v', 'V', 't', 'T', 'b', 'B', 'c', 'C', 'h', 'k', 'K'].includes(prefix)) {
    data = data.toLowerCase();
  }
  const enc = encoding(/** @type {BaseCode} */(data[0]));
  return enc.decode(data.substring(1))
}

/**
 * Is the given data multibase encoded?
 *
 * @param {Uint8Array|string} data
 */
function isEncoded (data) {
  if (data instanceof Uint8Array) {
    data = decodeText(data);
  }

  // Ensure bufOrString is a string
  if (Object.prototype.toString.call(data) !== '[object String]') {
    return false
  }

  try {
    const enc = encoding(/** @type {BaseCode} */(data[0]));
    return enc.name
  } catch (err) {
    return false
  }
}

/**
 * Validate encoded data
 *
 * @param {BaseNameOrCode} name
 * @param {Uint8Array} buf
 * @returns {void}
 * @throws {Error} Will throw if the encoding is not supported
 */
function validEncode (name, buf) {
  const enc = encoding(name);
  enc.decode(decodeText(buf));
}

/**
 * Get the encoding by name or code
 *
 * @param {BaseNameOrCode} nameOrCode
 * @returns {Base}
 * @throws {Error} Will throw if the encoding is not supported
 */
function encoding (nameOrCode) {
  if (Object.prototype.hasOwnProperty.call(constants_1.names, /** @type {BaseName} */(nameOrCode))) {
    return constants_1.names[/** @type {BaseName} */(nameOrCode)]
  } else if (Object.prototype.hasOwnProperty.call(constants_1.codes, /** @type {BaseCode} */(nameOrCode))) {
    return constants_1.codes[/** @type {BaseCode} */(nameOrCode)]
  } else {
    throw new Error(`Unsupported encoding: ${nameOrCode}`)
  }
}

/**
 * Get encoding from data
 *
 * @param {string|Uint8Array} data
 * @returns {Base}
 * @throws {Error} Will throw if the encoding is not supported
 */
function encodingFromData (data) {
  if (data instanceof Uint8Array) {
    data = decodeText(data);
  }

  return encoding(/** @type {BaseCode} */(data[0]))
}

exports = module.exports = multibase;
exports.encode = encode;
exports.decode = decode;
exports.isEncoded = isEncoded;
exports.encoding = encoding;
exports.encodingFromData = encodingFromData;
const names = Object.freeze(constants_1.names);
const codes = Object.freeze(constants_1.codes);
exports.names = names;
exports.codes = codes;
});

// SHA2-256 need to try 2^128 hashes to execute birthday attack.
// BTC network is doing 2^67 hashes/sec as per early 2023.
// Round constants:
// first 32 bits of the fractional parts of the cube roots of the first 64 primes 2..311)
// prettier-ignore
const SHA256_K = /* @__PURE__ */ new Uint32Array([
    0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
    0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3, 0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
    0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
    0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
    0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13, 0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
    0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
    0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
    0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208, 0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
]);
// Initial state:
// first 32 bits of the fractional parts of the square roots of the first 8 primes 2..19
// prettier-ignore
const SHA256_IV = /* @__PURE__ */ new Uint32Array([
    0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a, 0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19
]);
// Temporary buffer, not used to store anything between runs
// Named this way because it matches specification.
const SHA256_W = /* @__PURE__ */ new Uint32Array(64);
class SHA256 extends HashMD {
    constructor() {
        super(64, 32, 8, false);
        // We cannot use array here since array allows indexing by variable
        // which means optimizer/compiler cannot use registers.
        this.A = SHA256_IV[0] | 0;
        this.B = SHA256_IV[1] | 0;
        this.C = SHA256_IV[2] | 0;
        this.D = SHA256_IV[3] | 0;
        this.E = SHA256_IV[4] | 0;
        this.F = SHA256_IV[5] | 0;
        this.G = SHA256_IV[6] | 0;
        this.H = SHA256_IV[7] | 0;
    }
    get() {
        const { A, B, C, D, E, F, G, H } = this;
        return [A, B, C, D, E, F, G, H];
    }
    // prettier-ignore
    set(A, B, C, D, E, F, G, H) {
        this.A = A | 0;
        this.B = B | 0;
        this.C = C | 0;
        this.D = D | 0;
        this.E = E | 0;
        this.F = F | 0;
        this.G = G | 0;
        this.H = H | 0;
    }
    process(view, offset) {
        // Extend the first 16 words into the remaining 48 words w[16..63] of the message schedule array
        for (let i = 0; i < 16; i++, offset += 4)
            SHA256_W[i] = view.getUint32(offset, false);
        for (let i = 16; i < 64; i++) {
            const W15 = SHA256_W[i - 15];
            const W2 = SHA256_W[i - 2];
            const s0 = rotr(W15, 7) ^ rotr(W15, 18) ^ (W15 >>> 3);
            const s1 = rotr(W2, 17) ^ rotr(W2, 19) ^ (W2 >>> 10);
            SHA256_W[i] = (s1 + SHA256_W[i - 7] + s0 + SHA256_W[i - 16]) | 0;
        }
        // Compression function main loop, 64 rounds
        let { A, B, C, D, E, F, G, H } = this;
        for (let i = 0; i < 64; i++) {
            const sigma1 = rotr(E, 6) ^ rotr(E, 11) ^ rotr(E, 25);
            const T1 = (H + sigma1 + Chi(E, F, G) + SHA256_K[i] + SHA256_W[i]) | 0;
            const sigma0 = rotr(A, 2) ^ rotr(A, 13) ^ rotr(A, 22);
            const T2 = (sigma0 + Maj(A, B, C)) | 0;
            H = G;
            G = F;
            F = E;
            E = (D + T1) | 0;
            D = C;
            C = B;
            B = A;
            A = (T1 + T2) | 0;
        }
        // Add the compressed chunk to the current hash value
        A = (A + this.A) | 0;
        B = (B + this.B) | 0;
        C = (C + this.C) | 0;
        D = (D + this.D) | 0;
        E = (E + this.E) | 0;
        F = (F + this.F) | 0;
        G = (G + this.G) | 0;
        H = (H + this.H) | 0;
        this.set(A, B, C, D, E, F, G, H);
    }
    roundClean() {
        SHA256_W.fill(0);
    }
    destroy() {
        this.set(0, 0, 0, 0, 0, 0, 0, 0);
        this.buffer.fill(0);
    }
}
/**
 * SHA2-256 hash function
 * @param message - data that would be hashed
 */
const sha256$2 = /* @__PURE__ */ wrapConstructor(() => new SHA256());

// HMAC (RFC 2104)
class HMAC extends Hash {
    constructor(hash$1, _key) {
        super();
        this.finished = false;
        this.destroyed = false;
        hash(hash$1);
        const key = toBytes(_key);
        this.iHash = hash$1.create();
        if (typeof this.iHash.update !== 'function')
            throw new Error('Expected instance of class which extends utils.Hash');
        this.blockLen = this.iHash.blockLen;
        this.outputLen = this.iHash.outputLen;
        const blockLen = this.blockLen;
        const pad = new Uint8Array(blockLen);
        // blockLen can be bigger than outputLen
        pad.set(key.length > blockLen ? hash$1.create().update(key).digest() : key);
        for (let i = 0; i < pad.length; i++)
            pad[i] ^= 0x36;
        this.iHash.update(pad);
        // By doing update (processing of first block) of outer hash here we can re-use it between multiple calls via clone
        this.oHash = hash$1.create();
        // Undo internal XOR && apply outer XOR
        for (let i = 0; i < pad.length; i++)
            pad[i] ^= 0x36 ^ 0x5c;
        this.oHash.update(pad);
        pad.fill(0);
    }
    update(buf) {
        exists(this);
        this.iHash.update(buf);
        return this;
    }
    digestInto(out) {
        exists(this);
        bytes(out, this.outputLen);
        this.finished = true;
        this.iHash.digestInto(out);
        this.oHash.update(out);
        this.oHash.digestInto(out);
        this.destroy();
    }
    digest() {
        const out = new Uint8Array(this.oHash.outputLen);
        this.digestInto(out);
        return out;
    }
    _cloneInto(to) {
        // Create new instance without calling constructor since key already in state and we don't know it.
        to || (to = Object.create(Object.getPrototypeOf(this), {}));
        const { oHash, iHash, finished, destroyed, blockLen, outputLen } = this;
        to = to;
        to.finished = finished;
        to.destroyed = destroyed;
        to.blockLen = blockLen;
        to.outputLen = outputLen;
        to.oHash = oHash._cloneInto(to.oHash);
        to.iHash = iHash._cloneInto(to.iHash);
        return to;
    }
    destroy() {
        this.destroyed = true;
        this.oHash.destroy();
        this.iHash.destroy();
    }
}
/**
 * HMAC: RFC2104 message authentication code.
 * @param hash - function that would be used e.g. sha256
 * @param key - message key
 * @param message - message data
 */
const hmac = (hash, key, message) => new HMAC(hash, key).update(message).digest();
hmac.create = (hash, key) => new HMAC(hash, key);

/*! noble-curves - MIT License (c) 2022 Paul Miller (paulmillr.com) */
function validatePointOpts(curve) {
    const opts = validateBasic(curve);
    validateObject(opts, {
        a: 'field',
        b: 'field',
    }, {
        allowedPrivateKeyLengths: 'array',
        wrapPrivateKey: 'boolean',
        isTorsionFree: 'function',
        clearCofactor: 'function',
        allowInfinityPoint: 'boolean',
        fromBytes: 'function',
        toBytes: 'function',
    });
    const { endo, Fp, a } = opts;
    if (endo) {
        if (!Fp.eql(a, Fp.ZERO)) {
            throw new Error('Endomorphism can only be defined for Koblitz curves that have a=0');
        }
        if (typeof endo !== 'object' ||
            typeof endo.beta !== 'bigint' ||
            typeof endo.splitScalar !== 'function') {
            throw new Error('Expected endomorphism with beta: bigint and splitScalar: function');
        }
    }
    return Object.freeze({ ...opts });
}
// ASN.1 DER encoding utilities
const { bytesToNumberBE: b2n, hexToBytes: h2b } = ut;
const DER = {
    // asn.1 DER encoding utils
    Err: class DERErr extends Error {
        constructor(m = '') {
            super(m);
        }
    },
    _parseInt(data) {
        const { Err: E } = DER;
        if (data.length < 2 || data[0] !== 0x02)
            throw new E('Invalid signature integer tag');
        const len = data[1];
        const res = data.subarray(2, len + 2);
        if (!len || res.length !== len)
            throw new E('Invalid signature integer: wrong length');
        // https://crypto.stackexchange.com/a/57734 Leftmost bit of first byte is 'negative' flag,
        // since we always use positive integers here. It must always be empty:
        // - add zero byte if exists
        // - if next byte doesn't have a flag, leading zero is not allowed (minimal encoding)
        if (res[0] & 0b10000000)
            throw new E('Invalid signature integer: negative');
        if (res[0] === 0x00 && !(res[1] & 0b10000000))
            throw new E('Invalid signature integer: unnecessary leading zero');
        return { d: b2n(res), l: data.subarray(len + 2) }; // d is data, l is left
    },
    toSig(hex) {
        // parse DER signature
        const { Err: E } = DER;
        const data = typeof hex === 'string' ? h2b(hex) : hex;
        abytes(data);
        let l = data.length;
        if (l < 2 || data[0] != 0x30)
            throw new E('Invalid signature tag');
        if (data[1] !== l - 2)
            throw new E('Invalid signature: incorrect length');
        const { d: r, l: sBytes } = DER._parseInt(data.subarray(2));
        const { d: s, l: rBytesLeft } = DER._parseInt(sBytes);
        if (rBytesLeft.length)
            throw new E('Invalid signature: left bytes after parsing');
        return { r, s };
    },
    hexFromSig(sig) {
        // Add leading zero if first byte has negative bit enabled. More details in '_parseInt'
        const slice = (s) => (Number.parseInt(s[0], 16) & 0b1000 ? '00' + s : s);
        const h = (num) => {
            const hex = num.toString(16);
            return hex.length & 1 ? `0${hex}` : hex;
        };
        const s = slice(h(sig.s));
        const r = slice(h(sig.r));
        const shl = s.length / 2;
        const rhl = r.length / 2;
        const sl = h(shl);
        const rl = h(rhl);
        return `30${h(rhl + shl + 4)}02${rl}${r}02${sl}${s}`;
    },
};
// Be friendly to bad ECMAScript parsers by not using bigint literals
// prettier-ignore
const _0n$1 = BigInt(0), _1n$2 = BigInt(1); BigInt(2); const _3n = BigInt(3); BigInt(4);
function weierstrassPoints(opts) {
    const CURVE = validatePointOpts(opts);
    const { Fp } = CURVE; // All curves has same field / group length as for now, but they can differ
    const toBytes = CURVE.toBytes ||
        ((_c, point, _isCompressed) => {
            const a = point.toAffine();
            return concatBytes(Uint8Array.from([0x04]), Fp.toBytes(a.x), Fp.toBytes(a.y));
        });
    const fromBytes = CURVE.fromBytes ||
        ((bytes) => {
            // const head = bytes[0];
            const tail = bytes.subarray(1);
            // if (head !== 0x04) throw new Error('Only non-compressed encoding is supported');
            const x = Fp.fromBytes(tail.subarray(0, Fp.BYTES));
            const y = Fp.fromBytes(tail.subarray(Fp.BYTES, 2 * Fp.BYTES));
            return { x, y };
        });
    /**
     * y² = x³ + ax + b: Short weierstrass curve formula
     * @returns y²
     */
    function weierstrassEquation(x) {
        const { a, b } = CURVE;
        const x2 = Fp.sqr(x); // x * x
        const x3 = Fp.mul(x2, x); // x2 * x
        return Fp.add(Fp.add(x3, Fp.mul(x, a)), b); // x3 + a * x + b
    }
    // Validate whether the passed curve params are valid.
    // We check if curve equation works for generator point.
    // `assertValidity()` won't work: `isTorsionFree()` is not available at this point in bls12-381.
    // ProjectivePoint class has not been initialized yet.
    if (!Fp.eql(Fp.sqr(CURVE.Gy), weierstrassEquation(CURVE.Gx)))
        throw new Error('bad generator point: equation left != right');
    // Valid group elements reside in range 1..n-1
    function isWithinCurveOrder(num) {
        return typeof num === 'bigint' && _0n$1 < num && num < CURVE.n;
    }
    function assertGE(num) {
        if (!isWithinCurveOrder(num))
            throw new Error('Expected valid bigint: 0 < bigint < curve.n');
    }
    // Validates if priv key is valid and converts it to bigint.
    // Supports options allowedPrivateKeyLengths and wrapPrivateKey.
    function normPrivateKeyToScalar(key) {
        const { allowedPrivateKeyLengths: lengths, nByteLength, wrapPrivateKey, n } = CURVE;
        if (lengths && typeof key !== 'bigint') {
            if (isBytes$1(key))
                key = bytesToHex$1(key);
            // Normalize to hex string, pad. E.g. P521 would norm 130-132 char hex to 132-char bytes
            if (typeof key !== 'string' || !lengths.includes(key.length))
                throw new Error('Invalid key');
            key = key.padStart(nByteLength * 2, '0');
        }
        let num;
        try {
            num =
                typeof key === 'bigint'
                    ? key
                    : bytesToNumberBE(ensureBytes('private key', key, nByteLength));
        }
        catch (error) {
            throw new Error(`private key must be ${nByteLength} bytes, hex or bigint, not ${typeof key}`);
        }
        if (wrapPrivateKey)
            num = mod(num, n); // disabled by default, enabled for BLS
        assertGE(num); // num in range [1..N-1]
        return num;
    }
    const pointPrecomputes = new Map();
    function assertPrjPoint(other) {
        if (!(other instanceof Point))
            throw new Error('ProjectivePoint expected');
    }
    /**
     * Projective Point works in 3d / projective (homogeneous) coordinates: (x, y, z) ∋ (x=x/z, y=y/z)
     * Default Point works in 2d / affine coordinates: (x, y)
     * We're doing calculations in projective, because its operations don't require costly inversion.
     */
    class Point {
        constructor(px, py, pz) {
            this.px = px;
            this.py = py;
            this.pz = pz;
            if (px == null || !Fp.isValid(px))
                throw new Error('x required');
            if (py == null || !Fp.isValid(py))
                throw new Error('y required');
            if (pz == null || !Fp.isValid(pz))
                throw new Error('z required');
        }
        // Does not validate if the point is on-curve.
        // Use fromHex instead, or call assertValidity() later.
        static fromAffine(p) {
            const { x, y } = p || {};
            if (!p || !Fp.isValid(x) || !Fp.isValid(y))
                throw new Error('invalid affine point');
            if (p instanceof Point)
                throw new Error('projective point not allowed');
            const is0 = (i) => Fp.eql(i, Fp.ZERO);
            // fromAffine(x:0, y:0) would produce (x:0, y:0, z:1), but we need (x:0, y:1, z:0)
            if (is0(x) && is0(y))
                return Point.ZERO;
            return new Point(x, y, Fp.ONE);
        }
        get x() {
            return this.toAffine().x;
        }
        get y() {
            return this.toAffine().y;
        }
        /**
         * Takes a bunch of Projective Points but executes only one
         * inversion on all of them. Inversion is very slow operation,
         * so this improves performance massively.
         * Optimization: converts a list of projective points to a list of identical points with Z=1.
         */
        static normalizeZ(points) {
            const toInv = Fp.invertBatch(points.map((p) => p.pz));
            return points.map((p, i) => p.toAffine(toInv[i])).map(Point.fromAffine);
        }
        /**
         * Converts hash string or Uint8Array to Point.
         * @param hex short/long ECDSA hex
         */
        static fromHex(hex) {
            const P = Point.fromAffine(fromBytes(ensureBytes('pointHex', hex)));
            P.assertValidity();
            return P;
        }
        // Multiplies generator point by privateKey.
        static fromPrivateKey(privateKey) {
            return Point.BASE.multiply(normPrivateKeyToScalar(privateKey));
        }
        // "Private method", don't use it directly
        _setWindowSize(windowSize) {
            this._WINDOW_SIZE = windowSize;
            pointPrecomputes.delete(this);
        }
        // A point on curve is valid if it conforms to equation.
        assertValidity() {
            if (this.is0()) {
                // (0, 1, 0) aka ZERO is invalid in most contexts.
                // In BLS, ZERO can be serialized, so we allow it.
                // (0, 0, 0) is wrong representation of ZERO and is always invalid.
                if (CURVE.allowInfinityPoint && !Fp.is0(this.py))
                    return;
                throw new Error('bad point: ZERO');
            }
            // Some 3rd-party test vectors require different wording between here & `fromCompressedHex`
            const { x, y } = this.toAffine();
            // Check if x, y are valid field elements
            if (!Fp.isValid(x) || !Fp.isValid(y))
                throw new Error('bad point: x or y not FE');
            const left = Fp.sqr(y); // y²
            const right = weierstrassEquation(x); // x³ + ax + b
            if (!Fp.eql(left, right))
                throw new Error('bad point: equation left != right');
            if (!this.isTorsionFree())
                throw new Error('bad point: not in prime-order subgroup');
        }
        hasEvenY() {
            const { y } = this.toAffine();
            if (Fp.isOdd)
                return !Fp.isOdd(y);
            throw new Error("Field doesn't support isOdd");
        }
        /**
         * Compare one point to another.
         */
        equals(other) {
            assertPrjPoint(other);
            const { px: X1, py: Y1, pz: Z1 } = this;
            const { px: X2, py: Y2, pz: Z2 } = other;
            const U1 = Fp.eql(Fp.mul(X1, Z2), Fp.mul(X2, Z1));
            const U2 = Fp.eql(Fp.mul(Y1, Z2), Fp.mul(Y2, Z1));
            return U1 && U2;
        }
        /**
         * Flips point to one corresponding to (x, -y) in Affine coordinates.
         */
        negate() {
            return new Point(this.px, Fp.neg(this.py), this.pz);
        }
        // Renes-Costello-Batina exception-free doubling formula.
        // There is 30% faster Jacobian formula, but it is not complete.
        // https://eprint.iacr.org/2015/1060, algorithm 3
        // Cost: 8M + 3S + 3*a + 2*b3 + 15add.
        double() {
            const { a, b } = CURVE;
            const b3 = Fp.mul(b, _3n);
            const { px: X1, py: Y1, pz: Z1 } = this;
            let X3 = Fp.ZERO, Y3 = Fp.ZERO, Z3 = Fp.ZERO; // prettier-ignore
            let t0 = Fp.mul(X1, X1); // step 1
            let t1 = Fp.mul(Y1, Y1);
            let t2 = Fp.mul(Z1, Z1);
            let t3 = Fp.mul(X1, Y1);
            t3 = Fp.add(t3, t3); // step 5
            Z3 = Fp.mul(X1, Z1);
            Z3 = Fp.add(Z3, Z3);
            X3 = Fp.mul(a, Z3);
            Y3 = Fp.mul(b3, t2);
            Y3 = Fp.add(X3, Y3); // step 10
            X3 = Fp.sub(t1, Y3);
            Y3 = Fp.add(t1, Y3);
            Y3 = Fp.mul(X3, Y3);
            X3 = Fp.mul(t3, X3);
            Z3 = Fp.mul(b3, Z3); // step 15
            t2 = Fp.mul(a, t2);
            t3 = Fp.sub(t0, t2);
            t3 = Fp.mul(a, t3);
            t3 = Fp.add(t3, Z3);
            Z3 = Fp.add(t0, t0); // step 20
            t0 = Fp.add(Z3, t0);
            t0 = Fp.add(t0, t2);
            t0 = Fp.mul(t0, t3);
            Y3 = Fp.add(Y3, t0);
            t2 = Fp.mul(Y1, Z1); // step 25
            t2 = Fp.add(t2, t2);
            t0 = Fp.mul(t2, t3);
            X3 = Fp.sub(X3, t0);
            Z3 = Fp.mul(t2, t1);
            Z3 = Fp.add(Z3, Z3); // step 30
            Z3 = Fp.add(Z3, Z3);
            return new Point(X3, Y3, Z3);
        }
        // Renes-Costello-Batina exception-free addition formula.
        // There is 30% faster Jacobian formula, but it is not complete.
        // https://eprint.iacr.org/2015/1060, algorithm 1
        // Cost: 12M + 0S + 3*a + 3*b3 + 23add.
        add(other) {
            assertPrjPoint(other);
            const { px: X1, py: Y1, pz: Z1 } = this;
            const { px: X2, py: Y2, pz: Z2 } = other;
            let X3 = Fp.ZERO, Y3 = Fp.ZERO, Z3 = Fp.ZERO; // prettier-ignore
            const a = CURVE.a;
            const b3 = Fp.mul(CURVE.b, _3n);
            let t0 = Fp.mul(X1, X2); // step 1
            let t1 = Fp.mul(Y1, Y2);
            let t2 = Fp.mul(Z1, Z2);
            let t3 = Fp.add(X1, Y1);
            let t4 = Fp.add(X2, Y2); // step 5
            t3 = Fp.mul(t3, t4);
            t4 = Fp.add(t0, t1);
            t3 = Fp.sub(t3, t4);
            t4 = Fp.add(X1, Z1);
            let t5 = Fp.add(X2, Z2); // step 10
            t4 = Fp.mul(t4, t5);
            t5 = Fp.add(t0, t2);
            t4 = Fp.sub(t4, t5);
            t5 = Fp.add(Y1, Z1);
            X3 = Fp.add(Y2, Z2); // step 15
            t5 = Fp.mul(t5, X3);
            X3 = Fp.add(t1, t2);
            t5 = Fp.sub(t5, X3);
            Z3 = Fp.mul(a, t4);
            X3 = Fp.mul(b3, t2); // step 20
            Z3 = Fp.add(X3, Z3);
            X3 = Fp.sub(t1, Z3);
            Z3 = Fp.add(t1, Z3);
            Y3 = Fp.mul(X3, Z3);
            t1 = Fp.add(t0, t0); // step 25
            t1 = Fp.add(t1, t0);
            t2 = Fp.mul(a, t2);
            t4 = Fp.mul(b3, t4);
            t1 = Fp.add(t1, t2);
            t2 = Fp.sub(t0, t2); // step 30
            t2 = Fp.mul(a, t2);
            t4 = Fp.add(t4, t2);
            t0 = Fp.mul(t1, t4);
            Y3 = Fp.add(Y3, t0);
            t0 = Fp.mul(t5, t4); // step 35
            X3 = Fp.mul(t3, X3);
            X3 = Fp.sub(X3, t0);
            t0 = Fp.mul(t3, t1);
            Z3 = Fp.mul(t5, Z3);
            Z3 = Fp.add(Z3, t0); // step 40
            return new Point(X3, Y3, Z3);
        }
        subtract(other) {
            return this.add(other.negate());
        }
        is0() {
            return this.equals(Point.ZERO);
        }
        wNAF(n) {
            return wnaf.wNAFCached(this, pointPrecomputes, n, (comp) => {
                const toInv = Fp.invertBatch(comp.map((p) => p.pz));
                return comp.map((p, i) => p.toAffine(toInv[i])).map(Point.fromAffine);
            });
        }
        /**
         * Non-constant-time multiplication. Uses double-and-add algorithm.
         * It's faster, but should only be used when you don't care about
         * an exposed private key e.g. sig verification, which works over *public* keys.
         */
        multiplyUnsafe(n) {
            const I = Point.ZERO;
            if (n === _0n$1)
                return I;
            assertGE(n); // Will throw on 0
            if (n === _1n$2)
                return this;
            const { endo } = CURVE;
            if (!endo)
                return wnaf.unsafeLadder(this, n);
            // Apply endomorphism
            let { k1neg, k1, k2neg, k2 } = endo.splitScalar(n);
            let k1p = I;
            let k2p = I;
            let d = this;
            while (k1 > _0n$1 || k2 > _0n$1) {
                if (k1 & _1n$2)
                    k1p = k1p.add(d);
                if (k2 & _1n$2)
                    k2p = k2p.add(d);
                d = d.double();
                k1 >>= _1n$2;
                k2 >>= _1n$2;
            }
            if (k1neg)
                k1p = k1p.negate();
            if (k2neg)
                k2p = k2p.negate();
            k2p = new Point(Fp.mul(k2p.px, endo.beta), k2p.py, k2p.pz);
            return k1p.add(k2p);
        }
        /**
         * Constant time multiplication.
         * Uses wNAF method. Windowed method may be 10% faster,
         * but takes 2x longer to generate and consumes 2x memory.
         * Uses precomputes when available.
         * Uses endomorphism for Koblitz curves.
         * @param scalar by which the point would be multiplied
         * @returns New point
         */
        multiply(scalar) {
            assertGE(scalar);
            let n = scalar;
            let point, fake; // Fake point is used to const-time mult
            const { endo } = CURVE;
            if (endo) {
                const { k1neg, k1, k2neg, k2 } = endo.splitScalar(n);
                let { p: k1p, f: f1p } = this.wNAF(k1);
                let { p: k2p, f: f2p } = this.wNAF(k2);
                k1p = wnaf.constTimeNegate(k1neg, k1p);
                k2p = wnaf.constTimeNegate(k2neg, k2p);
                k2p = new Point(Fp.mul(k2p.px, endo.beta), k2p.py, k2p.pz);
                point = k1p.add(k2p);
                fake = f1p.add(f2p);
            }
            else {
                const { p, f } = this.wNAF(n);
                point = p;
                fake = f;
            }
            // Normalize `z` for both points, but return only real one
            return Point.normalizeZ([point, fake])[0];
        }
        /**
         * Efficiently calculate `aP + bQ`. Unsafe, can expose private key, if used incorrectly.
         * Not using Strauss-Shamir trick: precomputation tables are faster.
         * The trick could be useful if both P and Q are not G (not in our case).
         * @returns non-zero affine point
         */
        multiplyAndAddUnsafe(Q, a, b) {
            const G = Point.BASE; // No Strauss-Shamir trick: we have 10% faster G precomputes
            const mul = (P, a // Select faster multiply() method
            ) => (a === _0n$1 || a === _1n$2 || !P.equals(G) ? P.multiplyUnsafe(a) : P.multiply(a));
            const sum = mul(this, a).add(mul(Q, b));
            return sum.is0() ? undefined : sum;
        }
        // Converts Projective point to affine (x, y) coordinates.
        // Can accept precomputed Z^-1 - for example, from invertBatch.
        // (x, y, z) ∋ (x=x/z, y=y/z)
        toAffine(iz) {
            const { px: x, py: y, pz: z } = this;
            const is0 = this.is0();
            // If invZ was 0, we return zero point. However we still want to execute
            // all operations, so we replace invZ with a random number, 1.
            if (iz == null)
                iz = is0 ? Fp.ONE : Fp.inv(z);
            const ax = Fp.mul(x, iz);
            const ay = Fp.mul(y, iz);
            const zz = Fp.mul(z, iz);
            if (is0)
                return { x: Fp.ZERO, y: Fp.ZERO };
            if (!Fp.eql(zz, Fp.ONE))
                throw new Error('invZ was invalid');
            return { x: ax, y: ay };
        }
        isTorsionFree() {
            const { h: cofactor, isTorsionFree } = CURVE;
            if (cofactor === _1n$2)
                return true; // No subgroups, always torsion-free
            if (isTorsionFree)
                return isTorsionFree(Point, this);
            throw new Error('isTorsionFree() has not been declared for the elliptic curve');
        }
        clearCofactor() {
            const { h: cofactor, clearCofactor } = CURVE;
            if (cofactor === _1n$2)
                return this; // Fast-path
            if (clearCofactor)
                return clearCofactor(Point, this);
            return this.multiplyUnsafe(CURVE.h);
        }
        toRawBytes(isCompressed = true) {
            this.assertValidity();
            return toBytes(Point, this, isCompressed);
        }
        toHex(isCompressed = true) {
            return bytesToHex$1(this.toRawBytes(isCompressed));
        }
    }
    Point.BASE = new Point(CURVE.Gx, CURVE.Gy, Fp.ONE);
    Point.ZERO = new Point(Fp.ZERO, Fp.ONE, Fp.ZERO);
    const _bits = CURVE.nBitLength;
    const wnaf = wNAF(Point, CURVE.endo ? Math.ceil(_bits / 2) : _bits);
    // Validate if generator point is on curve
    return {
        CURVE,
        ProjectivePoint: Point,
        normPrivateKeyToScalar,
        weierstrassEquation,
        isWithinCurveOrder,
    };
}
function validateOpts(curve) {
    const opts = validateBasic(curve);
    validateObject(opts, {
        hash: 'hash',
        hmac: 'function',
        randomBytes: 'function',
    }, {
        bits2int: 'function',
        bits2int_modN: 'function',
        lowS: 'boolean',
    });
    return Object.freeze({ lowS: true, ...opts });
}
function weierstrass(curveDef) {
    const CURVE = validateOpts(curveDef);
    const { Fp, n: CURVE_ORDER } = CURVE;
    const compressedLen = Fp.BYTES + 1; // e.g. 33 for 32
    const uncompressedLen = 2 * Fp.BYTES + 1; // e.g. 65 for 32
    function isValidFieldElement(num) {
        return _0n$1 < num && num < Fp.ORDER; // 0 is banned since it's not invertible FE
    }
    function modN(a) {
        return mod(a, CURVE_ORDER);
    }
    function invN(a) {
        return invert(a, CURVE_ORDER);
    }
    const { ProjectivePoint: Point, normPrivateKeyToScalar, weierstrassEquation, isWithinCurveOrder, } = weierstrassPoints({
        ...CURVE,
        toBytes(_c, point, isCompressed) {
            const a = point.toAffine();
            const x = Fp.toBytes(a.x);
            const cat = concatBytes;
            if (isCompressed) {
                return cat(Uint8Array.from([point.hasEvenY() ? 0x02 : 0x03]), x);
            }
            else {
                return cat(Uint8Array.from([0x04]), x, Fp.toBytes(a.y));
            }
        },
        fromBytes(bytes) {
            const len = bytes.length;
            const head = bytes[0];
            const tail = bytes.subarray(1);
            // this.assertValidity() is done inside of fromHex
            if (len === compressedLen && (head === 0x02 || head === 0x03)) {
                const x = bytesToNumberBE(tail);
                if (!isValidFieldElement(x))
                    throw new Error('Point is not on curve');
                const y2 = weierstrassEquation(x); // y² = x³ + ax + b
                let y;
                try {
                    y = Fp.sqrt(y2); // y = y² ^ (p+1)/4
                }
                catch (sqrtError) {
                    const suffix = sqrtError instanceof Error ? ': ' + sqrtError.message : '';
                    throw new Error('Point is not on curve' + suffix);
                }
                const isYOdd = (y & _1n$2) === _1n$2;
                // ECDSA
                const isHeadOdd = (head & 1) === 1;
                if (isHeadOdd !== isYOdd)
                    y = Fp.neg(y);
                return { x, y };
            }
            else if (len === uncompressedLen && head === 0x04) {
                const x = Fp.fromBytes(tail.subarray(0, Fp.BYTES));
                const y = Fp.fromBytes(tail.subarray(Fp.BYTES, 2 * Fp.BYTES));
                return { x, y };
            }
            else {
                throw new Error(`Point of length ${len} was invalid. Expected ${compressedLen} compressed bytes or ${uncompressedLen} uncompressed bytes`);
            }
        },
    });
    const numToNByteStr = (num) => bytesToHex$1(numberToBytesBE(num, CURVE.nByteLength));
    function isBiggerThanHalfOrder(number) {
        const HALF = CURVE_ORDER >> _1n$2;
        return number > HALF;
    }
    function normalizeS(s) {
        return isBiggerThanHalfOrder(s) ? modN(-s) : s;
    }
    // slice bytes num
    const slcNum = (b, from, to) => bytesToNumberBE(b.slice(from, to));
    /**
     * ECDSA signature with its (r, s) properties. Supports DER & compact representations.
     */
    class Signature {
        constructor(r, s, recovery) {
            this.r = r;
            this.s = s;
            this.recovery = recovery;
            this.assertValidity();
        }
        // pair (bytes of r, bytes of s)
        static fromCompact(hex) {
            const l = CURVE.nByteLength;
            hex = ensureBytes('compactSignature', hex, l * 2);
            return new Signature(slcNum(hex, 0, l), slcNum(hex, l, 2 * l));
        }
        // DER encoded ECDSA signature
        // https://bitcoin.stackexchange.com/questions/57644/what-are-the-parts-of-a-bitcoin-transaction-input-script
        static fromDER(hex) {
            const { r, s } = DER.toSig(ensureBytes('DER', hex));
            return new Signature(r, s);
        }
        assertValidity() {
            // can use assertGE here
            if (!isWithinCurveOrder(this.r))
                throw new Error('r must be 0 < r < CURVE.n');
            if (!isWithinCurveOrder(this.s))
                throw new Error('s must be 0 < s < CURVE.n');
        }
        addRecoveryBit(recovery) {
            return new Signature(this.r, this.s, recovery);
        }
        recoverPublicKey(msgHash) {
            const { r, s, recovery: rec } = this;
            const h = bits2int_modN(ensureBytes('msgHash', msgHash)); // Truncate hash
            if (rec == null || ![0, 1, 2, 3].includes(rec))
                throw new Error('recovery id invalid');
            const radj = rec === 2 || rec === 3 ? r + CURVE.n : r;
            if (radj >= Fp.ORDER)
                throw new Error('recovery id 2 or 3 invalid');
            const prefix = (rec & 1) === 0 ? '02' : '03';
            const R = Point.fromHex(prefix + numToNByteStr(radj));
            const ir = invN(radj); // r^-1
            const u1 = modN(-h * ir); // -hr^-1
            const u2 = modN(s * ir); // sr^-1
            const Q = Point.BASE.multiplyAndAddUnsafe(R, u1, u2); // (sr^-1)R-(hr^-1)G = -(hr^-1)G + (sr^-1)
            if (!Q)
                throw new Error('point at infinify'); // unsafe is fine: no priv data leaked
            Q.assertValidity();
            return Q;
        }
        // Signatures should be low-s, to prevent malleability.
        hasHighS() {
            return isBiggerThanHalfOrder(this.s);
        }
        normalizeS() {
            return this.hasHighS() ? new Signature(this.r, modN(-this.s), this.recovery) : this;
        }
        // DER-encoded
        toDERRawBytes() {
            return hexToBytes$1(this.toDERHex());
        }
        toDERHex() {
            return DER.hexFromSig({ r: this.r, s: this.s });
        }
        // padded bytes of r, then padded bytes of s
        toCompactRawBytes() {
            return hexToBytes$1(this.toCompactHex());
        }
        toCompactHex() {
            return numToNByteStr(this.r) + numToNByteStr(this.s);
        }
    }
    const utils = {
        isValidPrivateKey(privateKey) {
            try {
                normPrivateKeyToScalar(privateKey);
                return true;
            }
            catch (error) {
                return false;
            }
        },
        normPrivateKeyToScalar: normPrivateKeyToScalar,
        /**
         * Produces cryptographically secure private key from random of size
         * (groupLen + ceil(groupLen / 2)) with modulo bias being negligible.
         */
        randomPrivateKey: () => {
            const length = getMinHashLength(CURVE.n);
            return mapHashToField(CURVE.randomBytes(length), CURVE.n);
        },
        /**
         * Creates precompute table for an arbitrary EC point. Makes point "cached".
         * Allows to massively speed-up `point.multiply(scalar)`.
         * @returns cached point
         * @example
         * const fast = utils.precompute(8, ProjectivePoint.fromHex(someonesPubKey));
         * fast.multiply(privKey); // much faster ECDH now
         */
        precompute(windowSize = 8, point = Point.BASE) {
            point._setWindowSize(windowSize);
            point.multiply(BigInt(3)); // 3 is arbitrary, just need any number here
            return point;
        },
    };
    /**
     * Computes public key for a private key. Checks for validity of the private key.
     * @param privateKey private key
     * @param isCompressed whether to return compact (default), or full key
     * @returns Public key, full when isCompressed=false; short when isCompressed=true
     */
    function getPublicKey(privateKey, isCompressed = true) {
        return Point.fromPrivateKey(privateKey).toRawBytes(isCompressed);
    }
    /**
     * Quick and dirty check for item being public key. Does not validate hex, or being on-curve.
     */
    function isProbPub(item) {
        const arr = isBytes$1(item);
        const str = typeof item === 'string';
        const len = (arr || str) && item.length;
        if (arr)
            return len === compressedLen || len === uncompressedLen;
        if (str)
            return len === 2 * compressedLen || len === 2 * uncompressedLen;
        if (item instanceof Point)
            return true;
        return false;
    }
    /**
     * ECDH (Elliptic Curve Diffie Hellman).
     * Computes shared public key from private key and public key.
     * Checks: 1) private key validity 2) shared key is on-curve.
     * Does NOT hash the result.
     * @param privateA private key
     * @param publicB different public key
     * @param isCompressed whether to return compact (default), or full key
     * @returns shared public key
     */
    function getSharedSecret(privateA, publicB, isCompressed = true) {
        if (isProbPub(privateA))
            throw new Error('first arg must be private key');
        if (!isProbPub(publicB))
            throw new Error('second arg must be public key');
        const b = Point.fromHex(publicB); // check for being on-curve
        return b.multiply(normPrivateKeyToScalar(privateA)).toRawBytes(isCompressed);
    }
    // RFC6979: ensure ECDSA msg is X bytes and < N. RFC suggests optional truncating via bits2octets.
    // FIPS 186-4 4.6 suggests the leftmost min(nBitLen, outLen) bits, which matches bits2int.
    // bits2int can produce res>N, we can do mod(res, N) since the bitLen is the same.
    // int2octets can't be used; pads small msgs with 0: unacceptatble for trunc as per RFC vectors
    const bits2int = CURVE.bits2int ||
        function (bytes) {
            // For curves with nBitLength % 8 !== 0: bits2octets(bits2octets(m)) !== bits2octets(m)
            // for some cases, since bytes.length * 8 is not actual bitLength.
            const num = bytesToNumberBE(bytes); // check for == u8 done here
            const delta = bytes.length * 8 - CURVE.nBitLength; // truncate to nBitLength leftmost bits
            return delta > 0 ? num >> BigInt(delta) : num;
        };
    const bits2int_modN = CURVE.bits2int_modN ||
        function (bytes) {
            return modN(bits2int(bytes)); // can't use bytesToNumberBE here
        };
    // NOTE: pads output with zero as per spec
    const ORDER_MASK = bitMask(CURVE.nBitLength);
    /**
     * Converts to bytes. Checks if num in `[0..ORDER_MASK-1]` e.g.: `[0..2^256-1]`.
     */
    function int2octets(num) {
        if (typeof num !== 'bigint')
            throw new Error('bigint expected');
        if (!(_0n$1 <= num && num < ORDER_MASK))
            throw new Error(`bigint expected < 2^${CURVE.nBitLength}`);
        // works with order, can have different size than numToField!
        return numberToBytesBE(num, CURVE.nByteLength);
    }
    // Steps A, D of RFC6979 3.2
    // Creates RFC6979 seed; converts msg/privKey to numbers.
    // Used only in sign, not in verify.
    // NOTE: we cannot assume here that msgHash has same amount of bytes as curve order, this will be wrong at least for P521.
    // Also it can be bigger for P224 + SHA256
    function prepSig(msgHash, privateKey, opts = defaultSigOpts) {
        if (['recovered', 'canonical'].some((k) => k in opts))
            throw new Error('sign() legacy options not supported');
        const { hash, randomBytes } = CURVE;
        let { lowS, prehash, extraEntropy: ent } = opts; // generates low-s sigs by default
        if (lowS == null)
            lowS = true; // RFC6979 3.2: we skip step A, because we already provide hash
        msgHash = ensureBytes('msgHash', msgHash);
        if (prehash)
            msgHash = ensureBytes('prehashed msgHash', hash(msgHash));
        // We can't later call bits2octets, since nested bits2int is broken for curves
        // with nBitLength % 8 !== 0. Because of that, we unwrap it here as int2octets call.
        // const bits2octets = (bits) => int2octets(bits2int_modN(bits))
        const h1int = bits2int_modN(msgHash);
        const d = normPrivateKeyToScalar(privateKey); // validate private key, convert to bigint
        const seedArgs = [int2octets(d), int2octets(h1int)];
        // extraEntropy. RFC6979 3.6: additional k' (optional).
        if (ent != null && ent !== false) {
            // K = HMAC_K(V || 0x00 || int2octets(x) || bits2octets(h1) || k')
            const e = ent === true ? randomBytes(Fp.BYTES) : ent; // generate random bytes OR pass as-is
            seedArgs.push(ensureBytes('extraEntropy', e)); // check for being bytes
        }
        const seed = concatBytes(...seedArgs); // Step D of RFC6979 3.2
        const m = h1int; // NOTE: no need to call bits2int second time here, it is inside truncateHash!
        // Converts signature params into point w r/s, checks result for validity.
        function k2sig(kBytes) {
            // RFC 6979 Section 3.2, step 3: k = bits2int(T)
            const k = bits2int(kBytes); // Cannot use fields methods, since it is group element
            if (!isWithinCurveOrder(k))
                return; // Important: all mod() calls here must be done over N
            const ik = invN(k); // k^-1 mod n
            const q = Point.BASE.multiply(k).toAffine(); // q = Gk
            const r = modN(q.x); // r = q.x mod n
            if (r === _0n$1)
                return;
            // Can use scalar blinding b^-1(bm + bdr) where b ∈ [1,q−1] according to
            // https://tches.iacr.org/index.php/TCHES/article/view/7337/6509. We've decided against it:
            // a) dependency on CSPRNG b) 15% slowdown c) doesn't really help since bigints are not CT
            const s = modN(ik * modN(m + r * d)); // Not using blinding here
            if (s === _0n$1)
                return;
            let recovery = (q.x === r ? 0 : 2) | Number(q.y & _1n$2); // recovery bit (2 or 3, when q.x > n)
            let normS = s;
            if (lowS && isBiggerThanHalfOrder(s)) {
                normS = normalizeS(s); // if lowS was passed, ensure s is always
                recovery ^= 1; // // in the bottom half of N
            }
            return new Signature(r, normS, recovery); // use normS, not s
        }
        return { seed, k2sig };
    }
    const defaultSigOpts = { lowS: CURVE.lowS, prehash: false };
    const defaultVerOpts = { lowS: CURVE.lowS, prehash: false };
    /**
     * Signs message hash with a private key.
     * ```
     * sign(m, d, k) where
     *   (x, y) = G × k
     *   r = x mod n
     *   s = (m + dr)/k mod n
     * ```
     * @param msgHash NOT message. msg needs to be hashed to `msgHash`, or use `prehash`.
     * @param privKey private key
     * @param opts lowS for non-malleable sigs. extraEntropy for mixing randomness into k. prehash will hash first arg.
     * @returns signature with recovery param
     */
    function sign(msgHash, privKey, opts = defaultSigOpts) {
        const { seed, k2sig } = prepSig(msgHash, privKey, opts); // Steps A, D of RFC6979 3.2.
        const C = CURVE;
        const drbg = createHmacDrbg(C.hash.outputLen, C.nByteLength, C.hmac);
        return drbg(seed, k2sig); // Steps B, C, D, E, F, G
    }
    // Enable precomputes. Slows down first publicKey computation by 20ms.
    Point.BASE._setWindowSize(8);
    // utils.precompute(8, ProjectivePoint.BASE)
    /**
     * Verifies a signature against message hash and public key.
     * Rejects lowS signatures by default: to override,
     * specify option `{lowS: false}`. Implements section 4.1.4 from https://www.secg.org/sec1-v2.pdf:
     *
     * ```
     * verify(r, s, h, P) where
     *   U1 = hs^-1 mod n
     *   U2 = rs^-1 mod n
     *   R = U1⋅G - U2⋅P
     *   mod(R.x, n) == r
     * ```
     */
    function verify(signature, msgHash, publicKey, opts = defaultVerOpts) {
        const sg = signature;
        msgHash = ensureBytes('msgHash', msgHash);
        publicKey = ensureBytes('publicKey', publicKey);
        if ('strict' in opts)
            throw new Error('options.strict was renamed to lowS');
        const { lowS, prehash } = opts;
        let _sig = undefined;
        let P;
        try {
            if (typeof sg === 'string' || isBytes$1(sg)) {
                // Signature can be represented in 2 ways: compact (2*nByteLength) & DER (variable-length).
                // Since DER can also be 2*nByteLength bytes, we check for it first.
                try {
                    _sig = Signature.fromDER(sg);
                }
                catch (derError) {
                    if (!(derError instanceof DER.Err))
                        throw derError;
                    _sig = Signature.fromCompact(sg);
                }
            }
            else if (typeof sg === 'object' && typeof sg.r === 'bigint' && typeof sg.s === 'bigint') {
                const { r, s } = sg;
                _sig = new Signature(r, s);
            }
            else {
                throw new Error('PARSE');
            }
            P = Point.fromHex(publicKey);
        }
        catch (error) {
            if (error.message === 'PARSE')
                throw new Error(`signature must be Signature instance, Uint8Array or hex string`);
            return false;
        }
        if (lowS && _sig.hasHighS())
            return false;
        if (prehash)
            msgHash = CURVE.hash(msgHash);
        const { r, s } = _sig;
        const h = bits2int_modN(msgHash); // Cannot use fields methods, since it is group element
        const is = invN(s); // s^-1
        const u1 = modN(h * is); // u1 = hs^-1 mod n
        const u2 = modN(r * is); // u2 = rs^-1 mod n
        const R = Point.BASE.multiplyAndAddUnsafe(P, u1, u2)?.toAffine(); // R = u1⋅G + u2⋅P
        if (!R)
            return false;
        const v = modN(R.x);
        return v === r;
    }
    return {
        CURVE,
        getPublicKey,
        getSharedSecret,
        sign,
        verify,
        ProjectivePoint: Point,
        Signature,
        utils,
    };
}

/*! noble-curves - MIT License (c) 2022 Paul Miller (paulmillr.com) */
// connects noble-curves to noble-hashes
function getHash(hash) {
    return {
        hash,
        hmac: (key, ...msgs) => hmac(hash, key, concatBytes$1(...msgs)),
        randomBytes: randomBytes$1,
    };
}
function createCurve(curveDef, defHash) {
    const create = (hash) => weierstrass({ ...curveDef, ...getHash(hash) });
    return Object.freeze({ ...create(defHash), create });
}

/*! noble-curves - MIT License (c) 2022 Paul Miller (paulmillr.com) */
const secp256k1P = BigInt('0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffefffffc2f');
const secp256k1N = BigInt('0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141');
const _1n$1 = BigInt(1);
const _2n$1 = BigInt(2);
const divNearest = (a, b) => (a + b / _2n$1) / b;
/**
 * √n = n^((p+1)/4) for fields p = 3 mod 4. We unwrap the loop and multiply bit-by-bit.
 * (P+1n/4n).toString(2) would produce bits [223x 1, 0, 22x 1, 4x 0, 11, 00]
 */
function sqrtMod(y) {
    const P = secp256k1P;
    // prettier-ignore
    const _3n = BigInt(3), _6n = BigInt(6), _11n = BigInt(11), _22n = BigInt(22);
    // prettier-ignore
    const _23n = BigInt(23), _44n = BigInt(44), _88n = BigInt(88);
    const b2 = (y * y * y) % P; // x^3, 11
    const b3 = (b2 * b2 * y) % P; // x^7
    const b6 = (pow2(b3, _3n, P) * b3) % P;
    const b9 = (pow2(b6, _3n, P) * b3) % P;
    const b11 = (pow2(b9, _2n$1, P) * b2) % P;
    const b22 = (pow2(b11, _11n, P) * b11) % P;
    const b44 = (pow2(b22, _22n, P) * b22) % P;
    const b88 = (pow2(b44, _44n, P) * b44) % P;
    const b176 = (pow2(b88, _88n, P) * b88) % P;
    const b220 = (pow2(b176, _44n, P) * b44) % P;
    const b223 = (pow2(b220, _3n, P) * b3) % P;
    const t1 = (pow2(b223, _23n, P) * b22) % P;
    const t2 = (pow2(t1, _6n, P) * b2) % P;
    const root = pow2(t2, _2n$1, P);
    if (!Fp$1.eql(Fp$1.sqr(root), y))
        throw new Error('Cannot find square root');
    return root;
}
const Fp$1 = Field(secp256k1P, undefined, undefined, { sqrt: sqrtMod });
const secp256k1 = createCurve({
    a: BigInt(0), // equation params: a, b
    b: BigInt(7), // Seem to be rigid: bitcointalk.org/index.php?topic=289795.msg3183975#msg3183975
    Fp: Fp$1, // Field's prime: 2n**256n - 2n**32n - 2n**9n - 2n**8n - 2n**7n - 2n**6n - 2n**4n - 1n
    n: secp256k1N, // Curve order, total count of valid points in the field
    // Base point (x, y) aka generator point
    Gx: BigInt('55066263022277343669578718895168534326250603453777594175500187360389116729240'),
    Gy: BigInt('32670510020758816978083085130507043184471273380659243275938904335757337482424'),
    h: BigInt(1), // Cofactor
    lowS: true, // Allow only low-S signatures by default in sign() and verify()
    /**
     * secp256k1 belongs to Koblitz curves: it has efficiently computable endomorphism.
     * Endomorphism uses 2x less RAM, speeds up precomputation by 2x and ECDH / key recovery by 20%.
     * For precomputed wNAF it trades off 1/2 init time & 1/3 ram for 20% perf hit.
     * Explanation: https://gist.github.com/paulmillr/eb670806793e84df628a7c434a873066
     */
    endo: {
        beta: BigInt('0x7ae96a2b657c07106e64479eac3434e99cf0497512f58995c1396c28719501ee'),
        splitScalar: (k) => {
            const n = secp256k1N;
            const a1 = BigInt('0x3086d221a7d46bcde86c90e49284eb15');
            const b1 = -_1n$1 * BigInt('0xe4437ed6010e88286f547fa90abfe4c3');
            const a2 = BigInt('0x114ca50f7a8e2f3f657c1108d9d44cfd8');
            const b2 = a1;
            const POW_2_128 = BigInt('0x100000000000000000000000000000000'); // (2n**128n).toString(16)
            const c1 = divNearest(b2 * k, n);
            const c2 = divNearest(-b1 * k, n);
            let k1 = mod(k - c1 * a1 - c2 * a2, n);
            let k2 = mod(-c1 * b1 - c2 * b2, n);
            const k1neg = k1 > POW_2_128;
            const k2neg = k2 > POW_2_128;
            if (k1neg)
                k1 = n - k1;
            if (k2neg)
                k2 = n - k2;
            if (k1 > POW_2_128 || k2 > POW_2_128) {
                throw new Error('splitScalar: Endomorphism failed, k=' + k);
            }
            return { k1neg, k1, k2neg, k2 };
        },
    },
}, sha256$2);
// Schnorr signatures are superior to ECDSA from above. Below is Schnorr-specific BIP0340 code.
// https://github.com/bitcoin/bips/blob/master/bip-0340.mediawiki
BigInt(0);

/*! noble-curves - MIT License (c) 2022 Paul Miller (paulmillr.com) */
// NIST secp256r1 aka p256
// https://www.secg.org/sec2-v2.pdf, https://neuromancer.sk/std/nist/P-256
const Fp = Field(BigInt('0xffffffff00000001000000000000000000000000ffffffffffffffffffffffff'));
const CURVE_A = Fp.create(BigInt('-3'));
const CURVE_B = BigInt('0x5ac635d8aa3a93e7b3ebbd55769886bc651d06b0cc53b0f63bce3c3e27d2604b');
// prettier-ignore
const p256 = createCurve({
    a: CURVE_A, // Equation params: a, b
    b: CURVE_B,
    Fp, // Field: 2n**224n * (2n**32n-1n) + 2n**192n + 2n**96n-1n
    // Curve order, total count of valid points in the field
    n: BigInt('0xffffffff00000000ffffffffffffffffbce6faada7179e84f3b9cac2fc632551'),
    // Base (generator) point (x, y)
    Gx: BigInt('0x6b17d1f2e12c4247f8bce6e563a440f277037d812deb33a0f4a13945d898c296'),
    Gy: BigInt('0x4fe342e2fe1a7f9b8ee7eb4a7c0f9e162bce33576b315ececbb6406837bf51f5'),
    h: BigInt(1),
    lowS: false,
}, sha256$2);

// https://homes.esat.kuleuven.be/~bosselae/ripemd160.html
// https://homes.esat.kuleuven.be/~bosselae/ripemd160/pdf/AB-9601/AB-9601.pdf
const Rho = /* @__PURE__ */ new Uint8Array([7, 4, 13, 1, 10, 6, 15, 3, 12, 0, 9, 5, 2, 14, 11, 8]);
const Id = /* @__PURE__ */ new Uint8Array(new Array(16).fill(0).map((_, i) => i));
const Pi = /* @__PURE__ */ Id.map((i) => (9 * i + 5) % 16);
let idxL = [Id];
let idxR = [Pi];
for (let i = 0; i < 4; i++)
    for (let j of [idxL, idxR])
        j.push(j[i].map((k) => Rho[k]));
const shifts = /* @__PURE__ */ [
    [11, 14, 15, 12, 5, 8, 7, 9, 11, 13, 14, 15, 6, 7, 9, 8],
    [12, 13, 11, 15, 6, 9, 9, 7, 12, 15, 11, 13, 7, 8, 7, 7],
    [13, 15, 14, 11, 7, 7, 6, 8, 13, 14, 13, 12, 5, 5, 6, 9],
    [14, 11, 12, 14, 8, 6, 5, 5, 15, 12, 15, 14, 9, 9, 8, 6],
    [15, 12, 13, 13, 9, 5, 8, 6, 14, 11, 12, 11, 8, 6, 5, 5],
].map((i) => new Uint8Array(i));
const shiftsL = /* @__PURE__ */ idxL.map((idx, i) => idx.map((j) => shifts[i][j]));
const shiftsR = /* @__PURE__ */ idxR.map((idx, i) => idx.map((j) => shifts[i][j]));
const Kl = /* @__PURE__ */ new Uint32Array([
    0x00000000, 0x5a827999, 0x6ed9eba1, 0x8f1bbcdc, 0xa953fd4e,
]);
const Kr = /* @__PURE__ */ new Uint32Array([
    0x50a28be6, 0x5c4dd124, 0x6d703ef3, 0x7a6d76e9, 0x00000000,
]);
// It's called f() in spec.
function f(group, x, y, z) {
    if (group === 0)
        return x ^ y ^ z;
    else if (group === 1)
        return (x & y) | (~x & z);
    else if (group === 2)
        return (x | ~y) ^ z;
    else if (group === 3)
        return (x & z) | (y & ~z);
    else
        return x ^ (y | ~z);
}
// Temporary buffer, not used to store anything between runs
const R_BUF = /* @__PURE__ */ new Uint32Array(16);
class RIPEMD160 extends HashMD {
    constructor() {
        super(64, 20, 8, true);
        this.h0 = 0x67452301 | 0;
        this.h1 = 0xefcdab89 | 0;
        this.h2 = 0x98badcfe | 0;
        this.h3 = 0x10325476 | 0;
        this.h4 = 0xc3d2e1f0 | 0;
    }
    get() {
        const { h0, h1, h2, h3, h4 } = this;
        return [h0, h1, h2, h3, h4];
    }
    set(h0, h1, h2, h3, h4) {
        this.h0 = h0 | 0;
        this.h1 = h1 | 0;
        this.h2 = h2 | 0;
        this.h3 = h3 | 0;
        this.h4 = h4 | 0;
    }
    process(view, offset) {
        for (let i = 0; i < 16; i++, offset += 4)
            R_BUF[i] = view.getUint32(offset, true);
        // prettier-ignore
        let al = this.h0 | 0, ar = al, bl = this.h1 | 0, br = bl, cl = this.h2 | 0, cr = cl, dl = this.h3 | 0, dr = dl, el = this.h4 | 0, er = el;
        // Instead of iterating 0 to 80, we split it into 5 groups
        // And use the groups in constants, functions, etc. Much simpler
        for (let group = 0; group < 5; group++) {
            const rGroup = 4 - group;
            const hbl = Kl[group], hbr = Kr[group]; // prettier-ignore
            const rl = idxL[group], rr = idxR[group]; // prettier-ignore
            const sl = shiftsL[group], sr = shiftsR[group]; // prettier-ignore
            for (let i = 0; i < 16; i++) {
                const tl = (rotl(al + f(group, bl, cl, dl) + R_BUF[rl[i]] + hbl, sl[i]) + el) | 0;
                al = el, el = dl, dl = rotl(cl, 10) | 0, cl = bl, bl = tl; // prettier-ignore
            }
            // 2 loops are 10% faster
            for (let i = 0; i < 16; i++) {
                const tr = (rotl(ar + f(rGroup, br, cr, dr) + R_BUF[rr[i]] + hbr, sr[i]) + er) | 0;
                ar = er, er = dr, dr = rotl(cr, 10) | 0, cr = br, br = tr; // prettier-ignore
            }
        }
        // Add the compressed chunk to the current hash value
        this.set((this.h1 + cl + dr) | 0, (this.h2 + dl + er) | 0, (this.h3 + el + ar) | 0, (this.h4 + al + br) | 0, (this.h0 + bl + cr) | 0);
    }
    roundClean() {
        R_BUF.fill(0);
    }
    destroy() {
        this.destroyed = true;
        this.buffer.fill(0);
        this.set(0, 0, 0, 0, 0);
    }
}
/**
 * RIPEMD-160 - a hash function from 1990s.
 * @param message - msg that would be hashed
 */
const ripemd160 = /* @__PURE__ */ wrapConstructor(() => new RIPEMD160());

// SHA3 (keccak) is based on a new design: basically, the internal state is bigger than output size.
// It's called a sponge function.
// Various per round constants calculations
const SHA3_PI = [];
const SHA3_ROTL = [];
const _SHA3_IOTA = [];
const _0n = /* @__PURE__ */ BigInt(0);
const _1n = /* @__PURE__ */ BigInt(1);
const _2n = /* @__PURE__ */ BigInt(2);
const _7n = /* @__PURE__ */ BigInt(7);
const _256n = /* @__PURE__ */ BigInt(256);
const _0x71n = /* @__PURE__ */ BigInt(0x71);
for (let round = 0, R = _1n, x = 1, y = 0; round < 24; round++) {
    // Pi
    [x, y] = [y, (2 * x + 3 * y) % 5];
    SHA3_PI.push(2 * (5 * y + x));
    // Rotational
    SHA3_ROTL.push((((round + 1) * (round + 2)) / 2) % 64);
    // Iota
    let t = _0n;
    for (let j = 0; j < 7; j++) {
        R = ((R << _1n) ^ ((R >> _7n) * _0x71n)) % _256n;
        if (R & _2n)
            t ^= _1n << ((_1n << /* @__PURE__ */ BigInt(j)) - _1n);
    }
    _SHA3_IOTA.push(t);
}
const [SHA3_IOTA_H, SHA3_IOTA_L] = /* @__PURE__ */ split(_SHA3_IOTA, true);
// Left rotation (without 0, 32, 64)
const rotlH = (h, l, s) => (s > 32 ? rotlBH(h, l, s) : rotlSH(h, l, s));
const rotlL = (h, l, s) => (s > 32 ? rotlBL(h, l, s) : rotlSL(h, l, s));
// Same as keccakf1600, but allows to skip some rounds
function keccakP(s, rounds = 24) {
    const B = new Uint32Array(5 * 2);
    // NOTE: all indices are x2 since we store state as u32 instead of u64 (bigints to slow in js)
    for (let round = 24 - rounds; round < 24; round++) {
        // Theta θ
        for (let x = 0; x < 10; x++)
            B[x] = s[x] ^ s[x + 10] ^ s[x + 20] ^ s[x + 30] ^ s[x + 40];
        for (let x = 0; x < 10; x += 2) {
            const idx1 = (x + 8) % 10;
            const idx0 = (x + 2) % 10;
            const B0 = B[idx0];
            const B1 = B[idx0 + 1];
            const Th = rotlH(B0, B1, 1) ^ B[idx1];
            const Tl = rotlL(B0, B1, 1) ^ B[idx1 + 1];
            for (let y = 0; y < 50; y += 10) {
                s[x + y] ^= Th;
                s[x + y + 1] ^= Tl;
            }
        }
        // Rho (ρ) and Pi (π)
        let curH = s[2];
        let curL = s[3];
        for (let t = 0; t < 24; t++) {
            const shift = SHA3_ROTL[t];
            const Th = rotlH(curH, curL, shift);
            const Tl = rotlL(curH, curL, shift);
            const PI = SHA3_PI[t];
            curH = s[PI];
            curL = s[PI + 1];
            s[PI] = Th;
            s[PI + 1] = Tl;
        }
        // Chi (χ)
        for (let y = 0; y < 50; y += 10) {
            for (let x = 0; x < 10; x++)
                B[x] = s[y + x];
            for (let x = 0; x < 10; x++)
                s[y + x] ^= ~B[(x + 2) % 10] & B[(x + 4) % 10];
        }
        // Iota (ι)
        s[0] ^= SHA3_IOTA_H[round];
        s[1] ^= SHA3_IOTA_L[round];
    }
    B.fill(0);
}
class Keccak extends Hash {
    // NOTE: we accept arguments in bytes instead of bits here.
    constructor(blockLen, suffix, outputLen, enableXOF = false, rounds = 24) {
        super();
        this.blockLen = blockLen;
        this.suffix = suffix;
        this.outputLen = outputLen;
        this.enableXOF = enableXOF;
        this.rounds = rounds;
        this.pos = 0;
        this.posOut = 0;
        this.finished = false;
        this.destroyed = false;
        // Can be passed from user as dkLen
        number(outputLen);
        // 1600 = 5x5 matrix of 64bit.  1600 bits === 200 bytes
        if (0 >= this.blockLen || this.blockLen >= 200)
            throw new Error('Sha3 supports only keccak-f1600 function');
        this.state = new Uint8Array(200);
        this.state32 = u32(this.state);
    }
    keccak() {
        if (!isLE)
            byteSwap32(this.state32);
        keccakP(this.state32, this.rounds);
        if (!isLE)
            byteSwap32(this.state32);
        this.posOut = 0;
        this.pos = 0;
    }
    update(data) {
        exists(this);
        const { blockLen, state } = this;
        data = toBytes(data);
        const len = data.length;
        for (let pos = 0; pos < len;) {
            const take = Math.min(blockLen - this.pos, len - pos);
            for (let i = 0; i < take; i++)
                state[this.pos++] ^= data[pos++];
            if (this.pos === blockLen)
                this.keccak();
        }
        return this;
    }
    finish() {
        if (this.finished)
            return;
        this.finished = true;
        const { state, suffix, pos, blockLen } = this;
        // Do the padding
        state[pos] ^= suffix;
        if ((suffix & 0x80) !== 0 && pos === blockLen - 1)
            this.keccak();
        state[blockLen - 1] ^= 0x80;
        this.keccak();
    }
    writeInto(out) {
        exists(this, false);
        bytes(out);
        this.finish();
        const bufferOut = this.state;
        const { blockLen } = this;
        for (let pos = 0, len = out.length; pos < len;) {
            if (this.posOut >= blockLen)
                this.keccak();
            const take = Math.min(blockLen - this.posOut, len - pos);
            out.set(bufferOut.subarray(this.posOut, this.posOut + take), pos);
            this.posOut += take;
            pos += take;
        }
        return out;
    }
    xofInto(out) {
        // Sha3/Keccak usage with XOF is probably mistake, only SHAKE instances can do XOF
        if (!this.enableXOF)
            throw new Error('XOF is not possible for this instance');
        return this.writeInto(out);
    }
    xof(bytes) {
        number(bytes);
        return this.xofInto(new Uint8Array(bytes));
    }
    digestInto(out) {
        output(out, this);
        if (this.finished)
            throw new Error('digest() was already called');
        this.writeInto(out);
        this.destroy();
        return out;
    }
    digest() {
        return this.digestInto(new Uint8Array(this.outputLen));
    }
    destroy() {
        this.destroyed = true;
        this.state.fill(0);
    }
    _cloneInto(to) {
        const { blockLen, suffix, outputLen, rounds, enableXOF } = this;
        to || (to = new Keccak(blockLen, suffix, outputLen, enableXOF, rounds));
        to.state32.set(this.state32);
        to.pos = this.pos;
        to.posOut = this.posOut;
        to.finished = this.finished;
        to.rounds = rounds;
        // Suffix can change in cSHAKE
        to.suffix = suffix;
        to.outputLen = outputLen;
        to.enableXOF = enableXOF;
        to.destroyed = this.destroyed;
        return to;
    }
}
const gen = (suffix, blockLen, outputLen) => wrapConstructor(() => new Keccak(blockLen, suffix, outputLen));
/**
 * keccak-256 hash function. Different from SHA3-256.
 * @param message - that would be hashed
 */
const keccak_256 = /* @__PURE__ */ gen(0x01, 136, 256 / 8);

/* jshint esversion: 6 */

var canonicalize = function serialize (object) {
  if (typeof object === 'number' && isNaN(object)) {
    throw new Error('NaN is not allowed');
  }

  if (typeof object === 'number' && !isFinite(object)) {
    throw new Error('Infinity is not allowed');
  }

  if (object === null || typeof object !== 'object') {
    return JSON.stringify(object);
  }

  if (object.toJSON instanceof Function) {
    return serialize(object.toJSON());
  }

  if (Array.isArray(object)) {
    const values = object.reduce((t, cv, ci) => {
      const comma = ci === 0 ? '' : ',';
      const value = cv === undefined || typeof cv === 'symbol' ? null : cv;
      return `${t}${comma}${serialize(value)}`;
    }, '');
    return `[${values}]`;
  }

  const values = Object.keys(object).sort().reduce((t, cv) => {
    if (object[cv] === undefined ||
        typeof object[cv] === 'symbol') {
      return t;
    }
    const comma = t.length === 0 ? '' : ',';
    return `${t}${comma}${serialize(cv)}:${serialize(object[cv])}`;
  }, '');
  return `{${values}}`;
};

// Copyright 2018 Consensys AG

function _catch$2(body, recover) {
  try {
    var result = body();
  } catch (e) {
    return recover(e);
  }
  if (result && result.then) {
    return result.then(void 0, recover);
  }
  return result;
}
function inMemoryCache() {
  const cache = new Map();
  return function (parsed, resolve) {
    try {
      let _exit;
      function _temp2(_result) {
        if (_exit) return _result;
        const cached = cache.get(parsed.didUrl);
        return cached !== undefined ? cached : Promise.resolve(resolve()).then(function (result) {
          var _result$didResolution;
          if (((_result$didResolution = result.didResolutionMetadata) == null ? void 0 : _result$didResolution.error) !== 'notFound') {
            cache.set(parsed.didUrl, result);
          }
          return result;
        });
      }
      const _temp = function () {
        if (parsed.params && parsed.params['no-cache'] === 'true') {
          return Promise.resolve(resolve()).then(function (_await$resolve) {
            _exit = 1;
            return _await$resolve;
          });
        }
      }();
      return Promise.resolve(_temp && _temp.then ? _temp.then(_temp2) : _temp2(_temp));
    } catch (e) {
      return Promise.reject(e);
    }
  };
}
function noCache(parsed, resolve) {
  return resolve();
}
const PCT_ENCODED = '(?:%[0-9a-fA-F]{2})';
const ID_CHAR = `(?:[a-zA-Z0-9._-]|${PCT_ENCODED})`;
const METHOD = '([a-z0-9]+)';
const METHOD_ID = `((?:${ID_CHAR}*:)*(${ID_CHAR}+))`;
const PARAM_CHAR = '[a-zA-Z0-9_.:%-]';
const PARAM = `;${PARAM_CHAR}+=${PARAM_CHAR}*`;
const PARAMS = `((${PARAM})*)`;
const PATH = `(/[^#?]*)?`;
const QUERY = `([?][^#]*)?`;
const FRAGMENT = `(#.*)?`;
const DID_MATCHER = new RegExp(`^did:${METHOD}:${METHOD_ID}${PARAMS}${PATH}${QUERY}${FRAGMENT}$`);
/**
 * Parses a DID URL and builds a {@link ParsedDID | ParsedDID object}
 *
 * @param didUrl - the DID URL string to be parsed
 * @returns a ParsedDID object, or null if the input is not a DID URL
 */
function parse(didUrl) {
  if (didUrl === '' || !didUrl) return null;
  const sections = didUrl.match(DID_MATCHER);
  if (sections) {
    const parts = {
      did: `did:${sections[1]}:${sections[2]}`,
      method: sections[1],
      id: sections[2],
      didUrl
    };
    if (sections[4]) {
      const params = sections[4].slice(1).split(';');
      parts.params = {};
      for (const p of params) {
        const kv = p.split('=');
        parts.params[kv[0]] = kv[1];
      }
    }
    if (sections[6]) parts.path = sections[6];
    if (sections[7]) parts.query = sections[7].slice(1);
    if (sections[8]) parts.fragment = sections[8].slice(1);
    return parts;
  }
  return null;
}
const EMPTY_RESULT = {
  didResolutionMetadata: {},
  didDocument: null,
  didDocumentMetadata: {}
};
function wrapLegacyResolver(resolve) {
  return function (did, parsed, resolver) {
    try {
      return Promise.resolve(_catch$2(function () {
        return Promise.resolve(resolve(did, parsed, resolver)).then(function (doc) {
          return {
            ...EMPTY_RESULT,
            didResolutionMetadata: {
              contentType: 'application/did+ld+json'
            },
            didDocument: doc
          };
          // eslint-disable-next-line @typescript-eslint/no-explicit-any
        });
      }, function (e) {
        return {
          ...EMPTY_RESULT,
          didResolutionMetadata: {
            error: 'notFound',
            message: e.toString() // This is not in spec, but may be helpful
          }
        };
      }));
    } catch (e) {
      return Promise.reject(e);
    }
  };
}
/**
 * This implementation of {@link Resolvable} bundles together multiple implementations of {@link DIDResolver} and
 * presents a single function call to users.
 */
class Resolver {
  constructor(registry = {}, options = {}) {
    this.registry = void 0;
    this.cache = void 0;
    this.registry = registry;
    this.cache = options.cache === true ? inMemoryCache() : options.cache || noCache;
    if (options.legacyResolvers) {
      Object.keys(options.legacyResolvers).map(methodName => {
        if (!this.registry[methodName]) {
          this.registry[methodName] = wrapLegacyResolver(
          // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
          options.legacyResolvers[methodName]);
        }
      });
    }
  }
  resolve(didUrl, options = {}) {
    try {
      const _this = this;
      const parsed = parse(didUrl);
      if (parsed === null) {
        return Promise.resolve({
          ...EMPTY_RESULT,
          didResolutionMetadata: {
            error: 'invalidDid'
          }
        });
      }
      const resolver = _this.registry[parsed.method];
      if (!resolver) {
        return Promise.resolve({
          ...EMPTY_RESULT,
          didResolutionMetadata: {
            error: 'unsupportedDidMethod'
          }
        });
      }
      return Promise.resolve(_this.cache(parsed, () => resolver(parsed.did, parsed, _this, options)));
    } catch (e) {
      return Promise.reject(e);
    }
  }
}

/*! scure-base - MIT License (c) 2022 Paul Miller (paulmillr.com) */
// Utilities
/**
 * @__NO_SIDE_EFFECTS__
 */
function assertNumber(n) {
    if (!Number.isSafeInteger(n))
        throw new Error(`Wrong integer: ${n}`);
}
function isBytes(a) {
    return (a instanceof Uint8Array ||
        (a != null && typeof a === 'object' && a.constructor.name === 'Uint8Array'));
}
/**
 * @__NO_SIDE_EFFECTS__
 */
function chain(...args) {
    const id = (a) => a;
    // Wrap call in closure so JIT can inline calls
    const wrap = (a, b) => (c) => a(b(c));
    // Construct chain of args[-1].encode(args[-2].encode([...]))
    const encode = args.map((x) => x.encode).reduceRight(wrap, id);
    // Construct chain of args[0].decode(args[1].decode(...))
    const decode = args.map((x) => x.decode).reduce(wrap, id);
    return { encode, decode };
}
/**
 * Encodes integer radix representation to array of strings using alphabet and back
 * @__NO_SIDE_EFFECTS__
 */
function alphabet(alphabet) {
    return {
        encode: (digits) => {
            if (!Array.isArray(digits) || (digits.length && typeof digits[0] !== 'number'))
                throw new Error('alphabet.encode input should be an array of numbers');
            return digits.map((i) => {
                assertNumber(i);
                if (i < 0 || i >= alphabet.length)
                    throw new Error(`Digit index outside alphabet: ${i} (alphabet: ${alphabet.length})`);
                return alphabet[i];
            });
        },
        decode: (input) => {
            if (!Array.isArray(input) || (input.length && typeof input[0] !== 'string'))
                throw new Error('alphabet.decode input should be array of strings');
            return input.map((letter) => {
                if (typeof letter !== 'string')
                    throw new Error(`alphabet.decode: not string element=${letter}`);
                const index = alphabet.indexOf(letter);
                if (index === -1)
                    throw new Error(`Unknown letter: "${letter}". Allowed: ${alphabet}`);
                return index;
            });
        },
    };
}
/**
 * @__NO_SIDE_EFFECTS__
 */
function join(separator = '') {
    if (typeof separator !== 'string')
        throw new Error('join separator should be string');
    return {
        encode: (from) => {
            if (!Array.isArray(from) || (from.length && typeof from[0] !== 'string'))
                throw new Error('join.encode input should be array of strings');
            for (let i of from)
                if (typeof i !== 'string')
                    throw new Error(`join.encode: non-string input=${i}`);
            return from.join(separator);
        },
        decode: (to) => {
            if (typeof to !== 'string')
                throw new Error('join.decode input should be string');
            return to.split(separator);
        },
    };
}
/**
 * Pad strings array so it has integer number of bits
 * @__NO_SIDE_EFFECTS__
 */
function padding(bits, chr = '=') {
    assertNumber(bits);
    if (typeof chr !== 'string')
        throw new Error('padding chr should be string');
    return {
        encode(data) {
            if (!Array.isArray(data) || (data.length && typeof data[0] !== 'string'))
                throw new Error('padding.encode input should be array of strings');
            for (let i of data)
                if (typeof i !== 'string')
                    throw new Error(`padding.encode: non-string input=${i}`);
            while ((data.length * bits) % 8)
                data.push(chr);
            return data;
        },
        decode(input) {
            if (!Array.isArray(input) || (input.length && typeof input[0] !== 'string'))
                throw new Error('padding.encode input should be array of strings');
            for (let i of input)
                if (typeof i !== 'string')
                    throw new Error(`padding.decode: non-string input=${i}`);
            let end = input.length;
            if ((end * bits) % 8)
                throw new Error('Invalid padding: string should have whole number of bytes');
            for (; end > 0 && input[end - 1] === chr; end--) {
                if (!(((end - 1) * bits) % 8))
                    throw new Error('Invalid padding: string has too much padding');
            }
            return input.slice(0, end);
        },
    };
}
/**
 * @__NO_SIDE_EFFECTS__
 */
function normalize(fn) {
    if (typeof fn !== 'function')
        throw new Error('normalize fn should be function');
    return { encode: (from) => from, decode: (to) => fn(to) };
}
const gcd = /* @__NO_SIDE_EFFECTS__ */ (a, b) => (!b ? a : gcd(b, a % b));
const radix2carry = /*@__NO_SIDE_EFFECTS__ */ (from, to) => from + (to - gcd(from, to));
/**
 * Implemented with numbers, because BigInt is 5x slower
 * @__NO_SIDE_EFFECTS__
 */
function convertRadix2(data, from, to, padding) {
    if (!Array.isArray(data))
        throw new Error('convertRadix2: data should be array');
    if (from <= 0 || from > 32)
        throw new Error(`convertRadix2: wrong from=${from}`);
    if (to <= 0 || to > 32)
        throw new Error(`convertRadix2: wrong to=${to}`);
    if (radix2carry(from, to) > 32) {
        throw new Error(`convertRadix2: carry overflow from=${from} to=${to} carryBits=${radix2carry(from, to)}`);
    }
    let carry = 0;
    let pos = 0; // bitwise position in current element
    const mask = 2 ** to - 1;
    const res = [];
    for (const n of data) {
        assertNumber(n);
        if (n >= 2 ** from)
            throw new Error(`convertRadix2: invalid data word=${n} from=${from}`);
        carry = (carry << from) | n;
        if (pos + from > 32)
            throw new Error(`convertRadix2: carry overflow pos=${pos} from=${from}`);
        pos += from;
        for (; pos >= to; pos -= to)
            res.push(((carry >> (pos - to)) & mask) >>> 0);
        carry &= 2 ** pos - 1; // clean carry, otherwise it will cause overflow
    }
    carry = (carry << (to - pos)) & mask;
    if (!padding && pos >= from)
        throw new Error('Excess padding');
    if (!padding && carry)
        throw new Error(`Non-zero padding: ${carry}`);
    if (padding && pos > 0)
        res.push(carry >>> 0);
    return res;
}
/**
 * If both bases are power of same number (like `2**8 <-> 2**64`),
 * there is a linear algorithm. For now we have implementation for power-of-two bases only.
 * @__NO_SIDE_EFFECTS__
 */
function radix2(bits, revPadding = false) {
    assertNumber(bits);
    if (bits <= 0 || bits > 32)
        throw new Error('radix2: bits should be in (0..32]');
    if (radix2carry(8, bits) > 32 || radix2carry(bits, 8) > 32)
        throw new Error('radix2: carry overflow');
    return {
        encode: (bytes) => {
            if (!isBytes(bytes))
                throw new Error('radix2.encode input should be Uint8Array');
            return convertRadix2(Array.from(bytes), 8, bits, !revPadding);
        },
        decode: (digits) => {
            if (!Array.isArray(digits) || (digits.length && typeof digits[0] !== 'number'))
                throw new Error('radix2.decode input should be array of numbers');
            return Uint8Array.from(convertRadix2(digits, bits, 8, revPadding));
        },
    };
}
/**
 * @__NO_SIDE_EFFECTS__
 */
function unsafeWrapper(fn) {
    if (typeof fn !== 'function')
        throw new Error('unsafeWrapper fn should be function');
    return function (...args) {
        try {
            return fn.apply(null, args);
        }
        catch (e) { }
    };
}
// RFC 4648 aka RFC 3548
// ---------------------
/* @__PURE__ */ chain(radix2(4), alphabet('0123456789ABCDEF'), join(''));
/* @__PURE__ */ chain(radix2(5), alphabet('ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'), padding(5), join(''));
/* @__PURE__ */ chain(radix2(5), alphabet('ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'), join(''));
/* @__PURE__ */ chain(radix2(5), alphabet('0123456789ABCDEFGHIJKLMNOPQRSTUV'), padding(5), join(''));
/* @__PURE__ */ chain(radix2(5), alphabet('0123456789ABCDEFGHIJKLMNOPQRSTUV'), join(''));
/* @__PURE__ */ chain(radix2(5), alphabet('0123456789ABCDEFGHJKMNPQRSTVWXYZ'), join(''), normalize((s) => s.toUpperCase().replace(/O/g, '0').replace(/[IL]/g, '1')));
/* @__PURE__ */ chain(radix2(6), alphabet('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'), padding(6), join(''));
/* @__PURE__ */ chain(radix2(6), alphabet('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'), join(''));
/* @__PURE__ */ chain(radix2(6), alphabet('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_'), padding(6), join(''));
/* @__PURE__ */ chain(radix2(6), alphabet('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_'), join(''));
const BECH_ALPHABET = /* @__PURE__ */ chain(alphabet('qpzry9x8gf2tvdw0s3jn54khce6mua7l'), join(''));
const POLYMOD_GENERATORS = [0x3b6a57b2, 0x26508e6d, 0x1ea119fa, 0x3d4233dd, 0x2a1462b3];
/**
 * @__NO_SIDE_EFFECTS__
 */
function bech32Polymod(pre) {
    const b = pre >> 25;
    let chk = (pre & 0x1ffffff) << 5;
    for (let i = 0; i < POLYMOD_GENERATORS.length; i++) {
        if (((b >> i) & 1) === 1)
            chk ^= POLYMOD_GENERATORS[i];
    }
    return chk;
}
/**
 * @__NO_SIDE_EFFECTS__
 */
function bechChecksum(prefix, words, encodingConst = 1) {
    const len = prefix.length;
    let chk = 1;
    for (let i = 0; i < len; i++) {
        const c = prefix.charCodeAt(i);
        if (c < 33 || c > 126)
            throw new Error(`Invalid prefix (${prefix})`);
        chk = bech32Polymod(chk) ^ (c >> 5);
    }
    chk = bech32Polymod(chk);
    for (let i = 0; i < len; i++)
        chk = bech32Polymod(chk) ^ (prefix.charCodeAt(i) & 0x1f);
    for (let v of words)
        chk = bech32Polymod(chk) ^ v;
    for (let i = 0; i < 6; i++)
        chk = bech32Polymod(chk);
    chk ^= encodingConst;
    return BECH_ALPHABET.encode(convertRadix2([chk % 2 ** 30], 30, 5, false));
}
/**
 * @__NO_SIDE_EFFECTS__
 */
function genBech32(encoding) {
    const ENCODING_CONST = encoding === 'bech32' ? 1 : 0x2bc830a3;
    const _words = radix2(5);
    const fromWords = _words.decode;
    const toWords = _words.encode;
    const fromWordsUnsafe = unsafeWrapper(fromWords);
    function encode(prefix, words, limit = 90) {
        if (typeof prefix !== 'string')
            throw new Error(`bech32.encode prefix should be string, not ${typeof prefix}`);
        if (!Array.isArray(words) || (words.length && typeof words[0] !== 'number'))
            throw new Error(`bech32.encode words should be array of numbers, not ${typeof words}`);
        if (prefix.length === 0)
            throw new TypeError(`Invalid prefix length ${prefix.length}`);
        const actualLength = prefix.length + 7 + words.length;
        if (limit !== false && actualLength > limit)
            throw new TypeError(`Length ${actualLength} exceeds limit ${limit}`);
        const lowered = prefix.toLowerCase();
        const sum = bechChecksum(lowered, words, ENCODING_CONST);
        return `${lowered}1${BECH_ALPHABET.encode(words)}${sum}`;
    }
    function decode(str, limit = 90) {
        if (typeof str !== 'string')
            throw new Error(`bech32.decode input should be string, not ${typeof str}`);
        if (str.length < 8 || (limit !== false && str.length > limit))
            throw new TypeError(`Wrong string length: ${str.length} (${str}). Expected (8..${limit})`);
        // don't allow mixed case
        const lowered = str.toLowerCase();
        if (str !== lowered && str !== str.toUpperCase())
            throw new Error(`String must be lowercase or uppercase`);
        const sepIndex = lowered.lastIndexOf('1');
        if (sepIndex === 0 || sepIndex === -1)
            throw new Error(`Letter "1" must be present between prefix and data only`);
        const prefix = lowered.slice(0, sepIndex);
        const data = lowered.slice(sepIndex + 1);
        if (data.length < 6)
            throw new Error('Data must be at least 6 characters long');
        const words = BECH_ALPHABET.decode(data).slice(0, -6);
        const sum = bechChecksum(prefix, words, ENCODING_CONST);
        if (!data.endsWith(sum))
            throw new Error(`Invalid checksum in ${str}: expected "${sum}"`);
        return { prefix, words };
    }
    const decodeUnsafe = unsafeWrapper(decode);
    function decodeToBytes(str) {
        const { prefix, words } = decode(str, false);
        return { prefix, words, bytes: fromWords(words) };
    }
    return { encode, decode, decodeToBytes, decodeUnsafe, fromWords, fromWordsUnsafe, toWords };
}
const bech32 = /* @__PURE__ */ genBech32('bech32');
/* @__PURE__ */ chain(radix2(4), alphabet('0123456789abcdef'), join(''), normalize((s) => {
    if (typeof s !== 'string' || s.length % 2)
        throw new TypeError(`hex.decode: expected string, got ${typeof s} with length ${s.length}`);
    return s.toLowerCase();
}));

const u8a = {
  toString,
  fromString,
  concat: concat$1
};
function bytesToBase64url(b) {
  return u8a.toString(b, 'base64url');
}
function base64ToBytes(s) {
  const inputBase64Url = s.replace(/\+/g, '-').replace(/\//g, '_').replace(/=/g, '');
  return u8a.fromString(inputBase64Url, 'base64url');
}
function base58ToBytes(s) {
  return u8a.fromString(s, 'base58btc');
}
function bytesToBase58(b) {
  return u8a.toString(b, 'base58btc');
}
const SUPPORTED_PUBLIC_KEY_TYPES = {
  ES256: ['JsonWebKey2020', 'Multikey', 'EcdsaSecp256r1VerificationKey2019'],
  ES256K: ['EcdsaSecp256k1VerificationKey2019',
  /**
   * Equivalent to EcdsaSecp256k1VerificationKey2019 when key is an ethereumAddress
   */
  'EcdsaSecp256k1RecoveryMethod2020',
  /**
   * @deprecated, supported for backward compatibility. Equivalent to EcdsaSecp256k1VerificationKey2019 when key is
   *   not an ethereumAddress
   */
  'Secp256k1VerificationKey2018',
  /**
   * @deprecated, supported for backward compatibility. Equivalent to EcdsaSecp256k1VerificationKey2019 when key is
   *   not an ethereumAddress
   */
  'Secp256k1SignatureVerificationKey2018',
  /**
   * @deprecated, supported for backward compatibility. Equivalent to EcdsaSecp256k1VerificationKey2019 when key is
   *   not an ethereumAddress
   */
  'EcdsaPublicKeySecp256k1',
  /**
   *  TODO - support R1 key as well
   *   'ConditionalProof2022',
   */
  'JsonWebKey2020', 'Multikey'],
  'ES256K-R': ['EcdsaSecp256k1VerificationKey2019',
  /**
   * Equivalent to EcdsaSecp256k1VerificationKey2019 when key is an ethereumAddress
   */
  'EcdsaSecp256k1RecoveryMethod2020',
  /**
   * @deprecated, supported for backward compatibility. Equivalent to EcdsaSecp256k1VerificationKey2019 when key is
   *   not an ethereumAddress
   */
  'Secp256k1VerificationKey2018',
  /**
   * @deprecated, supported for backward compatibility. Equivalent to EcdsaSecp256k1VerificationKey2019 when key is
   *   not an ethereumAddress
   */
  'Secp256k1SignatureVerificationKey2018',
  /**
   * @deprecated, supported for backward compatibility. Equivalent to EcdsaSecp256k1VerificationKey2019 when key is
   *   not an ethereumAddress
   */
  'EcdsaPublicKeySecp256k1', 'ConditionalProof2022', 'JsonWebKey2020', 'Multikey'],
  Ed25519: ['ED25519SignatureVerification', 'Ed25519VerificationKey2018', 'Ed25519VerificationKey2020', 'JsonWebKey2020', 'Multikey'],
  EdDSA: ['ED25519SignatureVerification', 'Ed25519VerificationKey2018', 'Ed25519VerificationKey2020', 'JsonWebKey2020', 'Multikey']
};
const VM_TO_KEY_TYPE = {
  Secp256k1SignatureVerificationKey2018: 'Secp256k1',
  Secp256k1VerificationKey2018: 'Secp256k1',
  EcdsaSecp256k1VerificationKey2019: 'Secp256k1',
  EcdsaPublicKeySecp256k1: 'Secp256k1',
  EcdsaSecp256k1RecoveryMethod2020: 'Secp256k1',
  EcdsaSecp256r1VerificationKey2019: 'P-256',
  Ed25519VerificationKey2018: 'Ed25519',
  Ed25519VerificationKey2020: 'Ed25519',
  ED25519SignatureVerification: 'Ed25519',
  X25519KeyAgreementKey2019: 'X25519',
  X25519KeyAgreementKey2020: 'X25519',
  ConditionalProof2022: undefined,
  JsonWebKey2020: undefined,
  Multikey: undefined // key type must be extracted from the multicodec

}; // this is from the multicodec table https://github.com/multiformats/multicodec/blob/master/table.csv

const supportedCodecs = {
  'ed25519-pub': 0xed,
  'x25519-pub': 0xec,
  'secp256k1-pub': 0xe7,
  'bls12_381-g1-pub': 0xea,
  'bls12_381-g2-pub': 0xeb,
  'p256-pub': 0x1200
};
const CODEC_TO_KEY_TYPE = {
  'bls12_381-g1-pub': 'Bls12381G1',
  'bls12_381-g2-pub': 'Bls12381G2',
  'ed25519-pub': 'Ed25519',
  'p256-pub': 'P-256',
  'secp256k1-pub': 'Secp256k1',
  'x25519-pub': 'X25519'
};
/**
 * Extracts the raw byte representation of a public key from a VerificationMethod along with an inferred key type
 * @param pk a VerificationMethod entry from a DIDDocument
 * @return an object containing the `keyBytes` of the public key and an inferred `keyType`
 */

function extractPublicKeyBytes(pk) {
  if (pk.publicKeyBase58) {
    return {
      keyBytes: base58ToBytes(pk.publicKeyBase58),
      keyType: VM_TO_KEY_TYPE[pk.type]
    };
  } else if (pk.publicKeyBase64) {
    return {
      keyBytes: base64ToBytes(pk.publicKeyBase64),
      keyType: VM_TO_KEY_TYPE[pk.type]
    };
  } else if (pk.publicKeyHex) {
    return {
      keyBytes: hexToBytes(pk.publicKeyHex),
      keyType: VM_TO_KEY_TYPE[pk.type]
    };
  } else if (pk.publicKeyJwk && pk.publicKeyJwk.crv === 'secp256k1' && pk.publicKeyJwk.x && pk.publicKeyJwk.y) {
    return {
      keyBytes: secp256k1.ProjectivePoint.fromAffine({
        x: bytesToBigInt(base64ToBytes(pk.publicKeyJwk.x)),
        y: bytesToBigInt(base64ToBytes(pk.publicKeyJwk.y))
      }).toRawBytes(false),
      keyType: 'Secp256k1'
    };
  } else if (pk.publicKeyJwk && pk.publicKeyJwk.crv === 'P-256' && pk.publicKeyJwk.x && pk.publicKeyJwk.y) {
    return {
      keyBytes: p256.ProjectivePoint.fromAffine({
        x: bytesToBigInt(base64ToBytes(pk.publicKeyJwk.x)),
        y: bytesToBigInt(base64ToBytes(pk.publicKeyJwk.y))
      }).toRawBytes(false),
      keyType: 'P-256'
    };
  } else if (pk.publicKeyJwk && pk.publicKeyJwk.kty === 'OKP' && ['Ed25519', 'X25519'].includes(pk.publicKeyJwk.crv ?? '') && pk.publicKeyJwk.x) {
    return {
      keyBytes: base64ToBytes(pk.publicKeyJwk.x),
      keyType: pk.publicKeyJwk.crv
    };
  } else if (pk.publicKeyMultibase) {
    const {
      keyBytes,
      keyType
    } = multibaseToBytes(pk.publicKeyMultibase);
    return {
      keyBytes,
      keyType: keyType ?? VM_TO_KEY_TYPE[pk.type]
    };
  }

  return {
    keyBytes: new Uint8Array()
  };
}
/**
 * Converts a multibase string to the Uint8Array it represents.
 * This method will assume the byte array that is multibase encoded is a multicodec and will attempt to decode it.
 *
 * @param s - the string to be converted
 *
 * @throws if the string is not formatted correctly.
 *
 * @public
 */

function multibaseToBytes(s) {
  const bytes = src.decode(s); // look for known key lengths first
  // Ed25519/X25519, secp256k1/P256 compressed or not, BLS12-381 G1/G2 compressed

  if ([32, 33, 48, 64, 65, 96].includes(bytes.length)) {
    return {
      keyBytes: bytes
    };
  } // then assume multicodec, otherwise return the bytes


  try {
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    const [codec, length] = decode$1(bytes);
    const possibleCodec = Object.entries(supportedCodecs).filter(([, code]) => code === codec)?.[0][0] ?? '';
    return {
      keyBytes: bytes.slice(length),
      keyType: CODEC_TO_KEY_TYPE[possibleCodec]
    };
  } catch (e) {
    // not a multicodec, return the bytes
    return {
      keyBytes: bytes
    };
  }
}
function hexToBytes(s, minLength) {
  let input = s.startsWith('0x') ? s.substring(2) : s;

  if (input.length % 2 !== 0) {
    input = `0${input}`;
  }

  if (minLength) {
    const paddedLength = Math.max(input.length, minLength * 2);
    input = input.padStart(paddedLength, '00');
  }

  return u8a.fromString(input.toLowerCase(), 'base16');
}
function encodeBase64url(s) {
  return bytesToBase64url(u8a.fromString(s));
}
function decodeBase64url(s) {
  return u8a.toString(base64ToBytes(s));
}
function bytesToHex(b) {
  return u8a.toString(b, 'base16');
}
function bytesToBigInt(b) {
  return BigInt(`0x` + u8a.toString(b, 'base16'));
}
function stringToBytes(s) {
  return u8a.fromString(s, 'utf-8');
}
function toJose({
  r,
  s,
  recoveryParam
}, recoverable) {
  const jose = new Uint8Array(recoverable ? 65 : 64);
  jose.set(u8a.fromString(r, 'base16'), 0);
  jose.set(u8a.fromString(s, 'base16'), 32);

  if (recoverable) {
    if (typeof recoveryParam === 'undefined') {
      throw new Error('Signer did not return a recoveryParam');
    }

    jose[64] = recoveryParam;
  }

  return bytesToBase64url(jose);
}
function fromJose(signature) {
  const signatureBytes = base64ToBytes(signature);

  if (signatureBytes.length < 64 || signatureBytes.length > 65) {
    throw new TypeError(`Wrong size for signature. Expected 64 or 65 bytes, but got ${signatureBytes.length}`);
  }

  const r = bytesToHex(signatureBytes.slice(0, 32));
  const s = bytesToHex(signatureBytes.slice(32, 64));
  const recoveryParam = signatureBytes.length === 65 ? signatureBytes[64] : undefined;
  return {
    r,
    s,
    recoveryParam
  };
}
function leftpad(data, size = 64) {
  if (data.length === size) return data;
  return '0'.repeat(size - data.length) + data;
}

function sha256$1(payload) {
  const data = typeof payload === 'string' ? fromString(payload) : payload;
  return sha256$2(data);
}
const keccak = keccak_256;
function toEthereumAddress(hexPublicKey) {
  const hashInput = fromString(hexPublicKey.slice(2), 'base16');
  return `0x${toString(keccak(hashInput).slice(-20), 'base16')}`;
}

/**
 *  Creates a configured signer function for signing data using the ES256K (secp256k1 + sha256) algorithm.
 *
 *  The signing function itself takes the data as a `Uint8Array` or `string` and returns a `base64Url`-encoded signature
 *
 *  @example
 *  ```typescript
 *  const sign: Signer = ES256KSigner(process.env.PRIVATE_KEY)
 *  const signature: string = await sign(data)
 *  ```
 *
 *  @param    {String}    privateKey   a private key as `Uint8Array`
 *  @param    {Boolean}   recoverable  an optional flag to add the recovery param to the generated signatures
 *  @return   {Function}               a configured signer function `(data: string | Uint8Array): Promise<string>`
 */

function ES256KSigner(privateKey, recoverable = false) {
  const privateKeyBytes = privateKey;

  if (privateKeyBytes.length !== 32) {
    throw new Error(`bad_key: Invalid private key format. Expecting 32 bytes, but got ${privateKeyBytes.length}`);
  }

  return function (data) {
    try {
      const signature = secp256k1.sign(sha256$1(data), privateKeyBytes);
      return Promise.resolve(toJose({
        r: leftpad(signature.r.toString(16)),
        s: leftpad(signature.s.toString(16)),
        recoveryParam: signature.recovery
      }, recoverable));
    } catch (e) {
      return Promise.reject(e);
    }
  };
}

/**
 *  Creates a configured signer function for signing data using the ES256 (secp256r1 + sha256) algorithm.
 *
 *  The signing function itself takes the data as a `Uint8Array` or `string` and returns a `base64Url`-encoded signature
 *
 *  @example
 *  ```typescript
 *  const sign: Signer = ES256Signer(process.env.PRIVATE_KEY)
 *  const signature: string = await sign(data)
 *  ```
 *
 *  @param    {String}    privateKey   a private key as `Uint8Array`
 *  @return   {Function}               a configured signer function `(data: string | Uint8Array): Promise<string>`
 */

function ES256Signer(privateKey) {
  if (privateKey.length !== 32) {
    throw new Error(`bad_key: Invalid private key format. Expecting 32 bytes, but got ${privateKey.length}`);
  }

  return function (data) {
    try {
      const signature = p256.sign(sha256$1(data), privateKey);
      return Promise.resolve(toJose({
        r: leftpad(signature.r.toString(16)),
        s: leftpad(signature.s.toString(16))
      }));
    } catch (e) {
      return Promise.reject(e);
    }
  };
}

function instanceOfEcdsaSignature(object) {
  return typeof object === 'object' && 'r' in object && 's' in object;
}

function ES256SignerAlg() {
  return function sign(payload, signer) {
    try {
      return Promise.resolve(signer(payload)).then(function (signature) {
        if (instanceOfEcdsaSignature(signature)) {
          return toJose(signature);
        } else {
          return signature;
        }
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };
}
function ES256KSignerAlg(recoverable) {
  return function sign(payload, signer) {
    try {
      return Promise.resolve(signer(payload)).then(function (signature) {
        if (instanceOfEcdsaSignature(signature)) {
          return toJose(signature, recoverable);
        } else {
          if (recoverable && typeof fromJose(signature).recoveryParam === 'undefined') {
            throw new Error(`not_supported: ES256K-R not supported when signer doesn't provide a recovery param`);
          }

          return signature;
        }
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };
}
function Ed25519SignerAlg() {
  return function sign(payload, signer) {
    try {
      return Promise.resolve(signer(payload)).then(function (signature) {
        if (!instanceOfEcdsaSignature(signature)) {
          return signature;
        } else {
          throw new Error('invalid_config: expected a signer function that returns a string instead of signature object');
        }
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };
}
const algorithms$1 = {
  ES256: ES256SignerAlg(),
  ES256K: ES256KSignerAlg(),
  // This is a non-standard algorithm but retained for backwards compatibility
  // see https://github.com/decentralized-identity/did-jwt/issues/146
  'ES256K-R': ES256KSignerAlg(true),
  // This is actually incorrect but retained for backwards compatibility
  // see https://github.com/decentralized-identity/did-jwt/issues/130
  Ed25519: Ed25519SignerAlg(),
  EdDSA: Ed25519SignerAlg()
};

function SignerAlg(alg) {
  const impl = algorithms$1[alg];
  if (!impl) throw new Error(`not_supported: Unsupported algorithm ${alg}`);
  return impl;
}

function publicKeyToAddress$1(publicKey, otherAddress) {
  // Use the same version/prefix byte as the given address.
  const version = bytesToHex(base58ToBytes(otherAddress).slice(0, 1));
  const publicKeyBuffer = hexToBytes(publicKey);
  const publicKeyHash = ripemd160(sha256$1(publicKeyBuffer));
  const step1 = version + bytesToHex(publicKeyHash);
  const step2 = sha256$1(hexToBytes(step1));
  const step3 = sha256$1(step2);
  const checksum = bytesToHex(step3).substring(0, 8);
  const step4 = step1 + checksum;
  return bytesToBase58(hexToBytes(step4));
}

function publicKeyToAddress(publicKey, prefix) {
  const publicKeyBuffer = secp256k1.ProjectivePoint.fromHex(publicKey).toRawBytes();
  const hash = ripemd160(sha256$1(publicKeyBuffer));
  const words = bech32.toWords(hash);
  return bech32.encode(prefix, words).replace(prefix, '');
}

function verifyBlockchainAccountId(publicKey, blockchainAccountId) {
  if (blockchainAccountId) {
    const chain = blockchainAccountId.split(':');

    switch (chain[0]) {
      case 'bip122':
        chain[chain.length - 1] = publicKeyToAddress$1(publicKey, chain[chain.length - 1]);
        break;

      case 'cosmos':
        chain[chain.length - 1] = publicKeyToAddress(publicKey, chain[1]);
        break;

      case 'eip155':
        chain[chain.length - 1] = toEthereumAddress(publicKey);
        break;

      default:
        return false;
    }

    return chain.join(':').toLowerCase() === blockchainAccountId.toLowerCase();
  }

  return false;
}

function toSignatureObject(signature, recoverable = false) {
  const rawSig = base64ToBytes(signature);

  if (rawSig.length !== (recoverable ? 65 : 64)) {
    throw new Error('wrong signature length');
  }

  const r = bytesToHex(rawSig.slice(0, 32));
  const s = bytesToHex(rawSig.slice(32, 64));
  const sigObj = {
    r,
    s
  };

  if (recoverable) {
    sigObj.recoveryParam = rawSig[64];
  }

  return sigObj;
}
function toSignatureObject2(signature, recoverable = false) {
  const bytes = base64ToBytes(signature);

  if (bytes.length !== (recoverable ? 65 : 64)) {
    throw new Error('wrong signature length');
  }

  return {
    compact: bytes.slice(0, 64),
    recovery: bytes[64]
  };
}
function verifyES256(data, signature, authenticators) {
  const hash = sha256$1(data);
  const sig = p256.Signature.fromCompact(toSignatureObject2(signature).compact);
  const fullPublicKeys = authenticators.filter(a => !a.ethereumAddress && !a.blockchainAccountId);
  const signer = fullPublicKeys.find(pk => {
    try {
      const {
        keyBytes
      } = extractPublicKeyBytes(pk);
      return p256.verify(sig, hash, keyBytes);
    } catch (err) {
      return false;
    }
  });
  if (!signer) throw new Error('invalid_signature: Signature invalid for JWT');
  return signer;
}
function verifyES256K(data, signature, authenticators) {
  const hash = sha256$1(data);
  const signatureNormalized = secp256k1.Signature.fromCompact(base64ToBytes(signature)).normalizeS();
  const fullPublicKeys = authenticators.filter(a => {
    return !a.ethereumAddress && !a.blockchainAccountId;
  });
  const blockchainAddressKeys = authenticators.filter(a => {
    return a.ethereumAddress || a.blockchainAccountId;
  });
  let signer = fullPublicKeys.find(pk => {
    try {
      const {
        keyBytes
      } = extractPublicKeyBytes(pk);
      return secp256k1.verify(signatureNormalized, hash, keyBytes);
    } catch (err) {
      return false;
    }
  });

  if (!signer && blockchainAddressKeys.length > 0) {
    signer = verifyRecoverableES256K(data, signature, blockchainAddressKeys);
  }

  if (!signer) throw new Error('invalid_signature: Signature invalid for JWT');
  return signer;
}
function verifyRecoverableES256K(data, signature, authenticators) {
  const signatures = [];

  if (signature.length > 86) {
    signatures.push(toSignatureObject2(signature, true));
  } else {
    const so = toSignatureObject2(signature, false);
    signatures.push({ ...so,
      recovery: 0
    });
    signatures.push({ ...so,
      recovery: 1
    });
  }

  const hash = sha256$1(data);

  const checkSignatureAgainstSigner = sigObj => {
    const signature = secp256k1.Signature.fromCompact(sigObj.compact).addRecoveryBit(sigObj.recovery || 0);
    const recoveredPublicKey = signature.recoverPublicKey(hash);
    const recoveredAddress = toEthereumAddress(recoveredPublicKey.toHex(false)).toLowerCase();
    const recoveredPublicKeyHex = recoveredPublicKey.toHex(false);
    const recoveredCompressedPublicKeyHex = recoveredPublicKey.toHex(true);
    return authenticators.find(a => {
      const {
        keyBytes
      } = extractPublicKeyBytes(a);
      const keyHex = bytesToHex(keyBytes);
      return keyHex === recoveredPublicKeyHex || keyHex === recoveredCompressedPublicKeyHex || a.ethereumAddress?.toLowerCase() === recoveredAddress || a.blockchainAccountId?.split('@eip155')?.[0].toLowerCase() === recoveredAddress || // CAIP-2
      verifyBlockchainAccountId(recoveredPublicKeyHex, a.blockchainAccountId) // CAIP-10
      ;
    });
  }; // Find first verification method


  for (const signature of signatures) {
    const verificationMethod = checkSignatureAgainstSigner(signature);
    if (verificationMethod) return verificationMethod;
  } // If no one found matching


  throw new Error('invalid_signature: Signature invalid for JWT');
}
function verifyEd25519(data, signature, authenticators) {
  const clear = stringToBytes(data);
  const signatureBytes = base64ToBytes(signature);
  const signer = authenticators.find(a => {
    const {
      keyBytes,
      keyType
    } = extractPublicKeyBytes(a);

    if (keyType === 'Ed25519') {
      return ed25519.verify(signatureBytes, clear, keyBytes);
    } else {
      return false;
    }
  });
  if (!signer) throw new Error('invalid_signature: Signature invalid for JWT');
  return signer;
}
const algorithms = {
  ES256: verifyES256,
  ES256K: verifyES256K,
  // This is a non-standard algorithm but retained for backwards compatibility
  // see https://github.com/decentralized-identity/did-jwt/issues/146
  'ES256K-R': verifyRecoverableES256K,
  // This is actually incorrect but retained for backwards compatibility
  // see https://github.com/decentralized-identity/did-jwt/issues/130
  Ed25519: verifyEd25519,
  EdDSA: verifyEd25519
};

function VerifierAlgorithm(alg) {
  const impl = algorithms[alg];
  if (!impl) throw new Error(`not_supported: Unsupported algorithm ${alg}`);
  return impl;
}

VerifierAlgorithm.toSignatureObject = toSignatureObject;

/**
 * Error prefixes used for known verification failure cases.
 *
 * For compatibility, these error prefixes match the existing error messages, but will be adjusted in a future major
 * version update to match the scenarios better.
 *
 * @beta
 */
const JWT_ERROR = {
  /**
   * Thrown when a JWT payload schema is unexpected or when validity period does not match
   */
  INVALID_JWT: 'invalid_jwt',

  /**
   * Thrown when the verifier audience does not match the one set in the JWT payload
   */
  INVALID_AUDIENCE: 'invalid_config',

  /**
   * Thrown when none of the public keys of the issuer match the signature of the JWT.
   *
   * This is equivalent to `NO_SUITABLE_KEYS` when the `proofPurpose` is NOT specified.
   */
  INVALID_SIGNATURE: 'invalid_signature',

  /**
   * Thrown when the DID document of the issuer does not have any keys that match the signature for the given
   * `proofPurpose`.
   *
   * This is equivalent to `invalid_signature`, when a `proofPurpose` is specified.
   */
  NO_SUITABLE_KEYS: 'no_suitable_keys',

  /**
   * Thrown when the `alg` of the JWT or the encoding of the key is not supported
   */
  NOT_SUPPORTED: 'not_supported',

  /**
   * Thrown when the DID resolver is unable to resolve the issuer DID.
   */
  RESOLVER_ERROR: 'resolver_error'
};

function _catch$1(body, recover) {
  try {
    var result = body();
  } catch (e) {
    return recover(e);
  }

  if (result && result.then) {
    return result.then(void 0, recover);
  }

  return result;
}

const verifyConditionDelegated = function (jwt, {
  header,
  payload,
  data,
  signature
}, authenticator, options) {
  try {
    if (!authenticator.conditionDelegated) {
      throw new Error('Expected conditionDelegated');
    }

    if (!options.resolver) {
      throw new Error('Expected resolver');
    }

    let foundSigner;
    const issuer = authenticator.conditionDelegated;
    return Promise.resolve(resolveAuthenticator(options.resolver, header.alg, issuer, options.proofPurpose)).then(function (didAuthenticator) {
      let _exit2;

      function _temp6(_result4) {
        if (_exit2) ;

        if (foundSigner) {
          return authenticator;
        }

        throw new Error(`${JWT_ERROR.INVALID_SIGNATURE}: condition for authenticator ${authenticator.id} is not met.`);
      }

      const didResolutionResult = didAuthenticator.didResolutionResult;

      if (!didResolutionResult?.didDocument) {
        throw new Error(`${JWT_ERROR.RESOLVER_ERROR}: Could not resolve delegated DID ${issuer}.`);
      }

      const delegatedAuthenticator = didAuthenticator.authenticators.find(authenticator => authenticator.id === issuer);

      if (!delegatedAuthenticator) {
        throw new Error(`${JWT_ERROR.NO_SUITABLE_KEYS}: Could not find delegated authenticator ${issuer} in it's DID Document`);
      }

      const _temp5 = function () {
        if (delegatedAuthenticator.type === CONDITIONAL_PROOF_2022) {
          return Promise.resolve(verifyJWT(jwt, { ...options,
            ...{
              didAuthenticator: {
                didResolutionResult,
                authenticators: [delegatedAuthenticator],
                issuer: delegatedAuthenticator.id
              }
            }
          })).then(function ({
            verified
          }) {
            if (verified) {
              foundSigner = delegatedAuthenticator;
            }
          });
        } else {
          try {
            foundSigner = verifyJWTDecoded({
              header,
              payload,
              data,
              signature
            }, delegatedAuthenticator);
          } catch (e) {
            if (!e.message.startsWith('invalid_signature:')) throw e;
          }
        }
      }();

      return _temp5 && _temp5.then ? _temp5.then(_temp6) : _temp6(_temp5);
    });
  } catch (e) {
    return Promise.reject(e);
  }
};

const _iteratorSymbol$1 = typeof Symbol !== "undefined" ? Symbol.iterator || (Symbol.iterator = Symbol("Symbol.iterator")) : "@@iterator";

function _settle$2(pact, state, value) {
  if (!pact.s) {
    if (value instanceof _Pact$2) {
      if (value.s) {
        if (state & 1) {
          state = value.s;
        }

        value = value.v;
      } else {
        value.o = _settle$2.bind(null, pact, state);
        return;
      }
    }

    if (value && value.then) {
      value.then(_settle$2.bind(null, pact, state), _settle$2.bind(null, pact, 2));
      return;
    }

    pact.s = state;
    pact.v = value;
    const observer = pact.o;

    if (observer) {
      observer(pact);
    }
  }
}

const _Pact$2 = /*#__PURE__*/function () {
  function _Pact() {}

  _Pact.prototype.then = function (onFulfilled, onRejected) {
    const result = new _Pact();
    const state = this.s;

    if (state) {
      const callback = state & 1 ? onFulfilled : onRejected;

      if (callback) {
        try {
          _settle$2(result, 1, callback(this.v));
        } catch (e) {
          _settle$2(result, 2, e);
        }

        return result;
      } else {
        return this;
      }
    }

    this.o = function (_this) {
      try {
        const value = _this.v;

        if (_this.s & 1) {
          _settle$2(result, 1, onFulfilled ? onFulfilled(value) : value);
        } else if (onRejected) {
          _settle$2(result, 1, onRejected(value));
        } else {
          _settle$2(result, 2, value);
        }
      } catch (e) {
        _settle$2(result, 2, e);
      }
    };

    return result;
  };

  return _Pact;
}();

function _isSettledPact$2(thenable) {
  return thenable instanceof _Pact$2 && thenable.s & 1;
}

function _forTo$2(array, body, check) {
  var i = -1,
      pact,
      reject;

  function _cycle(result) {
    try {
      while (++i < array.length && (!check || !check())) {
        result = body(i);

        if (result && result.then) {
          if (_isSettledPact$2(result)) {
            result = result.v;
          } else {
            result.then(_cycle, reject || (reject = _settle$2.bind(null, pact = new _Pact$2(), 2)));
            return;
          }
        }
      }

      if (pact) {
        _settle$2(pact, 1, result);
      } else {
        pact = result;
      }
    } catch (e) {
      _settle$2(pact || (pact = new _Pact$2()), 2, e);
    }
  }

  _cycle();

  return pact;
}

const verifyConditionWeightedThreshold = function (jwt, {
  header,
  payload,
  data,
  signature
}, authenticator, options) {
  try {
    let _exit;

    function _temp4(_result3) {
      if (_exit) return _result3;
      throw new Error(`${JWT_ERROR.INVALID_SIGNATURE}: condition for authenticator ${authenticator.id} is not met.`);
    }

    if (!authenticator.conditionWeightedThreshold || !authenticator.threshold) {
      throw new Error('Expected conditionWeightedThreshold and threshold');
    }

    const issuers = [];
    const threshold = authenticator.threshold;
    let weightCount = 0;

    const _temp3 = _forOf$1(authenticator.conditionWeightedThreshold, function (weightedCondition) {
      function _temp2(_result2) {
        if (_exit) return _result2;

        if (foundSigner && !issuers.includes(foundSigner.id)) {
          issuers.push(foundSigner.id);
          weightCount += weightedCondition.weight;

          if (weightCount >= threshold) {
            _exit = 1;
            return authenticator;
          }
        }
      }

      const currentCondition = weightedCondition.condition;
      let foundSigner;

      const _temp = _catch$1(function () {
        if (currentCondition.type === CONDITIONAL_PROOF_2022) {
          if (!options.didAuthenticator) {
            throw new Error('Expected didAuthenticator');
          }

          const newOptions = { ...options,
            didAuthenticator: {
              didResolutionResult: options.didAuthenticator?.didResolutionResult,
              authenticators: [currentCondition],
              issuer: currentCondition.id
            }
          };
          return Promise.resolve(verifyJWT(jwt, newOptions)).then(function ({
            verified
          }) {
            if (verified) {
              foundSigner = currentCondition;
            }
          });
        } else {
          return Promise.resolve(verifyJWTDecoded({
            header,
            payload,
            data,
            signature
          }, currentCondition)).then(function (_verifyJWTDecoded) {
            foundSigner = _verifyJWTDecoded;
          });
        }
      }, function (e) {
        if (!e.message.startsWith(JWT_ERROR.INVALID_SIGNATURE)) throw e;
      });

      return _temp && _temp.then ? _temp.then(_temp2) : _temp2(_temp);
    }, function () {
      return _exit;
    });

    return Promise.resolve(_temp3 && _temp3.then ? _temp3.then(_temp4) : _temp4(_temp3));
  } catch (e) {
    return Promise.reject(e);
  }
};

function _forOf$1(target, body, check) {
  if (typeof target[_iteratorSymbol$1] === "function") {
    var iterator = target[_iteratorSymbol$1](),
        step,
        pact,
        reject;

    function _cycle(result) {
      try {
        while (!(step = iterator.next()).done && (!check || !check())) {
          result = body(step.value);

          if (result && result.then) {
            if (_isSettledPact$2(result)) {
              result = result.v;
            } else {
              result.then(_cycle, reject || (reject = _settle$2.bind(null, pact = new _Pact$2(), 2)));
              return;
            }
          }
        }

        if (pact) {
          _settle$2(pact, 1, result);
        } else {
          pact = result;
        }
      } catch (e) {
        _settle$2(pact || (pact = new _Pact$2()), 2, e);
      }
    }

    _cycle();

    if (iterator.return) {
      var _fixup = function (value) {
        try {
          if (!step.done) {
            iterator.return();
          }
        } catch (e) {}

        return value;
      };

      if (pact && pact.then) {
        return pact.then(_fixup, function (e) {
          throw _fixup(e);
        });
      }

      _fixup();
    }

    return pact;
  } // No support for Symbol.iterator


  if (!("length" in target)) {
    throw new TypeError("Object is not iterable");
  } // Handle live collections properly


  var values = [];

  for (var i = 0; i < target.length; i++) {
    values.push(target[i]);
  }

  return _forTo$2(values, function (i) {
    return body(values[i]);
  }, check);
}

const verifyConditionalProof = function (jwt, {
  header,
  payload,
  signature,
  data
}, authenticator, options) {
  try {
    // Validate the condition according to its condition property
    if (authenticator.conditionWeightedThreshold) {
      return verifyConditionWeightedThreshold(jwt, {
        header,
        payload,
        data,
        signature
      }, authenticator, options);
    } else if (authenticator.conditionDelegated) {
      return verifyConditionDelegated(jwt, {
        header,
        payload,
        data,
        signature
      }, authenticator, options);
    } // TODO other conditions


    throw new Error(`${JWT_ERROR.INVALID_JWT}: conditional proof type did not find condition for authenticator ${authenticator.id}.`);
  } catch (e) {
    return Promise.reject(e);
  }
};
const verifyProof = function (jwt, {
  header,
  payload,
  signature,
  data
}, authenticator, options) {
  try {
    if (authenticator.type === CONDITIONAL_PROOF_2022) {
      return verifyConditionalProof(jwt, {
        payload,
        header,
        signature,
        data
      }, authenticator, options);
    } else {
      return Promise.resolve(verifyJWTDecoded({
        header,
        payload,
        data,
        signature
      }, [authenticator]));
    }
  } catch (e) {
    return Promise.reject(e);
  }
};
const CONDITIONAL_PROOF_2022 = 'ConditionalProof2022';

/**
 * Resolves relevant public keys or other authenticating material used to verify signature from the DID document of
 * provided DID
 *
 *  @example
 *  ```ts
 *  resolveAuthenticator(resolver, 'ES256K', 'did:uport:2nQtiQG6Cgm1GYTBaaKAgr76uY7iSexUkqX').then(obj => {
 *      const payload = obj.payload
 *      const profile = obj.profile
 *      const jwt = obj.jwt
 *      // ...
 *  })
 *  ```
 *
 *  @param resolver - {Resolvable} a DID resolver function that can obtain the `DIDDocument` for the `issuer`
 *  @param alg - {String} a JWT algorithm
 *  @param issuer - {String} a Decentralized Identifier (DID) to lookup
 *  @param proofPurpose - {ProofPurposeTypes} *Optional* Use the verificationMethod linked in that section of the
 *   issuer DID document
 *  @return {Promise<DIDAuthenticator>} a promise which resolves with an object containing an array of authenticators
 *   or rejects with an error if none exist
 */
function _settle$1(pact, state, value) {
  if (!pact.s) {
    if (value instanceof _Pact$1) {
      if (value.s) {
        if (state & 1) {
          state = value.s;
        }

        value = value.v;
      } else {
        value.o = _settle$1.bind(null, pact, state);
        return;
      }
    }

    if (value && value.then) {
      value.then(_settle$1.bind(null, pact, state), _settle$1.bind(null, pact, 2));
      return;
    }

    pact.s = state;
    pact.v = value;
    const observer = pact.o;

    if (observer) {
      observer(pact);
    }
  }
}

const _Pact$1 = /*#__PURE__*/function () {
  function _Pact() {}

  _Pact.prototype.then = function (onFulfilled, onRejected) {
    const result = new _Pact();
    const state = this.s;

    if (state) {
      const callback = state & 1 ? onFulfilled : onRejected;

      if (callback) {
        try {
          _settle$1(result, 1, callback(this.v));
        } catch (e) {
          _settle$1(result, 2, e);
        }

        return result;
      } else {
        return this;
      }
    }

    this.o = function (_this) {
      try {
        const value = _this.v;

        if (_this.s & 1) {
          _settle$1(result, 1, onFulfilled ? onFulfilled(value) : value);
        } else if (onRejected) {
          _settle$1(result, 1, onRejected(value));
        } else {
          _settle$1(result, 2, value);
        }
      } catch (e) {
        _settle$1(result, 2, e);
      }
    };

    return result;
  };

  return _Pact;
}();

function _isSettledPact$1(thenable) {
  return thenable instanceof _Pact$1 && thenable.s & 1;
}

function _forTo$1(array, body, check) {
  var i = -1,
      pact,
      reject;

  function _cycle(result) {
    try {
      while (++i < array.length && (!check || !check())) {
        result = body(i);

        if (result && result.then) {
          if (_isSettledPact$1(result)) {
            result = result.v;
          } else {
            result.then(_cycle, reject || (reject = _settle$1.bind(null, pact = new _Pact$1(), 2)));
            return;
          }
        }
      }

      if (pact) {
        _settle$1(pact, 1, result);
      } else {
        pact = result;
      }
    } catch (e) {
      _settle$1(pact || (pact = new _Pact$1()), 2, e);
    }
  }

  _cycle();

  return pact;
}

function _catch(body, recover) {
  try {
    var result = body();
  } catch (e) {
    return recover(e);
  }

  if (result && result.then) {
    return result.then(void 0, recover);
  }

  return result;
}

function _for$1(test, update, body) {
  var stage;

  for (;;) {
    var shouldContinue = test();

    if (_isSettledPact$1(shouldContinue)) {
      shouldContinue = shouldContinue.v;
    }

    if (!shouldContinue) {
      return result;
    }

    if (shouldContinue.then) {
      stage = 0;
      break;
    }

    var result = body();

    if (result && result.then) {
      if (_isSettledPact$1(result)) {
        result = result.s;
      } else {
        stage = 1;
        break;
      }
    }

    if (update) {
      var updateValue = update();

      if (updateValue && updateValue.then && !_isSettledPact$1(updateValue)) {
        stage = 2;
        break;
      }
    }
  }

  var pact = new _Pact$1();

  var reject = _settle$1.bind(null, pact, 2);

  (stage === 0 ? shouldContinue.then(_resumeAfterTest) : stage === 1 ? result.then(_resumeAfterBody) : updateValue.then(_resumeAfterUpdate)).then(void 0, reject);
  return pact;

  function _resumeAfterBody(value) {
    result = value;

    do {
      if (update) {
        updateValue = update();

        if (updateValue && updateValue.then && !_isSettledPact$1(updateValue)) {
          updateValue.then(_resumeAfterUpdate).then(void 0, reject);
          return;
        }
      }

      shouldContinue = test();

      if (!shouldContinue || _isSettledPact$1(shouldContinue) && !shouldContinue.v) {
        _settle$1(pact, 1, result);

        return;
      }

      if (shouldContinue.then) {
        shouldContinue.then(_resumeAfterTest).then(void 0, reject);
        return;
      }

      result = body();

      if (_isSettledPact$1(result)) {
        result = result.v;
      }
    } while (!result || !result.then);

    result.then(_resumeAfterBody).then(void 0, reject);
  }

  function _resumeAfterTest(shouldContinue) {
    if (shouldContinue) {
      result = body();

      if (result && result.then) {
        result.then(_resumeAfterBody).then(void 0, reject);
      } else {
        _resumeAfterBody(result);
      }
    } else {
      _settle$1(pact, 1, result);
    }
  }

  function _resumeAfterUpdate() {
    if (shouldContinue = test()) {
      if (shouldContinue.then) {
        shouldContinue.then(_resumeAfterTest).then(void 0, reject);
      } else {
        _resumeAfterTest(shouldContinue);
      }
    } else {
      _settle$1(pact, 1, result);
    }
  }
}

const resolveAuthenticator = function (resolver, alg, issuer, proofPurpose) {
  try {
    const types = SUPPORTED_PUBLIC_KEY_TYPES[alg];

    if (!types || types.length === 0) {
      throw new Error(`${JWT_ERROR.NOT_SUPPORTED}: No supported signature types for algorithm ${alg}`);
    }

    let didResult;
    return Promise.resolve(resolver.resolve(issuer, {
      accept: DID_JSON
    })).then(function (result) {
      // support legacy resolvers that do not produce DIDResolutionResult
      if (Object.getOwnPropertyNames(result).indexOf('didDocument') === -1) {
        didResult = {
          didDocument: result,
          didDocumentMetadata: {},
          didResolutionMetadata: {
            contentType: DID_JSON
          }
        };
      } else {
        didResult = result;
      }

      if (didResult.didResolutionMetadata?.error || didResult.didDocument == null) {
        const {
          error,
          message
        } = didResult.didResolutionMetadata;
        throw new Error(`${JWT_ERROR.RESOLVER_ERROR}: Unable to resolve DID document for ${issuer}: ${error}, ${message || ''}`);
      }

      const getPublicKeyById = (verificationMethods, pubid) => {
        const filtered = verificationMethods.filter(({
          id
        }) => pubid === id);
        return filtered.length > 0 ? filtered[0] : null;
      };

      let publicKeysToCheck = [...(didResult?.didDocument?.verificationMethod || []), ...(didResult?.didDocument?.publicKey || [])];

      if (typeof proofPurpose === 'string') {
        // support legacy DID Documents that do not list assertionMethod
        if (proofPurpose.startsWith('assertion') && !Object.getOwnPropertyNames(didResult?.didDocument).includes('assertionMethod')) {
          didResult.didDocument = { ...didResult.didDocument
          };
          didResult.didDocument.assertionMethod = [...publicKeysToCheck.map(pk => pk.id)];
        }

        publicKeysToCheck = (didResult.didDocument[proofPurpose] || []).map(verificationMethod => {
          if (typeof verificationMethod === 'string') {
            return getPublicKeyById(publicKeysToCheck, verificationMethod);
          } else if (typeof verificationMethod.publicKey === 'string') {
            // this is a legacy format
            return getPublicKeyById(publicKeysToCheck, verificationMethod.publicKey);
          } else {
            return verificationMethod;
          }
        }).filter(key => key != null);
      }

      const authenticators = publicKeysToCheck.filter(({
        type
      }) => types.find(supported => supported === type));

      if (typeof proofPurpose === 'string' && (!authenticators || authenticators.length === 0)) {
        throw new Error(`${JWT_ERROR.NO_SUITABLE_KEYS}: DID document for ${issuer} does not have public keys suitable for ${alg} with ${proofPurpose} purpose`);
      }

      if (!authenticators || authenticators.length === 0) {
        throw new Error(`${JWT_ERROR.NO_SUITABLE_KEYS}: DID document for ${issuer} does not have public keys for ${alg}`);
      }

      return {
        authenticators,
        issuer,
        didResolutionResult: didResult
      };
    });
  } catch (e) {
    return Promise.reject(e);
  }
};

/**
 *  Verifies given JWT. If the JWT is valid, the promise returns an object including the JWT, the payload of the JWT,
 *  and the DID document of the issuer of the JWT.
 *
 *  @example
 *  ```ts
 *  verifyJWT(
 *      'did:uport:eyJ0eXAiOiJKV1QiLCJhbGciOiJFUzI1NksifQ.eyJyZXF1Z....',
 *      {audience: '5A8bRWU3F7j3REx3vkJ...', callbackUrl: 'https://...'}
 *    ).then(obj => {
 *        const did = obj.did // DID of signer
 *        const payload = obj.payload
 *        const doc = obj.didResolutionResult.didDocument // DID Document of issuer
 *        const jwt = obj.jwt
 *        const signerKeyId = obj.signer.id // ID of key in DID document that signed JWT
 *        ...
 *    })
 *  ```
 *
 *  @param    {String}            jwt                a JSON Web Token to verify
 *  @param    {Object}            [options]           an unsigned credential object
 *  @param    {Boolean}           options.auth        Require signer to be listed in the authentication section of the
 *   DID document (for Authentication purposes)
 *  @param    {String}            options.audience    DID of the recipient of the JWT
 *  @param    {String}            options.callbackUrl callback url in JWT
 *  @return   {Promise<Object, Error>}               a promise which resolves with a response object or rejects with an
 *   error
 */
const verifyJWT = function (jwt, options = {
  resolver: undefined,
  auth: undefined,
  audience: undefined,
  callbackUrl: undefined,
  skewTime: undefined,
  proofPurpose: undefined,
  policies: {},
  didAuthenticator: undefined
}) {
  try {
    function _temp7() {
      let _exit;

      function _temp5(_result) {
        if (_exit) ;

        if (signer) {
          const now = typeof options.policies?.now === 'number' ? options.policies.now : Math.floor(Date.now() / 1000);
          const skewTime = typeof options.skewTime !== 'undefined' && options.skewTime >= 0 ? options.skewTime : NBF_SKEW;
          const nowSkewed = now + skewTime;

          if (options.policies?.nbf !== false && payload.nbf) {
            if (payload.nbf > nowSkewed) {
              throw new Error(`${JWT_ERROR.INVALID_JWT}: JWT not valid before nbf: ${payload.nbf}`);
            }
          } else if (options.policies?.iat !== false && payload.iat && payload.iat > nowSkewed) {
            throw new Error(`${JWT_ERROR.INVALID_JWT}: JWT not valid yet (issued in the future) iat: ${payload.iat}`);
          }

          if (options.policies?.exp !== false && payload.exp && payload.exp <= now - skewTime) {
            throw new Error(`${JWT_ERROR.INVALID_JWT}: JWT has expired: exp: ${payload.exp} < now: ${now}`);
          }

          if (options.policies?.aud !== false && payload.aud) {
            if (!options.audience && !options.callbackUrl) {
              throw new Error(`${JWT_ERROR.INVALID_AUDIENCE}: JWT audience is required but your app address has not been configured`);
            }

            const audArray = Array.isArray(payload.aud) ? payload.aud : [payload.aud];
            const matchedAudience = audArray.find(item => options.audience === item || options.callbackUrl === item);

            if (typeof matchedAudience === 'undefined') {
              throw new Error(`${JWT_ERROR.INVALID_AUDIENCE}: JWT audience does not match your DID or callback url`);
            }
          }

          return {
            verified: true,
            payload,
            didResolutionResult,
            issuer,
            signer,
            jwt,
            policies: options.policies
          };
        }

        throw new Error(`${JWT_ERROR.INVALID_SIGNATURE}: JWT not valid. issuer DID document does not contain a verificationMethod that matches the signature.`);
      }

      const {
        did
      } = parse(didUrl);
      let signer = null;

      const _temp4 = function () {
        if (did !== didUrl) {
          const authenticator = authenticators.find(auth => auth.id === didUrl);

          if (!authenticator) {
            throw new Error(`${JWT_ERROR.INVALID_JWT}: No authenticator found for did URL ${didUrl}`);
          }

          return Promise.resolve(verifyProof(jwt, {
            payload,
            header,
            signature,
            data
          }, authenticator, options)).then(function (_verifyProof) {
            signer = _verifyProof;
          });
        } else {
          let i = 0;
          return _for$1(function () {
            return !_exit && !signer && i < authenticators.length;
          }, void 0, function () {
            function _temp3(_result2) {
              if (_exit) ;
              i++;
            }

            const authenticator = authenticators[i];

            const _temp2 = _catch(function () {
              return Promise.resolve(verifyProof(jwt, {
                payload,
                header,
                signature,
                data
              }, authenticator, options)).then(function (_verifyProof2) {
                signer = _verifyProof2;
              });
            }, function (e) {
              if (!e.message.includes(JWT_ERROR.INVALID_SIGNATURE) || i === authenticators.length - 1) throw e;
            });

            return _temp2 && _temp2.then ? _temp2.then(_temp3) : _temp3(_temp2);
          });
        }
      }();

      return _temp4 && _temp4.then ? _temp4.then(_temp5) : _temp5(_temp4);
    }

    if (!options.resolver) throw new Error('missing_resolver: No DID resolver has been configured');
    const {
      payload,
      header,
      signature,
      data
    } = decodeJWT(jwt, false);
    const proofPurpose = Object.prototype.hasOwnProperty.call(options, 'auth') ? options.auth ? 'authentication' : undefined : options.proofPurpose;
    let didUrl;

    if (!payload.iss && !payload.client_id) {
      throw new Error(`${JWT_ERROR.INVALID_JWT}: JWT iss or client_id are required`);
    }

    if (options.didAuthenticator) {
      didUrl = options.didAuthenticator.issuer;
    } else if (payload.iss === SELF_ISSUED_V2 || payload.iss === SELF_ISSUED_V2_VC_INTEROP) {
      if (!payload.sub) {
        throw new Error(`${JWT_ERROR.INVALID_JWT}: JWT sub is required`);
      }

      if (typeof payload.sub_jwk === 'undefined') {
        didUrl = payload.sub;
      } else {
        didUrl = (header.kid || '').split('#')[0];
      }
    } else if (payload.iss === SELF_ISSUED_V0_1) {
      if (!payload.did) {
        throw new Error(`${JWT_ERROR.INVALID_JWT}: JWT did is required`);
      }

      didUrl = payload.did;
    } else if (!payload.iss && payload.scope === 'openid' && payload.redirect_uri) {
      // SIOP Request payload
      // https://identity.foundation/jwt-vc-presentation-profile/#self-issued-op-request-object
      if (!payload.client_id) {
        throw new Error(`${JWT_ERROR.INVALID_JWT}: JWT client_id is required`);
      }

      didUrl = payload.client_id;
    } else {
      didUrl = payload.iss;
    }

    if (!didUrl) {
      throw new Error(`${JWT_ERROR.INVALID_JWT}: No DID has been found in the JWT`);
    }

    let authenticators;
    let issuer;
    let didResolutionResult;

    const _temp6 = function () {
      if (options.didAuthenticator) {
        ({
          didResolutionResult,
          authenticators,
          issuer
        } = options.didAuthenticator);
      } else {
        return Promise.resolve(resolveAuthenticator(options.resolver, header.alg, didUrl, proofPurpose)).then(function (_resolveAuthenticator) {
          ({
            didResolutionResult,
            authenticators,
            issuer
          } = _resolveAuthenticator);
          // Add to options object for recursive reference
          options.didAuthenticator = {
            didResolutionResult,
            authenticators,
            issuer
          };
        });
      }
    }();

    return Promise.resolve(_temp6 && _temp6.then ? _temp6.then(_temp7) : _temp7(_temp6));
  } catch (e) {
    return Promise.reject(e);
  }
};

/**
 *  Creates a multi-signature signed JWT given multiple issuers and their corresponding signers, and a payload for
 * which the signature is over.
 *
 *  @example
 *  const signer = ES256KSigner(process.env.PRIVATE_KEY)
 *  createJWT({address: '5A8bRWU3F7j3REx3vkJ...', signer}, {key1: 'value', key2: ..., ... }).then(jwt => {
 *      ...
 *  })
 *
 *  @param    {Object}            payload               payload object
 *  @param    {Object}            [options]             an unsigned credential object
 *  @param    {boolean}           options.expiresIn     optional flag to denote the expiration time
 *  @param    {boolean}           options.canonicalize  optional flag to canonicalize header and payload before signing
 *  @param    {Object[]}          issuers               array of the issuers, their signers and algorithms
 *  @param    {string}            issuers[].issuer      The DID of the issuer (signer) of JWT
 *  @param    {Signer}            issuers[].signer      a `Signer` function, Please see `ES256KSigner` or `EdDSASigner`
 *  @param    {String}            issuers[].alg         [DEPRECATED] The JWT signing algorithm to use. Supports:
 *   [ES256K, ES256K-R, Ed25519, EdDSA], Defaults to: ES256K. Please use `header.alg` to specify the algorithm
 *  @return   {Promise<Object, Error>}                  a promise which resolves with a signed JSON Web Token or
 *   rejects with an error
 */
const createMultisignatureJWT = function (payload, {
  expiresIn,
  canonicalize
}, issuers) {
  try {
    if (issuers.length === 0) throw new Error('invalid_argument: must provide one or more issuers');
    let payloadResult = payload;
    let jwt = '';

    const _temp = _forTo$1(issuers, function (i) {
      const issuer = issuers[i];
      const header = {
        typ: 'JWT',
        alg: issuer.alg
      }; // Create nested JWT
      // See Point 5 of https://www.rfc-editor.org/rfc/rfc7519#section-7.1
      // After the first JWT is created (the first JWS), the next JWT is created by inputting the previous JWT as the
      // payload

      if (i !== 0) {
        header.cty = 'JWT';
      }

      return Promise.resolve(createJWT(payloadResult, { ...issuer,
        canonicalize,
        expiresIn
      }, header)).then(function (_createJWT) {
        jwt = _createJWT;
        payloadResult = {
          jwt
        };
      });
    });

    return Promise.resolve(_temp && _temp.then ? _temp.then(function () {
      return jwt;
    }) : jwt);
  } catch (e) {
    return Promise.reject(e);
  }
};

/**
 *  Creates a signed JWT given an address which becomes the issuer, a signer, and a payload for which the signature is
 * over.
 *
 *  @example
 *  const signer = ES256KSigner(process.env.PRIVATE_KEY)
 *  createJWT({address: '5A8bRWU3F7j3REx3vkJ...', signer}, {key1: 'value', key2: ..., ... }).then(jwt => {
 *      ...
 *  })
 *
 *  @param    {Object}            payload               payload object
 *  @param    {Object}            [options]             an unsigned credential object
 *  @param    {String}            options.issuer        The DID of the issuer (signer) of JWT
 *  @param    {String}            options.alg           [DEPRECATED] The JWT signing algorithm to use. Supports:
 *   [ES256K, ES256K-R, Ed25519, EdDSA], Defaults to: ES256K. Please use `header.alg` to specify the algorithm
 *  @param    {Signer}            options.signer        a `Signer` function, Please see `ES256KSigner` or `EdDSASigner`
 *  @param    {boolean}           options.canonicalize  optional flag to canonicalize header and payload before signing
 *  @param    {Object}            header                optional object to specify or customize the JWT header
 *  @return   {Promise<Object, Error>}                  a promise which resolves with a signed JSON Web Token or
 *   rejects with an error
 */
const createJWT = function (payload, {
  issuer,
  signer,
  alg,
  expiresIn,
  canonicalize
}, header = {}) {
  try {
    if (!signer) throw new Error('missing_signer: No Signer functionality has been configured');
    if (!issuer) throw new Error('missing_issuer: No issuing DID has been configured');
    if (!header.typ) header.typ = 'JWT';
    if (!header.alg) header.alg = alg;
    const timestamps = {
      iat: Math.floor(Date.now() / 1000),
      exp: undefined
    };

    if (expiresIn) {
      if (typeof expiresIn === 'number') {
        timestamps.exp = (payload.nbf || timestamps.iat) + Math.floor(expiresIn);
      } else {
        throw new Error('invalid_argument: JWT expiresIn is not a number');
      }
    }

    const fullPayload = { ...timestamps,
      ...payload,
      iss: issuer
    };
    return createJWS(fullPayload, signer, header, {
      canonicalize
    });
  } catch (e) {
    return Promise.reject(e);
  }
};

/**
 *  Creates a signed JWS given a payload, a signer, and an optional header.
 *
 *  @example
 *  const signer = ES256KSigner(process.env.PRIVATE_KEY)
 *  const jws = await createJWS({ my: 'payload' }, signer)
 *
 *  @param    {Object}            payload           payload object
 *  @param    {Signer}            signer            a signer, see `ES256KSigner or `EdDSASigner`
 *  @param    {Object}            header            optional object to specify or customize the JWS header
 *  @param    {Object}            options           can be used to trigger automatic canonicalization of header and
 *                                                    payload properties
 *  @return   {Promise<string>}                     a Promise which resolves to a JWS string or rejects with an error
 */
const createJWS = function (payload, signer, header = {}, options = {}) {
  try {
    if (!header.alg) header.alg = defaultAlg;
    const encodedPayload = typeof payload === 'string' ? payload : encodeSection(payload, options.canonicalize);
    const signingInput = [encodeSection(header, options.canonicalize), encodedPayload].join('.');
    const jwtSigner = SignerAlg(header.alg);
    return Promise.resolve(jwtSigner(signingInput, signer)).then(function (signature) {
      // JWS Compact Serialization
      // https://www.rfc-editor.org/rfc/rfc7515#section-7.1
      return [signingInput, signature].join('.');
    });
  } catch (e) {
    return Promise.reject(e);
  }
};
const SELF_ISSUED_V2 = 'https://self-issued.me/v2';
const SELF_ISSUED_V2_VC_INTEROP = 'https://self-issued.me/v2/openid-vc'; // https://identity.foundation/jwt-vc-presentation-profile/#id-token-validation

const SELF_ISSUED_V0_1 = 'https://self-issued.me';
const defaultAlg = 'ES256K';
const DID_JSON = 'application/did+json'; // eslint-disable-next-line @typescript-eslint/no-explicit-any

function encodeSection(data, shouldCanonicalize = false) {
  if (shouldCanonicalize) {
    return encodeBase64url(canonicalize(data));
  } else {
    return encodeBase64url(JSON.stringify(data));
  }
}

const NBF_SKEW = 300;

function decodeJWS(jws) {
  const parts = jws.match(/^([a-zA-Z0-9_-]+)\.([a-zA-Z0-9_-]+)\.([a-zA-Z0-9_-]+)$/);

  if (parts) {
    return {
      header: JSON.parse(decodeBase64url(parts[1])),
      payload: parts[2],
      signature: parts[3],
      data: `${parts[1]}.${parts[2]}`
    };
  }

  throw new Error('invalid_argument: Incorrect format JWS');
}
/**
 *  Decodes a JWT and returns an object representing the payload
 *
 *  @example
 *  decodeJWT('eyJ0eXAiOiJKV1QiLCJhbGciOiJFUzI1NksifQ.eyJpYXQiOjE1...')
 *
 *  @param    {String}            jwt                a JSON Web Token to verify
 * @param    {Object}            [recurse]          whether to recurse into the payload to decode any nested JWTs
 *  @return   {Object}                               a JS object representing the decoded JWT
 */


function decodeJWT(jwt, recurse = true) {
  if (!jwt) throw new Error('invalid_argument: no JWT passed into decodeJWT');

  try {
    const jws = decodeJWS(jwt);
    const decodedJwt = Object.assign(jws, {
      payload: JSON.parse(decodeBase64url(jws.payload))
    });
    const iss = decodedJwt.payload.iss;

    if (decodedJwt.header.cty === 'JWT' && recurse) {
      const innerDecodedJwt = decodeJWT(decodedJwt.payload.jwt);
      if (innerDecodedJwt.payload.iss !== iss) throw new Error(`${JWT_ERROR.INVALID_JWT}: multiple issuers`);
      return innerDecodedJwt;
    }

    return decodedJwt;
  } catch (e) {
    throw new Error('invalid_argument: Incorrect format JWT');
  }
}
function verifyJWTDecoded({
  header,
  payload,
  data,
  signature
}, pubKeys) {
  if (!Array.isArray(pubKeys)) pubKeys = [pubKeys];
  const iss = payload.iss;
  let recurse = true;

  do {
    if (iss !== payload.iss) throw new Error(`${JWT_ERROR.INVALID_JWT}: multiple issuers`);

    try {
      const result = VerifierAlgorithm(header.alg)(data, signature, pubKeys);
      return result;
    } catch (e) {
      if (!e.message.startsWith(JWT_ERROR.INVALID_SIGNATURE)) throw e;
    } // TODO probably best to create copy objects than replace reference objects


    if (header.cty !== 'JWT') {
      recurse = false;
    } else {
      ({
        payload,
        header,
        signature,
        data
      } = decodeJWT(payload.jwt, false));
    }
  } while (recurse);

  throw new Error(`${JWT_ERROR.INVALID_SIGNATURE}: no matching public key found`);
}

typeof Symbol !== "undefined" ? Symbol.iterator || (Symbol.iterator = Symbol("Symbol.iterator")) : "@@iterator";

function randomBytes(bytes) {
  return rb(new Uint8Array(bytes));
}
/*

/* Creates a signer from a private key that can be used to sign a JWT
 *
 * @param privateKey the private key to use to sign the JWT
 * @returns a signer (function) that can be used to sign a JWT
 */
function createSigner(privateKey) {
  if (privateKey.type === KeyType.K1) {
    return ES256KSigner(privateKey.data.array, true);
  }
  if (privateKey.type === KeyType.R1 || privateKey.type === KeyType.WA) {
    return ES256Signer(privateKey.data.array);
  }
  throw new Error('Unsupported key type');
}
function randomString(bytes) {
  const random = randomBytes(bytes);
  return Array.from(random).map(int2hex).join('');
}
function sha256(digest) {
  return Checksum256.hash(Bytes.from(encodeHex(digest), 'hex')).toString();
}
function int2hex(i) {
  return ('0' + i.toString(16)).slice(-2);
}
function encodeHex(str) {
  return str.split('').map(c => c.charCodeAt(0).toString(16).padStart(2, '0')).join('');
}
function generateRandomKeyPair() {
  const bytes = randomBytes(32);
  const privateKey = new PrivateKey(KeyType.K1, new Bytes(bytes));
  const publicKey = privateKey.toPublic();
  return {
    privateKey,
    publicKey
  };
}
function createVCSigner(keyManager, level) {
  return {
    async sign(data) {
      return await keyManager.signData({
        level,
        data,
        outputType: 'jwt'
      });
    }
  };
}

const JWT_ALG = 'ES256K';
const JWT_FORMAT = /^[A-Za-z0-9-_=]+\.[A-Za-z0-9-_=]+\.?[A-Za-z0-9-_.+/=]*$/;
const DEFAULT_CONTEXT = 'https://www.w3.org/2018/credentials/v1';
const DEFAULT_VC_TYPE = 'VerifiableCredential';
/**
 * The `JwtProof2020` is a synthetic proof type, usable for differentiating credentials by proof type when representing
 * JWT credentials as W3C VC JSON. It is not a registered W3C VC Data Model algorithm and should not be treated as
 * such.
 *
 * This proof type is only intended as a convenience and does not actually prove the validity of a VC/VP in JSON
 * representation. The actual verifiable credential or presentation is represented in the `jwt` property.
 */
const DEFAULT_JWT_PROOF_TYPE = 'JwtProof2020';

/**
 * Additional W3C VC fields:
 * These are defined as optional top-level properties in the W3C spec but are not mapped to top-level JWT names,
 * so they should be moved inside the "vc" object when transforming to a JWT.
 * Conversely, they should be moved out of the "vc" object when transforming from a JWT to W3C JSON.
 */
const additionalPropNames = ['evidence', 'termsOfUse', 'refreshService', 'credentialSchema', 'credentialStatus'];
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function asArray(arg) {
  return Array.isArray(arg) ? arg : [arg];
}
function deepCopy(source) {
  return Array.isArray(source) ? source.map(item => deepCopy(item)) : source instanceof Date ? new Date(source.getTime()) : source && typeof source === 'object' ? Object.getOwnPropertyNames(source).reduce((o, prop) => {
    Object.defineProperty(o, prop, Object.getOwnPropertyDescriptor(source, prop));
    o[prop] = deepCopy(source[prop]);
    return o;
  }, Object.create(Object.getPrototypeOf(source))) : source;
}
function notEmpty(value) {
  return value !== null && value !== undefined;
}
function cleanUndefined(input) {
  if (typeof input !== 'object' || input === null) {
    return input;
  }
  const obj = {
    ...input
  };
  Object.keys(obj).forEach(key => obj[key] === undefined && delete obj[key]);
  return obj;
}
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function isLegacyAttestationFormat(payload) {
  // payload is an object and has all the required fields of old attestation format
  return typeof payload === 'object' && payload.sub && payload.iss && payload.claim && payload.iat;
}
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function attestationToVcFormat(payload) {
  const {
    iat,
    nbf,
    claim,
    vc,
    ...rest
  } = payload;
  const result = {
    ...rest,
    nbf: nbf ? nbf : iat,
    vc: {
      '@context': [DEFAULT_CONTEXT],
      type: [DEFAULT_VC_TYPE],
      credentialSubject: claim
    }
  };
  if (vc) payload.issVc = vc;
  return result;
}
function normalizeJwtCredentialPayload(input, removeOriginalFields = true) {
  let result = deepCopy(input);
  if (isLegacyAttestationFormat(input)) {
    result = attestationToVcFormat(input);
  }
  // FIXME: handle case when credentialSubject(s) are not object types
  result.credentialSubject = {
    ...input.credentialSubject,
    ...input.vc?.credentialSubject
  };
  if (input.sub && !input.credentialSubject?.id && result.credentialSubject) {
    result.credentialSubject.id = input.sub;
    if (removeOriginalFields) {
      delete result.sub;
    }
  }
  if (removeOriginalFields) {
    delete result.vc?.credentialSubject;
  }
  if (typeof input.issuer === 'undefined' || typeof input.issuer === 'object') {
    result.issuer = cleanUndefined({
      id: input.iss,
      ...input.issuer
    });
    if (removeOriginalFields && !input.issuer?.id) {
      delete result.iss;
    }
  }
  if (!input.id && input.jti) {
    result.id = result.id || result.jti;
    if (removeOriginalFields) {
      delete result.jti;
    }
  }
  const types = [...asArray(result.type), ...asArray(result.vc?.type)].filter(notEmpty);
  result.type = [...new Set(types)];
  if (removeOriginalFields) {
    delete result.vc?.type;
  }
  for (const prop of additionalPropNames) {
    if (input.vc && input.vc[prop]) {
      if (!result[prop]) {
        result[prop] = input.vc[prop];
      }
      if (removeOriginalFields) {
        delete result.vc[prop];
      }
    }
  }
  const contextArray = [...asArray(input.context), ...asArray(input['@context']), ...asArray(input.vc?.['@context'])].filter(notEmpty);
  result['@context'] = [...new Set(contextArray)];
  if (removeOriginalFields) {
    delete result.context;
    delete result.vc?.['@context'];
  }
  if (!input.issuanceDate && (input.iat || input.nbf)) {
    result.issuanceDate = new Date((input.nbf || input.iat) * 1000).toISOString();
    if (removeOriginalFields) {
      if (input.nbf) {
        delete result.nbf;
      } else {
        delete result.iat;
      }
    }
  }
  if (!input.expirationDate && input.exp) {
    result.expirationDate = new Date(input.exp * 1000).toISOString();
    if (removeOriginalFields) {
      delete result.exp;
    }
  }
  if (removeOriginalFields) {
    if (result.vc && Object.keys(result.vc).length === 0) {
      delete result.vc;
    }
  }
  // FIXME: interpret `aud` property as `verifier`
  return result;
}
function normalizeJwtCredential(input, removeOriginalFields = true) {
  let decoded;
  try {
    decoded = decodeJWT(input);
  } catch (e) {
    throw new TypeError('unknown credential format');
  }
  return {
    ...normalizeJwtCredentialPayload(decoded.payload, removeOriginalFields),
    proof: {
      type: DEFAULT_JWT_PROOF_TYPE,
      jwt: input
    }
  };
}
/**
 * Normalizes a credential payload into an unambiguous W3C credential data type In case of conflict, existing W3C
 * Credential specific properties take precedence, except for arrays and object types which get merged.
 *
 * @param input - either a JWT or JWT payload, or a VerifiableCredential
 * @param removeOriginalFields - if true, removes all fields that were transformed according to the W3C mapping
 *
 * @see {@link https://www.w3.org/TR/vc-data-model/#jwt-encoding | VC JWT encoding }
 */
function normalizeCredential(input, removeOriginalFields = true) {
  if (typeof input === 'string') {
    if (JWT_FORMAT.test(input)) {
      return normalizeJwtCredential(input, removeOriginalFields);
    } else {
      let parsed;
      try {
        parsed = JSON.parse(input);
      } catch (e) {
        throw new TypeError('unknown credential format');
      }
      return normalizeCredential(parsed, removeOriginalFields);
    }
  } else if (input.proof?.jwt) {
    // TODO: test that it correctly propagates app specific proof properties
    return deepCopy({
      ...normalizeJwtCredential(input.proof.jwt, removeOriginalFields),
      proof: input.proof
    });
  } else {
    // TODO: test that it accepts JWT payload, CredentialPayload, VerifiableCredential
    // TODO: test that it correctly propagates proof, if any
    return {
      proof: {},
      ...normalizeJwtCredentialPayload(input, removeOriginalFields)
    };
  }
}
/**
 * Transforms a W3C Credential payload into a JWT compatible encoding.
 * The method accepts app specific fields and in case of collision, existing JWT properties will take precedence.
 * Also, `nbf`, `exp` and `jti` properties can be explicitly set to `undefined` and they will be kept intact.
 * @param input - either a JWT payload or a CredentialPayloadInput
 * @param removeOriginalFields - if true, removes original W3C fields from the resulting object
 *
 * @see {@link https://www.w3.org/TR/vc-data-model/#jwt-encoding | VC JWT encoding }
 */
function transformCredentialInput(input, removeOriginalFields = true) {
  if (Array.isArray(input.credentialSubject)) throw Error('credentialSubject of type array not supported');
  const result = deepCopy({
    vc: {
      ...input.vc
    },
    ...input
  });
  result.vc = result.vc;
  const credentialSubject = {
    ...input.credentialSubject,
    ...input.vc?.credentialSubject
  };
  if (!input.sub) {
    result.sub = input.credentialSubject?.id;
    if (removeOriginalFields) {
      delete credentialSubject.id;
    }
  }
  const contextEntries = [...asArray(input.context), ...asArray(input['@context']), ...asArray(input.vc?.['@context'])].filter(notEmpty);
  result.vc['@context'] = [...new Set(contextEntries)];
  if (removeOriginalFields) {
    delete result.context;
    delete result['@context'];
  }
  const types = [...asArray(input.type), ...asArray(input.vc?.type)].filter(notEmpty);
  result.vc.type = [...new Set(types)];
  if (removeOriginalFields) {
    delete result.type;
  }
  if (input.id && Object.getOwnPropertyNames(input).indexOf('jti') === -1) {
    result.jti = input.id;
    if (removeOriginalFields) {
      delete result.id;
    }
  }
  if (input.issuanceDate && Object.getOwnPropertyNames(input).indexOf('nbf') === -1) {
    const converted = Date.parse(input.issuanceDate);
    if (!isNaN(converted)) {
      result.nbf = Math.floor(converted / 1000);
      if (removeOriginalFields) {
        delete result.issuanceDate;
      }
    }
  }
  if (input.expirationDate && Object.getOwnPropertyNames(input).indexOf('exp') === -1) {
    const converted = Date.parse(input.expirationDate);
    if (!isNaN(converted)) {
      result.exp = Math.floor(converted / 1000);
      if (removeOriginalFields) {
        delete result.expirationDate;
      }
    }
  }
  if (input.issuer && Object.getOwnPropertyNames(input).indexOf('iss') === -1) {
    if (typeof input.issuer === 'object') {
      result.iss = input.issuer?.id;
      if (removeOriginalFields) {
        delete result.issuer.id;
        if (Object.keys(result.issuer).length === 0) {
          delete result.issuer;
        }
      }
    } else if (typeof input.issuer === 'string') {
      result.iss = input.iss || '' + input.issuer;
      if (removeOriginalFields) {
        delete result.issuer;
      }
    } else ;
  }
  result.vc.credentialSubject = credentialSubject;
  if (removeOriginalFields) {
    delete result.credentialSubject;
  }
  for (const prop of additionalPropNames) {
    if (input[prop]) {
      if (!result.vc[prop]) {
        result.vc[prop] = input[prop];
      }
      if (removeOriginalFields) {
        delete result[prop];
      }
    }
  }
  return result;
}

/**
 * Error prefixes used for known verification failure cases related to the
 * {@link https://www.w3.org/TR/vc-data-model/ | Verifiable Credential data model }
 */
const VC_ERROR = {
  /**
   * Thrown when the credential or presentation being verified does not conform to the data model defined by
   * {@link https://www.w3.org/TR/vc-data-model/ | the spec}
   */
  SCHEMA_ERROR: 'schema_error',
  /**
   * Thrown when the input is not a JWT string
   */
  FORMAT_ERROR: 'format_error',
  /**
   * Thrown when verifying a presentation where `challenge` and/or `domain` don't match the expected values.
   */
  AUTH_ERROR: 'auth_error'
};
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function isDateObject(input) {
  return input && !isNaN(input) && Object.prototype.toString.call(input) === '[object Date]';
}
// The main scenario we want to guard against is having a timestamp in milliseconds
// instead of seconds (ex: from new Date().getTime()).
// We will check the number of digits and assume that any number with 12 or more
// digits is a millisecond timestamp.
// 10 digits max is 9999999999 -> 11/20/2286 @ 5:46pm (UTC)
// 11 digits max is 99999999999 -> 11/16/5138 @ 9:46am (UTC)
// 12 digits max is 999999999999 -> 09/27/33658 @ 1:46am (UTC)
function validateTimestamp(value) {
  if (typeof value === 'number') {
    if (!(Number.isInteger(value) && value < 100000000000)) {
      throw new TypeError(`${VC_ERROR.SCHEMA_ERROR}: "${value}" is not a unix timestamp in seconds`);
    }
  } else if (typeof value === 'string') {
    validateTimestamp(Math.floor(new Date(value).valueOf() / 1000));
  } else if (!isDateObject(value)) {
    throw new TypeError(`${VC_ERROR.SCHEMA_ERROR}: "${value}" is not a valid time`);
  }
}
function validateContext(value) {
  const input = asArray(value);
  if (input.length < 1 || input.indexOf(DEFAULT_CONTEXT) === -1) {
    throw new TypeError(`${VC_ERROR.SCHEMA_ERROR}: @context is missing default context "${DEFAULT_CONTEXT}"`);
  }
}
function validateVcType(value) {
  const input = asArray(value);
  if (input.length < 1 || input.indexOf(DEFAULT_VC_TYPE) === -1) {
    throw new TypeError(`${VC_ERROR.SCHEMA_ERROR}: type is missing default "${DEFAULT_VC_TYPE}"`);
  }
}
function validateCredentialSubject(value) {
  if (Object.keys(value).length === 0) {
    throw new TypeError(`${VC_ERROR.SCHEMA_ERROR}: credentialSubject must not be empty`);
  }
}
/**
 * Verifies and validates a VerifiableCredential that is encoded as a JWT according to the W3C spec.
 *
 * @return a `Promise` that resolves to a `VerifiedCredential` or rejects with `TypeError` if the input is not
 * W3C compliant
 * @param vc - the credential to be verified. Currently only the JWT encoding is supported by this library
 * @param resolver - a configured `Resolver` (or an implementation of `Resolvable`) that can provide the DID document
 *   of the JWT issuer
 * @param options - optional tweaks to the verification process. These are forwarded to did-jwt.
 */
const verifyCredential = function (vc, resolver, options = {}) {
  try {
    const nbf = options?.policies?.issuanceDate === false ? false : undefined;
    const exp = options?.policies?.expirationDate === false ? false : undefined;
    options = {
      ...options,
      policies: {
        ...options?.policies,
        nbf,
        exp,
        iat: nbf
      }
    };
    return Promise.resolve(verifyJWT(vc, {
      resolver,
      ...options
    })).then(function (verified) {
      verified.verifiableCredential = normalizeCredential(verified.jwt, options?.removeOriginalFields);
      if (options?.policies?.format !== false) {
        validateCredentialPayload(verified.verifiableCredential);
      }
      return verified;
    });
  } catch (e) {
    return Promise.reject(e);
  }
};
/**
 * Creates a VerifiableCredential given a `CredentialPayload` or `JwtCredentialPayload` and an `Issuer`.
 *
 * This method transforms the payload into the [JWT encoding](https://www.w3.org/TR/vc-data-model/#jwt-encoding)
 * described in the [W3C VC spec](https://www.w3.org/TR/vc-data-model) and then validated to conform to the minimum
 * spec
 * required spec.
 *
 * The `issuer` is then used to assign an algorithm, override the `iss` field of the payload and then sign the JWT.
 *
 * @param payload - `CredentialPayload` or `JwtCredentialPayload`
 * @param issuer - `Issuer` the DID, signer and algorithm that will sign the token
 * @param options - Use these options to tweak the creation of the JWT Credential. These are forwarded to did-jwt.
 * @return a `Promise` that resolves to the JWT encoded verifiable credential or rejects with `TypeError` if the
 * `payload` is not W3C compliant
 */
const createVerifiableCredentialJwt = function (payload, issuer, options = {}) {
  try {
    const parsedPayload = {
      iat: undefined,
      ...transformCredentialInput(payload, options.removeOriginalFields)
    };
    validateJwtCredentialPayload(parsedPayload);
    if (!Array.isArray(issuer)) {
      return Promise.resolve(createJWT(parsedPayload, {
        ...options,
        issuer: issuer.did || parsedPayload.iss || '',
        signer: issuer.signer
      }, {
        ...options.header,
        alg: issuer.alg || options.header?.alg || JWT_ALG
      }));
    } else {
      const did = issuer[0].did;
      const issuers = [];
      for (const iss of issuer) {
        if (iss.did !== did) {
          throw new Error('All issuers must be the same did to comply with the Verifiable Conditions spec');
        }
        issuers.push({
          issuer: iss.did || parsedPayload.iss || '',
          signer: iss.signer,
          alg: iss.alg || options.header?.alg || JWT_ALG
        });
      }
      return Promise.resolve(createMultisignatureJWT(parsedPayload, {
        ...options
      }, issuers));
    }
  } catch (e) {
    return Promise.reject(e);
  }
};
function validateJwtCredentialPayload(payload) {
  validateContext(payload.vc['@context']);
  validateVcType(payload.vc.type);
  validateCredentialSubject(payload.vc.credentialSubject);
  if (payload.nbf) validateTimestamp(payload.nbf);
  if (payload.exp) validateTimestamp(payload.exp);
}
function validateCredentialPayload(payload) {
  validateContext(payload['@context']);
  validateVcType(payload.type);
  validateCredentialSubject(payload.credentialSubject);
  if (payload.issuanceDate) validateTimestamp(payload.issuanceDate);
  if (payload.expirationDate) validateTimestamp(payload.expirationDate);
}

const SECONDS_IN_MINUTE = 60;
const MINUTES_IN_HOUR = 60;
const HOURS_IN_DAY = 24;
const SECONDS_IN_HOUR = MINUTES_IN_HOUR * SECONDS_IN_MINUTE;
const SECONDS_IN_DAY = HOURS_IN_DAY * SECONDS_IN_HOUR;
/**
 * Takes the number of seconds since epoch and returns a Date object
 *
 * @param {number} secs - Number of seconds since epoch
 * @returns {Date} - Date object
 */
function toDateTime(secs) {
  const t = new Date(1970, 0, 1); // Epoch
  t.setSeconds(secs);
  return t;
}
/**
 * Sleeps for the given number of milliseconds
 *
 * @async
 * @param {number} milliseconds - Number of milliseconds to sleep
 */
async function sleep(milliseconds) {
  return new Promise(resolve => setTimeout(resolve, milliseconds));
}
/**
 * Adds a number of seconds and returns a new Date object
 *
 * @param {Date} date - Date object
 * @param {number} secs - Number of seconds to add
 */
function addSeconds(date, secs) {
  return new Date(date.getTime() + secs * 1000);
}
/**
 * Adds a number of microseconds and returns a new Date object
 *
 * @param {Date} date - Date object
 * @param {number} microseconds - Number of microseconds to add
 */
function addMicroseconds(date, microseconds) {
  return new Date(date.getTime() + microseconds / 1000);
}

/**
 * A W3C Verifiable Credential
 */
class VerifiableCredential {
  /**
   * @param {JWT} jwt - a JWT string
   *
   * @returns {VerifiableCredential} a VerifiableCredential object
   */
  constructor(jwt) {
    this.decodedJwt = void 0;
    this.jwt = void 0;
    this.decodedJwt = decodeJWT(jwt);
    this.jwt = jwt;
  }
  /**
   * Returns the payload of the JWT
   *
   * @returns {JWTPayload} the payload of the JWT
   */
  getPayload() {
    return this.decodedJwt.payload;
  }
  /**
   * Returns the vc object of the JWT
   *
   * @returns {JWTVCPayload} the vc object
   */
  getVc() {
    return this.getPayload().vc;
  }
  /**
   * Returns the credentialSubject object of the Verifiable Credential
   *
   * @returns {any} the credentialSubject object
   */
  getCredentialSubject() {
    return this.getVc().credentialSubject;
  }
  /**
   * Returns the issuer of the Verifiable Credential
   *
   * @returns {DIDurl | undefined} the issuer
   */
  getIssuer() {
    return this.getPayload().iss;
  }
  /**
   * Returns the subject of the Verifiable Credential
   *
   * @returns {DIDurl | undefined} the subject of the Verifiable Credential
   */
  getSubject() {
    return this.getPayload().sub;
  }
  /**
   * Returns the ID of the Verifiable Credential
   *
   * @returns {URL | undefined} the id
   */
  getId() {
    return this.getPayload().jti;
  }
  /**
   * Returns the audience of the Verifiable Credential
   *
   * @returns {DIDurl | DIDurl[] | undefined} the audience
   */
  getAudience() {
    return this.getPayload().aud;
  }
  /**
   * Returns the expiration of the Verifiable Credential
   *
   * @returns {Date | undefined} the expiration
   */
  getExpiration() {
    const secSinceEpoc = this.getPayload().exp;
    if (secSinceEpoc) {
      return toDateTime(secSinceEpoc);
    } else return undefined;
  }
  /**
   * Returns the issued time of the Verifiable Credential
   *
   * @returns {Date | undefined} the issued time
   */
  getIssuedAt() {
    const secSinceEpoc = this.getPayload().iat;
    if (secSinceEpoc) {
      return toDateTime(secSinceEpoc);
    } else return undefined;
  }
  /**
   * Returns the time the verifiable credential is valid after
   *
   * @returns {Date | undefined} the time the verifiable credential is valid after
   */
  getNotBefore() {
    const secSinceEpoc = this.getPayload().nbf;
    if (secSinceEpoc) {
      return toDateTime(secSinceEpoc);
    } else return undefined;
  }
  /**
   * Verifies the Verifiable Credential is signed by the issuer
   *
   * @returns {Promise<VerifiedCredential>} the verified credential
   * @throws {Error} if the Verifiable Credential is not signed correctly by the issuer
   */
  async verify() {
    const settings = getSettings();
    const resolver = new Resolver({
      ...getDidKeyResolver(),
      ...getResolver({
        antelopeChainUrl: settings.blockchainUrl,
        fetch: getFetch()
      })
    });
    return verifyCredential(this.jwt, resolver);
  }
  /**
   * Creates a new Verifiable Credential object signed by the issuer
   *
   * @param {DIDurl} id - the id of the Verifiable Credential
   * @param {string[]} type - the type of the Verifiable Credential
   * @param {object} credentialSubject - the credential subject of the Verifiable Credential
   * @param {Issuer} issuer - the issuer of the Verifiable Credential
   * @property {URL} [options.subject] - the subject of the Verifiable Credential
   *
   * @returns {Promise<VerifiableCredential>} the Verifiable Credential
   */
  static async sign(id, type, credentialSubject, issuer, options = {}) {
    const typeArray = Array.isArray(type) ? type : [type];
    if (!typeArray.includes('VerifiableCredential')) typeArray.push('VerifiableCredential');
    const vc = {
      '@context': ['https://www.w3.org/2018/credentials/v1'],
      id,
      type: typeArray,
      issuer: {
        id: issuer.did
      },
      issuanceDate: new Date().toISOString(),
      credentialSubject: credentialSubject
    };
    if (options.subject) vc.credentialSubject.id = options.subject;
    const jwt = await createVerifiableCredentialJwt(vc, issuer, {
      canonicalize: true
    });
    return new VerifiableCredential(jwt);
  }
  /**
   * Returns the JWT string
   *
   * @returns {string} the JWT string
   */
  toString() {
    return this.jwt;
  }
  /**
   * Returns the JSON string representation of the object - which is the jwt string
   *
   * @description This is used in JSON.stringify(). It is not recommended to use this method directly.
   * Only the jwt is returned, not the full object. This is all that is needed to reconstruct the
   * full object in the constructor, and thus minimizes the size of the JSON string.
   *
   * @override Serializable.toJSON
   * @returns {string} the jwt
   */
  toJSON() {
    return this.toString();
  }
}
/**
 * A wrapper that adds a type object to VCs to allow for identification, and presents a simper interface
 *
 * This is the base class. It is expected that extension classes will be created for the different VC types.
 *
 * @inheritdoc sign() is a protected alternative constructor. In child classes, a new alternative constructor
 * should be created which returns the child class type.
 * @inheritdoc protected "type" property should be overridden in child classes with the name of the class
 * @inheritdoc if the payload type requires ad-hoc decoding, override the constructor and decode the payload
 *
 * @example see an example of the above in the LoginRequestMessage class in `../..services/communication/messages.ts`
 */
class VerifiableCredentialWithType {
  /**
   *
   * @param vc the VC to wrap
   *
   * @inheritdoc override me if the payload type requires ad-hoc decoding
   */
  constructor(vc) {
    this.vc = void 0;
    this.decodedPayload = void 0;
    if (typeof vc === 'string') {
      this.vc = new VerifiableCredential(vc);
    } else if (vc instanceof VerifiableCredentialWithType) {
      this.vc = vc.getVc();
    } else {
      this.vc = vc;
    }
    this.decodedPayload = this.getVc().getCredentialSubject().payload;
  }
  /**
   * Creates a signed VC object
   *
   * @access protected
   *
   * @param {object} payload the payload
   * @param {Issuer} issuer the issuer id
   * @param {VerifiableCredentialOptions} options the options
   *
   * @returns a VC object
   */
  static async sign(payload, issuer, options = {}) {
    const payloadType = this.type;
    if (payloadType === 'VerifiableCredentialWithType' || payloadType === '') {
      throw new Error('class should be a derived class of VerifiableCredentialWithType to use the type property');
    }
    const credentialSubject = Object.assign({}, {
      payload,
      type: payloadType
    });
    const id = 'https://tonomy.foundation/vc/id/' + randomString(10);
    const vcType = ['VerifiableCredential', 'TonomyVerifiableCredentialWithType'];
    if (options.additionalTypes) {
      vcType.push(...options.additionalTypes);
    }
    const vc = await VerifiableCredential.sign(id, vcType, credentialSubject, issuer, options);
    return new VerifiableCredentialWithType(vc);
  }
  /**
   * Returns the type of the VC
   * @returns {string} the class type of the VC
   */
  static getType() {
    return this.type;
  }
  /**
   * Returns the internal Verifiable Credential
   * @returns {VerifiableCredential} the VC
   */
  getVc() {
    return this.vc;
  }
  /**
   * Returns the issuer of the payload
   * @returns {DIDurl} the payload issuer
   */
  getIssuer() {
    return this.getVc().getIssuer();
  }
  /**
   * Returns the payload
   * @returns {object} the payload
   */
  getPayload() {
    return this.decodedPayload;
  }
  /**
   * Returns the payload type used to determine what kind of payload it is
   * @returns {string | undefined} the payload type
   */
  getType() {
    return this.getVc().getCredentialSubject().type;
  }
  /**
   * Verifies that the VC is signed by the issuer
   *
   * @returns {Promise<boolean>} true if the VC is signed by the issuer
   * @throws {Error} if the VC is not signed by the issuer
   */
  async verify() {
    return (await this.getVc().verify()).verified;
  }
  /**
   * Returns the JWT string
   *
   * @returns {string} the JWT string
   */
  toString() {
    return this.getVc().toString();
  }
  /**
   * Returns the JSON string representation of the object - which is the jwt string
   *
   * @description This is used in JSON.stringify(). It is not recommended to use this method directly.
   * Only the jwt is returned, not the full object. This is all that is needed to reconstruct the
   * full object in the constructor, and thus minimizes the size of the JSON string.
   *
   * @override Serializable.toJSON
   * @returns {string} the jwt
   */
  toJSON() {
    return this.toString();
  }
}
VerifiableCredentialWithType.type = 'VerifiableCredentialWithType';

class WalletRequest extends VerifiableCredentialWithType {
  /**
   * @inheritdoc override me if the payload type requires ad-hoc decoding
   */
  constructor(vc) {
    super(vc);
  }
  /**
   * Alternative constructor that returns type WalletRequest. To be used in derived classes
   *
   * @access protected
   *
   * @param {object} payload the payload
   * @param {Issuer} issuer the issuer id
   * @param {VerifiableCredentialOptions} options the options
   *
   * @returns a WalletRequest object
   */
  static async signWalletRequest(payload, issuer, options = {}) {
    if (this.type === 'WalletRequest') {
      throw new Error('class should be a derived class of WalletRequest to use the type property');
    }
    const vc = await super.sign(payload, issuer, options);
    return new WalletRequest(vc);
  }
}
WalletRequest.type = 'WalletRequest';
class LoginRequest extends WalletRequest {
  /**
   * @override the VerifiableCredentialWithType constructor to decode the payload of type LoginRequestPayload
   */
  constructor(vc) {
    super(vc);
    this.decodedPayload.publicKey = PublicKey.from(super.getPayload().publicKey);
  }
  /**
   * Alternative constructor that returns type LoginRequest
   */
  static async signRequest(payload, issuer, options = {}) {
    const vc = await super.signWalletRequest(payload, issuer, options);
    return new LoginRequest(vc);
  }
}
LoginRequest.type = 'LoginRequest';
class DataSharingRequest extends WalletRequest {
  /**
   * Alternative constructor that returns type DataSharingRequest
   */
  static async signRequest(payload, issuer, options = {}) {
    const vc = await super.signWalletRequest(payload, issuer, options);
    return new DataSharingRequest(vc);
  }
}
DataSharingRequest.type = 'DataSharingRequest';

function getAccountNameFromDid(did) {
  const parsed = parseDid(did);
  const id = parsed.id.split(':');
  const accountName = id[id.length - 1];
  return Name.from(accountName);
}
function parseDid(did) {
  const parsed = parse(did);
  if (!parsed) throw new Error('Invalid DID');
  return parsed;
}

var AccountType;
(function (AccountType) {
  AccountType["PERSON"] = "PERSON";
  AccountType["ORG"] = "ORG";
  AccountType["APP"] = "APP";
  AccountType["GOV"] = "GOV";
})(AccountType || (AccountType = {}));
// eslint-disable-next-line @typescript-eslint/no-namespace
(function (AccountType) {
  /*
   * Returns the index of the enum value
   *
   * @param value The level to get the index of
   */
  function indexFor(value) {
    return Object.keys(AccountType).indexOf(value);
  }
  AccountType.indexFor = indexFor;
  /*
   * Creates an AccountType from a string or index of the level
   *
   * @param value The string or index
   */
  function from(value) {
    let index;
    if (typeof value !== 'number') {
      index = AccountType.indexFor(value);
    } else {
      index = value;
    }
    return Object.values(AccountType)[index];
  }
  AccountType.from = from;
  function getPreSuffix(value) {
    return value.toLowerCase();
  }
  AccountType.getPreSuffix = getPreSuffix;
})(AccountType || (AccountType = {}));
class TonomyUsername {
  constructor(username) {
    this.username = void 0;
    this.usernameHash = void 0;
    this.username = username;
    this.usernameHash = sha256(this.username);
  }
  static fromUsername(username, type, suffix) {
    const fullUsername = username.normalize('NFKC').toLowerCase() + '.' + AccountType.getPreSuffix(type) + suffix;
    return new TonomyUsername(fullUsername);
  }
  static fromFullUsername(username) {
    return new TonomyUsername(username);
  }
  getBaseUsername() {
    var _this$username;
    return (_this$username = this.username) == null ? void 0 : _this$username.split('.')[0];
  }
  /**
   * @returns the username hash
   */
  toString() {
    return this.username;
  }
  /**
   * Returns the JSON string representation of the object - which is the username string
   *
   * @description This is used in JSON.stringify(). It is not recommended to use this method directly.
   * Only the username is returned, not the full object. This is all that is needed to reconstruct the
   * full object in the constructor, and thus minimizes the size of the JSON string.
   *
   * @override Serializable.toJSON
   * @returns {string} the username
   */
  toJSON() {
    return this.toString();
  }
}

/* eslint-disable camelcase */
const debug$6 = Debug('tonomy-id-sdk:services:blockchain:contracts:token');
const CONTRACT_NAME$7 = 'eosio.token';
function assetToNumberString(asset, symbol) {
  if (!symbol) {
    symbol = getSettings().currencySymbol;
  }
  const [res, currency] = asset.split(' ');
  if (currency !== symbol) {
    debug$6(`Invalid currency symbol: expected ${symbol}, for asset ${asset}`);
    throw new Error(`Invalid currency symbol: expected ${symbol}, got ${currency}`);
  }
  return res;
}
// FIXME: Remove use of this function. We should never use a number to represent tokens as they are not precise and cannot handle large numbers.
/**
 * @deprecated use assetToDecimal instead
 */
function assetToAmount(asset) {
  return parseFloat(assetToNumberString(asset));
}
function assetToDecimal(asset) {
  return new Decimal(assetToNumberString(asset));
}
/**
 * @deprecated
 * see FIXME above
 */
function amountToAsset(amount, symbol, precision = 6) {
  return amount.toFixed(precision) + ' ' + symbol;
}
function amountToSupplyPercentage(amount) {
  return amount.mul(100).div(EosioTokenContract.TOTAL_SUPPLY).toFixed(8) + '%';
}
class EosioTokenContract {
  static get Instance() {
    return this.singletonInstande || (this.singletonInstande = new this());
  }
  async create(supply, signer) {
    const actions = [{
      account: CONTRACT_NAME$7,
      name: 'create',
      authorization: [{
        actor: CONTRACT_NAME$7,
        permission: 'active'
      }],
      data: {
        issuer: CONTRACT_NAME$7,
        maximum_supply: supply
      }
    }];
    return await transact(Name.from(CONTRACT_NAME$7), actions, signer);
  }
  async issue(to, quantity, signer) {
    const actions = [{
      account: CONTRACT_NAME$7,
      name: 'issue',
      authorization: [{
        actor: CONTRACT_NAME$7,
        permission: 'active'
      }],
      data: {
        to,
        quantity,
        memo: 'issued'
      }
    }];
    return await transact(Name.from(CONTRACT_NAME$7), actions, signer);
  }
  async transfer(from, to, quantity, memo, signer) {
    const actions = [{
      account: CONTRACT_NAME$7,
      name: 'transfer',
      authorization: [{
        actor: from.toString(),
        permission: 'active'
      }],
      data: {
        from,
        to,
        quantity,
        memo
      }
    }];
    return await transact(Name.from(CONTRACT_NAME$7), actions, signer);
  }
  async getBalance(account) {
    const assets = await (await getApi()).v1.chain.get_currency_balance(CONTRACT_NAME$7, account, getSettings().currencySymbol);
    if (assets.length === 0) return 0;
    return assets[0].value;
  }
  async getBalanceDecimal(account) {
    const assets = await (await getApi()).v1.chain.get_currency_balance(CONTRACT_NAME$7, account, getSettings().currencySymbol);
    if (assets.length === 0) return new Decimal(0);
    return new Decimal(assets[0].quantity);
  }
}
EosioTokenContract.singletonInstande = void 0;
EosioTokenContract.TOTAL_SUPPLY = 50000000000.0;

/* eslint-disable camelcase */
const CONTRACT_NAME$6 = 'vesting.tmy';
const TONO_SEED_ROUND_PRICE = 0.0001;
const TONO_SEED_LATE_ROUND_PRICE = 0.0002;
const TONO_PUBLIC_SALE_PRICE = 0.0006;
const TONO_CURRENT_PRICE = TONO_SEED_ROUND_PRICE;
const MICROSECONDS_PER_SECOND = 1000000;
const SECONDS_PER_HOUR = 3600;
const MICROSECONDS_PER_DAY = 24 * SECONDS_PER_HOUR * MICROSECONDS_PER_SECOND;
const MICROSECONDS_PER_MONTH = 30 * MICROSECONDS_PER_DAY;
const MICROSECONDS_PER_YEAR = 365 * MICROSECONDS_PER_DAY;
const vestingCategories = new Map([[999,
// Testing Category
{
  startDelay: 10 * MICROSECONDS_PER_SECOND,
  cliffPeriod: 10 * MICROSECONDS_PER_SECOND,
  vestingPeriod: 20 * MICROSECONDS_PER_SECOND,
  tgeUnlock: 0.0,
  name: 'Testing Category (no unlock)'
}], [998,
// Testing Category
{
  startDelay: 10 * MICROSECONDS_PER_SECOND,
  cliffPeriod: 10 * MICROSECONDS_PER_SECOND,
  vestingPeriod: 20 * MICROSECONDS_PER_SECOND,
  tgeUnlock: 0.5,
  name: 'Testing Category (50% unlock)'
}], [1,
// Seed Private Sale (DEPRECIATED)
{
  startDelay: 0 * MICROSECONDS_PER_DAY,
  cliffPeriod: 6 * 30 * MICROSECONDS_PER_DAY,
  vestingPeriod: 2 * 365 * MICROSECONDS_PER_DAY,
  tgeUnlock: 0.0,
  name: 'Seed Private Sale (DEPRECIATED)'
}], [2,
// Strategic Partnerships Private Sale (DEPRECIATED)
{
  startDelay: 6 * 30 * MICROSECONDS_PER_DAY,
  cliffPeriod: 6 * 30 * MICROSECONDS_PER_DAY,
  vestingPeriod: 2 * 365 * MICROSECONDS_PER_DAY,
  tgeUnlock: 0.0,
  name: 'Strategic Partnerships Private Sale (DEPRECIATED)'
}],
// Unchanged:
[3,
// Public Sale (DO NOT USED YET)
{
  startDelay: 0 * MICROSECONDS_PER_DAY,
  cliffPeriod: 0 * MICROSECONDS_PER_DAY,
  vestingPeriod: 0 * MICROSECONDS_PER_DAY,
  tgeUnlock: 0.0,
  name: 'Public Sale (DEPRECIATED)'
}], [4,
// Team
{
  startDelay: 1 * 365 * MICROSECONDS_PER_DAY,
  cliffPeriod: 0 * MICROSECONDS_PER_DAY,
  vestingPeriod: 5 * 365 * MICROSECONDS_PER_DAY,
  tgeUnlock: 0.0,
  name: 'Team'
}], [5,
// Legal and Compliance
{
  startDelay: 0 * MICROSECONDS_PER_DAY,
  cliffPeriod: 0 * MICROSECONDS_PER_DAY,
  vestingPeriod: 1 * 365 * MICROSECONDS_PER_DAY,
  tgeUnlock: 0.0,
  name: 'Legal and Compliance'
}], [6,
// Reserves, Partnerships
{
  startDelay: 0 * MICROSECONDS_PER_DAY,
  cliffPeriod: 0 * MICROSECONDS_PER_DAY,
  vestingPeriod: 2 * 365 * MICROSECONDS_PER_DAY,
  tgeUnlock: 0.0,
  name: 'Reserves, Partnerships'
}], [7,
// Community and Marketing, Platform Dev, Infra Rewards, Ecosystem
{
  startDelay: 0 * MICROSECONDS_PER_DAY,
  cliffPeriod: 0 * MICROSECONDS_PER_DAY,
  vestingPeriod: 5 * 365 * MICROSECONDS_PER_DAY,
  tgeUnlock: 0.0,
  name: 'Community and Marketing, Platform Dev, Infra Rewards, Ecosystem'
}],
// New (replacing depreciated):
[8,
// Seed
{
  startDelay: 6 * MICROSECONDS_PER_MONTH,
  cliffPeriod: 0 * MICROSECONDS_PER_DAY,
  vestingPeriod: 12 * MICROSECONDS_PER_MONTH,
  tgeUnlock: 0.05,
  name: 'Seed'
}], [9,
// Pre-Sale
{
  startDelay: 4 * MICROSECONDS_PER_MONTH,
  cliffPeriod: 0 * MICROSECONDS_PER_DAY,
  vestingPeriod: 12 * MICROSECONDS_PER_MONTH,
  tgeUnlock: 0.075,
  name: 'Pre-Sale'
}], [10,
// Public (TGE)
{
  startDelay: 1 * MICROSECONDS_PER_MONTH,
  cliffPeriod: 0 * MICROSECONDS_PER_DAY,
  vestingPeriod: 3 * MICROSECONDS_PER_MONTH,
  tgeUnlock: 0.25,
  name: 'Public (TGE)'
}], [11,
// Private
{
  startDelay: 3 * MICROSECONDS_PER_MONTH,
  cliffPeriod: 0 * MICROSECONDS_PER_DAY,
  vestingPeriod: 9 * MICROSECONDS_PER_MONTH,
  tgeUnlock: 0.125,
  name: 'Private'
}], [12,
// KOL
{
  startDelay: 1 * MICROSECONDS_PER_MONTH,
  cliffPeriod: 0 * MICROSECONDS_PER_DAY,
  vestingPeriod: 3 * MICROSECONDS_PER_MONTH,
  tgeUnlock: 0.25,
  name: 'KOL'
}], [13,
// Incubator
{
  startDelay: 0 * MICROSECONDS_PER_MONTH,
  cliffPeriod: 0 * MICROSECONDS_PER_DAY,
  vestingPeriod: 6 * MICROSECONDS_PER_DAY,
  tgeUnlock: 0.7,
  name: 'Incubator'
}], [14,
// Liquidity
{
  startDelay: 0 * MICROSECONDS_PER_MONTH,
  cliffPeriod: 0 * MICROSECONDS_PER_DAY,
  vestingPeriod: 6 * MICROSECONDS_PER_DAY,
  tgeUnlock: 0.25,
  name: 'Liquidity'
}]]);
class VestingContract {
  static get Instance() {
    return this.singletonInstance || (this.singletonInstance = new this());
  }
  static calculateVestingPeriod(settings, allocation) {
    const vestingCategory = vestingCategories.get(allocation.vesting_category_type);
    if (!vestingCategory) throw new Error('Invalid vesting category');
    const launchDate = new Date(settings.launch_date + 'Z');
    const vestingStart = addMicroseconds(launchDate, vestingCategory.startDelay);
    const cliffEnd = addMicroseconds(vestingStart, vestingCategory.cliffPeriod);
    const vestingEnd = addMicroseconds(vestingStart, vestingCategory.vestingPeriod);
    return {
      launchDate,
      vestingStart,
      cliffEnd,
      vestingEnd
    };
  }
  constructor(contractName = CONTRACT_NAME$6) {
    this.contractName = CONTRACT_NAME$6;
    this.contractName = contractName;
  }
  async setSettings(salesDateStr, launchDateStr, signer) {
    const action = {
      authorization: [{
        actor: CONTRACT_NAME$6,
        permission: 'active'
      }],
      account: CONTRACT_NAME$6,
      name: 'setsettings',
      data: {
        sales_start_date: salesDateStr,
        launch_date: launchDateStr
      }
    };
    return await transact(Name.from(CONTRACT_NAME$6), [action], signer);
  }
  async assignTokens(sender, holder, amount, categoryId, signer) {
    const action = {
      authorization: [{
        actor: sender.toString(),
        permission: 'active'
      }],
      account: CONTRACT_NAME$6,
      name: 'assigntokens',
      data: {
        sender,
        holder,
        amount,
        category: categoryId
      }
    };
    return await transact(Name.from(CONTRACT_NAME$6), [action], signer);
  }
  async withdraw(holder, signer) {
    const action = {
      authorization: [{
        actor: holder.toString(),
        permission: 'active'
      }],
      account: CONTRACT_NAME$6,
      name: 'withdraw',
      data: {
        holder
      }
    };
    return await transact(Name.from(CONTRACT_NAME$6), [action], signer);
  }
  async migrateAllocation(sender, holder, allocationId, oldAmount, newAmount, oldCategoryId, newCategoryId, signer) {
    const action = {
      authorization: [{
        actor: sender.toString(),
        permission: 'active'
      }, {
        actor: CONTRACT_NAME$6,
        permission: 'active'
      }],
      account: CONTRACT_NAME$6,
      name: 'migratealloc',
      data: {
        sender,
        holder,
        allocation_id: allocationId,
        old_amount: oldAmount,
        new_amount: newAmount,
        old_category_id: oldCategoryId,
        new_category_id: newCategoryId
      }
    };
    return await transact(Name.from(CONTRACT_NAME$6), [action], signer);
  }
  async getSettings() {
    const res = await (await getApi()).v1.chain.get_table_rows({
      code: 'vesting.tmy',
      scope: 'vesting.tmy',
      table: 'settings',
      json: true
    });
    if (res.rows.length === 0) throw new Error('Vesting settings have not yet been set');
    return res.rows[0];
  }
  async getAllocations(account) {
    const res = await (await getApi()).v1.chain.get_table_rows({
      code: 'vesting.tmy',
      scope: account.toString(),
      table: 'allocation',
      json: true,
      limit: VestingContract.getMaxAllocations() + 1
    });
    return res.rows;
  }
  async getBalance(account) {
    const allocations = await this.getAllocations(account);
    let totalBalance = new Decimal(0);
    for (const allocation of allocations) {
      const tokens = assetToDecimal(allocation.tokens_allocated);
      totalBalance = totalBalance.add(tokens);
    }
    return totalBalance.toNumber();
  }
  getVestingCategory(categoryId) {
    const vestingCategory = vestingCategories.get(categoryId);
    if (!vestingCategory) {
      throw new Error(`Vesting category ${categoryId} not found`);
    }
    return vestingCategory;
  }
  getVestingPeriodYears(categoryId) {
    const vestingCategory = vestingCategories.get(categoryId);
    if (!vestingCategory) {
      throw new Error(`Vesting category ${categoryId} not found`);
    }
    const vestingPeriod = vestingCategory.vestingPeriod;
    // Convert to seconds for categories 999 and 998, otherwise to years
    if (categoryId === 999 || categoryId === 998) {
      return `${(vestingPeriod / MICROSECONDS_PER_SECOND).toFixed(2)}`;
    } else {
      return `${(vestingPeriod / MICROSECONDS_PER_DAY).toFixed(2)}`;
    }
  }
  getVestingPeriod(categoryId) {
    const vestingCategory = vestingCategories.get(categoryId);
    if (!vestingCategory) {
      throw new Error(`Vesting category ${categoryId} not found`);
    }
    const vestingPeriod = vestingCategory.vestingPeriod;
    const vestingPeriodInDays = vestingPeriod / MICROSECONDS_PER_DAY;
    if (vestingPeriodInDays < 1) {
      // If less than a day, check if it's less than an hour (in seconds)
      const vestingPeriodInSeconds = vestingPeriod / MICROSECONDS_PER_SECOND;
      if (vestingPeriodInSeconds < 60) {
        // Return seconds if it's less than a minute
        return `${vestingPeriodInSeconds.toFixed(0)} seconds`;
      } else {
        // Return hours if it's more than a minute but less than a day
        return `${(vestingPeriodInSeconds / SECONDS_PER_HOUR).toFixed(1)} hours`;
      }
    } else if (vestingPeriodInDays < 30) {
      // Return days if it's less than 30 days
      return `${vestingPeriodInDays.toFixed(1)} days`;
    } else {
      // Calculate months or years if it's 30 days or more
      const vestingPeriodInMonths = vestingPeriod / MICROSECONDS_PER_MONTH;
      const vestingPeriodInYears = vestingPeriod / MICROSECONDS_PER_YEAR;
      if (vestingPeriodInMonths < 12) {
        return `${vestingPeriodInMonths.toFixed(1)} months`;
      } else {
        return `${vestingPeriodInYears.toFixed(1)} years`;
      }
    }
  }
  async getVestingAllocations(account) {
    const allocations = await this.getAllocations(account);
    const currentTime = new Date();
    const allocationsDetails = [];
    for (const allocation of allocations) {
      const tokensAllocated = assetToAmount(allocation.tokens_allocated);
      const unlocked = assetToAmount(allocation.tokens_claimed);
      const settings = await this.getSettings();
      const vestingPeriods = VestingContract.calculateVestingPeriod(settings, allocation);
      const {
        vestingStart,
        cliffEnd,
        vestingEnd
      } = vestingPeriods;
      const vestingCategory = this.getVestingCategory(allocation.vesting_category_type);
      let claimable = 0;
      if (currentTime >= cliffEnd) {
        // Calculate the percentage of the vesting period that has passed
        const timeSinceVestingStart = (currentTime.getTime() - vestingStart.getTime()) / 1000;
        const vestingDuration = (vestingEnd.getTime() - vestingStart.getTime()) / 1000;
        const vestingProgress = Math.min(timeSinceVestingStart / vestingDuration, 1.0); // Ensure it doesn't exceed 100%
        // Calculate unlockable amount considering TGE unlock
        claimable = tokensAllocated * ((1.0 - vestingCategory.tgeUnlock) * vestingProgress + vestingCategory.tgeUnlock);
      }
      const unlockable = claimable - unlocked;
      const locked = tokensAllocated - unlocked;
      const unlockAtVestingStart = vestingCategory.tgeUnlock;
      const saleStart = new Date(settings.sales_start_date);
      allocationsDetails.push({
        totalAllocation: tokensAllocated,
        unlockable,
        unlocked,
        locked,
        vestingStart,
        unlockAtVestingStart,
        allocationDate: new Date(saleStart.getTime() + allocation.time_since_sale_start._count / 1000),
        vestingPeriod: this.getVestingPeriod(allocation.vesting_category_type),
        categoryId: allocation.vesting_category_type
      });
    }
    const totalAllocation = allocationsDetails.reduce((sum, allocation) => sum + allocation.totalAllocation, 0);
    const totalUnlockable = allocationsDetails.reduce((sum, allocation) => sum + allocation.unlockable, 0);
    const totalUnlocked = allocationsDetails.reduce((sum, allocation) => sum + allocation.unlocked, 0);
    const totalLocked = allocationsDetails.reduce((sum, allocation) => sum + allocation.locked, 0);
    return {
      totalAllocation,
      unlockable: totalUnlockable,
      unlocked: totalUnlocked,
      locked: totalLocked,
      allocationsDetails
    };
  }
}
VestingContract.singletonInstance = void 0;
VestingContract.getMaxAllocations = () => getSettings().environment === 'test' || getSettings().environment === 'staging' ? 5 : 150;
VestingContract.SALE_START_DATE = '2024-04-30T12:00:00';
VestingContract.VESTING_START_DATE = '2030-01-01T00:00:00';

/* eslint-disable camelcase */
const debug$5 = Debug('tonomy-sdk:tonomy-contract');
const CONTRACT_NAME$5 = 'tonomy';
const GOVERNANCE_ACCOUNT_NAME = 'tonomy';
var PermissionLevel;
(function (PermissionLevel) {
  PermissionLevel["OWNER"] = "OWNER";
  PermissionLevel["ACTIVE"] = "ACTIVE";
  PermissionLevel["PASSWORD"] = "PASSWORD";
  PermissionLevel["PIN"] = "PIN";
  PermissionLevel["BIOMETRIC"] = "BIOMETRIC";
  PermissionLevel["LOCAL"] = "LOCAL";
})(PermissionLevel || (PermissionLevel = {}));
var AccountTypeEnum;
(function (AccountTypeEnum) {
  AccountTypeEnum[AccountTypeEnum["Person"] = 0] = "Person";
  AccountTypeEnum[AccountTypeEnum["Organization"] = 1] = "Organization";
  AccountTypeEnum[AccountTypeEnum["App"] = 2] = "App";
  AccountTypeEnum[AccountTypeEnum["Gov"] = 3] = "Gov";
  AccountTypeEnum[AccountTypeEnum["Service"] = 4] = "Service";
})(AccountTypeEnum || (AccountTypeEnum = {}));
// eslint-disable-next-line @typescript-eslint/no-namespace
(function (PermissionLevel) {
  /*
   * Returns the index of the enum value
   *
   * @param value The value to get the index of
   */
  function indexFor(value) {
    return Object.keys(PermissionLevel).indexOf(value);
  }
  PermissionLevel.indexFor = indexFor;
  /*
   * Creates an PermissionLevel from a string or index of the level
   *
   * @param value The string or index
   */
  function from(value) {
    let index;
    if (typeof value !== 'number') {
      index = PermissionLevel.indexFor(value);
    } else {
      index = value;
    }
    return Object.values(PermissionLevel)[index];
  }
  PermissionLevel.from = from;
})(PermissionLevel || (PermissionLevel = {}));
function calculateRamPrice() {
  // See https://docs.google.com/spreadsheets/d/1_S0S7Gu-PHzt-IzCqNl3CaWnniAt1KwaXDB50roTZUQ/edit?gid=1773951365#gid=1773951365&range=C84
  const ramPricePerGb = 7; // $7.00 per GB of RAM taken from standard AWS EC2 pricing
  const numberOfNodes = 29;
  const costOverhead = 1; // 100% overhead
  const totalRamPrice = ramPricePerGb * numberOfNodes * (1 + costOverhead); // $ / Gb
  const totalRamPriceBytes = totalRamPrice / (1024 * 1024 * 1024); // $ / byte
  return TONO_PUBLIC_SALE_PRICE / totalRamPriceBytes; // bytes / TONO
}
const RAM_PRICE = calculateRamPrice(); // bytes / token
const RAM_FEE = 0.25 / 100; // 0.25%
const TOTAL_RAM_AVAILABLE = 8 * 1024 * 1024 * 1024; // 8 GB
/**
 * Converts bytes to tokens.
 *
 * @param bytes The number of bytes to convert.
 * @returns The converted value in tokens.
 */
function bytesToTokens(bytes) {
  return (bytes * (1 + RAM_FEE) / RAM_PRICE).toFixed(6) + ` ${getSettings().currencySymbol}`;
}
class TonomyContract {
  static get Instance() {
    return this.singletonInstance || (this.singletonInstance = new this());
  }
  /**
   * Buys RAM for an account
   *
   * @param  daoOwner - The owner of the DAO (Name is assumed to be a class that represents an EOSIO account name)
   * @param account - The name of the app buying RAM (Name is assumed to be a class that represents an EOSIO account name)
   * @param quant - The quantity of RAM to buy (Asset is assumed to be a class that represents an EOSIO asset)
   */
  async buyRam(dao_owner, app, quant, signer) {
    const actions = [{
      account: CONTRACT_NAME$5,
      name: 'buyram',
      authorization: [{
        actor: app,
        permission: 'active'
      }, {
        actor: dao_owner,
        permission: 'active'
      }],
      data: {
        dao_owner,
        app,
        quant
      }
    }];
    return await transact(Name.from(CONTRACT_NAME$5), actions, signer);
  }
  /**
   * Sets the resource parameters
   *
   * @param ram_price - The price of RAM (bytes per token)
   * @param total_ram_available - The total available RAM.
   * @param ram_fee - The fee for RAM.
  
   */
  async setResourceParams(ram_price, total_ram_available, ram_fee, signer) {
    const action = {
      authorization: [{
        actor: GOVERNANCE_ACCOUNT_NAME,
        permission: 'active'
      }],
      account: CONTRACT_NAME$5,
      name: 'setresparams',
      data: {
        ram_price,
        total_ram_available,
        ram_fee
      }
    };
    return await transact(Name.from(CONTRACT_NAME$5), [action], signer);
  }
  async newperson(username_hash, password_key, password_salt, signer) {
    const action = {
      authorization: [{
        actor: CONTRACT_NAME$5,
        permission: 'active'
      }],
      account: CONTRACT_NAME$5,
      name: 'newperson',
      data: {
        username_hash,
        password_key,
        password_salt
      }
    };
    return await transact(Name.from(CONTRACT_NAME$5), [action], signer);
  }
  async updatekeysper(account, keys, signer) {
    const accountInfo = await getAccount(account);
    const actions = [];
    if (Object.keys(keys).length === 0) throwError('At least one key must be provided', SdkErrors.UpdateKeysTransactionNoKeys);
    for (const key in keys) {
      const permission = PermissionLevel.from(key);
      // "keys as any" fixes typescript issue see https://stackoverflow.com/a/57192972
      const publicKey = keys[key];
      let link_auth = true;
      try {
        const accountPermission = accountInfo.getPermission(permission.toLowerCase());
        if (accountPermission && accountPermission.linked_actions.find(a => a.account.equals(CONTRACT_NAME$5) && a.action.equals('loginwithapp'))) {
          link_auth = false;
        }
      } catch (e) {
        if (!e.message.startsWith('Unknown permission ')) {
          throw e;
        }
      }
      actions.push({
        authorization: [{
          actor: account,
          permission: 'active'
        }],
        account: CONTRACT_NAME$5,
        name: 'updatekeyper',
        data: {
          account,
          permission: PermissionLevel.indexFor(permission),
          key: publicKey,
          link_auth
        }
      });
    }
    return await transact(Name.from(CONTRACT_NAME$5), actions, signer);
  }
  async updateactive(account, active, signer) {
    const action = {
      authorization: [{
        actor: account,
        permission: 'active'
      }],
      account: CONTRACT_NAME$5,
      name: 'updateactive',
      data: {
        account,
        active
      }
    };
    return await transact(Name.from(CONTRACT_NAME$5), [action], signer);
  }
  async newapp(app_name, description, username_hash, logo_url, origin, background_color, accent_color, key, signer) {
    /^(((http:\/\/)|(https:\/\/))?)(([a-zA-Z0-9.])+)((:{1}[0-9]+)?)$/g.test(origin);
    /^(((http:\/\/)|(https:\/\/))?)(([a-zA-Z0-9.])+)((:{1}[0-9]+)?)([?#/a-zA-Z0-9.]*)$/g.test(logo_url);
    // Combine the individual fields into a JSON string
    const json_data = JSON.stringify({
      app_name: app_name,
      description: description,
      logo_url: logo_url,
      background_color: background_color,
      accent_color: accent_color
    });
    const action = {
      authorization: [{
        actor: CONTRACT_NAME$5,
        permission: 'active'
      }],
      account: CONTRACT_NAME$5,
      name: 'newapp',
      data: {
        json_data,
        username_hash,
        origin: origin,
        key
      }
    };
    return await transact(Name.from(CONTRACT_NAME$5), [action], signer);
  }
  async loginwithapp(account, app, parent, key, signer) {
    const action = {
      authorization: [{
        actor: account,
        permission: parent
      }],
      account: CONTRACT_NAME$5,
      name: 'loginwithapp',
      data: {
        account,
        app,
        parent,
        key
      }
    };
    return await transact(Name.from(CONTRACT_NAME$5), [action], signer);
  }
  async getPerson(account) {
    let data;
    const api = await getApi();
    if (account instanceof TonomyUsername) {
      // this is a username
      const usernameHash = account.usernameHash;
      data = await api.v1.chain.get_table_rows({
        code: CONTRACT_NAME$5,
        scope: CONTRACT_NAME$5,
        table: 'people',
        lower_bound: Checksum256.from(usernameHash),
        limit: 1,
        index_position: 'secondary'
      });
      if (!data || !data.rows) throwError('No data found', SdkErrors.DataQueryNoRowDataFound);
      if (data.rows.length === 0 || data.rows[0].username_hash.toString() !== usernameHash) {
        throwError('Person with username "' + account.username + '" not found', SdkErrors.UsernameNotFound);
      }
    } else {
      // use the account name directly
      data = await api.v1.chain.get_table_rows({
        code: CONTRACT_NAME$5,
        scope: CONTRACT_NAME$5,
        table: 'people',
        lower_bound: account,
        limit: 1
      });
      if (!data || !data.rows) throwError('No data found', SdkErrors.DataQueryNoRowDataFound);
      if (data.rows.length === 0 || data.rows[0].account_name !== account.toString()) {
        throwError('Person with account name "' + account.toString() + '" not found', SdkErrors.AccountDoesntExist);
      }
    }
    const idData = data.rows[0];
    return {
      account_name: Name.from(idData.account_name),
      status: idData.status,
      username_hash: Checksum256.from(idData.username_hash),
      password_salt: Checksum256.from(idData.password_salt),
      version: idData.version
    };
  }
  async getApp(account) {
    let data;
    const api = await getApi();
    if (account instanceof TonomyUsername) {
      // this is a username
      const usernameHash = account.usernameHash;
      data = await api.v1.chain.get_table_rows({
        code: CONTRACT_NAME$5,
        scope: CONTRACT_NAME$5,
        table: 'appsv2',
        lower_bound: Checksum256.from(usernameHash),
        limit: 1,
        index_position: 'secondary'
      });
      if (!data || !data.rows) throwError('No data found', SdkErrors.DataQueryNoRowDataFound);
      if (data.rows.length === 0 || data.rows[0].username_hash.toString() !== usernameHash) {
        throwError('Account with username "' + account.username + '" not found', SdkErrors.UsernameNotFound);
      }
    } else if (account instanceof Name) {
      // use the account name directly
      data = await api.v1.chain.get_table_rows({
        code: CONTRACT_NAME$5,
        scope: CONTRACT_NAME$5,
        table: 'appsv2',
        lower_bound: account,
        limit: 1
      });
      if (!data || !data.rows) throwError('No data found', SdkErrors.DataQueryNoRowDataFound);
      if (data.rows.length === 0 || data.rows[0].account_name !== account.toString()) {
        throwError('Account "' + account.toString() + '" not found', SdkErrors.AccountDoesntExist);
      }
    } else {
      // account is the origin
      const origin = account;
      const originHash = sha256(origin);
      data = await api.v1.chain.get_table_rows({
        code: CONTRACT_NAME$5,
        scope: CONTRACT_NAME$5,
        table: 'appsv2',
        lower_bound: Checksum256.from(originHash),
        limit: 1,
        index_position: 'tertiary'
      });
      if (!data || !data.rows) throwError('No data found', SdkErrors.DataQueryNoRowDataFound);
      debug$5('getApp table row', origin, JSON.stringify(data.rows, null, 2));
      if (data.rows.length === 0 || data.rows[0].origin !== origin) {
        throwError('Account with origin "' + origin + '" not found', SdkErrors.OriginNotFound);
      }
    }
    const idData = data.rows[0];
    debug$5('getApp idData', JSON.stringify(idData, null, 2));
    const appInfo = JSON.parse(idData.json_data);
    return {
      app_name: appInfo.app_name,
      description: appInfo.description,
      logo_url: appInfo.logo_url,
      origin: idData.origin,
      account_name: Name.from(idData.account_name),
      background_color: appInfo.background_color,
      accent_color: appInfo.accent_color,
      username_hash: Checksum256.from(idData.username_hash),
      version: idData.version
    };
  }
  async getApps() {
    const api = await getApi();
    const data = await api.v1.chain.get_table_rows({
      code: CONTRACT_NAME$5,
      scope: CONTRACT_NAME$5,
      table: 'apps',
      limit: 1000
    });
    if (!data || !data.rows) throwError('No data found', SdkErrors.DataQueryNoRowDataFound);
    if (data.rows.length === 0) {
      throwError('No apps found', SdkErrors.DataQueryNoRowDataFound);
    }
    return data.rows.map(row => ({
      app_name: row.app_name,
      description: row.description,
      logo_url: row.logo_url,
      origin: row.origin,
      account_name: Name.from(row.account_name),
      username_hash: Checksum256.from(row.username_hash),
      background_color: row.background_color,
      accent_color: row.accent_color,
      version: row.version
    }));
  }
  async adminSetApp(accountName, appName, description, usernameHash, logoUrl, origin, backgroundColor, accentColor, signer) {
    const json_data = JSON.stringify({
      app_name: appName,
      description: description,
      logo_url: logoUrl,
      background_color: backgroundColor,
      accent_color: accentColor
    });
    const action = {
      authorization: [{
        actor: CONTRACT_NAME$5,
        permission: 'active'
      }],
      account: CONTRACT_NAME$5,
      name: 'adminsetapp',
      data: {
        account_name: Name.from(accountName),
        json_data,
        username_hash: usernameHash,
        origin
      }
    };
    return await transact(Name.from(CONTRACT_NAME$5), [action], signer);
  }
}
TonomyContract.singletonInstance = void 0;
async function getAccountNameFromUsername(username) {
  return (await TonomyContract.Instance.getPerson(username)).account_name;
}

var AppStatusEnum;
(function (AppStatusEnum) {
  AppStatusEnum["PENDING"] = "PENDING";
  AppStatusEnum["CREATING"] = "CREATING";
  AppStatusEnum["READY"] = "READY";
  AppStatusEnum["DEACTIVATED"] = "DEACTIVATED";
})(AppStatusEnum || (AppStatusEnum = {}));
// eslint-disable-next-line @typescript-eslint/no-namespace
(function (AppStatusEnum) {
  /*
   * Returns the index of the enum value
   *
   * @param value The level to get the index of
   */
  function indexFor(value) {
    return Object.keys(AppStatusEnum).indexOf(value);
  }
  AppStatusEnum.indexFor = indexFor;
  /*
   * Creates an AppStatusEnum from a string or index of the level
   *
   * @param value The string or index
   */
  function from(value) {
    let index;
    if (typeof value !== 'number') {
      index = AppStatusEnum.indexFor(value);
    } else {
      index = value;
    }
    return Object.values(AppStatusEnum)[index];
  }
  AppStatusEnum.from = from;
})(AppStatusEnum || (AppStatusEnum = {}));

const tonomyContract$a = TonomyContract.Instance;
class App {
  constructor(options) {
    this.accountName = void 0;
    this.appName = void 0;
    this.username = void 0;
    this.usernameHash = void 0;
    this.description = void 0;
    this.logoUrl = void 0;
    this.origin = void 0;
    this.version = void 0;
    this.status = void 0;
    this.accentColor = void 0;
    this.backgroundColor = void 0;
    this.accountName = options.accountName;
    this.appName = options.appName;
    this.username = options.username;
    this.description = options.description;
    this.logoUrl = options.logoUrl;
    this.origin = options.origin;
    this.version = options.version;
    this.status = options.status;
    this.accentColor = options.accentColor;
    this.backgroundColor = options.backgroundColor;
    if (options.usernameHash) {
      this.usernameHash = options.usernameHash;
    } else if (options.username) {
      this.usernameHash = Checksum256.from(options.username.usernameHash);
    } else {
      throw new Error('Either username or usernameHash must be provided');
    }
  }
  static async create(options) {
    const username = TonomyUsername.fromUsername(options.usernamePrefix, AccountType.APP, getSettings().accountSuffix);
    const res = await tonomyContract$a.newapp(options.appName, options.description, username.usernameHash, options.logoUrl, options.origin, options.backgroundColor, options.accentColor, options.publicKey, options.signer);
    const newAccountAction = res.processed.action_traces[0].inline_traces[0].act;
    return new App({
      ...options,
      accountName: Name.from(newAccountAction.data.name),
      username,
      version: newAccountAction.data.version,
      status: AppStatusEnum.READY
    });
  }
  static async getApp(origin) {
    const contractAppData = await tonomyContract$a.getApp(origin);
    return new App({
      accountName: contractAppData.account_name,
      appName: contractAppData.app_name,
      usernameHash: contractAppData.username_hash,
      description: contractAppData.description,
      logoUrl: contractAppData.logo_url,
      origin: contractAppData.origin,
      version: contractAppData.version,
      status: AppStatusEnum.READY,
      accentColor: contractAppData.accent_color,
      backgroundColor: contractAppData.background_color
    });
  }
}

/* eslint-disable camelcase */
const CONTRACT_NAME$4 = 'eosio';
class EosioContract {
  static get Instance() {
    return this.singletonInstance || (this.singletonInstance = new this());
  }
  constructor(contractName = CONTRACT_NAME$4) {
    this.contractName = CONTRACT_NAME$4;
    this.contractName = contractName;
  }
  /**
   * Deploys a contract at the specified address
   *
   * @param account - Account where the contract will be deployed
   * @param wasmFileContents - wasmFile after reading with fs.readFileSync(path) or equivalent
   * @param abiFileContents - abiFile after reading with fs.readFileSync(path, `utf8`) or equivalent
   * @param signer - Signer to sign the transaction
   * @param [extraAuthorization] - Extra authorization to be added to the transaction
   */
  async deployContract(account, wasmFileContent, abiFileContent, signer, options = {}) {
    // 1. Prepare SETCODE
    // read the file and make a hex string out of it
    const wasm = wasmFileContent.toString(`hex`);
    // 2. Prepare SETABI
    const abi = JSON.parse(abiFileContent);
    const abiDef = ABI.from(abi);
    const abiSerializedHex = Serializer.encode({
      object: abiDef
    }).hexString;
    // 3. Send transaction with both setcode and setabi actions
    const setCodeAction = {
      account: CONTRACT_NAME$4,
      name: 'setcode',
      authorization: [{
        actor: account.toString(),
        permission: 'active'
      }],
      data: {
        account: account.toString(),
        vmtype: 0,
        vmversion: 0,
        code: wasm
      }
    };
    if (options.extraAuthorization) setCodeAction.authorization.push(options.extraAuthorization);
    const setAbiAction = {
      account: CONTRACT_NAME$4,
      name: 'setabi',
      authorization: [{
        actor: account.toString(),
        permission: 'active'
      }],
      data: {
        account,
        abi: abiSerializedHex
      }
    };
    if (options.extraAuthorization) setAbiAction.authorization.push(options.extraAuthorization);
    const actions = [setCodeAction, setAbiAction];
    return await transact(Name.from(CONTRACT_NAME$4), actions, signer);
  }
  async updateauth(account, permission, parent, auth, signer) {
    const action = {
      authorization: [{
        actor: account,
        permission: parent // all higher parents, and permission, work as authorization. though permission is supposed to be the authorization that works
      }],
      account: CONTRACT_NAME$4,
      name: 'updateauth',
      data: {
        account,
        permission,
        parent: permission === 'owner' ? '' : parent,
        auth
      }
    };
    return await transact(Name.from(CONTRACT_NAME$4), [action], signer);
  }
  async newaccount(creator, name, owner, active, signer) {
    const action = {
      authorization: [{
        actor: creator.toString(),
        permission: 'active'
      }],
      account: CONTRACT_NAME$4,
      name: 'newaccount',
      data: {
        creator,
        name,
        owner,
        active
      }
    };
    return await transact(Name.from(CONTRACT_NAME$4), [action], signer);
  }
  /**
   * @param account - the permission's owner to be linked and the payer of the RAM needed to store this link,
   * @param code - the owner of the action to be linked,
   * @param type - the action to be linked,
   * @param requirement - the permission to be linked.
   */
  async linkAuth(account, code, type, requirement, signer) {
    const action = {
      authorization: [{
        actor: account,
        permission: 'active'
      }],
      account: CONTRACT_NAME$4,
      name: 'linkauth',
      data: {
        account,
        code,
        type,
        requirement
      }
    };
    return await transact(Name.from(CONTRACT_NAME$4), [action], signer);
  }
  async setPriv(account, isPriv, signer) {
    const action = {
      authorization: [{
        actor: CONTRACT_NAME$4,
        permission: 'active'
      }],
      account: CONTRACT_NAME$4,
      name: 'setpriv',
      data: {
        account,
        is_priv: isPriv
      }
    };
    return await transact(Name.from(CONTRACT_NAME$4), [action], signer);
  }
  async setParams(blockchainParameters = defaultBlockchainParams, signer) {
    const action = {
      authorization: [{
        actor: CONTRACT_NAME$4,
        permission: 'active'
      }],
      account: CONTRACT_NAME$4,
      name: 'setparams',
      data: {
        params: blockchainParameters
      }
    };
    return await transact(Name.from(CONTRACT_NAME$4), [action], signer);
  }
}
EosioContract.singletonInstance = void 0;
const defaultBlockchainParams = {
  // NET
  max_block_net_usage: 1024 * 1024,
  target_block_net_usage_pct: 8000,
  // 80%
  max_transaction_net_usage: 1024 * 512,
  base_per_transaction_net_usage: 12,
  net_usage_leeway: 500,
  context_free_discount_net_usage_num: 20,
  context_free_discount_net_usage_den: 100,
  // CPU
  max_block_cpu_usage: 200000,
  target_block_cpu_usage_pct: 8000,
  // 80%
  max_transaction_cpu_usage: 50000,
  min_transaction_cpu_usage: 100,
  // Other
  max_transaction_lifetime: 3600,
  deferred_trx_expiration_window: 600,
  max_transaction_delay: 3888000,
  max_inline_action_size: 512 * 1024,
  max_inline_action_depth: 4,
  max_authority_depth: 6
};

/* eslint-disable camelcase */
const CONTRACT_NAME$3 = 'tonomy';
class TonomyEosioProxyContract {
  static get Instance() {
    return this.singletonInstance || (this.singletonInstance = new this());
  }
  /**
   * Deploys a contract at the specified address
   *
   * @param account - Account where the contract will be deployed
   * @param wasmFileContents - wasmFile after reading with fs.readFileSync(path) or equivalent
   * @param abiFileContents - abiFile after reading with fs.readFileSync(path, `utf8`) or equivalent
   * @param signer - Signer or Signer[] to sign the transaction
   * @param extraAuthorization - Extra authorization to be added to the transaction
   */
  async deployContract(account, wasmFileContent, abiFileContent, signer, options = {}) {
    // 1. Prepare SETCODE
    // read the file and make a hex string out of it
    const wasm = wasmFileContent.toString(`hex`);
    // 2. Prepare SETABI
    const abi = JSON.parse(abiFileContent);
    const abiDef = ABI.from(abi);
    const abiSerializedHex = Serializer.encode({
      object: abiDef
    }).hexString;
    // 3. Send transaction with both setcode and setabi actions
    const setCodeAction = {
      account: CONTRACT_NAME$3,
      name: 'setcode',
      authorization: [{
        actor: account.toString(),
        permission: 'active'
      }],
      data: {
        account: account.toString(),
        vmtype: 0,
        vmversion: 0,
        code: wasm
      }
    };
    if (options.extraAuthorization) setCodeAction.authorization.push(options.extraAuthorization);
    const setAbiAction = {
      account: CONTRACT_NAME$3,
      name: 'setabi',
      authorization: [{
        actor: account.toString(),
        permission: 'active'
      }],
      data: {
        account,
        abi: abiSerializedHex
      }
    };
    if (options.extraAuthorization) setAbiAction.authorization.push(options.extraAuthorization);
    const actions = [setCodeAction, setAbiAction];
    return await transact(Name.from(CONTRACT_NAME$3), actions, signer);
  }
  async updateauth(account, permission, parent, auth, signer, options = {}) {
    var _options$authParent;
    const action = {
      authorization: [{
        actor: GOVERNANCE_ACCOUNT_NAME,
        permission: 'active'
      }, {
        actor: account,
        permission: parent // all higher parents, and permission, work as authorization. though permission is supposed to be the authorization that works
      }],
      account: CONTRACT_NAME$3,
      name: 'updateauth',
      data: {
        account,
        permission,
        parent: permission === 'owner' ? '' : parent,
        auth,
        auth_parent: (_options$authParent = options.authParent) != null ? _options$authParent : false // should be true when a new permission is being created, otherwise false
      }
    };
    return await transact(Name.from(CONTRACT_NAME$3), [action], signer);
  }
  /**
   * @param account - the permission's owner to be linked and the payer of the RAM needed to store this link,
   * @param code - the owner of the action to be linked,
   * @param type - the action to be linked,
   * @param requirement - the permission to be linked.
   */
  async linkAuth(account, code, type, requirement, signer) {
    const action = {
      authorization: [{
        actor: account,
        permission: 'active'
      }],
      account: CONTRACT_NAME$3,
      name: 'linkauth',
      data: {
        account,
        code,
        type,
        requirement
      }
    };
    return await transact(Name.from(CONTRACT_NAME$3), [action], signer);
  }
  async setpriv(account, isPriv, signer) {
    const action = {
      authorization: [{
        actor: GOVERNANCE_ACCOUNT_NAME,
        permission: 'active'
      }, {
        actor: account,
        permission: 'active'
      }],
      account: CONTRACT_NAME$3,
      name: 'setpriv',
      data: {
        account,
        is_priv: isPriv
      }
    };
    return await transact(Name.from(CONTRACT_NAME$3), [action], signer);
  }
}
TonomyEosioProxyContract.singletonInstance = void 0;

/* eslint-disable camelcase */
const CONTRACT_NAME$2 = 'eosio.msig';
class EosioMsigContract {
  static get Instance() {
    return this.singletonInstance || (this.singletonInstance = new this());
  }
  constructor(contractName = CONTRACT_NAME$2) {
    this.contractName = CONTRACT_NAME$2;
    this.contractName = contractName;
  }
  async propose(proposer, proposalName, requested, actions, signer) {
    // Serialize the actions
    const serializedActions = await Promise.all(actions.map(async action => {
      if (!action.account || !action.name) throw new Error('Invalid action');
      return {
        account: action.account,
        name: action.name,
        authorization: action.authorization,
        data: await serializeActionData(action.account, action.name, action.data)
      };
    }));
    // Determine expiration
    const now = new Date();
    const expireInSeconds = 60 * 60 * 24 * 7; // 7 days
    const expiration = new Date(now.getTime() + expireInSeconds * 1000);
    const expirationString = expiration.toISOString().split('.')[0];
    const info = await (await getApi()).v1.chain.get_info();
    const trx = {
      expiration: expirationString,
      ref_block_num: info.getTransactionHeader().ref_block_num,
      ref_block_prefix: info.getTransactionHeader().ref_block_prefix,
      max_net_usage_words: 0,
      max_cpu_usage_ms: 0,
      delay_sec: 0,
      context_free_actions: [],
      actions: serializedActions,
      transaction_extensions: []
    };
    const proposalTrx = Transaction.from(trx);
    const proposalHash = proposalTrx.id;
    // Action data (not sure if this is right format)
    const data = {
      proposer,
      proposal_name: proposalName,
      requested,
      trx
    };
    // Propose action
    const action = {
      authorization: [{
        actor: proposer.toString(),
        permission: 'active'
      }],
      account: CONTRACT_NAME$2,
      name: 'propose',
      data
    };
    const myTrx = await transact(Name.from(CONTRACT_NAME$2), [action], signer);
    return {
      proposalHash,
      transaction: myTrx
    };
  }
  async approve(proposer, proposalName, approver, proposalHash, signer) {
    const auth = {
      actor: approver.toString(),
      permission: 'active'
    };
    const actions = [{
      account: CONTRACT_NAME$2,
      name: 'approve',
      authorization: [auth],
      data: {
        proposer,
        proposal_name: proposalName,
        level: auth,
        proposal_hash: proposalHash
      }
    }];
    return await transact(Name.from(CONTRACT_NAME$2), actions, signer);
  }
  async exec(proposer, proposalName, executer, signer) {
    const actions = [{
      account: CONTRACT_NAME$2,
      name: 'exec',
      authorization: [{
        actor: executer.toString(),
        permission: 'active'
      }],
      data: {
        proposer,
        proposal_name: proposalName,
        executer
      }
    }];
    return await transact(Name.from(CONTRACT_NAME$2), actions, signer);
  }
  async cancel(proposer, proposalName, canceler, signer) {
    const actions = [{
      account: CONTRACT_NAME$2,
      name: 'cancel',
      authorization: [{
        actor: canceler.toString(),
        permission: 'active'
      }],
      data: {
        proposer,
        proposal_name: proposalName,
        canceler
      }
    }];
    return await transact(Name.from(CONTRACT_NAME$2), actions, signer);
  }
}
EosioMsigContract.singletonInstance = void 0;

/* eslint-disable camelcase */
const tonomyContract$9 = TonomyContract.Instance;
class DemoTokenContract {
  constructor() {
    this.contractName = void 0;
  }
  static get Instance() {
    return this.singletonInstande || (this.singletonInstande = new this());
  }
  static atAccount(account) {
    const instance = new this();
    instance.contractName = account.toString();
    return instance;
  }
  async getContractName() {
    if (!this.contractName) {
      const username = TonomyUsername.fromUsername('demo', AccountType.APP, getSettings().accountSuffix);
      const app = await tonomyContract$9.getApp(username);
      this.contractName = app.account_name.toString();
    }
    return this.contractName;
  }
  async create(supply, signer) {
    const actions = [{
      account: await this.getContractName(),
      name: 'create',
      authorization: [{
        actor: await this.getContractName(),
        permission: 'active'
      }],
      data: {
        issuer: await this.getContractName(),
        maximum_supply: supply
      }
    }];
    return await transact(Name.from(await this.getContractName()), actions, signer);
  }
  async issue(quantity, signer) {
    const actions = [{
      account: await this.getContractName(),
      name: 'issue',
      authorization: [{
        actor: await this.getContractName(),
        permission: 'active'
      }],
      data: {
        to: await this.getContractName(),
        quantity,
        memo: 'issued'
      }
    }];
    return await transact(Name.from(await this.getContractName()), actions, signer);
  }
  async selfIssue(to, quantity, signer) {
    const actions = [{
      account: await this.getContractName(),
      name: 'issue',
      authorization: [{
        actor: await this.getContractName(),
        permission: 'active'
      }],
      data: {
        to,
        quantity,
        memo: 'self issued'
      }
    }];
    return await transact(Name.from(await this.getContractName()), actions, signer);
  }
  async getBalance(account) {
    const assets = await (await getApi()).v1.chain.get_currency_balance(await this.getContractName(), account, getSettings().currencySymbol);
    if (assets.length === 0) return 0;
    return assets[0].value;
  }
}
DemoTokenContract.singletonInstande = void 0;

class Authority {
  constructor(threshold, keys, accounts, waits) {
    this.threshold = void 0;
    this.keys = void 0;
    this.accounts = void 0;
    this.waits = void 0;
    this.threshold = threshold;
    this.keys = keys;
    this.accounts = accounts;
    this.waits = waits;
  }
  static fromKey(key) {
    const keys = [{
      key,
      weight: 1
    }];
    return new this(1, keys, [], []);
  }
  static fromAccount(permission) {
    if (!Name.pattern.test(permission.actor)) throw new Error(`Invalid account name ${permission.actor}`);
    if (!Name.pattern.test(permission.permission)) throw new Error(`Invalid account permission ${permission.permission}`);
    const accounts = [{
      permission,
      weight: 1
    }];
    return new this(1, [], accounts, []);
  }
  static fromAccountArray(accounts, permission, threshold = 1) {
    const authority = Authority.fromAccount({
      actor: accounts[0],
      permission
    });
    if (accounts.length > 1) {
      for (const arg of accounts.slice(1)) {
        authority.addAccount({
          actor: arg,
          permission
        });
      }
    }
    authority.setThreshold(threshold);
    authority.sort();
    return authority;
  }
  static fromAccountPermission(permission) {
    return new Authority(permission.required_auth.threshold.toNumber(), permission.required_auth.keys.map(keyWeight => ({
      key: keyWeight.key.toString(),
      weight: keyWeight.weight.toNumber()
    })), permission.required_auth.accounts.map(permissionWeight => ({
      permission: {
        actor: permissionWeight.permission.actor.toString(),
        permission: permissionWeight.permission.permission.toString()
      },
      weight: permissionWeight.weight.toNumber()
    })), permission.required_auth.waits.map(waitWeight => ({
      // eslint-disable-next-line camelcase
      wait_sec: waitWeight.wait_sec.toNumber(),
      weight: waitWeight.weight.toNumber()
    })));
  }
  // to add the eosio.code authority for smart contracts
  // https://developers.eos.io/welcome/v2.1/smart-contract-guides/adding-inline-actions#step-1-adding-eosiocode-to-permissions
  addCodePermission(account) {
    this.accounts.push({
      permission: {
        actor: account,
        permission: 'eosio.code'
      },
      weight: 1
    });
    this.sort();
    return this;
  }
  addAccount(permission) {
    this.accounts.push({
      permission,
      weight: 1
    });
    this.sort();
    return this;
  }
  addKey(key, weight) {
    this.keys.push({
      key,
      weight
    });
    this.sort();
    return this;
  }
  setThreshold(threshold) {
    this.threshold = threshold;
    return this;
  }
  /**
   * Sorts the authority weights in place, should be called before including the authority in a `updateauth` action or it might be rejected.
   */
  sort() {
    this.keys.sort((a, b) => String(a.key).localeCompare(String(b.key)));
    this.accounts.sort((a, b) => comparePermissionLevelWeight(a, b));
    this.waits.sort((a, b) => a.wait_sec - b.wait_sec);
  }
}
function comparePermissionLevelWeight(lhs, rhs) {
  if (compareNameType(lhs.permission.actor, rhs.permission.actor) !== 0) {
    return compareNameType(lhs.permission.actor, rhs.permission.actor);
  }
  if (compareNameType(lhs.permission.permission, rhs.permission.permission) !== 0) {
    return compareNameType(lhs.permission.permission, rhs.permission.permission);
  }
  return lhs.weight - rhs.weight;
}
function compareNameType(lhs, rhs) {
  const lhsValue = Name.from(lhs).value.value;
  const rhsValue = Name.from(rhs).value.value;
  return lhsValue.cmp(rhsValue);
}

var _StakingContract;
const debug$4 = Debug('tonomy-sdk:services:blockchain:contracts:staking');
const tonomyContract$8 = TonomyContract.Instance;
const CONTRACT_NAME$1 = 'staking.tmy';
class StakingContract {
  static get Instance() {
    return this.singletonInstance || (this.singletonInstance = new this());
  }
  constructor(contractName = CONTRACT_NAME$1) {
    this.contractName = CONTRACT_NAME$1;
    this.contractName = contractName;
  }
  async stakeTokens(staker, quantity, signer) {
    const account = await getAccount(staker.toString());
    const activePermission = account.getPermission('active');
    if (!activePermission.required_auth.accounts.some(acc => acc.permission.actor.toString() === CONTRACT_NAME$1 && acc.permission.permission.toString() === 'eosio.code')) {
      const newPermission = Authority.fromAccountPermission(activePermission);
      newPermission.addCodePermission(CONTRACT_NAME$1);
      debug$4('Adding staking.tmy@eosio.code to active permission', JSON.stringify(newPermission, null, 2));
      await tonomyContract$8.updateactive(staker.toString(), newPermission, signer);
    }
    const action = {
      authorization: [{
        actor: staker.toString(),
        permission: 'active'
      }],
      account: CONTRACT_NAME$1,
      name: 'staketokens',
      data: {
        account_name: staker.toString(),
        quantity
      }
    };
    return await transact(Name.from(CONTRACT_NAME$1), [action], signer);
  }
  async requestUnstake(staker, allocationId, signer) {
    const action = {
      authorization: [{
        actor: staker.toString(),
        permission: 'active'
      }],
      account: CONTRACT_NAME$1,
      name: 'requnstake',
      data: {
        account_name: staker.toString(),
        allocation_id: allocationId
      }
    };
    return await transact(Name.from(CONTRACT_NAME$1), [action], signer);
  }
  async releaseToken(staker, allocationId, signer) {
    const action = {
      authorization: [{
        actor: staker.toString(),
        permission: 'active'
      }],
      account: CONTRACT_NAME$1,
      name: 'releasetoken',
      data: {
        account_name: staker.toString(),
        allocation_id: allocationId
      }
    };
    return await transact(Name.from(CONTRACT_NAME$1), [action], signer);
  }
  async cron(signer) {
    const action = {
      authorization: [{
        actor: CONTRACT_NAME$1,
        permission: 'active'
      }],
      account: CONTRACT_NAME$1,
      name: 'cron',
      data: {}
    };
    return await transact(Name.from(CONTRACT_NAME$1), [action], signer);
  }
  async resetAll(signer) {
    const action = {
      authorization: [{
        actor: CONTRACT_NAME$1,
        permission: 'active'
      }],
      account: CONTRACT_NAME$1,
      name: 'resetall',
      data: {}
    };
    return await transact(Name.from(CONTRACT_NAME$1), [action], signer);
  }
  async addYield(sender, quantity, signer) {
    const action = {
      authorization: [{
        actor: sender.toString(),
        permission: 'active'
      }],
      account: CONTRACT_NAME$1,
      name: 'addyield',
      data: {
        sender: sender.toString(),
        quantity
      }
    };
    return await transact(Name.from(CONTRACT_NAME$1), [action], signer);
  }
  async setSettings(yearlyStakePool, signer) {
    const action = {
      authorization: [{
        actor: CONTRACT_NAME$1,
        permission: 'active'
      }],
      account: CONTRACT_NAME$1,
      name: 'setsettings',
      data: {
        yearly_stake_pool: yearlyStakePool
      }
    };
    return await transact(Name.from(CONTRACT_NAME$1), [action], signer);
  }
  async getAllocationsData(staker) {
    const res = await (await getApi()).v1.chain.get_table_rows({
      code: CONTRACT_NAME$1,
      scope: staker.toString(),
      table: 'stakingalloc',
      json: true,
      limit: StakingContract.getMaxAllocations() + 1
    });
    return res.rows;
  }
  /**
   * Get staking allocation details with improved types and calculated monthly yield.
   * @param staker - account name.
   * @param settings - current staking settings (including APY).
   */
  async getAllocations(staker, settings) {
    const allocations = await this.getAllocationsData(staker);
    const allocationDetails = [];
    for (const allocation of allocations) {
      const stakedTime = new Date(allocation.stake_time.toString() + 'Z');
      const unstakeTime = new Date(allocation.unstake_time.toString() + 'Z');
      const monthlyYield = amountToAsset(allocation.unstake_requested ? 0 : await this.calculateMonthlyYield(assetToAmount(allocation.tokens_staked), settings), 'TONO'); // Monthly yield from yearly APY.
      const yieldSoFar = amountToAsset(assetToAmount(allocation.tokens_staked) - assetToAmount(allocation.initial_stake), 'TONO');
      allocationDetails.push({
        id: allocation.id,
        staker: staker.toString(),
        initialStake: allocation.initial_stake,
        staked: allocation.tokens_staked,
        yieldSoFar,
        stakedTime,
        unstakeableTime: addSeconds(stakedTime, StakingContract.getLockedDays() * SECONDS_IN_DAY),
        unstakeTime,
        releaseTime: addSeconds(unstakeTime, StakingContract.getReleaseDays() * SECONDS_IN_DAY),
        unstakeRequested: allocation.unstake_requested === 1,
        monthlyYield
      });
    }
    return allocationDetails;
  }
  async getSettingsData() {
    const res = await (await getApi()).v1.chain.get_table_rows({
      code: CONTRACT_NAME$1,
      scope: CONTRACT_NAME$1,
      table: 'settings',
      json: true,
      limit: 1
    });
    if (res.rows.length === 0) throw new Error('Staking settings have not yet been set');
    return res.rows[0];
  }
  /**
   * Get staking settings with improved types and calculated APY.
   *
   * Calculated APY = min(yearlyStakePool / totalStaked, 2.0).
   * (If totalStaked is zero, APY is zero.)
   */
  async getSettings() {
    const settings = await this.getSettingsData();
    const yearlyStakePoolAmount = assetToAmount(settings.yearly_stake_pool);
    const totalStakedAmount = assetToAmount(settings.total_staked);
    const calculatedApy = totalStakedAmount > 0 ? Math.min(yearlyStakePoolAmount / totalStakedAmount, StakingContract.MAX_APY) : StakingContract.MAX_APY;
    return {
      currentYieldPool: settings.current_yield_pool,
      yearlyStakePool: settings.yearly_stake_pool,
      totalStaked: settings.total_staked,
      totalReleasing: settings.total_releasing,
      apy: calculatedApy
    };
  }
  async getAccountData(account) {
    const res = await (await getApi()).v1.chain.get_table_rows({
      code: CONTRACT_NAME$1,
      scope: CONTRACT_NAME$1,
      table: 'stakingaccou',
      json: true,
      lower_bound: Name.from(account),
      limit: 1
    });
    if (res.rows.length === 0 || res.rows[0].staker !== account.toString()) throwError('Account not found in staking contract', SdkErrors.AccountNotFound);
    return res.rows[0];
  }
  async getAccount(account) {
    const raw = await this.getAccountData(account);
    return {
      staker: raw.staker,
      totalYield: raw.total_yield,
      lastPayout: new Date(raw.last_payout + 'Z'),
      payments: raw.payments,
      version: raw.version
    };
  }
  /**
   * Returns full staking data for an account including:
   * - Typed staking account info.
   * - An array of typed allocations.
   * - Aggregated metrics:
   *    • estimatedMonthlyYield: Sum of monthly yields from active allocations.
   *    • lastPayout: Last time the account received a payout.
   *    • version: Account version.
   *    • totalYield: Total yield from all allocations.
   *    • totalStaked: Sum of active (non-unstaking) allocation amounts.
   *    • totalMonthlyYield: Sum of monthly yields from active allocations.
   *    • totalUnlockable: Sum of allocations that have completed the release period.
   *    • totalUnlocking: Sum of allocations still within the release period.
   */
  async getAccountState(account) {
    const settings = await this.getSettings();
    const allocations = await this.getAllocations(account, settings);
    const stakingAccount = await this.getAccount(account);
    let totalStaked = 0;
    let estimatedMonthlyYield = 0;
    const totalUnlockable = 0;
    let totalUnlocking = 0;
    for (const alloc of allocations) {
      const stakedAmount = assetToAmount(alloc.staked);
      if (!alloc.unstakeRequested) {
        totalStaked += stakedAmount;
        estimatedMonthlyYield += assetToAmount(alloc.monthlyYield);
      } else {
        // If unstake has been requested, check if release time has passed.
        totalUnlocking += stakedAmount;
      }
    }
    return {
      ...stakingAccount,
      allocations,
      totalStaked,
      totalUnlockable,
      totalUnlocking,
      estimatedMonthlyYield,
      settings
    };
  }
  async calculateMonthlyYield(amount, settings) {
    return amount * (Math.pow(1 + settings.apy, 1 / 12) - 1);
  }
}
_StakingContract = StakingContract;
StakingContract.singletonInstance = void 0;
StakingContract.getLockedDays = () => getSettings().environment === 'test' || getSettings().environment === 'staging' ? 10 / SECONDS_IN_DAY : 14;
// 14 days or 10 seconds
StakingContract.getReleaseDays = () => getSettings().environment === 'test' || getSettings().environment === 'staging' ? 5 / SECONDS_IN_DAY : 5;
// 5 days or 5 seconds
StakingContract.getMinimumTransfer = () => getSettings().environment === 'test' || getSettings().environment === 'staging' ? 1 : 1000;
// 1000 TONO or 1 TONO
StakingContract.getMaxAllocations = () => getSettings().environment === 'test' || getSettings().environment === 'staging' ? 5 : 20;
// 100 allocations or 5 allocations
StakingContract.getStakingCycleHours = () => getSettings().environment === 'test' || getSettings().environment === 'staging' ? 1 / 60 : 24;
// 24 hours or 1 minute
StakingContract.MAX_APY = 1.0;
StakingContract.STAKING_APY_TARGET = 50 / 100;
// 50%
// Use the TGE unlock: https://docs.google.com/spreadsheets/d/1uyvpgXC0th3Z1_bz4m18dJKy2yyVfYFmcaEyS9fveeA/edit?gid=1074294213#gid=1074294213&range=Q34
StakingContract.STAKING_ESTIMATED_STAKED_PERCENT = 15.1 / 100;
// 15.1%
StakingContract.yearlyStakePool = _StakingContract.STAKING_APY_TARGET * _StakingContract.STAKING_ESTIMATED_STAKED_PERCENT * EosioTokenContract.TOTAL_SUPPLY;

async function apps$1(args) {
  if (args[0] === 'create') {
    var _res$username;
    console.log('Creating new app');
    const options = {
      appName: args[1],
      usernamePrefix: args[2],
      description: args[3],
      logoUrl: args[4],
      origin: args[5],
      backgroundColor: args[6],
      accentColor: args[7],
      publicKey: PublicKey.from(args[8]),
      signer: createSigner$1(getTonomyOperationsKey())
    };
    const res = await App.create(options);
    console.log('New app created with username: ', (_res$username = res.username) == null ? void 0 : _res$username.username);
    console.log('and account name: ', res.accountName.toString());
  } else {
    throw new Error(`Unknown command ${args[0]}`);
  }
}

/**
 * Utility function that works like `Object.apply`, but copies getters and setters properly as well.  Additionally gives
 * the option to exclude properties by name.
 */
const copyProps = (dest, src, exclude = []) => {
    const props = Object.getOwnPropertyDescriptors(src);
    for (let prop of exclude)
        delete props[prop];
    Object.defineProperties(dest, props);
};
/**
 * Returns the full chain of prototypes up until Object.prototype given a starting object.  The order of prototypes will
 * be closest to farthest in the chain.
 */
const protoChain = (obj, currentChain = [obj]) => {
    const proto = Object.getPrototypeOf(obj);
    if (proto === null)
        return currentChain;
    return protoChain(proto, [...currentChain, proto]);
};
/**
 * Identifies the nearest ancestor common to all the given objects in their prototype chains.  For most unrelated
 * objects, this function should return Object.prototype.
 */
const nearestCommonProto = (...objs) => {
    if (objs.length === 0)
        return undefined;
    let commonProto = undefined;
    const protoChains = objs.map(obj => protoChain(obj));
    while (protoChains.every(protoChain => protoChain.length > 0)) {
        const protos = protoChains.map(protoChain => protoChain.pop());
        const potentialCommonProto = protos[0];
        if (protos.every(proto => proto === potentialCommonProto))
            commonProto = potentialCommonProto;
        else
            break;
    }
    return commonProto;
};
/**
 * Creates a new prototype object that is a mixture of the given prototypes.  The mixing is achieved by first
 * identifying the nearest common ancestor and using it as the prototype for a new object.  Then all properties/methods
 * downstream of this prototype (ONLY downstream) are copied into the new object.
 *
 * The resulting prototype is more performant than softMixProtos(...), as well as ES5 compatible.  However, it's not as
 * flexible as updates to the source prototypes aren't captured by the mixed result.  See softMixProtos for why you may
 * want to use that instead.
 */
const hardMixProtos = (ingredients, constructor, exclude = []) => {
    var _a;
    const base = (_a = nearestCommonProto(...ingredients)) !== null && _a !== void 0 ? _a : Object.prototype;
    const mixedProto = Object.create(base);
    // Keeps track of prototypes we've already visited to avoid copying the same properties multiple times.  We init the
    // list with the proto chain below the nearest common ancestor because we don't want any of those methods mixed in
    // when they will already be accessible via prototype access.
    const visitedProtos = protoChain(base);
    for (let prototype of ingredients) {
        let protos = protoChain(prototype);
        // Apply the prototype chain in reverse order so that old methods don't override newer ones.
        for (let i = protos.length - 1; i >= 0; i--) {
            let newProto = protos[i];
            if (visitedProtos.indexOf(newProto) === -1) {
                copyProps(mixedProto, newProto, ['constructor', ...exclude]);
                visitedProtos.push(newProto);
            }
        }
    }
    mixedProto.constructor = constructor;
    return mixedProto;
};
const unique = (arr) => arr.filter((e, i) => arr.indexOf(e) == i);

// Keeps track of constituent classes for every mixin class created by ts-mixer.
const mixins = new WeakMap();
const getMixinsForClass = (clazz) => mixins.get(clazz);
const registerMixins = (mixedClass, constituents) => mixins.set(mixedClass, constituents);

const mergeObjectsOfDecorators = (o1, o2) => {
    var _a, _b;
    const allKeys = unique([...Object.getOwnPropertyNames(o1), ...Object.getOwnPropertyNames(o2)]);
    const mergedObject = {};
    for (let key of allKeys)
        mergedObject[key] = unique([...((_a = o1 === null || o1 === void 0 ? void 0 : o1[key]) !== null && _a !== void 0 ? _a : []), ...((_b = o2 === null || o2 === void 0 ? void 0 : o2[key]) !== null && _b !== void 0 ? _b : [])]);
    return mergedObject;
};
const mergePropertyAndMethodDecorators = (d1, d2) => {
    var _a, _b, _c, _d;
    return ({
        property: mergeObjectsOfDecorators((_a = d1 === null || d1 === void 0 ? void 0 : d1.property) !== null && _a !== void 0 ? _a : {}, (_b = d2 === null || d2 === void 0 ? void 0 : d2.property) !== null && _b !== void 0 ? _b : {}),
        method: mergeObjectsOfDecorators((_c = d1 === null || d1 === void 0 ? void 0 : d1.method) !== null && _c !== void 0 ? _c : {}, (_d = d2 === null || d2 === void 0 ? void 0 : d2.method) !== null && _d !== void 0 ? _d : {}),
    });
};
const mergeDecorators = (d1, d2) => {
    var _a, _b, _c, _d, _e, _f;
    return ({
        class: unique([...(_a = d1 === null || d1 === void 0 ? void 0 : d1.class) !== null && _a !== void 0 ? _a : [], ...(_b = d2 === null || d2 === void 0 ? void 0 : d2.class) !== null && _b !== void 0 ? _b : []]),
        static: mergePropertyAndMethodDecorators((_c = d1 === null || d1 === void 0 ? void 0 : d1.static) !== null && _c !== void 0 ? _c : {}, (_d = d2 === null || d2 === void 0 ? void 0 : d2.static) !== null && _d !== void 0 ? _d : {}),
        instance: mergePropertyAndMethodDecorators((_e = d1 === null || d1 === void 0 ? void 0 : d1.instance) !== null && _e !== void 0 ? _e : {}, (_f = d2 === null || d2 === void 0 ? void 0 : d2.instance) !== null && _f !== void 0 ? _f : {}),
    });
};
const decorators = new Map();
const findAllConstituentClasses = (...classes) => {
    var _a;
    const allClasses = new Set();
    const frontier = new Set([...classes]);
    while (frontier.size > 0) {
        for (let clazz of frontier) {
            const protoChainClasses = protoChain(clazz.prototype).map(proto => proto.constructor);
            const mixinClasses = (_a = getMixinsForClass(clazz)) !== null && _a !== void 0 ? _a : [];
            const potentiallyNewClasses = [...protoChainClasses, ...mixinClasses];
            const newClasses = potentiallyNewClasses.filter(c => !allClasses.has(c));
            for (let newClass of newClasses)
                frontier.add(newClass);
            allClasses.add(clazz);
            frontier.delete(clazz);
        }
    }
    return [...allClasses];
};
const deepDecoratorSearch = (...classes) => {
    const decoratorsForClassChain = findAllConstituentClasses(...classes)
        .map(clazz => decorators.get(clazz))
        .filter(decorators => !!decorators);
    if (decoratorsForClassChain.length == 0)
        return {};
    if (decoratorsForClassChain.length == 1)
        return decoratorsForClassChain[0];
    return decoratorsForClassChain.reduce((d1, d2) => mergeDecorators(d1, d2));
};

function Mixin(...constructors) {
    var _a, _b, _c;
    const prototypes = constructors.map(constructor => constructor.prototype);
    function MixedClass(...args) {
        for (const constructor of constructors)
            // @ts-ignore: potentially abstract class
            copyProps(this, new constructor(...args));
    }
    MixedClass.prototype = hardMixProtos(prototypes, MixedClass)
        ;
    Object.setPrototypeOf(MixedClass, hardMixProtos(constructors, null, ['prototype'])
        );
    let DecoratedMixedClass = MixedClass;
    {
        const classDecorators = deepDecoratorSearch(...constructors)
            ;
        for (let decorator of (_a = classDecorators === null || classDecorators === void 0 ? void 0 : classDecorators.class) !== null && _a !== void 0 ? _a : []) {
            const result = decorator(DecoratedMixedClass);
            if (result) {
                DecoratedMixedClass = result;
            }
        }
        applyPropAndMethodDecorators((_b = classDecorators === null || classDecorators === void 0 ? void 0 : classDecorators.static) !== null && _b !== void 0 ? _b : {}, DecoratedMixedClass);
        applyPropAndMethodDecorators((_c = classDecorators === null || classDecorators === void 0 ? void 0 : classDecorators.instance) !== null && _c !== void 0 ? _c : {}, DecoratedMixedClass.prototype);
    }
    registerMixins(DecoratedMixedClass, constructors);
    return DecoratedMixedClass;
}
const applyPropAndMethodDecorators = (propAndMethodDecorators, target) => {
    const propDecorators = propAndMethodDecorators.property;
    const methodDecorators = propAndMethodDecorators.method;
    if (propDecorators)
        for (let key in propDecorators)
            for (let decorator of propDecorators[key])
                decorator(target, key);
    if (methodDecorators)
        for (let key in methodDecorators)
            for (let decorator of methodDecorators[key])
                decorator(target, key, Object.getOwnPropertyDescriptor(target, key));
};

var KeyManagerLevel;
(function (KeyManagerLevel) {
  KeyManagerLevel["PASSWORD"] = "PASSWORD";
  KeyManagerLevel["ACTIVE"] = "ACTIVE";
  KeyManagerLevel["PIN"] = "PIN";
  KeyManagerLevel["BIOMETRIC"] = "BIOMETRIC";
  KeyManagerLevel["LOCAL"] = "LOCAL";
  KeyManagerLevel["BROWSER_LOCAL_STORAGE"] = "BROWSER_LOCAL_STORAGE";
  KeyManagerLevel["BROWSER_SESSION_STORAGE"] = "BROWSER_SESSION_STORAGE";
})(KeyManagerLevel || (KeyManagerLevel = {}));
// eslint-disable-next-line @typescript-eslint/no-namespace
(function (KeyManagerLevel) {
  /*
   * Returns the index of the enum value
   *
   * @param value The level to get the index of
   */
  function indexFor(value) {
    return Object.keys(KeyManagerLevel).indexOf(value);
  }
  KeyManagerLevel.indexFor = indexFor;
  /*
   * Creates an AuthenticatorLevel from a string or index of the level
   *
   * @param value The string or index
   */
  function from(value) {
    let index;
    if (typeof value !== 'number') {
      index = KeyManagerLevel.indexFor(value);
    } else {
      index = value;
    }
    return Object.values(KeyManagerLevel)[index];
  }
  KeyManagerLevel.from = from;
  function validate(value) {
    if (KeyManagerLevel.indexFor(value) === -1) {
      throwError('Invalid level', SdkErrors.InvalidKeyLevel);
    }
  }
  KeyManagerLevel.validate = validate;
})(KeyManagerLevel || (KeyManagerLevel = {}));
/**
 * @param {KeyManagerLevel} level - The security level of the key
 * @param {PrivateKey} privateKey - The private key to be stored
 * @param {string} [challenge] - A challenge that needs to be presented in order for the key to be used
 */
class StoreKeyOptions {
  constructor() {
    this.level = void 0;
    this.privateKey = void 0;
    this.challenge = void 0;
  }
  static validate(options) {
    KeyManagerLevel.validate(options.level); // throws if level not valid
    if (options.challenge !== undefined) {
      if (typeof options.challenge !== 'string' || options.challenge.length === 0) {
        throwError('Invalid challenge', SdkErrors.InvalidChallenge);
      }
    }
  }
}
/**
 * @param {KeyManagerLevel} level - The security level of the key
 * @param {string | Checksum256} data - The data that will be used to create a digital signature
 * @param {string} [challenge] - A challenge that needs to be presented in order for the key to be used
 * @param {'jwt' | 'transaction'} [outputType] - The type of output to return
 */
class SignDataOptions {
  constructor() {
    this.level = void 0;
    this.data = void 0;
    this.challenge = void 0;
    this.outputType = void 0;
  }
  static validate(options) {
    KeyManagerLevel.validate(options.level); // throws if level not valid
    if (options.challenge !== undefined) {
      if (typeof options.challenge !== 'string' || options.challenge.length === 0) {
        throwError('Invalid challenge', SdkErrors.InvalidChallenge);
      }
    }
    if (options.outputType) {
      if (options.outputType !== 'jwt' && options.outputType !== 'transaction') {
        throwError('Invalid output type', SdkErrors.InvalidData);
      }
    }
  }
}
/**
 * @param {KeyManagerLevel} level - The security level of the key
 */
class GetKeyOptions {
  constructor() {
    this.level = void 0;
  }
  static validate(options) {
    KeyManagerLevel.validate(options.level); // throws if level not valid
  }
}
/**
 * @param {KeyManagerLevel} level - The security level of the key
 * @param {string} challenge - the challenge to check
 */
class CheckKeyOptions {
  constructor() {
    this.level = void 0;
    this.challenge = void 0;
  }
  static validate(options) {
    KeyManagerLevel.validate(options.level); // throws if level not valid
    if (options.challenge !== undefined) {
      if (typeof options.challenge !== 'string' || options.challenge.length === 0) {
        throwError('Invalid challenge', SdkErrors.InvalidChallenge);
      }
    }
  }
}

/**
 * A proxy handler that will create magic getters and setters for the storage
 */
const storageProxyHandler = {
  /**
   * return the called property from the storage if it exists
   * @param target - The target object
   * @param key - The property key
   * @returns The value of the property from the storage or cached value
   * @throws {Error} If the data could not be retrieved
   */
  get: (target, key) => {
    if (key === 'scope') throwError('Scope is a reserved key');
    if (key === 'cache') throwError('Cache is a reserved key');
    const scopedKey = target.scope + key;
    if (key in target) {
      if (key === 'clear') {
        target.cache = {};
      }
      return function () {
        target[key]();
      };
    }
    if (target.cache[scopedKey]) return target.cache[scopedKey];
    return target.retrieve(scopedKey).then(data => {
      target.cache[scopedKey] = data; // cache the data
      return data;
    }).catch(e => {
      throwError(`Could not get ${scopedKey} from storage - ${e}`);
    });
  },
  /**
   * store the value in the storage
   * @param target - The target object
   * @param key - The property key
   * @param value - The value to store
   * @returns true if the value was stored
   * @throws {Error} If the data could not be stored
   */
  set: async function (target, key, value) {
    const scopedKey = target.scope + key;
    return target.store(scopedKey, value).then(() => {
      target.cache[scopedKey] = value;
      return true;
    }).catch(() => {
      return false;
      // throw new Error(`Could not store data - ${e}`);
    });
  }
};
const STORAGE_NAMESPACE = 'tonomy.id.';
function createStorage(scope, storageFactory) {
  const storage = storageFactory(scope);
  storage.cache = {};
  const proxy = new Proxy(storage, storageProxyHandler);
  return proxy;
}

class UserBase {
  constructor(_keyManager, storageFactory) {
    this.keyManager = void 0;
    this.storage = void 0;
    this.keyManager = _keyManager;
    this.storage = createStorage(STORAGE_NAMESPACE + 'user.', storageFactory);
  }
  async getStatus() {
    return await this.storage.status;
  }
  async getAccountName() {
    return await this.storage.accountName;
  }
  async getUsername() {
    const username = await this.storage.username;
    if (!username) throwError('Username not set', SdkErrors.InvalidData);else if (username instanceof TonomyUsername) {
      return username;
    } else if (typeof username === 'string') {
      return new TonomyUsername(username);
    } else {
      throwError('Username not in expected format', SdkErrors.InvalidData);
    }
  }
  async getDid() {
    return await this.storage.did;
  }
  async getIssuer() {
    const did = await this.getDid();
    const signer = createVCSigner(this.keyManager, KeyManagerLevel.LOCAL);
    return {
      did: did + '#local',
      signer: signer.sign,
      alg: 'ES256K-R'
    };
  }
  async getSigner(level) {
    return createKeyManagerSigner(this.keyManager, level);
  }
}

class WalletRequestResponse extends VerifiableCredentialWithType {
  /**
   * @inheritdoc override me if the payload type requires ad-hoc decoding
   */
  constructor(vc) {
    super(vc);
  }
  /**
   * Alternative constructor that returns type WalletRequestResponse. To be used in derived classes
   *
   * @access protected
   *
   * @param {object} payload the payload
   * @param {Issuer} issuer the issuer id
   * @param {VerifiableCredentialOptions} options the options
   *
   * @returns a WalletRequestResponse object
   */
  static async signWalletResponse(payload, issuer, options = {}) {
    if (this.type === 'WalletRequest') {
      throw new Error('class should be a derived class of WalletRequest to use the type property');
    }
    const vc = await super.sign(payload, issuer, options);
    return new WalletRequestResponse(vc);
  }
}
WalletRequestResponse.type = 'WalletRequestResponse';
class LoginRequestResponse extends WalletRequestResponse {
  /**
   * @override the VerifiableCredentialWithType constructor to decode the payload of type LoginRequestPayload
   */
  constructor(vc) {
    super(vc);
    this.decodedPayload.accountName = Name.from(super.getPayload().accountName);
  }
  /**
   * Alternative constructor that returns type LoginRequestResponse
   */
  static async signResponse(payload, issuer, options = {}) {
    const vc = await super.signWalletResponse(payload, issuer, options);
    return new LoginRequestResponse(vc);
  }
}
LoginRequestResponse.type = 'LoginRequestResponse';
class DataRequestResponse extends WalletRequestResponse {
  /**
   * @override the VerifiableCredentialWithType constructor to decode the payload of type DataSharingRequestResponsePayload
   */
  constructor(vc) {
    super(vc);
    if (this.decodedPayload.data.username && typeof this.decodedPayload.data.username === 'string') {
      this.decodedPayload.data.username = new TonomyUsername(this.decodedPayload.data.username);
    }
  }
  /**
   * Alternative constructor that returns type DataRequestResponse
   */
  static async signResponse(payload, issuer, options = {}) {
    const vc = await super.signWalletResponse(payload, issuer, options);
    return new DataRequestResponse(vc);
  }
}
DataRequestResponse.type = 'DataRequestResponse';

const debug$3 = Debug('tonomy-sdk:requestsManager');
class RequestsManager {
  constructor(requests) {
    this.requests = [];
    if (typeof requests[0] === 'string') {
      this.fromStrings(requests);
    } else {
      this.from(requests);
    }
  }
  from(requests) {
    const checkedRequests = this.checkArrayAndFilterNull(requests);
    const classInitializedRequests = [];
    for (const request of checkedRequests) {
      debug$3('RequestsManager from', request.getType(), request.getPayload());
      classInitializedRequests.push(castToWalletRequestSubclass(request));
    }
    this.requests = classInitializedRequests;
  }
  fromStrings(requests) {
    const checkedRequests = this.checkArrayAndFilterNull(requests);
    this.from(checkedRequests.map(request => new WalletRequest(request)));
  }
  checkArrayAndFilterNull(array) {
    if (!array || !Array.isArray(array)) {
      throwError('No requests found', SdkErrors.RequestsNotFound);
    }
    const response = array.filter(request => request !== null);
    if (response.length === 0) {
      throwError('No requests found in array', SdkErrors.RequestsNotFound);
    }
    return response;
  }
  getRequests() {
    return this.requests;
  }
  /**
   * Verifies the WalletRequests are valid requests signed by valid DIDs
   * @throws if the requests are not valid
   */
  async verify() {
    for (const request of this.requests) {
      if (!(await request.verify())) {
        if (request instanceof LoginRequest) {
          throwError(`Invalid request for ${request.getType()} ${request.getPayload().origin}`, SdkErrors.JwtNotValid);
        } else if (request instanceof DataSharingRequest) {
          throwError(`Invalid request for ${request.getType()} `, SdkErrors.JwtNotValid);
        }
      }
    }
  }
  checkReferrerOrigin() {
    const docReferrer = document.referrer;
    if (!docReferrer) throwError('No referrer found', SdkErrors.ReferrerEmpty);
    const referrerOrigin = new URL(docReferrer).origin;
    const myRequest = this.getLoginRequestsOrThrow().find(request => request.getPayload().origin === referrerOrigin);
    if (!myRequest) {
      const msg = `No origins from: ${this.getLoginRequestsOrThrow().map(r => r.getPayload().origin)} ` + `match referrer: ${referrerOrigin}`;
      throwError(msg, SdkErrors.WrongOrigin);
    }
  }
  getLoginRequests() {
    const response = this.requests.filter(request => request instanceof LoginRequest);
    return response.length > 0 ? response : undefined;
  }
  getLoginRequestsOrThrow() {
    const response = this.getLoginRequests();
    if (!response) {
      throwError('No LoginRequests found', SdkErrors.RequestsNotFound);
    }
    return response;
  }
  getDataSharingRequests() {
    const response = this.requests.filter(request => request instanceof DataSharingRequest);
    return response.length > 0 ? response : undefined;
  }
  getDataSharingRequestsOrThrow() {
    const response = this.getDataSharingRequests();
    if (!response) {
      throwError('No DataSharingRequests found', SdkErrors.RequestsNotFound);
    }
    return response;
  }
  getLoginRequestWithSameOriginOrThrow() {
    const myOrigin = window.location.origin;
    const response = this.getLoginRequestsOrThrow().find(request => request.getPayload().origin === myOrigin);
    if (!response) {
      throwError(`No origin from ${myOrigin} found`, SdkErrors.OriginNotFound);
    }
    return response;
  }
  getLoginRequestWithDifferentOriginOrThrow() {
    const myOrigin = window.location.origin;
    const response = this.getLoginRequestsOrThrow().find(request => request.getPayload().origin !== myOrigin);
    if (!response) {
      throwError(`No origin different from ${myOrigin} found`, SdkErrors.OriginNotFound);
    }
    return response;
  }
  getRequestsDifferentOriginOrThrow() {
    const loginRequest = this.getLoginRequestWithDifferentOriginOrThrow();
    const issuer = loginRequest.getIssuer();
    const dataSharingRequest = this.requests.find(request => request instanceof DataSharingRequest && request.getIssuer() === issuer);
    if (dataSharingRequest) {
      return [loginRequest, dataSharingRequest];
    } else {
      return [loginRequest];
    }
  }
}
function castToWalletRequestSubclass(request) {
  if (request.getType() === LoginRequest.getType()) {
    return new LoginRequest(request);
  } else if (request.getType() === DataSharingRequest.getType()) {
    return new DataSharingRequest(request);
  } else {
    throwError('Invalid WalletRequest Type', SdkErrors.InvalidRequestType);
  }
}

Debug('tonomy-sdk:responsesManager');
class WalletRequestAndResponseObject {
  constructor(arg) {
    this.request = void 0;
    this.meta = void 0;
    this.response = void 0;
    if (typeof arg === 'string') {
      const parsed = JSON.parse(arg);
      const request = new WalletRequest(parsed.request);
      const response = new WalletRequestResponse(parsed.response);
      this.setRequest(castToWalletRequestSubclass(request));
      this.setResponse(castToWalletRequestResponseSubclass(response));
    } else if (arg instanceof WalletRequestAndResponse) {
      this.setRequest(castToWalletRequestSubclass(arg.request));
      this.setResponse(castToWalletRequestResponseSubclass(arg.response));
    } else if (WalletRequestAndResponse.isInstance(arg)) {
      this.setRequest(castToWalletRequestSubclass(arg.request));
      this.setResponse(castToWalletRequestResponseSubclass(arg.response));
    } else if (arg instanceof WalletRequest) {
      this.setRequest(castToWalletRequestSubclass(arg));
    } else if (WalletRequestAndResponseStrings.isInstance(arg)) {
      this.setRequest(castToWalletRequestSubclass(new WalletRequest(arg.request)));
      this.setResponse(castToWalletRequestResponseSubclass(new WalletRequestResponse(arg.response)));
    } else {
      throwError('Invalid argument type', SdkErrors.InvalidArgumentType);
    }
  }
  setRequest(request) {
    this.request = request;
  }
  setMeta(meta) {
    this.meta = meta;
  }
  setResponse(response) {
    this.response = response;
  }
  getRequestType() {
    return this.request.getType();
  }
  getRequest() {
    return this.request;
  }
  getMetaOrThrow() {
    if (!this.meta) {
      throwError('Response meta not set', SdkErrors.ResponsesNotFound);
    }
    return this.meta;
  }
  getAppOrThrow() {
    return this.getMetaOrThrow().app;
  }
  getResponse() {
    if (!this.response) {
      throwError('Response not set', SdkErrors.ResponsesNotFound);
    }
    return this.response;
  }
  getRequestAndResponse() {
    return {
      request: this.getRequest(),
      response: this.getResponse()
    };
  }
  toString() {
    return JSON.stringify(this.getRequestAndResponse());
  }
  toJSON() {
    return this.toString();
  }
}
class WalletRequestAndResponse {
  constructor() {
    this.request = void 0;
    this.response = void 0;
  }
  static isInstance(arg) {
    return arg instanceof Object && arg.request && arg.request instanceof WalletRequest && arg.response && arg.response instanceof WalletRequestResponse;
  }
}
class WalletRequestAndResponseStrings {
  constructor() {
    this.request = void 0;
    this.response = void 0;
  }
  static isInstance(arg) {
    return arg instanceof Object && arg.request && typeof arg.request === 'string' && arg.response && typeof arg.response === 'string';
  }
}
function castToWalletRequestResponseSubclass(response) {
  if (response.getType() === LoginRequestResponse.getType()) {
    return new LoginRequestResponse(response);
  } else if (response.getType() === DataRequestResponse.getType()) {
    return new DataRequestResponse(response);
  } else {
    throwError('Invalid WalletRequestResponse Type', SdkErrors.InvalidRequestResponseType);
  }
}

/**
 * A message that can be sent between two Tonomy identities
 *
 * @inheritdoc signMessageWithRecipient() is a protected alternative constructor. In child classes,
 * a new alternative constructor called signMessage() should be created which returns the child class type.
 * @inheritdoc protected "type" property should be overridden in child classes with the name of the class
 * @inheritdoc if the payload type requires ad-hoc decoding, override the constructor and decode the payload
 *
 * @example see an example of the above in the LoginRequestMessage class
 */
class Message extends VerifiableCredentialWithType {
  /**
   * @inheritdoc override me if the payload type requires ad-hoc decoding
   */
  constructor(vc) {
    super(vc);
  }
  /**
   * Alternative constructor that returns type Message
   *
   * @access protected
   *
   * @param {object} message the message
   * @param {Issuer} issuer the issuer id
   * @param {DIDUrl} recipient the recipient id
   * @param {VerifiableCredentialOptions} options the options
   *
   * @returns a message object
   */
  static async signMessageWithRecipient(message, issuer, recipient, options = {}) {
    if (this.type === 'Message') {
      throw new Error('class should be a derived class of Message to use the type property');
    }
    const newOptions = {
      subject: recipient,
      additionalTypes: ['TonomyMessage'],
      ...options
    };
    const request = await super.sign(message, issuer, newOptions);
    return new Message(request.getVc());
  }
  /**
   * Returns the recipient of the message
   * @returns {DIDur} the message recipient
   */
  getRecipient() {
    return this.getVc().getSubject();
  }
  /**
   * Returns the sender of the message
   * @returns {DIDur} the message sender
   */
  getSender() {
    return super.getIssuer();
  }
}
Message.type = 'Message';
class AuthenticationMessage extends Message {
  /**
   * Alternative constructor that returns type AuthenticationMessage
   */
  static async signMessageWithoutRecipient(message, issuer, options) {
    const vc = await super.signMessageWithRecipient(message, issuer, '', options);
    return new AuthenticationMessage(vc);
  }
}
AuthenticationMessage.type = 'AuthenticationMessage';
class IdentifyMessage extends Message {
  /**
   * Alternative constructor that returns type IdentifyMessage
   */
  static async signMessage(message, issuer, recipient, options = {}) {
    const vc = await super.signMessageWithRecipient(message, issuer, recipient, options);
    return new IdentifyMessage(vc);
  }
}
IdentifyMessage.type = 'IdentifyMessage';
class LoginRequestsMessage extends Message {
  /**
   * @override the Message constructor to decode the payload of type LoginRequestsMessagePayload
   */
  constructor(vc) {
    super(vc);
    const payload = this.getVc().getPayload().vc.credentialSubject.payload;
    if (!payload.requests) {
      throw new Error('LoginRequestsMessage must have a requests property');
    }
    const requests = payload.requests.map(request => new WalletRequest(request));
    const requestsManager = new RequestsManager(requests);
    this.decodedPayload = {
      requests: requestsManager.getRequests()
    };
  }
  /**
   * Alternative constructor that returns type LoginRequestsMessage
   */
  static async signMessage(message, issuer, recipient, options = {}) {
    const vc = await super.signMessageWithRecipient(message, issuer, recipient, options);
    return new LoginRequestsMessage(vc);
  }
}
LoginRequestsMessage.type = 'LoginRequestsMessage';
class LoginRequestResponseMessage extends Message {
  /**
   * @override the Message constructor to decode the payload of type LoginRequestResponseMessagePayload
   */
  constructor(vc) {
    super(vc);
    const payload = this.getVc().getPayload().vc.credentialSubject.payload;
    if (payload.success) {
      if (!payload.response) {
        throw new Error('LoginRequestsResponseMessage must have a response property');
      }
      const responses = payload.response.map(response => new WalletRequestAndResponseObject(response).getRequestAndResponse());
      this.decodedPayload = {
        ...payload,
        response: responses
      };
    } else {
      if (!payload.error) {
        throw new Error('LoginRequestsResponseMessage must have an error property');
      }
      const error = payload.error;
      const requests = error.requests.map(request => new WalletRequest(request));
      const requestsManager = new RequestsManager(requests);
      this.decodedPayload = {
        ...payload,
        error: {
          ...error,
          requests: requestsManager.getRequests()
        }
      };
    }
  }
  /**
   * Alternative constructor that returns type LoginRequestResponseMessage
   */
  static async signMessage(message, issuer, recipient, options = {}) {
    const vc = await super.signMessageWithRecipient(message, issuer, recipient, options);
    return new LoginRequestResponseMessage(vc);
  }
}
LoginRequestResponseMessage.type = 'LoginRequestResponseMessage';
class LinkAuthRequestMessage extends Message {
  /**
   * @override the Message constructor to decode the payload of type LinkAuthRequestMessagePayload
   */
  constructor(vc) {
    super(vc);
    const payload = this.getVc().getPayload().vc.credentialSubject.payload;
    this.decodedPayload = {
      contract: Name.from(payload.contract),
      action: Name.from(payload.action)
    };
  }
  /**
   * Alternative constructor that returns type LoginRequestResponseMessage
   */
  static async signMessage(message, issuer, recipient, options = {}) {
    const vc = await super.signMessageWithRecipient(message, issuer, recipient, options);
    return new LinkAuthRequestMessage(vc);
  }
}
LinkAuthRequestMessage.type = 'LinkAuthRequestMessage';
class LinkAuthRequestResponseMessage extends Message {
  /**
   * Alternative constructor that returns type LinkAuthRequestResponseMessage
   */
  static async signMessage(message, issuer, recipient, options = {}) {
    const vc = await super.signMessageWithRecipient(message, issuer, recipient, options);
    return new LinkAuthRequestResponseMessage(vc);
  }
}
LinkAuthRequestResponseMessage.type = 'LinkAuthRequestResponseMessage';

const tonomyContract$7 = TonomyContract.Instance;
async function getAccountInfo(account) {
  let accountName;
  if (account instanceof TonomyUsername) {
    const idData = await tonomyContract$7.getPerson(account);
    accountName = idData.account_name;
  } else {
    accountName = account;
  }
  return await getAccount(accountName);
}
/**
 * Initialize and return the user object
 * @param keyManager  the key manager
 * @param storage  the storage
 * @returns the user object
 */
function createUserObject(keyManager, storageFactory) {
  return new User(keyManager, storageFactory);
}
/**
 * Checks that a key exists in the key manager that has been authorized on the DID
 *
 * @description This is called on the callback page to verify that the user has logged in correctly
 *
 * @param {string} [accountName] - the account name to check the key on
 * @param {PublicKey} [publicKey] - the public key to check. if not supplied it will try lookup the app from window.location.origin
 * @param {KeyManager} [keyManager] - the key manager to check the key in
 * @param {KeyManagerLevel} [keyManagerLevel=BROWSER_LOCAL_STORAGE] - the level to check the key in
 * @returns {Promise<Name>} - the name of the permission that the key is authorized on
 *
 * @throws {SdkError} - if the key doesn't exist or isn't authorized
 */
async function verifyKeyExistsForApp(accountName, options) {
  const account = await getAccountInfo(accountName);
  if (!account) throwError("couldn't fetch account", SdkErrors.AccountNotFound);
  if (options.publicKey) {
    const pubKey = options.publicKey;
    const permissionWithKey = account.permissions.find(p => p.required_auth.keys[0].key.toString() === pubKey.toString());
    if (!permissionWithKey) throwError(`No permission found with key ${pubKey}`, SdkErrors.UserNotLoggedInWithThisApp);
    return permissionWithKey.perm_name;
  } else {
    if (!options.keyManager) throwError('keyManager missing', SdkErrors.MissingParams);
    const pubKey = await options.keyManager.getKey({
      level: KeyManagerLevel.BROWSER_LOCAL_STORAGE
    });
    const app = await App.getApp(window.location.origin);
    try {
      const permission = account.getPermission(app.accountName);
      const publicKey = permission.required_auth.keys[0].key;
      if (pubKey.toString() !== publicKey.toString()) throwError('key not authorized', SdkErrors.KeyNotFound);
    } catch (e) {
      if (e.message.startsWith('Unknown permission ')) throwError(`No permission found for app ${app.accountName}`, SdkErrors.UserNotLoggedInWithThisApp);else throw e;
    }
    return app.accountName;
  }
}

const tonomyContract$6 = TonomyContract.Instance;
class UserAuthorization extends UserBase {
  async savePassword(masterPassword, options) {
    let privateKey;
    let salt;
    if (options.salt) {
      salt = options.salt;
      const res = await options.keyFromPasswordFn(masterPassword, salt);
      privateKey = res.privateKey;
    } else {
      const res = await options.keyFromPasswordFn(masterPassword);
      privateKey = res.privateKey;
      salt = res.salt;
    }
    this.storage.salt = salt;
    await this.storage.salt; // wait for magic setter on storage
    await this.keyManager.storeKey({
      level: KeyManagerLevel.PASSWORD,
      privateKey,
      challenge: masterPassword
    });
    await this.keyManager.storeKey({
      level: KeyManagerLevel.ACTIVE,
      privateKey
      // eventually this should be different than the password key, but for now Antelope protocol doesn't support it
      // ideally we would have a different structure, and active key will be linked to local key
    });
  }
  async checkPassword(password, options) {
    const username = await this.getAccountName();
    const idData = await tonomyContract$6.getPerson(username);
    const salt = idData.password_salt;
    await this.savePassword(password, {
      ...options,
      salt
    });
    const passwordKey = await this.keyManager.getKey({
      level: KeyManagerLevel.PASSWORD
    });
    const accountData = await getAccountInfo(idData.account_name);
    const onchainKey = accountData.getPermission('owner').required_auth.keys[0].key; // TODO: change to active/other permissions when we make the change
    if (passwordKey.toString() !== onchainKey.toString()) throwError('Password is incorrect', SdkErrors.PasswordInvalid);
    return true;
  }
  async savePIN(pin) {
    const privateKey = generateRandomKeyPair().privateKey;
    await this.keyManager.storeKey({
      level: KeyManagerLevel.PIN,
      privateKey,
      challenge: pin
    });
  }
  async checkPin(pin) {
    const pinKey = await this.keyManager.checkKey({
      level: KeyManagerLevel.PIN,
      challenge: pin
    });
    if (!pinKey) throwError('Pin is incorrect', SdkErrors.PinInvalid);
    return true;
  }
  async saveFingerprint() {
    const privateKey = generateRandomKeyPair().privateKey;
    await this.keyManager.storeKey({
      level: KeyManagerLevel.BIOMETRIC,
      privateKey
    });
  }
  async saveLocal() {
    const privateKey = generateRandomKeyPair().privateKey;
    await this.keyManager.storeKey({
      level: KeyManagerLevel.LOCAL,
      privateKey
    });
  }
}

class UserCaptcha extends UserBase {
  async getCaptchaToken() {
    return await this.storage.captchaToken;
  }
  async saveCaptchaToken(captchaToken) {
    this.storage.captchaToken = captchaToken;
    await this.storage.captchaToken;
  }
}

async function createAccount(data) {
  const url = getSettings().accountsServiceUrl;
  const response = await fetch$1(`${url}/accounts`, {
    method: 'POST',
    headers: {
      'Content-Type': 'application/json'
    },
    body: JSON.stringify(data)
  });
  const resData = await response.json();
  if (response.status !== 201) {
    if (response.status === 400) {
      return throwError('Communication Service error: ' + resData.message + ', errors: ' + resData.errors, SdkErrors.AccountServiceError);
    }
    throwError('Communication Service error: ' + resData.message + ', status: ' + response.status, SdkErrors.AccountServiceError);
  }
  return {
    transactionId: resData.transactionId,
    accountName: Name.from(resData.accountName)
  };
}

var UserStatusEnum;
(function (UserStatusEnum) {
  UserStatusEnum["CREATING_ACCOUNT"] = "CREATING_ACCOUNT";
  UserStatusEnum["LOGGING_IN"] = "LOGGING_IN";
  UserStatusEnum["READY"] = "READY";
  UserStatusEnum["DEACTIVATED"] = "DEACTIVATED";
})(UserStatusEnum || (UserStatusEnum = {}));
// eslint-disable-next-line @typescript-eslint/no-namespace
(function (UserStatusEnum) {
  /*
   * Returns the index of the enum value
   *
   * @param value The level to get the index of
   */
  function indexFor(value) {
    return Object.keys(UserStatusEnum).indexOf(value);
  }
  UserStatusEnum.indexFor = indexFor;
  /*
   * Creates an AuthenticatorLevel from a string or index of the level
   *
   * @param value The string or index
   */
  function from(value) {
    let index;
    if (typeof value !== 'number') {
      index = UserStatusEnum.indexFor(value);
    } else {
      index = value;
    }
    return Object.values(UserStatusEnum)[index];
  }
  UserStatusEnum.from = from;
})(UserStatusEnum || (UserStatusEnum = {}));

const debug$2 = Debug('tonomy-sdk:services:communication:communication');
const SOCKET_TIMEOUT = 5000;
class Communication {
  constructor(singleton = true) {
    this.socketServer = void 0;
    // TODO: fix problem: if server restarts, this will be lost and all clients will need to reconnect
    this.subscribers = new Map();
    this.authMessage = void 0;
    this.loggedIn = false;
    this.url = void 0;
    if (Communication.singleton && singleton) return Communication.singleton;
    this.url = getSettings().communicationUrl;
    this.socketServer = io(this.url, {
      transports: ['websocket'],
      autoConnect: false
    });
    Communication.singleton = this;
  }
  isConnected() {
    return this.socketServer && this.socketServer.connected;
  }
  isLoggedIn() {
    return this.loggedIn && this.authMessage !== undefined && typeof this.authMessage === 'object';
  }
  /**
   * Connects to the Tonomy Communication server
   *
   * @returns {Promise<void>}
   * @throws {SdkError} CommunicationNotConnected
   */
  async connect() {
    if (this.isConnected()) return;
    this.socketServer.connect();
    await new Promise((resolve, reject) => {
      this.socketServer.on('connect', async () => {
        if (this.isLoggedIn()) {
          await this.login(this.authMessage);
        }
        resolve(true);
        return;
      });
      setTimeout(() => {
        if (this.isConnected()) return;
        reject(createSdkError('Could not connect to Tonomy Communication server', SdkErrors.CommunicationNotConnected));
      }, SOCKET_TIMEOUT);
    });
  }
  /**
   * Sends a Message object through a websocket connection to the Tonomy Communication server
   *
   * @param {string} event - the name of the event to emit
   * @param {Message} message - the Message object to send
   * @returns {Promise<boolean>} - true if successful and acknowledged by the server
   * @throws {SdkError} - CommunicationTimeout
   */
  async emitMessage(event, message) {
    debug$2('emitMessage', event, message.getType(), message.getSender(), message.getRecipient(), message.getPayload());
    const ack = await new Promise((resolve, reject) => {
      this.socketServer.timeout(SOCKET_TIMEOUT).emit(event, {
        message: message.toString()
      }, (error, response) => {
        debug$2('emitMessage response', JSON.stringify(error, null, 2), JSON.stringify(response, null, 2));
        if (response != null && response.error) {
          var _response$exception;
          if ((response == null || (_response$exception = response.exception) == null ? void 0 : _response$exception.name) === 'HttpException') {
            const communicationError = new CommunicationError(response);
            reject(communicationError);
            return;
          }
          reject(response);
          return;
        }
        if (error) {
          reject(error);
          return;
        }
        resolve(response);
        return;
      });
    });
    if (ack.status !== 200) {
      throw new CommunicationError({
        exception: {
          response: ack.error,
          name: 'HttpException',
          status: ack.status,
          message: ack.error
        },
        name: 'CommunicationError',
        message: ack.error
      });
    }
    return ack.details;
  }
  /**
   * connects to the Tonomy Communication server, authenticates with it's DID
   * @param {AuthenticationMessage} authorization - the VC the user sent
   *
   * @returns {boolean} - true if successful
   */
  async login(authorization) {
    await this.connect();
    const result = await this.emitMessage('login', authorization);
    if (result) {
      this.loggedIn = true;
      this.authMessage = authorization;
    }
    return result;
  }
  /* sends a message to another DID
   * create a Message object from the message argument
   * the message is used as the `vc` property of a VC signed by the User's key
   */
  async sendMessage(message) {
    if (!this.isLoggedIn()) {
      throwError('You need to login before sending a messages', SdkErrors.CommunicationNotLoggedIn);
    }
    return await this.emitMessage('message', message);
  }
  /**
   * function that adds a new subscriber, which is called every time a message is received
   *
   * @param {Subscriber} subscriber - the message object
   * @param {string} [type] - the type of message to subscribe to
   * @returns {number} - identifier which will be used for unsubscribe
   */
  subscribeMessage(subscriber, type) {
    Communication.identifier++;
    const messageHandler = message => {
      var _payload$requests;
      const msg = new Message(message);
      const payload = msg.getPayload();
      debug$2('receiveMessage', msg.getType(), msg.getSender(), msg.getRecipient(), (_payload$requests = payload.requests) == null ? void 0 : _payload$requests.length);
      if (!type || msg.getType() === type) {
        subscriber(msg);
      }
      return this;
    };
    this.socketServer.on('message', messageHandler);
    this.subscribers.set(Communication.identifier, messageHandler);
    return Communication.identifier;
  }
  /**
   * unsubscribes a function from the receiving a message
   *
   * @param {number} id - identifier which will be used for unsubscribe
   *
   */
  unsubscribeMessage(id) {
    const subscriber = this.subscribers.get(id);
    if (subscriber) {
      this.socketServer.off('message', subscriber);
      this.subscribers.delete(id);
    }
  }
  disconnect() {
    this.loggedIn = false;
    delete this.authMessage;
    if (this.isConnected()) {
      this.socketServer.disconnect();
    }
  }
}
Communication.singleton = void 0;
Communication.identifier = void 0;

class UserCommunication extends UserAuthorization {
  constructor(_keyManager, storageFactory) {
    super(_keyManager, storageFactory);
    this.communication = void 0;
    this.communication = new Communication(false);
  }
  async loginCommunication(authorization) {
    return await this.communication.login(authorization);
  }
  unsubscribeMessage(id) {
    this.communication.unsubscribeMessage(id);
  }
  subscribeMessage(subscriber, type) {
    return this.communication.subscribeMessage(subscriber, type);
  }
  async sendMessage(message) {
    return await this.communication.sendMessage(message);
  }
  disconnectCommunication() {
    this.communication.disconnect();
  }
}

const debug$1 = Debug('tonomy-sdk:controllers:user-onboarding');
const tonomyContract$5 = TonomyContract.Instance;
class UserOnboarding extends UserCommunication {
  constructor(...args) {
    super(...args);
    this.chainID = void 0;
  }
  validateUsername(username) {
    if (typeof username !== 'string' || username.length === 0) throwError('Username must be a string', SdkErrors.InvalidData);
    // Allow only letters, numbers, underscore and dash (1 to 50 characters)
    if (!/^[A-Za-z0-9_-]{1,100}$/g.test(username)) throwError('Username contains invalid characters', SdkErrors.InvalidUsername);
  }
  async createDid() {
    if (!this.chainID) {
      this.chainID = (await getChainInfo()).chain_id;
    }
    const accountName = await this.getAccountName();
    this.storage.did = `did:antelope:${this.chainID}:${accountName.toString()}`;
    await this.storage.did;
    return this.storage.did;
  }
  async login(username, password, options) {
    this.validateUsername(username.getBaseUsername());
    const {
      keyManager
    } = this;
    const idData = await tonomyContract$5.getPerson(username);
    const salt = idData.password_salt;
    await this.savePassword(password, {
      ...options,
      salt
    });
    const passwordKey = await keyManager.getKey({
      level: KeyManagerLevel.PASSWORD
    });
    const accountData = await getAccountInfo(idData.account_name);
    const onchainKey = accountData.getPermission('owner').required_auth.keys[0].key; // TODO: change to active/other permissions when we make the change
    if (passwordKey.toString() !== onchainKey.toString()) throwError('Password is incorrect', SdkErrors.PasswordInvalid);
    this.storage.accountName = Name.from(idData.account_name);
    this.storage.username = username;
    this.storage.status = UserStatusEnum.LOGGING_IN;
    await this.storage.accountName;
    await this.storage.username;
    await this.storage.status;
    await this.createDid();
    return idData;
  }
  async isLoggedIn() {
    return (await this.getStatus()) === UserStatusEnum.READY;
  }
  async createPerson() {
    const {
      keyManager
    } = this;
    const username = await this.getUsername();
    const usernameHash = username.usernameHash;
    const publicKey = await keyManager.getKey({
      level: KeyManagerLevel.PASSWORD
    });
    const salt = await this.storage.salt;
    const captchaToken = await this.storage.captchaToken;
    try {
      const res = await createAccount({
        usernameHash: usernameHash,
        publicKey,
        salt,
        captchaToken
      });
      this.storage.accountName = res.accountName;
    } catch (e) {
      if (e.status === 400 && e.message === 'Username is taken') {
        throwError('Username is taken', SdkErrors.UsernameTaken);
      }
      throw e;
    }
    await this.storage.accountName;
    this.storage.status = UserStatusEnum.CREATING_ACCOUNT;
    await this.storage.status;
    await this.createDid();
    debug$1('Created account', {
      accountName: (await this.getAccountName()).toString(),
      username: (await this.getUsername()).getBaseUsername(),
      did: await this.getDid()
    });
  }
  async saveUsername(username) {
    this.validateUsername(username);
    let user;
    const fullUsername = TonomyUsername.fromUsername(username, AccountType.PERSON, getSettings().accountSuffix);
    try {
      user = await getAccountInfo(fullUsername);
      if (user) throwError('Username is taken', SdkErrors.UsernameTaken);
    } catch (e) {
      if (!(e instanceof SdkError && e.code === SdkErrors.UsernameNotFound)) {
        throw e;
      }
    }
    this.storage.username = fullUsername;
    await this.storage.username;
  }
  async usernameExists(username) {
    const fullUsername = TonomyUsername.fromUsername(username, AccountType.PERSON, getSettings().accountSuffix);
    try {
      await getAccountInfo(fullUsername);
      return true;
    } catch (e) {
      if (e instanceof SdkError && e.code === SdkErrors.UsernameNotFound) {
        return false;
      }
      throw e;
    }
  }
  async updateKeys(password) {
    const status = await this.getStatus();
    if (status === UserStatusEnum.DEACTIVATED) {
      throwError("Can't update keys for deactivated user", SdkErrors.UserDeactivated);
    }
    const {
      keyManager
    } = this;
    // TODO:
    // use status in smart contract to lock the account till finished creating
    const keys = {};
    try {
      const pinKey = await keyManager.getKey({
        level: KeyManagerLevel.PIN
      });
      keys.PIN = pinKey.toString();
    } catch (e) {
      if (!(e instanceof SdkError) || e.code !== SdkErrors.KeyNotFound) throw e;
    }
    try {
      const biometricKey = await keyManager.getKey({
        level: KeyManagerLevel.BIOMETRIC
      });
      keys.BIOMETRIC = biometricKey.toString();
    } catch (e) {
      if (!(e instanceof SdkError) || e.code !== SdkErrors.KeyNotFound) throw e;
    }
    try {
      const localKey = await keyManager.getKey({
        level: KeyManagerLevel.LOCAL
      });
      keys.LOCAL = localKey.toString();
    } catch (e) {
      if (!(e instanceof SdkError) || e.code !== SdkErrors.KeyNotFound) throw e;
    }
    const signer = createKeyManagerSigner(keyManager, KeyManagerLevel.PASSWORD, password);
    const accountName = await this.getAccountName();
    await tonomyContract$5.updatekeysper(accountName.toString(), keys, signer);
    this.storage.status = UserStatusEnum.READY;
    await this.storage.status;
  }
  async checkKeysStillValid() {
    // Account been created, or has not finished being created yet
    if ((await this.getStatus()) !== UserStatusEnum.READY) throwError('User is not ready', SdkErrors.AccountDoesntExist);
    const accountInfo = await getAccountInfo(await this.storage.accountName);
    const checkPairs = [{
      level: KeyManagerLevel.PIN,
      permission: 'pin'
    }, {
      level: KeyManagerLevel.BIOMETRIC,
      permission: 'biometric'
    }, {
      level: KeyManagerLevel.LOCAL,
      permission: 'local'
    }, {
      level: KeyManagerLevel.PASSWORD,
      permission: 'active'
    }, {
      level: KeyManagerLevel.PASSWORD,
      permission: 'owner'
    }];
    for (const pair of checkPairs) {
      let localKey;
      try {
        localKey = await this.keyManager.getKey({
          level: pair.level
        });
      } catch (e) {
        localKey = null;
      }
      let blockchainPermission;
      try {
        blockchainPermission = accountInfo.getPermission(pair.permission);
      } catch (e) {
        blockchainPermission = null;
      }
      if (!localKey && blockchainPermission) {
        // User probably logged into another device and finished create account flow there
        throwError(`${pair.level} key was not found in the keyManager, but was found on the blockchain`, SdkErrors.KeyNotFound);
      }
      if (localKey && !blockchainPermission) {
        // User probably hasn't finished create account flow yet
        throwError(`${pair.level} keys was not found on the blockchain, but was found in the keyManager`, SdkErrors.KeyNotFound);
      }
      if (localKey && blockchainPermission && localKey.toString() !== blockchainPermission.required_auth.keys[0].key.toString()) {
        // User has logged in on another device
        throwError(`${pair.level} keys do not match`, SdkErrors.KeyNotFound);
      }
    }
    return true;
  }
  async logout() {
    // remove all keys
    for (const level of Object.keys(KeyManagerLevel)) {
      try {
        await this.keyManager.getKey({
          level: KeyManagerLevel.from(level)
        });
        this.keyManager.removeKey({
          level: KeyManagerLevel.from(level)
        });
      } catch (e) {
        if (!(e instanceof SdkError) || e.code !== SdkErrors.KeyNotFound && e.code !== SdkErrors.InvalidKeyLevel) throw e;
      }
    }
    // clear storage data
    this.storage.clear();
    this.disconnectCommunication();
  }
  async initializeFromStorage() {
    const accountName = await this.getAccountName();
    if (accountName) {
      return await this.checkKeysStillValid();
    } else {
      throwError('Account "' + accountName + '" not found', SdkErrors.AccountDoesntExist);
    }
  }
}

const debug = Debug('tonomy-sdk:UserRequestsManager');
const tonomyEosioProxyContract$2 = TonomyEosioProxyContract.Instance;
const tonomyContract$4 = TonomyContract.Instance;
class UserRequestsManager extends UserCommunication {
  async handleLinkAuthRequestMessage(message) {
    const linkAuthRequestMessage = new LinkAuthRequestMessage(message);
    try {
      if (!getAccountNameFromDid(message.getSender()).equals(await this.getAccountName())) throwError('Message not sent from authorized account', SdkErrors.SenderNotAuthorized);
      const payload = linkAuthRequestMessage.getPayload();
      const contract = payload.contract;
      const action = payload.action;
      const permission = parseDid(message.getSender()).fragment;
      if (!permission) throwError('DID does not contain App permission', SdkErrors.MissingParams);
      await tonomyContract$4.getApp(Name.from(permission));
      // Throws SdkErrors.DataQueryNoRowDataFound error if app does not exist
      // which cannot happen in theory, as the user is already logged in
      // if (permission !== contract.toString())
      //     throwError('Contract and permission do not match', SdkErrors.MismatchedParams);
      const signer = createKeyManagerSigner(this.keyManager, KeyManagerLevel.ACTIVE);
      await tonomyEosioProxyContract$2.linkAuth((await this.getAccountName()).toString(), contract.toString(), action.toString(), permission, signer);
      const linkAuthRequestResponseMessage = await LinkAuthRequestResponseMessage.signMessage({
        requestId: linkAuthRequestMessage.getVc().getId(),
        success: true
      }, await this.getIssuer(), linkAuthRequestMessage.getSender());
      await this.sendMessage(linkAuthRequestResponseMessage);
    } catch (e) {
      if (e instanceof SdkError && e.code === SdkErrors.SenderNotAuthorized) {
        // somebody may be trying to DoS the user, drop
        return;
      } else {
        // all other errors are Tonomy software errors, so throw to bubble up
        throw e;
      }
    }
  }
  async loginWithApp(app, key) {
    const myAccount = await this.getAccountName();
    const appRecord = {
      app,
      added: new Date(),
      status: AppStatusEnum.PENDING
    };
    let apps = await this.storage.appRecords;
    if (!apps) {
      apps = [];
    }
    apps.push(appRecord);
    this.storage.appRecords = apps;
    await this.storage.appRecords;
    const localSigner = createKeyManagerSigner(this.keyManager, KeyManagerLevel.LOCAL);
    let linkAuth = false;
    try {
      (await getAccountInfo(myAccount)).getPermission(app.accountName);
    } catch (e) {
      if (e.message === `Unknown permission ${app.accountName.toString()} on account ${myAccount.toString()}.`) {
        linkAuth = true;
      }
    }
    debug('loginWithApp key', key);
    await tonomyContract$4.loginwithapp(myAccount.toString(), app.accountName.toString(), 'local', key, localSigner);
    // If the permission was only just created, we link it to the app (using its account name)
    // so that this permission can be used to sign transactions in the app immediately
    // (otherwise a LinkAuthRequestMessage needs to be sent and executed with ExternalUser.checkLinkAuthRequirements())
    // when a transaction attempt is made in the app
    if (linkAuth) {
      await sleep(1000); // wait for the blockchain to catch up
      const activeSigner = createKeyManagerSigner(this.keyManager, KeyManagerLevel.ACTIVE);
      await tonomyEosioProxyContract$2.linkAuth(myAccount.toString(), app.accountName.toString(), '', app.accountName.toString(), activeSigner);
    }
    appRecord.status = AppStatusEnum.READY;
    this.storage.appRecords = apps;
    await this.storage.appRecords;
  }
  async acceptLoginRequest(responsesManager, platform, options) {
    debug('acceptLoginRequest() options', options);
    const finalResponses = await responsesManager.createResponses(this);
    const responsePayload = {
      success: true,
      response: finalResponses
    };
    if (platform === 'mobile') {
      if (!options.callbackPath || !options.callbackOrigin) throwError('Missing callback origin or path', SdkErrors.MissingParams);
      let callbackUrl = options.callbackOrigin + options.callbackPath + '?';
      callbackUrl += 'payload=' + objToBase64Url(responsePayload);
      return callbackUrl;
    } else {
      if (!options.messageRecipient) throwError('Missing message recipient', SdkErrors.MissingParams);
      const issuer = await this.getIssuer();
      const message = await LoginRequestResponseMessage.signMessage(responsePayload, issuer, options.messageRecipient);
      await this.sendMessage(message);
    }
  }
  async checkLoginRequests(requests) {
    const managedRequests = new RequestsManager(requests);
    await managedRequests.verify();
    const response = [];
    for (const request of managedRequests.getLoginRequestsOrThrow()) {
      const payload = request.getPayload();
      const app = await App.getApp(payload.origin);
      let requiresLogin = true;
      try {
        await verifyKeyExistsForApp(await this.getAccountName(), {
          publicKey: payload.publicKey
        });
        requiresLogin = false;
      } catch (e) {
        if (e instanceof SdkError && e.code === SdkErrors.UserNotLoggedInWithThisApp) {
          // Never consented
          requiresLogin = true;
        } else {
          throw e;
        }
      }
      response.push({
        request,
        app,
        requiresLogin,
        ssoApp: payload.origin === getSettings().ssoWebsiteOrigin,
        requestDid: request.getIssuer()
      });
    }
    return response;
  }
}

class User extends Mixin(UserBase, UserAuthorization, UserCaptcha, UserOnboarding, UserCommunication, UserRequestsManager) {}

// eslint-disable-next-line @typescript-eslint/ban-ts-comment
Debug('tonomy-sdk:util:ssi:veramo');

const KEY_STORAGE_NAMESPACE = STORAGE_NAMESPACE + 'key.';
class JsKeyManager {
  constructor() {
    this.keyStorage = {};
  }
  // creates a new secure key and returns the public key
  async storeKey(options) {
    StoreKeyOptions.validate(options);
    const keyStore = {
      privateKey: options.privateKey,
      publicKey: options.privateKey.toPublic()
    };
    switch (options.level) {
      case KeyManagerLevel.ACTIVE:
      case KeyManagerLevel.LOCAL:
      case KeyManagerLevel.BIOMETRIC:
        break;
      case KeyManagerLevel.PASSWORD:
      case KeyManagerLevel.PIN:
        if (!options.challenge) throwError('Challenge missing', SdkErrors.MissingChallenge);
        keyStore.salt = randomString(32);
        keyStore.hashedSaltedChallenge = sha256(options.challenge + keyStore.salt);
        break;
      case KeyManagerLevel.BROWSER_LOCAL_STORAGE:
        localStorage.setItem(KEY_STORAGE_NAMESPACE + options.level, JSON.stringify(keyStore));
        break;
      case KeyManagerLevel.BROWSER_SESSION_STORAGE:
        sessionStorage.setItem(KEY_STORAGE_NAMESPACE + options.level, JSON.stringify(keyStore));
        break;
      default:
        throwError('Invalid level', SdkErrors.InvalidKeyLevel);
    }
    this.keyStorage[options.level] = keyStore;
    return keyStore.publicKey;
  }
  fetchKey(options) {
    GetKeyOptions.validate(options);
    const keyStore = this.keyStorage[options.level];
    if (keyStore) return keyStore;
    if (options.level === KeyManagerLevel.BROWSER_LOCAL_STORAGE || options.level === KeyManagerLevel.BROWSER_SESSION_STORAGE) {
      const storage = options.level === KeyManagerLevel.BROWSER_LOCAL_STORAGE ? localStorage.getItem(KEY_STORAGE_NAMESPACE + options.level) : sessionStorage.getItem(KEY_STORAGE_NAMESPACE + options.level);
      if (!storage) throwError(`No key for level ${options.level}`, SdkErrors.KeyNotFound);
      const keyStore = JSON.parse(storage);
      this.keyStorage[options.level] = {
        privateKey: PrivateKey.from(keyStore.privateKey),
        publicKey: PublicKey.from(keyStore.publicKey)
      };
      if (keyStore.hashedSaltedChallenge) this.keyStorage[options.level].hashedSaltedChallenge = keyStore.hashedSaltedChallenge;
      if (keyStore.salt) this.keyStorage[options.level].salt = keyStore.salt;
      return this.keyStorage[options.level];
    }
    throwError(`No key for level ${options.level}`, SdkErrors.KeyNotFound);
  }
  async signData(options) {
    SignDataOptions.validate(options);
    const keyStore = this.fetchKey(options);
    if (options.level === KeyManagerLevel.PASSWORD || options.level === KeyManagerLevel.PIN) {
      if (!options.challenge) throwError('Challenge missing', SdkErrors.MissingChallenge);
      const validChallenge = await this.checkKey({
        level: options.level,
        challenge: options.challenge
      });
      if (!validChallenge && options.level === KeyManagerLevel.PASSWORD) throwError('Invalid password', SdkErrors.PasswordInvalid);
      if (!validChallenge && options.level === KeyManagerLevel.PIN) throwError('Invalid PIN', SdkErrors.PinInvalid);
    }
    const privateKey = keyStore.privateKey;
    if (options.outputType === 'jwt') {
      if (typeof options.data !== 'string') throw throwError('Data must be a string', SdkErrors.InvalidData);
      const signer = createSigner(privateKey);
      return await signer(options.data);
    } else {
      let digest;
      if (typeof options.data === 'string') {
        digest = Checksum256.hash(Buffer.from(options.data));
      } else {
        digest = options.data;
      }
      const signature = privateKey.signDigest(digest);
      return signature;
    }
  }
  async getKey(options) {
    GetKeyOptions.validate(options);
    const keyStore = this.fetchKey(options);
    return keyStore.publicKey;
  }
  async checkKey(options) {
    CheckKeyOptions.validate(options);
    const keyStore = this.fetchKey(options);
    if (options.level === KeyManagerLevel.PIN || options.level === KeyManagerLevel.PASSWORD) {
      if (!options.challenge) throwError('Challenge is missing', SdkErrors.MissingChallenge);
      const hashedSaltedChallenge = sha256((options == null ? void 0 : options.challenge) + keyStore.salt);
      return hashedSaltedChallenge === keyStore.hashedSaltedChallenge;
    } else throw throwError('Invalid Level', SdkErrors.InvalidKeyLevel);
  }
  async removeKey(options) {
    GetKeyOptions.validate(options);
    delete this.keyStorage[options.level];
    if (options.level === KeyManagerLevel.BROWSER_LOCAL_STORAGE || options.level === KeyManagerLevel.BROWSER_SESSION_STORAGE) {
      options.level === KeyManagerLevel.BROWSER_LOCAL_STORAGE ? localStorage.removeItem(KEY_STORAGE_NAMESPACE + options.level) : sessionStorage.removeItem(KEY_STORAGE_NAMESPACE + options.level);
    }
  }
}

Debug('tonomy-sdk:externalUser');

async function keys(args) {
  if (args[0] === 'create') {
    console.log('Creating new key\n');
    const keyPair = generateRandomKeyPair();
    console.log('Public key: ', keyPair.publicKey.toString());
    console.log('Private key: ', keyPair.privateKey.toString());
  } else if (args[0] === 'convert') {
    console.log('Converting key formats\n');
    try {
      const publicKey = PublicKey.from(args[1]);
      console.log('Public key: ', publicKey.toString());
      console.log('Public key (legacy): ', publicKey.toLegacyString());
    } catch (e) {
      if (e.message === 'Checksum mismatch' || e.message === 'Invalid Base58 character encountered') {
        const privateKey = PrivateKey.from(args[1]);
        const publicKey = privateKey.toPublic();
        console.log('Private key: ', privateKey.toString());
        console.log('Private key (legacy): ', privateKey.toWif());
        console.log('Public key: ', publicKey.toString());
        console.log('Public key (legacy): ', publicKey.toLegacyString());
      } else throw e;
    }
  } else {
    throw new Error(`Unknown command ${args[0]}`);
  }
}

async function authority(args) {
  let authority;
  try {
    const publicKey = PublicKey.from(args[0]);
    authority = Authority.fromKey(publicKey.toString());
  } catch (e) {
    if (e.message === 'Invalid public key string') {
      const permission = 'owner';
      authority = Authority.fromAccountArray(args, permission);
    } else throw e;
  }
  console.log('Authority: ', JSON.stringify(authority));
}

// https://medium.com/coinmonks/setcode-and-setabi-with-eos-js-dd83480ba234
const eosioContract$3 = EosioContract.Instance;
const tonomyContract$3 = TonomyEosioProxyContract.Instance;
function getDeployableFilesFromDir(dir) {
  const dirCont = fs.readdirSync(dir);
  const wasmFileName = dirCont.find(filePath => filePath.match(/.*\.(wasm)$/gi));
  const abiFileName = dirCont.find(filePath => filePath.match(/.*\.(abi)$/gi));
  if (!wasmFileName) throw new Error(`Cannot find a ".wasm file" in ${dir}`);
  if (!abiFileName) throw new Error(`Cannot find an ".abi file" in ${dir}`);
  return {
    wasmPath: path.join(dir, wasmFileName),
    abiPath: path.join(dir, abiFileName)
  };
}
async function deployContract$1({
  account,
  contractDir
}, signer, options) {
  const {
    wasmPath,
    abiPath
  } = getDeployableFilesFromDir(contractDir);
  const wasmFile = fs.readFileSync(wasmPath);
  const abiFile = fs.readFileSync(abiPath, 'utf8');
  const contract = options != null && options.throughTonomyProxy ? tonomyContract$3 : eosioContract$3;
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  await contract.deployContract(Name.from(account), wasmFile, abiFile, signer, options);
}

const eosioContract$2 = EosioContract.Instance;
const tonomyEosioProxyContract$1 = TonomyEosioProxyContract.Instance;
/**
 * creates a key based on secure (hashing) key generation algorithm Argon2
 */
async function generatePrivateKeyFromPassword(password, salt) {
  if (!salt) salt = Checksum256.from(randomBytes(32));
  const hash = await argon2.hash(password, {
    salt: Buffer.from(salt.hexString),
    type: argon2.argon2id,
    raw: true,
    timeCost: 40,
    memoryCost: 64 * 1024,
    parallelism: 1,
    hashLength: 32
  });
  const bytes = Bytes.from(hash);
  const privateKey = new PrivateKey(KeyType.K1, bytes);
  return {
    privateKey,
    salt
  };
}
function getSigner() {
  return createSigner$1(getDefaultAntelopePrivateKey());
}
function getDefaultAntelopePrivateKey() {
  // This is the default private key used by an Antelope node when it is first started
  return PrivateKey.from('PVT_K1_2bfGi9rYsXQSXXTvJbDAPhHLQUojjaNLomdm3cEJ1XTzMqUt3V');
  // PUB_K1_6MRyAjQq8ud7hVNYcfnVPJqcVpscN5So8BhtHuGYqET5BoDq63
}
function getDefaultAntelopePublicKey() {
  return getDefaultAntelopePrivateKey().toPublic();
}
async function updateAccountKey(account, newPublicKey, addCodePermission = false) {
  const authority = Authority.fromKey(newPublicKey.toString());
  if (addCodePermission) authority.addCodePermission(account.toString());
  await eosioContract$2.updateauth(account.toString(), 'active', 'owner', authority, getSigner());
  await eosioContract$2.updateauth(account.toString(), 'owner', 'owner', authority, getSigner());
}
/**
 * Updates the control by account, modifying the active and owner authorities.
 *
 * @param {NameType} account - The account name to update.
 * @param {string} controllerAccount - The account name(s) with controller permissions.
 * @param {boolean} [options.addCodePermission=false] - Whether to add code permission to the authorities. To add the eosio.code authority for smart contracts change this to [true]
 * @param {boolean} [options.replaceActive=true] - Whether to replace the active authority with the owner authority.
 * @param {boolean} [options.useTonomyContract=false] - Whether to use the tonomy eosio proxy contract instead of the eosio contract.
 * @returns {Promise<void>} A Promise that resolves when the update is complete.
 */
async function updateControlByAccount(account, controllerAccount, signer, options = {}) {
  var _options$useTonomyCon, _options$replaceActiv;
  if (!Array.isArray(controllerAccount)) controllerAccount = [controllerAccount];
  const activeAuthority = Authority.fromAccount({
    actor: controllerAccount[0],
    permission: 'active'
  });
  const ownerAuthority = Authority.fromAccount({
    actor: controllerAccount[0],
    permission: 'owner'
  });
  if (options.addCodePermission) {
    if (Array.isArray(options.addCodePermission)) {
      for (const permission of options.addCodePermission) {
        activeAuthority.addCodePermission(permission);
        ownerAuthority.addCodePermission(permission);
      }
    } else {
      activeAuthority.addCodePermission(options.addCodePermission);
      ownerAuthority.addCodePermission(options.addCodePermission);
    }
  }
  // If multiple keys provided, make a 2/3 multisig
  if (controllerAccount.length > 1) {
    for (let i = 1; i < controllerAccount.length; i++) {
      activeAuthority.addAccount({
        actor: controllerAccount[i],
        permission: 'active'
      });
      ownerAuthority.addAccount({
        actor: controllerAccount[i],
        permission: 'owner'
      });
    }
    const threshold = Math.ceil(controllerAccount.length * 2 / 3);
    ownerAuthority.setThreshold(threshold);
    activeAuthority.setThreshold(threshold);
  }
  let contract = eosioContract$2;
  if ((_options$useTonomyCon = options.useTonomyContract) != null ? _options$useTonomyCon : false) {
    // @ts-expect-error contract does not have some of the functions of eosioContract type
    contract = tonomyEosioProxyContract$1;
  }
  if ((_options$replaceActiv = options.replaceActive) != null ? _options$replaceActiv : true) {
    await contract.updateauth(account.toString(), 'active', 'owner', activeAuthority, signer);
  }
  await contract.updateauth(account.toString(), 'owner', 'owner', ownerAuthority, signer);
}

const eosioContract$1 = EosioContract.Instance;
async function createAntelopeAccount({
  account,
  key
}, signer) {
  const accountKey = key != null ? key : getDefaultAntelopePublicKey().toString();
  const ownerAuth = Authority.fromKey(accountKey);
  const activeAuth = Authority.fromKey(accountKey);
  // need to add the eosio.code authority as well so that it can call eosio from the smart contract
  ownerAuth.addCodePermission(account);
  activeAuth.addCodePermission(account);
  await eosioContract$1.newaccount('eosio', account, ownerAuth, activeAuth, signer);
}
async function createApp(options) {
  var _res$username;
  const res = await App.create(options);
  console.log('New app created with username: ', (_res$username = res.username) == null ? void 0 : _res$username.username);
  return res;
}

function getDefaultInterfaceName() {
    let val = 'eth';
    const platform = os.platform();
    if (platform === 'darwin') {
        val = 'en';
    }
    else if (platform === 'win32') {
        val = undefined;
    }
    return val;
}
// typeof os.networkInterfaces family is a number (v18.0.0)
// types: 'IPv4' | 'IPv6' => 4 | 6
// @see https://github.com/nodejs/node/issues/42861
function matchName(actualFamily, expectedFamily) {
    if (expectedFamily === 'IPv4') {
        return actualFamily === 'IPv4' || actualFamily === 4;
    }
    if (expectedFamily === 'IPv6') {
        return actualFamily === 'IPv6' || actualFamily === 6;
    }
    return actualFamily === expectedFamily;
}
function findAddressFromInterface(items, expectedFamily, ignoreLoAddress = false) {
    let firstMatchItem;
    for (const item of items) {
        if (matchName(item.family, expectedFamily)) {
            if (ignoreLoAddress && item.address.startsWith('127.')) {
                continue;
            }
            if (expectedFamily === 'IPv6') {
                // find the scopeid = 0 item
                if (item.scopeid === 0)
                    return item;
                if (!firstMatchItem) {
                    firstMatchItem = item;
                }
            }
            else {
                return item;
            }
        }
    }
    return firstMatchItem;
}
function getInterfaceAddress(family, name) {
    const interfaces = os.networkInterfaces();
    const noName = !name;
    name = name || getDefaultInterfaceName();
    family = family || 'IPv4';
    if (name) {
        for (let i = -1; i < 8; i++) {
            const interfaceName = name + (i >= 0 ? i : ''); // support 'lo' and 'lo0'
            const items = interfaces[interfaceName];
            if (items) {
                const item = findAddressFromInterface(items, family);
                if (item) {
                    return item;
                }
            }
        }
    }
    if (noName) {
        // filter all loopback or local addresses
        for (const k in interfaces) {
            const items = interfaces[k];
            if (items) {
                // all 127 addresses are local and should be ignored
                const item = findAddressFromInterface(items, family, true);
                if (item) {
                    return item;
                }
            }
        }
    }
    return;
}
/**
 * Get current machine IPv4
 *
 * interfaceName: interface name, default is 'eth' on linux, 'en' on mac os.
 */
function ip(interfaceName) {
    const item = getInterfaceAddress('IPv4', interfaceName);
    return item?.address;
}

// cannot use NODE_ENV as it is always "production" on `yarn run build`
const env = process.env.NODE_ENV || 'development';
const ipAddress = ip();
const defaultConfig = {
  environment: 'development',
  ssoWebsiteOrigin: `http://${ipAddress}:3000`,
  demoWebsiteOrigin: `http://${ipAddress}:3001`,
  consoleWebsiteOrigin: `http://${ipAddress}:3002`,
  blockchainUrl: `http://${ipAddress}:8888`,
  accountSuffix: '.stag.tonomy.id',
  communicationUrl: 'ws://localhost:5000',
  accountsServiceUrl: 'http://localhost:5000',
  tonomyIdSchema: 'tonomy-id-development://',
  loggerLevel: 'info',
  ecosystemName: 'Tonomy - Development',
  currencySymbol: 'TONO'
};
const stagingConfig = {
  environment: 'staging',
  ssoWebsiteOrigin: `https://accounts.staging.tonomy.foundation`,
  demoWebsiteOrigin: `https://demo.staging.tonomy.foundation`,
  consoleWebsiteOrigin: `https://console.developers.staging.tonomy.foundation`,
  blockchainUrl: `https://blockchain-api-staging.tonomy.foundation`,
  accountSuffix: '.stag.tonomy.id',
  communicationUrl: 'wss://communication.staging.tonomy.foundation',
  accountsServiceUrl: 'http://communication.staging.tonomy.foundation',
  tonomyIdSchema: 'tonomy-id-staging://',
  loggerLevel: 'info',
  ecosystemName: 'Tonomy - Staging',
  currencySymbol: 'TONO'
};
const testnetConfig = {
  environment: 'testnet',
  ssoWebsiteOrigin: `https://accounts.testnet.tonomy.io`,
  demoWebsiteOrigin: `https://demo.testnet.tonomy.io`,
  consoleWebsiteOrigin: `https://console.developers.testnet.tonomy.io`,
  blockchainUrl: `https://blockchain-api-testnet.tonomy.io`,
  accountSuffix: '.testnet.pangea',
  communicationUrl: 'wss://communication.testnet.tonomy.io',
  accountsServiceUrl: 'http://communication.testnet.tonomy.io',
  tonomyIdSchema: 'pangea-testnet://',
  loggerLevel: 'info',
  ecosystemName: 'Tonomy Testnet',
  currencySymbol: 'TONO'
};
const productionConfig = {
  environment: 'production',
  ssoWebsiteOrigin: `https://accounts.tonomy.io`,
  demoWebsiteOrigin: `https://demo.tonomy.io`,
  consoleWebsiteOrigin: `https://console.developers.tonomy.io`,
  blockchainUrl: `https://blockchain-api.tonomy.io`,
  accountSuffix: '.pangea.id',
  communicationUrl: 'wss://communication.tonomy.io',
  accountsServiceUrl: 'http://communication.tonomy.io',
  tonomyIdSchema: 'united-wallet://',
  loggerLevel: 'info',
  ecosystemName: 'Tonomy',
  currencySymbol: 'TONO'
};
let config;
const settings = {
  env,
  isProduction: () => settings.env === 'production'
};
switch (env) {
  case 'test':
  case 'local':
  case 'development':
    config = defaultConfig;
    break;
  case 'staging':
    config = stagingConfig;
    break;
  case 'testnet':
    config = testnetConfig;
    break;
  case 'production':
    config = productionConfig;
    break;
  default:
    throw new Error('Unknown environment: ' + env);
}
settings.config = config;

class JsStorage {
  constructor(scope) {
    this._storage = {};
    this.scope = void 0;
    this.scope = scope;
    this._storage = {};
  }
  async retrieve(key) {
    return this._storage[key];
  }
  async store(key, value) {
    this._storage[key] = value;
  }
  async clear() {
    this._storage = {};
  }
}
function jsStorageFactory(scope) {
  return new JsStorage(scope);
}

const tonomyContract$2 = TonomyContract.Instance;
const defaultCreateAccount = createAccount;
// Mock the createAccount function to not directly create the account, and thus bypass the hCaptcha requirement
// createAccount() is called during user.createPerson()
function mockCreateAccount() {
  console.log('Mocking createAccount()');
  // eslint-disable-next-line @typescript-eslint/ban-ts-comment
  // @ts-ignore
  createAccount = async function (data) {
    console.log('Calling mocked createAccount()');
    const res = await tonomyContract$2.newperson(data.usernameHash.toString(), data.publicKey.toString(), data.salt.toString(), getSigner());
    const newAccountAction = res.processed.action_traces[0].inline_traces[0].act;
    const accountName = Name.from(newAccountAction.data.name);
    return {
      transactionId: res.transaction_id,
      accountName
    };
  };
}
function restoreCreateAccountFromMock() {
  console.log('Restoring createAccount() to original function');
  // eslint-disable-next-line @typescript-eslint/ban-ts-comment
  // @ts-ignore
  createAccount = defaultCreateAccount;
}
async function createUser(username, password) {
  const auth = new JsKeyManager();
  const user = createUserObject(auth, jsStorageFactory);
  await user.saveUsername(username);
  await user.savePassword(password, {
    keyFromPasswordFn: generatePrivateKeyFromPassword
  });
  await user.createPerson();
  console.log('New user:', (await user.getUsername()).toString());
  return {
    user,
    password,
    auth
  };
}

const demoTokenContract = DemoTokenContract.Instance;
const tokenContract$1 = EosioTokenContract.Instance;
const tonomyContract$1 = TonomyContract.Instance;
const eosioContract = EosioContract.Instance;
const vestingContract$1 = VestingContract.Instance;
const stakeContract = StakingContract.Instance;
const __dirname$1 = path.dirname(fileURLToPath(import.meta.url));
const signer = getSigner();
async function bootstrap() {
  try {
    if (!process.env.TONOMY_OPS_PRIVATE_KEY) throw new Error('Missing TONOMY_OPS_PRIVATE_KEY');
    if (!process.env.TONOMY_BOARD_PUBLIC_KEYS || !Array.isArray(JSON.parse(process.env.TONOMY_BOARD_PUBLIC_KEYS).keys)) throw new Error('Missing TONOMY_BOARD_PUBLIC_KEYS');
    if (!process.env.TONOMY_TEST_ACCOUNTS_PASSPHRASE) throw new Error('Missing TONOMY_TEST_ACCOUNTS_PASSPHRASE');
    const newPrivateKey = PrivateKey.from(process.env.TONOMY_OPS_PRIVATE_KEY);
    const newPublicKey = newPrivateKey.toPublic();
    const newSigner = createSigner$1(newPrivateKey);
    const tonomyGovKeys = JSON.parse(process.env.TONOMY_BOARD_PUBLIC_KEYS).keys;
    const passphrase = process.env.TONOMY_TEST_ACCOUNTS_PASSPHRASE;
    await createAccounts(tonomyGovKeys);
    await setPrivilegedAccounts();
    await setBlockchainParameters();
    await deployEosioMsig();
    await deployVesting();
    await deployStaking();
    await createNativeToken();
    await createTokenDistribution();
    await createTonomyContractAndSetResources();
    await createUsers(passphrase);
    await createTonomyApps(newPublicKey, newSigner);
    await configureDemoToken(newSigner);
    await updateAccountControllers(tonomyGovKeys, newPublicKey);
    await setupVestingAndStaking(newSigner);
    await deployEosioTonomy(newSigner);
    await updateMsigControl(tonomyGovKeys, newSigner);
    console.log('Bootstrap complete');
  } catch (e) {
    console.error('Bootstrap error', e.message, JSON.stringify(e, null, 2));
    process.exit(1);
  }
}
const foundAccount = 'found.tmy';
const foundControlledAccounts = ['gov.tmy', 'team.tmy', 'prod1.tmy', 'prod2.tmy', 'prod3.tmy'];
const govControlledAccounts = ['ops.tmy'];
const operationsAccount = 'ops.tmy';
const opsControlledAccounts = ['tonomy', 'ecosystm.tmy', 'coinsale.tmy', 'eosio.token', 'eosio.msig', 'demo.tmy', 'vesting.tmy', 'staking.tmy', 'legal.tmy', 'reserves.tmy', 'partners.tmy', 'liquidty.tmy', 'marketng.tmy', 'infra.tmy'];
const systemAccount = 'eosio';
async function createAccounts(govKeys) {
  console.log('Create accounts');
  await createAntelopeAccount({
    account: foundAccount
  }, signer);
  // found.tmy should control the following accounts
  for (const account of foundControlledAccounts) {
    await createAntelopeAccount({
      account
    }, signer);
  }
  // gov.tmy should control the following accounts
  for (const account of govControlledAccounts) {
    await createAntelopeAccount({
      account
    }, signer);
  }
  // opts.tmy should control the following accounts
  for (const account of opsControlledAccounts) {
    await createAntelopeAccount({
      account
    }, signer);
  }
  console.log('Create accounts for msig control');
  for (let i = 0; i < govKeys.length; i++) {
    console.log('Create account', indexToAccountName(i));
    await createAntelopeAccount({
      account: indexToAccountName(i)
    }, signer);
  }
}
// Account names can only create digits 1-5
function indexToAccountName(index) {
  if (index > 23) throw new Error('Number too large for current algorithm');
  const indexBase5 = (index + 1).toString(5);
  return indexBase5 + '.found.tmy';
}
async function setPrivilegedAccounts() {
  console.log('Set privileged accounts');
  await eosioContract.setPriv('tonomy', 1, signer);
  await eosioContract.setPriv('eosio.msig', 1, signer);
}
async function setBlockchainParameters() {
  console.log('Set blockchain parameters');
  await eosioContract.setParams(defaultBlockchainParams, signer);
}
async function deployEosioMsig() {
  console.log('Deploy eosio.msig contract');
  await deployContract$1({
    account: 'eosio.msig',
    contractDir: path.join(__dirname$1, '../../Tonomy-Contracts/contracts/eosio.msig')
  }, signer);
}
async function configureDemoToken(newSigner) {
  await demoTokenContract.create(`1000000000 DEMO`, newSigner);
  await demoTokenContract.issue(`10000 DEMO`, newSigner);
}
async function deployVesting() {
  console.log('Deploy vesting.tmy contract');
  await deployContract$1({
    account: 'vesting.tmy',
    contractDir: path.join(__dirname$1, '../../Tonomy-Contracts/contracts/vesting.tmy')
  }, signer);
}
async function deployStaking() {
  console.log('Deploy staking.tmy contract');
  await deployContract$1({
    account: 'staking.tmy',
    contractDir: path.join(__dirname$1, '../../Tonomy-Contracts/contracts/staking.tmy')
  }, signer);
}
async function setupVestingAndStaking(newSigner) {
  await vestingContract$1.setSettings(VestingContract.SALE_START_DATE, VestingContract.VESTING_START_DATE, newSigner);
  await stakeContract.setSettings(amountToAsset(StakingContract.yearlyStakePool, 'TONO'), newSigner);
  await sleep(1000);
  await stakeContract.addYield('infra.tmy', amountToAsset(StakingContract.yearlyStakePool / 2, 'TONO'), newSigner); // 6 months budget in the account
  console.log('Staking settings', await stakeContract.getSettings());
}
async function createNativeToken() {
  console.log('Create and deploy native token contract');
  await deployContract$1({
    account: 'eosio.token',
    contractDir: path.join(__dirname$1, '../../Tonomy-Contracts/contracts/eosio.token')
  }, signer);
  console.log('Create and issue native token');
  await tokenContract$1.create(`50000000000.000000 ${getSettings().currencySymbol}`, signer);
  console.log('Issue native token');
  await tokenContract$1.issue('eosio.token', `50000000000.000000 ${getSettings().currencySymbol}`, signer);
}
const allocations = [['coinsale.tmy', 0.025],
// Seed Private Sale
['coinsale.tmy', 0.055],
// Strategic Partners Private Sale
['coinsale.tmy', 0.07],
// Public Sale
['team.tmy', 0.15],
// Team and Advisors
['legal.tmy', 0.02],
// Legal and Compliance
['reserves.tmy', 0.03],
// Reserves
['partners.tmy', 0.05],
// Partnerships
['liquidty.tmy', 0.05],
// Liquidity Allocation
['marketng.tmy', 0.1],
// Community and Marketing
['ops.tmy', 0.05],
// Platform Operations
['infra.tmy', 0.1],
// Infrastructure Rewards
['ecosystm.tmy', 0.3] // Ecosystem
];
const addCodePermissionTo = allocations.map(allocation => allocation[0]);
async function createTokenDistribution() {
  console.log('Create token distribution');
  let totalPercentage = 0;
  for (const allocation of allocations) {
    const account = allocation[0];
    const percentage = allocation[1];
    totalPercentage += percentage;
    console.log('Allocate', ((percentage * 100).toFixed(1) + '% to').padStart(8), account.padEnd(13), (percentage * EosioTokenContract.TOTAL_SUPPLY).toFixed(0) + `.000000 ${getSettings().currencySymbol}`);
    await tokenContract$1.transfer('eosio.token', account, (percentage * EosioTokenContract.TOTAL_SUPPLY).toFixed(0) + `.000000 ${getSettings().currencySymbol}`, 'Initial allocation', signer);
  }
  if (totalPercentage.toFixed(4) !== '1.0000') {
    throw new Error('Total percentage should be 100% but it is ' + totalPercentage.toFixed(4));
  }
}
async function createTonomyContractAndSetResources() {
  console.log('Deploy Tonomy system contract');
  await deployContract$1({
    account: 'tonomy',
    contractDir: path.join(__dirname$1, '../../Tonomy-Contracts/contracts/tonomy')
  }, signer);
  console.log('Set App account type');
  await tonomyContract$1.adminSetApp(systemAccount, 'System Contract', 'Antelope blockchain system governance contract', getAppUsernameHash('system'), createSubdomainOnOrigin(getSettings().ssoWebsiteOrigin, 'eosio') + '/tonomy-logo1024.png', createSubdomainOnOrigin(getSettings().ssoWebsiteOrigin, 'eosio'), '#251950', '#BA54D3', signer);
  await tonomyContract$1.adminSetApp('eosio.token', 'Native Currency', 'Ecosystem native currency', getAppUsernameHash('currency'), createSubdomainOnOrigin(getSettings().ssoWebsiteOrigin, 'eosio.token') + '/tonomy-logo1024.png', createSubdomainOnOrigin(getSettings().ssoWebsiteOrigin, 'eosio.token'), '#251950', '#BA54D3', signer);
  await tonomyContract$1.adminSetApp('tonomy', 'Tonomy System', 'Tonomy system contract', getAppUsernameHash('tonomy'), createSubdomainOnOrigin(getSettings().ssoWebsiteOrigin, 'tonomy') + '/tonomy-logo1024.png', createSubdomainOnOrigin(getSettings().ssoWebsiteOrigin, 'tonomy'), '#251950', '#BA54D3', signer);
  await tonomyContract$1.adminSetApp('vesting.tmy', 'TONO Vesting', 'TONO Vesting contract', getAppUsernameHash('vesting'), createSubdomainOnOrigin(getSettings().ssoWebsiteOrigin, 'vesting') + '/tonomy-logo1024.png', createSubdomainOnOrigin(getSettings().ssoWebsiteOrigin, 'vesting'), '#251950', '#BA54D3', signer);
  await tonomyContract$1.adminSetApp('staking.tmy', 'TONO Staking', 'TONO Staking contract', getAppUsernameHash('staking'), createSubdomainOnOrigin(getSettings().ssoWebsiteOrigin, 'staking') + '/tonomy-logo1024.png', createSubdomainOnOrigin(getSettings().ssoWebsiteOrigin, 'staking'), '#251950', '#BA54D3', signer);
  console.log('Set Tonomy system contract params and allocate RAM');
  console.log('Set resource params', RAM_PRICE, TOTAL_RAM_AVAILABLE, RAM_FEE);
  await tonomyContract$1.setResourceParams(RAM_PRICE, TOTAL_RAM_AVAILABLE, RAM_FEE, signer);
  console.log('Allocate operational tokens to accounts');
  await tokenContract$1.transfer('ops.tmy', 'tonomy', bytesToTokens(3750000), 'Initial allocation', signer);
  console.log('Allocate RAM to system accounts');
  // See calculation: https://docs.google.com/spreadsheets/d/17cd4wt3oDHp6p7hty9njKsuukTTn9BYJ5z3Ab0N6pMM/edit?pli=1#gid=0&range=D30
  const ramAllocations = [[systemAccount, 3750000], ['eosio.token', 2400000], ['tonomy', 4680000], ['vesting.tmy', 4680000], ['staking.tmy', 4680000]];
  for (const allocation of ramAllocations) {
    const account = allocation[0];
    const tokens = bytesToTokens(allocation[1]);
    console.log(`Buying ${allocation[1] / 1000}KB of RAM for ${account} for ${tokens}`);
    await tokenContract$1.transfer('ops.tmy', account, tokens, 'Initial allocation', signer);
    await tonomyContract$1.buyRam('ops.tmy', account, tokens, signer);
  }
}
function getAppUsernameHash(username) {
  const fullUername = TonomyUsername.fromUsername(username, AccountType.APP, getSettings().accountSuffix);
  return Checksum256.from(fullUername.usernameHash);
}
function createSubdomainOnOrigin(origin, subdomain) {
  const url = new URL(origin);
  return url.protocol + '//' + subdomain + '.' + url.host;
}
async function createUsers(passphrase) {
  mockCreateAccount();
  // Google and Apple app store managers needs to have a test user for their review. That is this user.
  await createUser('testuser', passphrase);
  // Create users for the demo website
  await createUser('lovesboost', passphrase);
  await createUser('sweetkristy', passphrase);
  await createUser('cheesecakeophobia', passphrase);
  await createUser('ultimateBeast', passphrase);
  await createUser('tomtom', passphrase);
  await createUser('readingpro', passphrase);
  restoreCreateAccountFromMock();
}
async function createTonomyApps(newPublicKey, newSigner) {
  console.log('Create Tonomy apps');
  const demo = await createApp({
    appName: `${settings.config.ecosystemName} Demo`,
    usernamePrefix: 'demo',
    description: `Demo of ${settings.config.ecosystemName} login and features`,
    origin: settings.config.demoWebsiteOrigin,
    logoUrl: settings.config.demoWebsiteOrigin + '/market.com.png',
    backgroundColor: '#251950',
    accentColor: '#BA54D3',
    publicKey: newPublicKey,
    signer
  });
  console.log('Deploy demo contract');
  await deployContract$1({
    account: demo.accountName,
    contractDir: path.join(__dirname$1, '../../Tonomy-Contracts/contracts/demo.tmy')
  }, newSigner);
  await createApp({
    appName: `${settings.config.ecosystemName} Website`,
    usernamePrefix: 'tonomy-sso',
    description: `${settings.config.ecosystemName} website to manager your ID and Data`,
    origin: settings.config.ssoWebsiteOrigin,
    logoUrl: settings.config.ssoWebsiteOrigin + '/tonomy-logo1024.png',
    backgroundColor: '#251950',
    accentColor: '#BA54D3',
    publicKey: newPublicKey,
    signer
  });
  await createApp({
    appName: 'Developers Console',
    usernamePrefix: 'developer-console',
    description: `Developer console to manage ${settings.config.ecosystemName} applications and infrastucture`,
    origin: settings.config.consoleWebsiteOrigin,
    logoUrl: settings.config.consoleWebsiteOrigin + '/tonomy-logo1024.png',
    backgroundColor: '#251950',
    accentColor: '#BA54D3',
    publicKey: newPublicKey,
    signer
  });
}
async function updateAccountControllers(govKeys, newPublicKey) {
  console.log('Change the key of the accounts to the new key', newPublicKey.toString());
  await updateAccountKey('found.tmy', newPublicKey);
  function updateControlByOptions(account) {
    return {
      addCodePermission: addCodePermissionTo.includes(account) ? 'vesting.tmy' : undefined
    };
  }
  // accounts controlled by found.tmy
  for (const account of foundControlledAccounts) {
    await updateControlByAccount(account, 'found.tmy', signer, updateControlByOptions(account));
  }
  // ops.tmy account controlled by gov.tmy
  const activeAuthority = Authority.fromAccount({
    actor: 'gov.tmy',
    permission: 'active'
  });
  const ownerAuthority = Authority.fromAccount({
    actor: 'gov.tmy',
    permission: 'owner'
  });
  activeAuthority.addKey(newPublicKey.toString(), 1);
  activeAuthority.addCodePermission('vesting.tmy');
  await eosioContract.updateauth(operationsAccount, 'active', 'owner', activeAuthority, signer);
  await eosioContract.updateauth(operationsAccount, 'owner', 'owner', ownerAuthority, signer);
  // accounts controlled by ops.tmy
  for (const account of opsControlledAccounts.filter(account => !['vesting.tmy', 'staking.tmy', 'tonomy'].includes(account))) {
    if (account === 'infra.tmy') {
      // infra.tmy also have be able to call staking.tmy::addyield() which transfers tokens
      await updateControlByAccount(account, 'ops.tmy', signer, {
        addCodePermission: ['vesting.tmy', 'staking.tmy']
      });
    } else {
      await updateControlByAccount(account, 'ops.tmy', signer, updateControlByOptions(account));
    }
  }
  // (contracts that are called by inline actions need eosio.code permission)
  await updateControlByAccount('vesting.tmy', 'ops.tmy', signer, {
    addCodePermission: 'vesting.tmy'
  });
  await updateControlByAccount('staking.tmy', 'ops.tmy', signer, {
    addCodePermission: 'staking.tmy'
  });
  await updateControlByAccount('tonomy', 'ops.tmy', signer, {
    addCodePermission: 'tonomy'
  });
  // Update the system contract
  await updateControlByAccount(systemAccount, 'tonomy', signer);
  for (let i = 0; i < govKeys.length; i++) {
    await updateAccountKey(indexToAccountName(i), govKeys[i]);
  }
}
async function deployEosioTonomy(signer) {
  console.log('Deploy eosio.tonomy contract');
  await deployContract$1({
    account: 'eosio',
    contractDir: path.join(__dirname$1, '../../Tonomy-Contracts/contracts/eosio.tonomy')
  }, signer, {
    extraAuthorization: {
      actor: 'tonomy',
      permission: 'active'
    }
  });
}
const tonomyEosioProxyContract = TonomyEosioProxyContract.Instance;
async function updateMsigControl(govKeys, signer) {
  console.log('Update found.tmy msig control');
  const govAccounts = govKeys.map(key => indexToAccountName(govKeys.indexOf(key)));
  const activeAuthority = Authority.fromAccount({
    actor: 'found.tmy',
    permission: 'owner'
  });
  const threshold = Math.ceil(govAccounts.length * 2 / 3);
  const ownerAuthority = Authority.fromAccountArray(govAccounts, 'active', threshold);
  await tonomyEosioProxyContract.updateauth('found.tmy', 'active', 'owner', activeAuthority, signer);
  await tonomyEosioProxyContract.updateauth('found.tmy', 'owner', 'owner', ownerAuthority, signer);
}

async function govMigrate(args, options) {
  const threshold = settings.isProduction() ? 3 : 2;
  const action = {
    account: 'tonomy',
    name: 'updateauth',
    authorization: [{
      actor: 'tonomy',
      permission: 'active'
    }, {
      actor: 'tonomy',
      permission: 'owner'
    }],
    data: {
      account: 'found.tmy',
      permission: 'owner',
      parent: '',
      auth: Authority.fromAccountArray(args.newGovernanceAccounts, 'active', threshold),
      // eslint-disable-next-line camelcase
      auth_parent: false // should be true when a new permission is being created, otherwise false
    }
  };
  const proposalHash = await createProposal(options.proposer, options.proposalName, [action], options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}

// @ts-expect-error args not used
async function newAccount(args, options) {
  const accountName = 'login.hypha';
  const active = Authority.fromKey('EOS5DMPJ4DsJ2Vc4f7g5o8z9o5HswcpXrE4C58r7wxxzZgYxQn8rB');
  active.addAccount({
    actor: 'gov.tmy',
    permission: 'active'
  });
  const owner = active;
  const action = {
    account: 'tonomy',
    name: 'newaccount',
    authorization: [{
      actor: 'tonomy',
      permission: 'owner'
    }, {
      actor: 'tonomy',
      permission: 'active'
    }],
    data: {
      creator: 'tonomy',
      name: accountName,
      owner,
      active
    }
  };
  const proposalHash = await createProposal(options.proposer, options.proposalName, [action], options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}

async function transfer$1(options) {
  const from = 'ecosystm.tmy';
  const to = 'coinsale.tmy';
  // const amountUsd = 100000;
  // const price = TONO_PUBLIC_SALE_PRICE;
  // const amount = amountUsd / price;
  // const quantity = amount.toFixed(0) + '.000000 TONO';
  const quantity = '5000000000.000000 TONO';
  const amount = assetToDecimal(quantity);
  const balance = await EosioTokenContract.Instance.getBalanceDecimal(from);
  if (balance.lessThan(amount)) {
    throw new Error(`Insufficient balance. Required: ${amount}, Available: ${balance}`);
  }
  const fraction = amountToSupplyPercentage(amount);
  console.log(`Transferring ${quantity} (${fraction}) from ${from} to ${to}`);
  const action = {
    account: 'eosio.token',
    name: 'transfer',
    authorization: [{
      actor: from,
      permission: 'active'
    }],
    data: {
      from,
      to,
      quantity,
      memo: ''
    }
  };
  const proposalHash = await createProposal(options.proposer, options.proposalName, [action], options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}

async function addAuth(args, options) {
  var _args$useParentAuth, _args$useParentAuth2;
  const accountInfo = await getAccountInfo(Name.from(args.account));
  const perm = accountInfo.getPermission(args.permission);
  const newAuthority = Authority.fromAccountPermission(perm);
  newAuthority.addAccount({
    actor: args.newDelegate,
    permission: 'active'
  });
  const permission = ((_args$useParentAuth = args.useParentAuth) != null ? _args$useParentAuth : false) ? perm.parent.toString() : args.permission;
  const action = {
    account: 'tonomy',
    name: 'updateauth',
    authorization: [{
      actor: args.account,
      permission
    }, {
      actor: 'tonomy',
      permission: 'active'
    }, {
      actor: 'tonomy',
      permission: 'owner'
    }],
    data: {
      account: args.account,
      permission: args.permission,
      parent: perm.parent,
      auth: newAuthority,
      // eslint-disable-next-line camelcase
      auth_parent: (_args$useParentAuth2 = args.useParentAuth) != null ? _args$useParentAuth2 : false
    }
  };
  const requested = [...options.requested, {
    actor: args.account,
    permission
  }];
  const proposalHash = await createProposal(options.proposer, options.proposalName, [action], options.privateKey, requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}

const __filename = fileURLToPath(import.meta.url);
const __dirname = dirname(__filename);
const thisFileDirectory = __dirname;
const defaultContractDirectory = path.join(thisFileDirectory, '..', '..', 'Tonomy-Contracts', 'contracts');
console.log('defaultContractDirectory', defaultContractDirectory);
async function deployContract(options) {
  var _options$directory, _options$returnAction;
  const contractName = Name.from(options.contract);
  const contractDir = `${(_options$directory = options.directory) != null ? _options$directory : defaultContractDirectory}/${contractName.toString()}`;
  if (!contractName) {
    throw new Error('Contract name must be provided for deploy-contract proposal');
  }
  const contractInfo = {
    account: contractName,
    contractDir
  };
  const {
    wasmPath,
    abiPath
  } = getDeployableFilesFromDir(contractInfo.contractDir);
  const wasmFile = fs.readFileSync(wasmPath);
  const abiFile = fs.readFileSync(abiPath, 'utf8');
  const wasm = wasmFile.toString(`hex`);
  // 2. Prepare SETABI
  const abi = JSON.parse(abiFile);
  const abiDef = ABI.from(abi);
  const abiSerializedHex = Serializer.encode({
    object: abiDef
  }).hexString;
  // Prepare SETCODE action
  const setCodeAction = {
    account: 'tonomy',
    name: 'setcode',
    authorization: [{
      actor: contractName.toString(),
      permission: 'active'
    }, {
      actor: 'tonomy',
      permission: 'active'
    }, {
      actor: 'tonomy',
      permission: 'owner'
    }],
    data: {
      account: contractName.toString(),
      vmtype: 0,
      vmversion: 0,
      code: wasm
    }
  };
  // Prepare SETABI action
  const setAbiAction = {
    account: 'tonomy',
    name: 'setabi',
    authorization: [{
      actor: contractName.toString(),
      permission: 'active'
    }, {
      actor: 'tonomy',
      permission: 'active'
    }, {
      actor: 'tonomy',
      permission: 'owner'
    }],
    data: {
      account: contractName.toString(),
      abi: abiSerializedHex
    }
  };
  const actions = [setCodeAction, setAbiAction];
  if ((_options$returnAction = options.returnActions) != null ? _options$returnAction : false) return actions;
  const proposalHash = await createProposal(options.proposer, options.proposalName, actions, options.privateKey, options.requested, options.dryRun);
  if (!options.dryRun && options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
  return;
}

async function addEosioCode(options) {
  const actions = [];
  const accountsToUpdate = [...addCodePermissionTo, 'advteam.tmy'];
  for (const account of accountsToUpdate) {
    const accountInfo = await getAccountInfo(Name.from(account));
    const ownerPermission = accountInfo.getPermission('owner');
    const activePermission = accountInfo.getPermission('active');
    const ownerAuthority = Authority.fromAccountPermission(ownerPermission);
    const activeAuthority = Authority.fromAccountPermission(activePermission);
    activeAuthority.addCodePermission('vesting.tmy');
    activeAuthority.addCodePermission('staking.tmy');
    ownerAuthority.addCodePermission('vesting.tmy');
    ownerAuthority.addCodePermission('staking.tmy');
    actions.push({
      account: 'tonomy',
      name: 'updateauth',
      authorization: [{
        actor: account,
        permission: 'owner'
      }, {
        actor: 'tonomy',
        permission: 'active'
      }, {
        actor: 'tonomy',
        permission: 'owner'
      }],
      data: {
        account: account,
        permission: 'owner',
        parent: '',
        auth: ownerAuthority,
        // eslint-disable-next-line camelcase
        auth_parent: false
      }
    });
    actions.push({
      account: 'tonomy',
      name: 'updateauth',
      authorization: [{
        actor: account,
        permission: 'active'
      }, {
        actor: 'tonomy',
        permission: 'active'
      }, {
        actor: 'tonomy',
        permission: 'owner'
      }],
      data: {
        account: account,
        permission: 'active',
        parent: 'owner',
        auth: activeAuthority,
        // eslint-disable-next-line camelcase
        auth_parent: false
      }
    });
  }
  const proposalHash = await createProposal(options.proposer, options.proposalName, actions, options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}

async function vestingBulk(args, options) {
  const csvFilePath = '/home/dev/Downloads/allocate.csv';
  console.log('Reading file: ', csvFilePath);
  const sender = settings.isProduction() ? 'advteam.tmy' : 'team.tmy';
  const requiredAuthority = options.autoExecute ? args.governanceAccounts[2] : '11.found.tmy';
  const categoryId = 7; // Community and Marketing, Platform Dev, Infra Rewards
  // https://github.com/Tonomy-Foundation/Tonomy-Contracts/blob/master/contracts/vesting.tmy/include/vesting.tmy/vesting.tmy.hpp#L31
  const records = parse$1(fs.readFileSync(csvFilePath, 'utf8'), {
    columns: true,
    // eslint-disable-next-line camelcase
    skip_empty_lines: true
  });
  const results = [];
  const unfoundAccounts = [];
  await Promise.all(records.map(async data => {
    // accountName, usdQuantity
    if (!data.accountName || !data.usdQuantity) {
      throw new Error(`Invalid CSV format on line ${results.length + 1}: ${data}`);
    }
    try {
      let accountName = data.accountName;
      if (accountName.startsWith('@')) {
        const usernameInstance = TonomyUsername.fromUsername(accountName.split('@')[1], AccountType.PERSON, settings.config.accountSuffix);
        accountName = (await getAccountNameFromUsername(usernameInstance)).toString();
      } else {
        await getAccount(accountName);
      }
      data.accountName = accountName;
      data.usdQuantity = Number(data.usdQuantity);
      if (isNaN(data.usdQuantity)) {
        throw new Error(`Invalid quantity type on line ${results.length + 1}: ${data}`);
      }
      if (data.usdQuantity <= 0 || data.usdQuantity > 100000) {
        throw new Error(`Invalid quantity on line ${results.length + 1}: ${data}`);
      }
      results.push(data);
    } catch (e) {
      if (e instanceof SdkError && (e.code === SdkErrors.AccountDoesntExist || e.code === SdkErrors.UsernameNotFound)) {
        unfoundAccounts.push(data.accountName);
      } else {
        throw e;
      }
    }
  }));
  if (unfoundAccounts.length > 0) {
    console.error(`${unfoundAccounts.length} accounts were not found in environment ${settings.env}:`, unfoundAccounts);
    process.exit(1);
  }
  const actions = results.map(data => {
    const tonoNumber = data.usdQuantity / TONO_CURRENT_PRICE;
    const tonoQuantity = tonoNumber.toFixed(0) + '.000000 TONO';
    console.log(`Assigning: ${tonoQuantity} ($${data.usdQuantity} USD) vested in category ${categoryId} to ${data.accountName} at rate of $${TONO_CURRENT_PRICE}/TONO`);
    return {
      account: 'vesting.tmy',
      name: 'assigntokens',
      authorization: [{
        actor: sender.toString(),
        permission: 'active'
      }],
      data: {
        sender,
        holder: data.accountName,
        amount: tonoQuantity,
        category: categoryId
      }
    };
  });
  console.log(`Total ${actions.length} accounts to be paid`);
  const proposalHash = await createProposal(options.proposer, options.proposalName, actions, options.privateKey, [requiredAuthority], options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}

/* eslint-disable camelcase */
// @ts-expect-error args not used
async function setResourceConfig(args, options) {
  const action = {
    authorization: [{
      actor: 'tonomy',
      permission: 'active'
    }],
    account: 'tonomy',
    name: 'setresparams',
    data: {
      ram_price: RAM_PRICE,
      total_ram_available: TOTAL_RAM_AVAILABLE,
      ram_fee: RAM_FEE
    }
  };
  const proposalHash = await createProposal(options.proposer, options.proposalName, [action], options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}

// @ts-expect-error args not used
async function setBlockchainConfig(args, options) {
  const action = {
    authorization: [{
      actor: 'tonomy',
      permission: 'active'
    }, {
      actor: 'tonomy',
      permission: 'owner'
    }],
    account: 'tonomy',
    name: 'setparams',
    data: {
      params: defaultBlockchainParams
    }
  };
  const proposalHash = await createProposal(options.proposer, options.proposalName, [action], options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}

// @ts-expect-error any not used
async function addProd(args, options) {
  const producer = '1.found.tmy';
  const signingKey = PublicKey.from('EOS6A3TosyQZPa9g186tqVFa52AfLdkvaosy1XVEEgziuAyp5PMUj');
  const newSchedule = await getSchedule();
  if (newSchedule.find(p => p.producer_name.equals(producer))) throw new Error('Producer already in the schedule');
  newSchedule.push({
    // eslint-disable-next-line camelcase
    producer_name: Name.from(producer),
    authority: ['block_signing_authority_v0', {
      threshold: 1,
      keys: [{
        key: signingKey,
        weight: Weight.from(1)
      }]
    }]
  });
  const action = createSetProdsAction(newSchedule);
  const proposalHash = await createProposal(options.proposer, options.proposalName, [action], options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}
// @ts-expect-error any not used
async function updateProd(args, options) {
  const producer = 'eosiodetroit';
  const signingKey = PublicKey.from('PUB_K1_7EyKQTyVABvDdmnS3VZceuqCCTTUZft6ZaPqbwZBYW8DFCxwKe');
  let newSchedule = await getSchedule();
  if (!newSchedule.find(p => p.producer_name.equals(producer))) throw new Error('Producer not in the schedule');
  newSchedule = newSchedule.filter(p => !p.producer_name.equals(producer));
  newSchedule.push({
    // eslint-disable-next-line camelcase
    producer_name: Name.from(producer),
    authority: ['block_signing_authority_v0', {
      threshold: 1,
      keys: [{
        key: signingKey,
        weight: Weight.from(1)
      }]
    }]
  });
  const action = createSetProdsAction(newSchedule);
  const proposalHash = await createProposal(options.proposer, options.proposalName, [action], options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}
// @ts-expect-error any not used
async function removeProd(args, options) {
  const producer = '1.found.tmy';
  // fetch the existing schedule and their keys
  let newSchedule = await getSchedule();
  if (!newSchedule.find(p => p.producer_name.equals(producer))) throw new Error('Producer not in the schedule');
  newSchedule = newSchedule.filter(p => !p.producer_name.equals(producer));
  const action = createSetProdsAction(newSchedule);
  const proposalHash = await createProposal(options.proposer, options.proposalName, [action], options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}
// @ts-expect-error any not used
async function changeProds(args, options) {
  const newProducers = [{
    producer: 'prod1.tmy',
    signingKey: PublicKey.from('PUB_K1_4uTNHX8hdWxyhYppW1pSESS9FeYP4nAipozUerQCTvXFwN9vq5')
  }, {
    producer: 'prod2.tmy',
    signingKey: PublicKey.from('PUB_K1_7THhTGLvEm5MB9YD1JEyDfQmpQtXhh2PXfY3vLtNJ72z7zZ7HN')
  }, {
    producer: 'prod3.tmy',
    signingKey: PublicKey.from('PUB_K1_56SRzctzrZTC5xtVnNYqS5kwVPxMi5QbTMZEQgpMrNFGatUrUX')
  }];
  const oldSchedule = await getSchedule();
  for (const {
    producer
  } of newProducers) {
    if (oldSchedule.find(p => p.producer_name.equals(producer))) throw new Error('Producer already in the schedule');
  }
  const newSchedule = newProducers.map(({
    producer,
    signingKey
  }) => {
    return {
      // eslint-disable-next-line camelcase
      producer_name: Name.from(producer),
      authority: ['block_signing_authority_v0', {
        threshold: 1,
        keys: [{
          key: signingKey,
          weight: Weight.from(1)
        }]
      }]
    };
  });
  const action = createSetProdsAction(newSchedule);
  const proposalHash = await createProposal(options.proposer, options.proposalName, [action], options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}
async function getSchedule() {
  const {
    pending,
    proposed,
    active
  } = await getProducers();
  if (pending || proposed) throw new Error("Can't add a producer while there is a pending schedule");
  console.log('Fetched schedule:');
  for (const producer of active.producers) {
    console.log(`- ${producer.producer_name.toString().padEnd(13, ' ')} with signing key ${producer.authority[1].keys[0].key}`);
  }
  return active.producers.map(p => {
    return {
      // eslint-disable-next-line camelcase
      producer_name: p.producer_name,
      authority: ['block_signing_authority_v0', {
        threshold: 1,
        keys: p.authority[1].keys.map(k => {
          return {
            key: k.key,
            weight: k.weight
          };
        })
      }]
    };
  });
}
function createSetProdsAction(schedule) {
  console.log('New schedule:');
  for (const producer of schedule) {
    console.log(`- ${producer.producer_name.toString().padEnd(13, ' ')} with signing key ${producer.authority[1].keys[0].key}`);
  }
  return {
    account: 'tonomy',
    name: 'setprods',
    authorization: [{
      actor: 'tonomy',
      permission: 'owner'
    }, {
      actor: 'tonomy',
      permission: 'active'
    }],
    data: {
      schedule
    }
  };
}

/* eslint-disable camelcase */
// @ts-expect-error args not used
async function hyphaAccountsCreate(args, options) {
  function createNewAccountAction(name, active, owner) {
    return {
      account: 'tonomy',
      name: 'newaccount',
      authorization: [{
        actor: 'tonomy',
        permission: 'owner'
      }, {
        actor: 'tonomy',
        permission: 'active'
      }],
      data: {
        creator: 'tonomy',
        name,
        owner,
        active
      }
    };
  }
  const tonomyGovActivePermission = {
    actor: 'gov.tmy',
    permission: 'active'
  };
  const daoHyphaAction = createNewAccountAction('dao.hypha', Authority.fromKey('EOS7pMcjCS15XKxEssnFqKeq5HUJVTuoXMBbjYfGrQJrya2TSxgNV').addAccount(tonomyGovActivePermission), Authority.fromKey('EOS7pMcjCS15XKxEssnFqKeq5HUJVTuoXMBbjYfGrQJrya2TSxgNV').addAccount(tonomyGovActivePermission).addCodePermission('dao.hypha'));
  const voiceHyphaAction = createNewAccountAction('voice.hypha', Authority.fromKey('EOS55WiNqboivthg3uERZkdj4saxbSt8D33KQzWqaMBCXASoqxFsP').addAccount(tonomyGovActivePermission).addCodePermission('dao.hypha').addCodePermission('voice.hypha'), Authority.fromKey('EOS55WiNqboivthg3uERZkdj4saxbSt8D33KQzWqaMBCXASoqxFsP').addAccount(tonomyGovActivePermission));
  const hyphaHyphaAction = createNewAccountAction('hypha.hypha', Authority.fromKey('EOS6KTU2YYtnN8ZRXDcG4cdFpRGkRssrBgR1u7Y3YqUX3KYYcfpRP').addAccount(tonomyGovActivePermission).addCodePermission('dao.hypha').addCodePermission('hypha.hypha'), Authority.fromKey('EOS6KTU2YYtnN8ZRXDcG4cdFpRGkRssrBgR1u7Y3YqUX3KYYcfpRP').addAccount(tonomyGovActivePermission));
  const husdHyphaAction = createNewAccountAction('husd.hypha', Authority.fromKey('EOS8ftihCrL5nVJDUo4Y5qCzcJpeHuLA1jezP38j3jHm8tGDiAsbQ').addAccount(tonomyGovActivePermission).addCodePermission('dao.hypha').addCodePermission('husd.hypha'), Authority.fromKey('EOS8ftihCrL5nVJDUo4Y5qCzcJpeHuLA1jezP38j3jHm8tGDiAsbQ').addAccount(tonomyGovActivePermission));
  const kvHyphaAction = createNewAccountAction('kv.hypha', Authority.fromKey('EOS8kcBzk6JrzDgoAqYMHzNFrgdkxJMM595kjJxWUPEfVtnHsdKLK').addAccount(tonomyGovActivePermission), Authority.fromKey('EOS8kcBzk6JrzDgoAqYMHzNFrgdkxJMM595kjJxWUPEfVtnHsdKLK').addAccount(tonomyGovActivePermission).addCodePermission('kv.hypha'));
  const joinHyphaAction = createNewAccountAction('join.hypha', Authority.fromKey('EOS52LPbWzd8iXUS3DWnn2CkU4HE4kVRSHVtyvWK4dDMLnAUYnogB').addAccount(tonomyGovActivePermission), Authority.fromKey('EOS52LPbWzd8iXUS3DWnn2CkU4HE4kVRSHVtyvWK4dDMLnAUYnogB').addAccount(tonomyGovActivePermission).addCodePermission('join.hypha'));
  const srviceHyphaAction = createNewAccountAction('srvice.hypha', Authority.fromKey('EOS7ammTdz8NCdpfPwwXqYck6d93iUVZaE4hgjgTyxXX1nYFEHMzJ').addAccount(tonomyGovActivePermission), Authority.fromKey('EOS7ammTdz8NCdpfPwwXqYck6d93iUVZaE4hgjgTyxXX1nYFEHMzJ').addAccount(tonomyGovActivePermission));
  const loginHyphaAction = createNewAccountAction('login.hypha', Authority.fromKey('EOS7ktQxW7cPTF17VhTV6U64WVdGCtMoxM7kbszQmPYTsi5nMVfYi').addAccount(tonomyGovActivePermission), Authority.fromKey('EOS7ktQxW7cPTF17VhTV6U64WVdGCtMoxM7kbszQmPYTsi5nMVfYi').addAccount(tonomyGovActivePermission));
  const actions = [daoHyphaAction, voiceHyphaAction, hyphaHyphaAction, husdHyphaAction, kvHyphaAction, joinHyphaAction, srviceHyphaAction, loginHyphaAction];
  const proposalHash = await createProposal(options.proposer, options.proposalName, actions, options.privateKey, [...options.requested], options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}
// @ts-expect-error args not used
async function hyphaAddAccountPermissions(args, options) {
  const tonomyActivePermission = {
    actor: 'tonomy',
    permission: 'active'
  };
  const tonomyOwnerPermission = {
    actor: 'tonomy',
    permission: 'owner'
  };
  const hyphaDaoSchedulerPermission = {
    account: 'tonomy',
    name: 'updateauth',
    authorization: [{
      actor: 'dao.hypha',
      permission: 'active'
    }, tonomyOwnerPermission, tonomyActivePermission],
    data: {
      account: 'dao.hypha',
      permission: 'scheduler',
      parent: 'active',
      auth: {
        threshold: 1,
        waits: [],
        accounts: [],
        keys: [{
          key: 'EOS8Et6mKgaNwQDBTSwPig9oDbgXKsjLyojbJg6rBNQ25M8YAoUiJ',
          weight: 1
        }]
      },
      auth_parent: true
    }
  };
  const hyphaDaoLinkauthScheduler = {
    account: 'tonomy',
    name: 'linkauth',
    authorization: [{
      actor: 'dao.hypha',
      permission: 'active'
    }],
    data: {
      account: 'dao.hypha',
      code: 'dao.hypha',
      type: 'removedtx',
      requirement: 'scheduler'
    }
  };
  const hyphaDaoAutoenrollPermission = {
    account: 'tonomy',
    name: 'updateauth',
    authorization: [{
      actor: 'dao.hypha',
      permission: 'active'
    }, tonomyOwnerPermission, tonomyActivePermission],
    data: {
      account: 'dao.hypha',
      permission: 'autoenroll',
      parent: 'active',
      auth: {
        threshold: 1,
        waits: [],
        accounts: [{
          permission: {
            actor: 'join.hypha',
            permission: 'eosio.code'
          },
          weight: 1
        }],
        keys: []
      },
      auth_parent: true
    }
  };
  const hyphaDaoLinkauthAutoenroll = {
    account: 'tonomy',
    name: 'linkauth',
    authorization: [{
      actor: 'dao.hypha',
      permission: 'active'
    }],
    data: {
      account: 'dao.hypha',
      code: 'dao.hypha',
      type: 'autoenroll',
      requirement: 'autoenroll'
    }
  };
  const actions = [hyphaDaoSchedulerPermission, hyphaDaoLinkauthScheduler, hyphaDaoAutoenrollPermission, hyphaDaoLinkauthAutoenroll];
  const proposalHash = await createProposal(options.proposer, options.proposalName, actions, options.privateKey, [...options.requested], options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}
// @ts-expect-error args not used
async function hyphaContractSet(args, options) {
  const ramKb = 60000;
  const name = 'dao';
  // const ramKb = 2000;
  // const name = 'husd';
  // const name = 'hypha';
  // const name = 'join';
  // const name = 'kv';
  // const name = 'login';
  // const name = 'voice';
  const contract = `${name}.hypha`;
  const appName = `Hypha ${name.toUpperCase()}`;
  const username = `${name}hypha`;
  const description = `Hypha ${name} contract`;
  const logoUrl = 'https://hypha.earth/wp-content/themes/hypha2023/img/logos/logo-white.svg';
  const origin = `https://${name}.pangea.hypha.earth`;
  const contractDir = `/home/dev/Downloads/pangea-hypha-deploy/${contract}`;
  const tonomyUsername = TonomyUsername.fromUsername(username, AccountType.APP, getSettings().accountSuffix);
  const tokens = bytesToTokens(ramKb * 1000);
  console.log(`Setting up hypha contract "${contract}" with ${tokens} tokens to buy ${ramKb}KB of RAM`);
  const adminSetAppAction = {
    authorization: [{
      actor: 'tonomy',
      permission: 'active'
    }],
    account: 'tonomy',
    name: 'adminsetapp',
    data: {
      account_name: Name.from(contract),
      app_name: appName,
      description,
      username_hash: tonomyUsername.usernameHash,
      logo_url: logoUrl,
      origin
    }
  };
  const transferTokensAction = {
    authorization: [{
      actor: 'partners.tmy',
      permission: 'active'
    }],
    account: 'eosio.token',
    name: 'transfer',
    data: {
      from: 'partners.tmy',
      to: contract,
      quantity: tokens,
      memo: `RAM for ${contract}`
    }
  };
  const buyRamAction = {
    account: 'tonomy',
    name: 'buyram',
    authorization: [{
      actor: contract,
      permission: 'active'
    }],
    data: {
      dao_owner: contract,
      app: contract,
      quant: tokens
    }
  };
  const deployActions = await deployContract({
    contract: contract,
    directory: contractDir,
    returnActions: true,
    ...options
  });
  if (!deployActions) throw new Error('Expected deployActions to be defined');
  const actions = [adminSetAppAction, transferTokensAction, buyRamAction, ...deployActions];
  const proposalHash = await createProposal(options.proposer, options.proposalName, actions, options.privateKey, [...options.requested, contract], options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}

const vestingContract = VestingContract.Instance;
async function getAllUniqueHolders(print = false) {
  const action = 'assigntokens';
  const contract = 'vesting.tmy';
  const limit = 100;
  let skip = 0;
  let actionsFound = 0;
  const uniqueHolders = new Set();
  let host = 'pangea.eosusa.io';
  if (getSettings().environment === 'testnet') {
    host = 'test.pangea.eosusa.io';
  } else if (getSettings().environment !== 'production') {
    throw new Error(`environment ${getSettings().environment} not supported for fetching all vesting holders`);
  }
  do {
    var _data$actions;
    const url = `https://${host}/v2/history/get_actions?act.name=${action}&sort=desc&skip=${skip}&limit=${limit}&account=${contract}&global_sequence=0-45539775`;
    const res = await fetch(url);
    const data = await res.json();
    const actions = data.actions;
    actionsFound = (data == null || (_data$actions = data.actions) == null ? void 0 : _data$actions.length) || 0;
    for (const action of actions) {
      const {
        sender,
        holder,
        amount,
        category
      } = action.act.data;
      uniqueHolders.add(holder);
      if (print) console.log(`${action.timestamp}: Sent ${amount} from ${sender} to ${holder} in category ${category}`);
    }
    skip += limit;
  } while (actionsFound > 0);
  return uniqueHolders;
}
async function getAllAllocations(accounts, print = false) {
  const allocations = [];
  for (const account of accounts) {
    const accountAllocations = await vestingContract.getAllocations(account);
    for (const allocation of accountAllocations) {
      // eslint-disable-next-line camelcase
      const {
        id,
        tokens_allocated,
        vesting_category_type
      } = allocation;
      if (print) console.log(
      // eslint-disable-next-line camelcase
      `Holder ${account}: Allocation ${id}: ${tokens_allocated} in category ${vesting_category_type}`);
      allocations.push({
        account,
        ...allocation
      });
    }
  }
  return allocations;
}
async function vesting(args) {
  if (args[0] === 'assign') {
    const privateKey = PrivateKey.from(process.env.SIGNING_KEY || '');
    const signer = createSigner$1(privateKey);
    const sender = args[1];
    let recipient = args[2];
    const quantity = args[3];
    const categoryId = Number(args[4]);
    if (recipient.startsWith('@')) {
      console.log('Searching for username: ', recipient);
      const usernameInstance = TonomyUsername.fromUsername(recipient.slice(1), AccountType.PERSON, settings.config.accountSuffix);
      recipient = (await getAccountNameFromUsername(usernameInstance)).toString();
    } else {
      console.log('Searching for account name: ', recipient);
      await getAccount(recipient);
    }
    console.log('Account name: ', recipient.toString());
    console.log('Assigning tokens to: ', {
      sender,
      holder: recipient,
      quantity,
      categoryId
    });
    const vestingSettings = await vestingContract.getSettings();
    console.log('settings', vestingSettings);
    const res = await vestingContract.assignTokens(sender, recipient, quantity, categoryId, signer);
    console.log('Transaction ID: ', JSON.stringify(res, null, 2));
  } else if (args[0] === 'audit') {
    const uniqueHolders = await getAllUniqueHolders(true);
    console.log('');
    console.log('Unique holders: ', uniqueHolders);
    console.log('');
    const allAllocations = await getAllAllocations(uniqueHolders, true);
    const totalVested = allAllocations.reduce((previous, allocation) => previous += assetToAmount(allocation.tokens_allocated), 0);
    console.log('');
    console.log('Total unique holders: ', uniqueHolders.size);
    console.log('Total vesting allocations: ', allAllocations.length);
    console.log(`Total vested: ${totalVested} TONO (${(100 * totalVested / EosioTokenContract.TOTAL_SUPPLY).toFixed(2)}%)`);
    console.log('');
  } else if (args[0] === 'setsettings') {
    const privateKey = PrivateKey.from(process.env.SIGNING_KEY || '');
    const signer = createSigner$1(privateKey);
    await vestingContract.setSettings('2024-04-30T12:00:00', '2030-01-01T00:00:00', signer);
  } else {
    printCliHelp();
    throw new Error(`Unknown command ${args[0]}`);
  }
}

/* eslint-disable no-console */
const tokenContract = EosioTokenContract.Instance;
async function transfer(args) {
  const privateKey = PrivateKey.from(process.env.SIGNING_KEY || '');
  const signer = createSigner$1(privateKey);
  const sender = args[1];
  let recipient = args[2];
  const quantity = args[3];
  const memo = args[4] || '';
  if (recipient.startsWith('@')) {
    console.log('Searching for username: ', recipient);
    const usernameInstance = TonomyUsername.fromUsername(recipient.slice(1), AccountType.PERSON, settings.config.accountSuffix);
    recipient = (await getAccountNameFromUsername(usernameInstance)).toString();
  } else {
    console.log('Searching for account name: ', recipient);
    await getAccount(recipient);
  }
  console.log('Account name: ', recipient.toString());
  console.log('Assigning tokens to: ', {
    from: sender,
    to: recipient,
    quantity,
    memo
  });
  const res = await tokenContract.transfer(sender, recipient, quantity, memo, signer);
  console.log('Transaction ID: ', JSON.stringify(res, null, 2));
}
const ZERO_DECIMAL = new Decimal(0);
const symbol = 'TONO';
async function audit() {
  console.log('Token symbol:', symbol);
  setSettings({
    ...settings.config,
    currencySymbol: symbol
  });
  console.log('');
  console.log('Fetching tokens for bootstrap accounts');
  const bootstrappedAccounts = new Set();
  bootstrappedAccounts.add(foundAccount);
  bootstrappedAccounts.add(operationsAccount);
  bootstrappedAccounts.add(systemAccount);
  if (settings.env === 'production') bootstrappedAccounts.add('advteam.tmy');
  for (const account of foundControlledAccounts) bootstrappedAccounts.add(account);
  for (const account of opsControlledAccounts) bootstrappedAccounts.add(account);
  const bootstrappedData = (await Promise.all(Array.from(bootstrappedAccounts).map(async account => {
    const balance = await tokenContract.getBalanceDecimal(account);
    return {
      account,
      description: 'Bootstrapped',
      tokens: balance,
      vested: ZERO_DECIMAL
    };
  }))).sort((a, b) => b.tokens.cmp(a.tokens));
  bootstrappedData.forEach(({
    account,
    tokens
  }) => {
    const fraction = amountToSupplyPercentage(tokens);
    console.log(`${account.padEnd(14)} ${tokens.toFixed(4).padStart(16)} ${symbol} (${fraction.padStart(12)})`);
  });
  const totalBoostrappedTokens = bootstrappedData.reduce((previous, account) => previous.add(account.tokens), ZERO_DECIMAL);
  console.log(`Total bootstrapped tokens:  ${totalBoostrappedTokens.toFixed(4).padStart(14)} ${symbol} (${amountToSupplyPercentage(totalBoostrappedTokens).padStart(10)})`);
  console.log('');
  console.log('Fetching apps tokens');
  const apps = await TonomyContract.Instance.getApps();
  const appAccounts = (await Promise.all(apps.map(async app => {
    const balance = await tokenContract.getBalanceDecimal(app.account_name);
    const account = await getAccount(app.account_name);
    return {
      account: app.account_name.toString(),
      description: app.description,
      tokens: balance,
      vested: ZERO_DECIMAL,
      ramQuota: account.ram_quota.toNumber(),
      ramUsage: account.ram_usage.toNumber()
    };
  }))).sort((a, b) => b.tokens.cmp(a.tokens));
  const totalAllTokens = appAccounts.reduce((previous, app) => previous.add(app.tokens), ZERO_DECIMAL);
  console.log('Total apps', apps.length);
  console.log(`Total app tokens:  ${totalAllTokens.toFixed(4).padStart(14)} ${symbol} (${amountToSupplyPercentage(totalAllTokens).padStart(10)})`);
  for (const app of appAccounts) {
    var _app$tokens;
    if ((_app$tokens = app.tokens) != null && _app$tokens.eq(ZERO_DECIMAL)) continue;
    const fraction = amountToSupplyPercentage(app.tokens);
    console.log(`> ${app.account.padEnd(14)} ${app.tokens.toFixed(4).padStart(16)} ${symbol} (${fraction.padStart(12)}) ${app.description}`);
  }
  console.log('');
  console.log('Fetching people tokens');
  const people = await getAllPeople();
  const peopleAccounts = [];
  const batchSize = 100;
  for (let i = 0; i < people.length; i += batchSize) {
    const batch = people.slice(i, i + batchSize);
    const batchResults = await Promise.all(batch.map(async person => {
      const balance = await tokenContract.getBalanceDecimal(person.account_name);
      return {
        account: person.account_name.toString(),
        description: person.status,
        tokens: balance,
        vested: ZERO_DECIMAL
      };
    }));
    peopleAccounts.push(...batchResults);
  }
  peopleAccounts.sort((a, b) => b.tokens.cmp(a.tokens));
  const totalPeopleTokens = peopleAccounts.reduce((previous, person) => previous.add(person.tokens), ZERO_DECIMAL);
  console.log('Total people', people.length);
  console.log(`Total people tokens:  ${totalPeopleTokens.toFixed(4).padStart(14)} ${symbol} (${amountToSupplyPercentage(totalPeopleTokens).padStart(10)})`);
  for (const person of peopleAccounts) {
    if (person.tokens.eq(ZERO_DECIMAL)) continue;
    const fraction = amountToSupplyPercentage(person.tokens);
    console.log(`> ${person.account.toString().padEnd(14)} ${person.tokens.toFixed(4).padStart(15)} ${symbol} (${fraction.padStart(12)})`);
  }
  console.log('');
  console.log('Fetching vested tokens');
  // const vestingHolders = await getAllUniqueHolders();
  const vestingHolders = people.reduce((previous, person) => {
    previous.add(person.account_name.toString());
    return previous;
  }, new Set());
  const vestingAllocations = await getAllAllocations(vestingHolders, false);
  const vestingCategories$1 = vestingAllocations.reduce((previous, allocation) => {
    if (previous.includes(allocation.vesting_category_type)) return previous;else return [...previous, allocation.vesting_category_type];
  }, []);
  const vestedTokensPerCategory = vestingCategories$1.reduce((map, category) => map.set(category, new Decimal(0)), new Map());
  for (const allocation of vestingAllocations) {
    const categoryTokens = vestedTokensPerCategory.get(allocation.vesting_category_type);
    if (!categoryTokens) throw new Error('categoryTokens undefined');
    const allocationTokens = assetToDecimal(allocation.tokens_allocated);
    vestedTokensPerCategory.set(allocation.vesting_category_type, categoryTokens.add(allocationTokens));
  }
  const totalVested = vestingAllocations.reduce((previous, allocation) => previous += assetToAmount(allocation.tokens_allocated), 0);
  console.log('Total unique holders: ', vestingHolders.size);
  console.log('Total vesting allocations: ', vestingAllocations.length);
  console.log(`Total vested:  ${totalVested.toFixed(4).padStart(15)} ${symbol} (${(100 * totalVested / EosioTokenContract.TOTAL_SUPPLY).toFixed(8).padStart(11)}%)`);
  vestedTokensPerCategory.forEach((tokens, category) => {
    var _vestingCategoriesLis;
    const fraction = amountToSupplyPercentage(tokens);
    const categoryName = (_vestingCategoriesLis = vestingCategories.get(category)) == null ? void 0 : _vestingCategoriesLis.name;
    console.log(`> category ${category.toString().padStart(2)}: ${tokens.toFixed(4).padStart(15)} ${symbol} (${fraction.padStart(12)}) ${categoryName}`);
  });
  await checkMissedVestingAllocations(bootstrappedAccounts, vestingHolders, apps, people);
  // TODO: check block producer accounts
  // TODO: check all staking allocations in staking contract
}
async function checkMissedVestingAllocations(bootstrappedAccounts, vestingHolders, apps, people) {
  console.log('');
  console.log('Checking if any people have allocations that were not considered above');
  const peopleNotInVestingHolders = new Set();
  for (const person of people) {
    if (!vestingHolders.has(person.account_name.toString())) {
      peopleNotInVestingHolders.add(person.account_name.toString());
    }
  }
  const peopleNotInVestingHoldersAllocations = await getAllAllocations(peopleNotInVestingHolders, true);
  const totalPeopleNotInVestingHolders = peopleNotInVestingHoldersAllocations.reduce((previous, allocation) => assetToDecimal(allocation.tokens_allocated).add(previous), ZERO_DECIMAL);
  const uniquePeopleNotInVestingHolders = peopleNotInVestingHoldersAllocations.reduce((previous, allocation) => {
    if (!previous.includes(allocation.account)) {
      return [...previous, allocation.account];
    } else {
      return previous;
    }
  }, []);
  console.log('People not in vesting holders: ', Array.from(peopleNotInVestingHolders).length);
  console.log('Total allocations: ', peopleNotInVestingHoldersAllocations.length);
  console.log(`Total tokens: ${totalPeopleNotInVestingHolders.toFixed(4).padStart(14)} ${symbol} (${amountToSupplyPercentage(totalPeopleNotInVestingHolders).padStart(10)})`);
  console.log('Unique people not in vesting holders: ', uniquePeopleNotInVestingHolders.length);
  console.log('');
  console.log('Checking if any apps have allocations that were not considered above');
  const appsNotInVestingHolders = new Set();
  for (const app of apps) {
    if (!vestingHolders.has(app.account_name.toString())) {
      appsNotInVestingHolders.add(app.account_name.toString());
    }
  }
  const appsNotInVestingHoldersAllocations = await getAllAllocations(appsNotInVestingHolders, true);
  const totalAppsNotInVestingHolders = appsNotInVestingHoldersAllocations.reduce((previous, allocation) => assetToDecimal(allocation.tokens_allocated).add(previous), ZERO_DECIMAL);
  const uniqueAppsNotInVestingHolders = appsNotInVestingHoldersAllocations.reduce((previous, allocation) => {
    if (!previous.includes(allocation.account)) {
      return [...previous, allocation.account];
    } else {
      return previous;
    }
  }, []);
  console.log('Apps not in vesting holders: ', Array.from(appsNotInVestingHolders).length);
  console.log('Total allocations: ', appsNotInVestingHoldersAllocations.length);
  console.log(`Total tokens: ${totalAppsNotInVestingHolders.toFixed(4).padStart(14)} ${symbol} (${amountToSupplyPercentage(totalAppsNotInVestingHolders).padStart(10)})`);
  console.log('Unique apps not in vesting holders: ', uniqueAppsNotInVestingHolders.length);
  console.log('');
  console.log('Checking if any bootstrapped accounts have allocations that were not considered above');
  const bootstrappedNotInVestingHolders = new Set();
  for (const account of bootstrappedAccounts) {
    if (!vestingHolders.has(account)) {
      peopleNotInVestingHolders.add(account);
    }
  }
  const bootstrappedNotInVestingHoldersAllocations = await getAllAllocations(bootstrappedNotInVestingHolders, true);
  const totalBootstrappedNotInVestingHolders = bootstrappedNotInVestingHoldersAllocations.reduce((previous, allocation) => assetToDecimal(allocation.tokens_allocated).add(previous), ZERO_DECIMAL);
  const uniqueBootstrappedNotInVestingHolders = bootstrappedNotInVestingHoldersAllocations.reduce((previous, allocation) => {
    if (!previous.includes(allocation.account)) {
      return [...previous, allocation.account];
    } else {
      return previous;
    }
  }, []);
  console.log('Bootstrapped accounts not in vesting holders: ', Array.from(bootstrappedNotInVestingHolders).length);
  console.log('Total allocations: ', bootstrappedNotInVestingHoldersAllocations.length);
  console.log(`Total tokens: ${totalBootstrappedNotInVestingHolders.toFixed(4).padStart(14)} ${symbol} (${amountToSupplyPercentage(totalBootstrappedNotInVestingHolders).padStart(10)})`);
  console.log('Unique bootstrapped accounts not in vesting holders: ', uniqueBootstrappedNotInVestingHolders.length);
}
async function getAllPeople(print = false) {
  const api = await getApi();
  const limit = 100;
  let lowerBound = Name.from('1');
  let peopleFound = 0;
  const people = [];
  do {
    if (print) console.log(`get_table_rows: people, ${limit}, ${lowerBound.toString()}`);
    const data = await api.v1.chain.get_table_rows({
      code: 'tonomy',
      scope: 'tonomy',
      table: 'people',
      limit,
      lower_bound: lowerBound
    });
    const rows = data.rows;
    peopleFound = rows.length;
    people.push(...rows.map(row => {
      return {
        account_name: row.account_name,
        status: row.status,
        username_hash: row.username_hash,
        password_salt: row.password_salt,
        version: row.version
      };
    }));
    lowerBound = Name.from(rows[rows.length - 1].account_name);
  } while (peopleFound === limit);
  return people;
}

async function vestingMigrate(options) {
  // Testnet list
  const migrateAccounts = ['p42pwxofd1vy',
  // 36 allocations
  'pwgvt1xn4ce5',
  // 10 allocations
  'p2mbvozcqp2l' // 5 allocations
  ];
  const actions = [];
  for (const account of migrateAccounts) {
    const accountActions = await createAccountActions(account);
    actions.push(...accountActions);
  }
  const proposalHash = await createProposal(options.proposer, options.proposalName, actions, options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}
function toBase6Plus1(num) {
  const base6 = num.toString(6);
  return base6.split('').map(digit => (parseInt(digit, 6) + 1).toString()).join('');
}
async function vestingMigrate2(options) {
  let count = 0;
  let proposals = 0;
  const priceMultiplier = 2.0;
  const uniqueHolders = await getAllUniqueHolders();
  const allAllocations = await getAllAllocations(uniqueHolders);
  const batchSize = 100;
  for (let i = 0; i < allAllocations.length; i += batchSize) {
    const batch = allAllocations.slice(i, i + batchSize);
    const actions = batch.map(allocation => {
      const newAmount = assetToDecimal(allocation.tokens_allocated).mul(priceMultiplier);
      const newAsset = `${newAmount.toFixed(6)} TONO`;
      console.log(`Migrating account ${allocation.holder} allocation ${allocation.id} in category ${allocation.vesting_category_type} from ${allocation.tokens_allocated} to ${newAsset}`);
      count++;
      return createMigrateAction('coinsale.tmy', allocation.holder, allocation.id, allocation.tokens_allocated, newAsset, allocation.vesting_category_type, allocation.vesting_category_type);
    });
    const proposalName = Name.from(`${options.proposalName}${toBase6Plus1(Math.floor(i / batchSize))}`);
    console.log(`Creating proposal ${proposalName.toString()} with ${actions.length} actions: ${i} - ${i + batchSize}`);
    console.log('----------------------------------------');
    proposals++;
    const proposalHash = await createProposal(options.proposer, proposalName, actions, options.privateKey, options.requested, options.dryRun);
    if (!options.dryRun && options.autoExecute) await executeProposal(options.proposer, proposalName, proposalHash);
    console.log('----------------------------------------');
  }
  console.log(`Batch size: ${batchSize}`);
  console.log(`Proposals created: ${proposals}`);
  console.log(`Processed ${count} / ${allAllocations.length} allocations`);
}
async function vestingMigrate3(options) {
  let count = 0;
  let proposals = 0;
  const priceMultiplier = 2.0;
  const people = await getAllPeople();
  const vestingHolders = await getAllUniqueHolders();
  const peopleNotInVestingHolders = new Set();
  for (const person of people) {
    if (!vestingHolders.has(person.account_name.toString())) {
      peopleNotInVestingHolders.add(person.account_name.toString());
    }
  }
  const missedAllocations = await getAllAllocations(peopleNotInVestingHolders);
  const batchSize = 100;
  for (let i = 0; i < missedAllocations.length; i += batchSize) {
    const batch = missedAllocations.slice(i, i + batchSize);
    const actions = batch.map(allocation => {
      const newAmount = assetToDecimal(allocation.tokens_allocated).mul(priceMultiplier);
      const newAsset = `${newAmount.toFixed(6)} TONO`;
      console.log(`Migrating account ${allocation.holder} allocation ${allocation.id} in category ${allocation.vesting_category_type} from ${allocation.tokens_allocated} to ${newAsset}`);
      count++;
      return createMigrateAction('coinsale.tmy', allocation.holder, allocation.id, allocation.tokens_allocated, newAsset, allocation.vesting_category_type, allocation.vesting_category_type);
    });
    const proposalName = Name.from(`${options.proposalName}${toBase6Plus1(Math.floor(i / batchSize))}`);
    console.log(`Creating proposal ${proposalName.toString()} with ${actions.length} actions: ${i} - ${i + batchSize}`);
    console.log('----------------------------------------');
    proposals++;
    const proposalHash = await createProposal(options.proposer, proposalName, actions, options.privateKey, options.requested, options.dryRun);
    if (!options.dryRun && options.autoExecute) await executeProposal(options.proposer, proposalName, proposalHash);
    console.log('----------------------------------------');
  }
  console.log(`Batch size: ${batchSize}`);
  console.log(`Proposals created: ${proposals}`);
  console.log(`Processed ${count} / ${missedAllocations.length} allocations`);
}
async function createAccountActions(account) {
  const allocations = await VestingContract.Instance.getAllocations(account);
  const actions = [];
  for (const allocation of allocations) {
    const oldCategory = allocation.vesting_category_type;
    let newCategory = 8;
    let newAmount = 0;
    if (oldCategory === 1 || oldCategory === 2) {
      if (oldCategory === 1) {
        const oldTokenPrice = 0.002;
        const newTokenPrice = TONO_SEED_ROUND_PRICE;
        const amount = assetToAmount(allocation.tokens_allocated);
        newAmount = amount * oldTokenPrice / newTokenPrice;
      } else {
        const oldTokenPrice = 0.004;
        const newTokenPrice = TONO_SEED_LATE_ROUND_PRICE;
        const amount = assetToAmount(allocation.tokens_allocated);
        newAmount = amount * oldTokenPrice / newTokenPrice;
        newCategory = 9;
      }
      const oldAmount = allocation.tokens_allocated;
      const newAsset = `${newAmount.toFixed(6)} TONO`;
      console.log(`Migrating account ${account.toString()} allocation ${allocation.id} from ${oldAmount} to ${newAsset}`);
      actions.push(createMigrateAction('coinsale.tmy', account.toString(), allocation.id, oldAmount, newAsset, oldCategory, newCategory));
    }
  }
  return actions;
}
function createMigrateAction(sender, holder, allocationId, oldAmount, newAmount, oldCategoryId, newCategoryId) {
  return {
    account: 'vesting.tmy',
    name: 'migratealloc',
    authorization: [{
      actor: sender,
      permission: 'owner'
    }, {
      actor: 'vesting.tmy',
      permission: 'active'
    }],
    data: {
      sender,
      holder,
      // eslint-disable-next-line camelcase
      allocation_id: allocationId,
      // eslint-disable-next-line camelcase
      old_amount: oldAmount,
      // eslint-disable-next-line camelcase
      new_amount: newAmount,
      // eslint-disable-next-line camelcase
      old_category_id: oldCategoryId,
      // eslint-disable-next-line camelcase
      new_category_id: newCategoryId
    }
  };
}

async function newApp(options) {
  const appName = 'Fiddl.Art';
  const description = 'Create and Earn with AI Art';
  const logoUrl = 'https://fiddl.art/fiddlLogoWithText.png';
  const origin = 'https://fiddl.art';
  const usernameShort = 'fiddlart';
  const username = TonomyUsername.fromUsername(usernameShort, AccountType.APP, getSettings().accountSuffix);
  const key = 'EOS4xnrCGUT688wFvinQoCuiu7E3Qpn8Phq76TRKNTb87XFMjzsJu';
  const action = {
    account: 'tonomy',
    name: 'newapp',
    authorization: [{
      actor: 'tonomy',
      permission: 'owner'
    }, {
      actor: 'tonomy',
      permission: 'active'
    }],
    data: {
      // eslint-disable-next-line camelcase
      app_name: appName,
      description,
      // eslint-disable-next-line camelcase
      logo_url: logoUrl,
      origin: origin,
      // eslint-disable-next-line camelcase
      username_hash: username.usernameHash,
      key
    }
  };
  const proposalHash = await createProposal(options.proposer, options.proposalName, [action], options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}

/* eslint-disable camelcase */
//create staking.tmy account controlled by ops.tmy
async function createStakingTmyAccount(options) {
  function createNewAccountAction(name, active, owner) {
    return {
      account: 'tonomy',
      name: 'newaccount',
      authorization: [{
        actor: 'tonomy',
        permission: 'owner'
      }, {
        actor: 'tonomy',
        permission: 'active'
      }],
      data: {
        creator: 'tonomy',
        name,
        owner,
        active
      }
    };
  }
  const activeAuthority = Authority.fromAccount({
    actor: 'ops.tmy',
    permission: 'active'
  });
  const ownerAuthority = Authority.fromAccount({
    actor: 'ops.tmy',
    permission: 'owner'
  });
  activeAuthority.addCodePermission('staking.tmy');
  ownerAuthority.addCodePermission('staking.tmy');
  const action = createNewAccountAction('staking.tmy', activeAuthority, ownerAuthority);
  //add staking permission to infra.tmy
  const accountInfo = await getAccountInfo(Name.from('infra.tmy'));
  const ownerPermission = accountInfo.getPermission('owner');
  const activePermission = accountInfo.getPermission('active');
  const ownerAuthorityInfra = Authority.fromAccountPermission(ownerPermission);
  const activeAuthorityInfra = Authority.fromAccountPermission(activePermission);
  // Preserve existing eosio.code permission for vesting.tmy
  ownerAuthorityInfra.addCodePermission('vesting.tmy');
  activeAuthorityInfra.addCodePermission('vesting.tmy');
  // Add new eosio.code permission for staking.tmy
  ownerAuthorityInfra.addCodePermission('staking.tmy');
  activeAuthorityInfra.addCodePermission('staking.tmy');
  const updateInfraOwnerPermission = {
    account: 'tonomy',
    name: 'updateauth',
    authorization: [{
      actor: 'infra.tmy',
      permission: 'owner'
    }, {
      actor: 'tonomy',
      permission: 'active'
    }, {
      actor: 'tonomy',
      permission: 'owner'
    }],
    data: {
      account: 'infra.tmy',
      permission: 'owner',
      parent: '',
      auth: ownerAuthority,
      auth_parent: false
    }
  };
  const updateInfraActivePermission = {
    account: 'tonomy',
    name: 'updateauth',
    authorization: [{
      actor: 'infra.tmy',
      permission: 'active'
    }, {
      actor: 'tonomy',
      permission: 'active'
    }, {
      actor: 'tonomy',
      permission: 'owner'
    }],
    data: {
      account: 'infra.tmy',
      permission: 'active',
      parent: 'owner',
      auth: activeAuthority,
      auth_parent: false
    }
  };
  const proposalHash = await createProposal(options.proposer, options.proposalName, [action, updateInfraOwnerPermission, updateInfraActivePermission], options.privateKey, [...options.requested], options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}
// call tonomy::adminSetApp() on the staking.tmy account
// call tonomy::buyRam() for the staking.tmy account
async function stakingContractSetup(options) {
  const ramKb = 4680000;
  const contract = 'staking.tmy';
  const tonomyUsername = getAppUsernameHash('staking');
  const tokens = bytesToTokens(ramKb * 1000);
  console.log(`Setting up hypha contract "${contract}" with ${tokens} tokens to buy ${ramKb}KB of RAM`);
  const adminSetAppAction = {
    authorization: [{
      actor: 'tonomy',
      permission: 'active'
    }],
    account: 'tonomy',
    name: 'adminsetapp',
    data: {
      account_name: 'staking.tmy',
      app_name: 'TONO Staking',
      description: 'TONO Staking contract',
      username_hash: tonomyUsername,
      logo_url: createSubdomainOnOrigin('https://accounts.testnet.tonomy.io', 'staking') + '/tonomy-logo1024.png',
      origin: createSubdomainOnOrigin('https://accounts.testnet.tonomy.io', 'staking')
    }
  };
  // const buyRamAction = {
  //     account: 'tonomy',
  //     name: 'buyram',
  //     authorization: [
  //         {
  //             actor: contract,
  //             permission: 'active',
  //         },
  //     ],
  //     data: {
  //         dao_owner: 'ops.tmy',
  //         app: contract,
  //         quant: tokens,
  //     },
  // };
  const setres = {
    authorization: [{
      actor: 'tonomy',
      permission: 'active'
    }],
    account: 'tonomy',
    name: 'setresparams',
    data: {
      ram_price: RAM_PRICE,
      total_ram_available: TOTAL_RAM_AVAILABLE,
      ram_fee: RAM_FEE
    }
  };
  const actions = [adminSetAppAction, setres];
  const proposalHash = await createProposal(options.proposer, options.proposalName, actions, options.privateKey, [...options.requested, contract], options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}
async function buyRam(options) {
  const ramKb = 4680000;
  const contract = 'staking.tmy';
  const tokens = bytesToTokens(ramKb * 1000);
  console.log(`Setting up hypha contract "${contract}" with ${tokens} tokens to buy ${ramKb}KB of RAM`);
  const buyRamAction = {
    account: 'tonomy',
    name: 'buyram',
    authorization: [{
      actor: contract,
      permission: 'active'
    }],
    data: {
      dao_owner: 'ops.tmy',
      app: contract,
      quant: tokens
    }
  };
  const actions = [buyRamAction];
  const proposalHash = await createProposal(options.proposer, options.proposalName, actions, options.privateKey, [...options.requested, contract], options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}
// deploy the new staking.tmy contract
async function deployStakingContract(options) {
  await deployContract({
    ...options,
    contract: 'staking.tmy'
  });
}
// re-deploy the vesting contract
async function reDeployVestingContract(options) {
  await deployContract({
    ...options,
    contract: 'vesting.tmy'
  });
}
// re-deploy the eosio contract
async function reDeployEosioContract(options) {
  await deployContract({
    ...options,
    contract: 'eosio'
  });
}
// re-deploy the tonomy contracts
async function reDeployTonomyContract(options) {
  await deployContract({
    ...options,
    contract: 'tonomy'
  });
}
//setup staking by calling setSettings() and addYield()
async function stakingSettings(options) {
  const setSettings = {
    authorization: [{
      actor: 'staking.tmy',
      permission: 'active'
    }],
    account: 'staking.tmy',
    name: 'setsettings',
    data: {
      yearly_stake_pool: amountToAsset(StakingContract.yearlyStakePool, 'TONO')
    }
  };
  const setYearlyYield = {
    authorization: [{
      actor: 'infra.tmy',
      permission: 'active'
    }],
    account: 'staking.tmy',
    name: 'addyield',
    data: {
      sender: 'infra.tmy',
      quantity: amountToAsset(StakingContract.yearlyStakePool / 2, 'TONO')
    }
  };
  const proposalHash = await createProposal(options.proposer, options.proposalName, [setSettings, setYearlyYield], options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}

/* eslint-disable camelcase */
const CONTRACT_NAME = 'tonomy';
function getDefaultColors() {
  return {
    background: '#251950',
    // white background
    branding: '#BA54D3' // grey color
  };
}
// Define a mapping for apps with custom colors
const customColorMapping = {
  'Tonomy - Development Demo': {
    background: '#251950',
    branding: '#BA54D3'
  }
  // Add other app mappings here
};
async function migrateApps(options) {
  const api = await getApi();
  // Step 1: Query the old table apps
  const data = await api.v1.chain.get_table_rows({
    json: true,
    code: CONTRACT_NAME,
    scope: CONTRACT_NAME,
    table: 'apps'
  });
  console.log('All existing apps', data.rows);
  // Step 2: For each row, call adminSetApp with default colors
  const actions = [];
  for (const row of data.rows) {
    // Retrieve default color values.
    const colors = customColorMapping[row.app_name] || getDefaultColors();
    const {
      background,
      branding
    } = colors;
    console.log(`Migrating app: ${row.app_name}`);
    const json_data = JSON.stringify({
      app_name: row.app_name,
      description: row.description,
      logo_url: row.logo_url,
      background_color: background,
      accent_color: branding
    });
    actions.push({
      authorization: [{
        actor: CONTRACT_NAME,
        permission: 'active'
      }],
      account: CONTRACT_NAME,
      name: 'adminsetapp',
      data: {
        account_name: row.account_name,
        json_data,
        username_hash: row.username_hash,
        origin: row.origin
      }
    });
  }
  // Step 3: Optionally, clear the old table entries
  const eraseAppAction = {
    account: CONTRACT_NAME,
    name: 'eraseoldapps',
    authorization: [{
      actor: CONTRACT_NAME,
      permission: 'owner'
    }],
    data: {}
  };
  const proposalHash = await createProposal(options.proposer, options.proposalName, [...actions, eraseAppAction], options.privateKey, options.requested, options.dryRun);
  if (options.dryRun) return;
  if (options.autoExecute) await executeProposal(options.proposer, options.proposalName, proposalHash);
}

/* eslint-disable camelcase */
async function symbolMigrate(options) {
  await redoployContracts(options);
  console.log('');
  await migrateStaking(options);
  console.log('');
  await migrateEosioToken(options);
  console.log('');
  await migrateVesting(options);
  console.log('');
  await migrateRebrandApps(options);
  console.log('');
  console.log('Migration complete');
}
// // @ts-expect-error options not used
async function redoployContracts(options) {
  console.log('### Redploying contracts');
  await deployContract({
    ...options,
    proposalName: Name.from(options.proposalName.toString() + '1a'),
    contract: 'eosio.token'
  });
  await deployContract({
    ...options,
    proposalName: Name.from(options.proposalName.toString() + '1b'),
    contract: 'tonomy'
  });
  await deployContract({
    ...options,
    proposalName: Name.from(options.proposalName.toString() + '1c'),
    contract: 'vesting.tmy'
  });
  await deployContract({
    ...options,
    proposalName: Name.from(options.proposalName.toString() + '1d'),
    contract: 'staking.tmy'
  });
}
async function migrateEosioToken(options) {
  console.log('### Migrating eosio.token');
  const bootstrappedAccounts = new Set();
  bootstrappedAccounts.add(foundAccount);
  bootstrappedAccounts.add(operationsAccount);
  bootstrappedAccounts.add(systemAccount);
  if (settings.env === 'production') bootstrappedAccounts.add('advteam.tmy');
  for (const account of foundControlledAccounts) bootstrappedAccounts.add(account);
  for (const account of opsControlledAccounts) bootstrappedAccounts.add(account);
  const actions = Array.from(bootstrappedAccounts).map(account => {
    console.log(`eosio.token::migrateacc(${account})`);
    return {
      account: 'eosio.token',
      name: 'migrateacc',
      authorization: [{
        actor: 'eosio.token',
        permission: 'active'
      }],
      data: {
        account
      }
    };
  });
  console.log(`Total accounts to migrate: ${actions.length}`);
  console.log(`eosio.token::migratestats()`);
  actions.push({
    account: 'eosio.token',
    name: 'migratestats',
    authorization: [{
      actor: 'eosio.token',
      permission: 'active'
    }],
    data: {}
  });
  const proposalName = Name.from(options.proposalName.toString() + '2');
  const proposalHash = await createProposal(options.proposer, proposalName, actions, options.privateKey, options.requested, options.dryRun);
  if (!options.dryRun && options.autoExecute) await executeProposal(options.proposer, proposalName, proposalHash);
}
async function migrateVesting(options) {
  console.log('### Migrating vesting.tmy');
  console.log('Fetching vested tokens');
  const vestingHolders = getSettings().environment === 'development' ? new Set(['team.tmy', 'found.tmy']) : await getAllUniqueHolders();
  const actions = Array.from(vestingHolders).map(holder => {
    console.log(`vesting.tmy::migrateacc(${holder})`);
    return {
      account: 'vesting.tmy',
      name: 'migrateacc',
      authorization: [{
        actor: 'vesting.tmy',
        permission: 'active'
      }],
      data: {
        account: holder
      }
    };
  });
  console.log(`Total accounts to migrate: ${actions.length}`);
  const proposalName = Name.from(options.proposalName.toString() + '3');
  const proposalHash = await createProposal(options.proposer, proposalName, actions, options.privateKey, options.requested, options.dryRun);
  if (!options.dryRun && options.autoExecute) await executeProposal(options.proposer, proposalName, proposalHash);
}
async function migrateStaking(options) {
  console.log('### Migrating staking.tmy');
  const action = {
    account: 'staking.tmy',
    name: 'resetall',
    authorization: [{
      actor: 'staking.tmy',
      permission: 'active'
    }],
    data: {}
  };
  const proposalName = Name.from(options.proposalName.toString() + '4');
  const proposalHash = await createProposal(options.proposer, proposalName, [action], options.privateKey, options.requested, options.dryRun);
  if (!options.dryRun && options.autoExecute) await executeProposal(options.proposer, proposalName, proposalHash);
}
async function migrateRebrandApps(options) {
  console.log('### Migrating rebranding');
  const apps = await TonomyContract.Instance.getApps();
  const actions = await apps.filter(app => {
    const name = app.app_name.toLowerCase();
    const description = app.description.toLowerCase();
    const origin = app.origin.toLowerCase();
    return name.includes('pangea') || name.includes('leos') || name.includes('sales') && description.includes(`sales${settings.env === 'production' ? '' : ' testnet'} platform`) || description.includes('pangea') || description.includes('leos') || origin.includes('pangea.web4.world');
  }).map(app => {
    let newAppName = app.app_name.replace('Pangea', 'Tonomy').replace('LEOS', 'TONO');
    let newDescription = app.description.replace('Pangea', 'Tonomy').replace('LEOS', 'TONO');
    let newOrigin = app.origin.replace('pangea.web4.world', 'tonomy.io');
    let newLogoUrl = app.logo_url.replace('pangea.web4.world', 'tonomy.io');
    if (newAppName.toLowerCase().includes('sales') && newDescription.toLowerCase().includes(`sales${settings.env === 'production' ? '' : ' testnet'} platform`)) {
      newAppName = `Tonomy${settings.env === 'production' ? '' : ' Testnet'} Launchpad`;
      newDescription = `Tonomy${settings.env === 'production' ? '' : ' Testnet'} Launchpad`;
      newOrigin = `https://launchpad${settings.env === 'production' ? '' : '.testnet'}.tonomy.io`;
    }
    if (newLogoUrl.includes('LEOS%20256x256.png')) {
      newLogoUrl = 'https://cdn.prod.website-files.com/67ea90b224287f4cbb2dd180/67ef991d349ec01179aec16d_icon1.png';
    }
    return {
      account: 'tonomy',
      name: 'adminsetapp',
      authorization: [{
        actor: 'tonomy',
        permission: 'active'
      }],
      data: {
        account_name: app.account_name,
        app_name: newAppName,
        description: newDescription,
        username_hash: app.username_hash,
        logo_url: newLogoUrl,
        origin: newOrigin
      }
    };
  });
  const proposalName = Name.from(options.proposalName.toString() + '5');
  const proposalHash = await createProposal(options.proposer, proposalName, actions, options.privateKey, options.requested, options.dryRun);
  if (!options.dryRun && options.autoExecute) {
    await executeProposal(options.proposer, proposalName, proposalHash);
  }
}

const eosioMsigContract = EosioMsigContract.Instance;
const governanceAccounts = ['1.found.tmy', '2.found.tmy', '3.found.tmy'];
let newGovernanceAccounts = ['13.found.tmy', '5.found.tmy', '11.found.tmy', '12.found.tmy', '14.found.tmy'];
if (!settings.isProduction()) {
  newGovernanceAccounts = governanceAccounts;
}
async function msig(args) {
  let autoExecute = false,
    dryRun = false;
  for (const arg of args) {
    if (arg.includes('--help')) {
      printMsigHelp();
      return;
    } else if (arg.includes('--auto-execute')) {
      console.log('Will attempt to execute the proposal');
      if (settings.isProduction()) {
        throw new Error(`Cannot test proposal on a live environment. Environment: ${settings.env}`);
      }
      autoExecute = true;
      args.splice(args.indexOf(arg), 1);
      newGovernanceAccounts = governanceAccounts;
    } else if (arg.includes('--dry-run')) {
      console.log('Dry run proposal');
      dryRun = true;
      args.splice(args.indexOf(arg), 1);
    }
  }
  const proposer = newGovernanceAccounts[0];
  let signingKey = process.env.SIGNING_KEY;
  const signingAccount = proposer;
  if (!signingKey) {
    if (!process.env.TONOMY_BOARD_PRIVATE_KEYS) throw new Error('Neither SIGNING_KEY or TONOMY_BOARD_PRIVATE_KEYS are set');
    const tonomyGovKeys = JSON.parse(process.env.TONOMY_BOARD_PRIVATE_KEYS).keys;
    signingKey = tonomyGovKeys[0];
  }
  const privateKey = PrivateKey.from(signingKey);
  const signer = createSigner$1(privateKey);
  if (args[0] === 'cancel') {
    const proposalName = Name.from(args[1]);
    try {
      const transaction = await eosioMsigContract.cancel(proposer, proposalName, signingAccount, signer);
      console.log('Transaction: ', JSON.stringify(transaction, null, 2));
      console.error('Transaction succeeded');
    } catch (e) {
      console.error('Error: ', JSON.stringify(e, null, 2));
      console.error('Transaction failed');
    }
  } else if (args[0] === 'propose') {
    const proposalType = args[1];
    const proposalSubtype = args[2];
    const proposalName = !proposalSubtype ? Name.from(args[2]) : Name.from(args[3]);
    const options = {
      proposer,
      proposalName,
      privateKey,
      requested: newGovernanceAccounts,
      autoExecute,
      dryRun
    };
    if (proposalType === 'account') {
      if (proposalSubtype === 'create') {
        await newAccount({
          governanceAccounts
        }, options);
      } else printMsigHelp();
    } else if (proposalType === 'transfer') {
      if (proposalSubtype === 'transfer') {
        await transfer$1(options);
      } else printMsigHelp();
    } else if (proposalType === 'deploy-contract') {
      var _args$;
      const contractName = (_args$ = args[3]) != null ? _args$ : 'tonomy';
      await deployContract({
        contract: contractName,
        ...options
      });
    } else if (proposalType === 'auth') {
      if (proposalSubtype === 'add-eosiocode') {
        await addEosioCode(options);
      } else if (proposalSubtype === 'create') {
        await addAuth({
          account: 'srvice.hypha',
          permission: 'active',
          newDelegate: 'gov.tmy',
          useParentAuth: true
        }, options);
      } else if (proposalSubtype === 'gov-migrate') {
        await govMigrate({
          newGovernanceAccounts
        }, {
          ...options,
          requested: governanceAccounts
        });
      } else printMsigHelp();
    } else if (proposalType === 'vesting-migrate') {
      if (proposalSubtype === 'migrate') {
        await vestingMigrate(options);
      } else if (proposalSubtype === 'migrate2') {
        await vestingMigrate2(options);
      } else if (proposalSubtype === 'migrate3') {
        await vestingMigrate3(options);
      } else if (proposalSubtype === 'bulk') {
        await vestingBulk({
          governanceAccounts
        }, options);
      } else printMsigHelp();
    } else if (proposalType === 'producers') {
      if (proposalSubtype === 'add') {
        await addProd({}, options);
      } else if (proposalSubtype === 'remove') {
        await removeProd({}, options);
      } else if (proposalSubtype === 'update') {
        await updateProd({}, options);
      } else if (proposalSubtype === 'change') {
        await changeProds({}, options);
      } else printMsigHelp();
    } else if (proposalType === 'hypha') {
      if (proposalSubtype === 'accounts-create') {
        await hyphaAccountsCreate({}, options);
      } else if (proposalSubtype === 'add-permissions') {
        await hyphaAddAccountPermissions({}, options);
      } else if (proposalSubtype === 'contract-set') {
        await hyphaContractSet({}, options);
      } else printMsigHelp();
    } else if (proposalType === 'res-config-set') {
      await setResourceConfig({}, options);
    } else if (proposalType === 'set-chain-config') {
      await setBlockchainConfig({}, options);
    } else if (proposalType === 'new-app') {
      await newApp(options);
    } else if (proposalType === 'migrate-appsv2') {
      await migrateApps(options);
    } else if (proposalType === 'app') {
      if (proposalSubtype === 'create') {
        await newApp(options);
      } else printMsigHelp();
    } else if (proposalType === 'staking') {
      if (proposalSubtype === 'account') {
        await createStakingTmyAccount(options);
      } else if (proposalSubtype === 'contract') {
        await stakingContractSetup(options);
      } else if (proposalSubtype === 'buyram') {
        await buyRam(options);
      } else if (proposalSubtype === 'deploy-staking-contract') {
        await deployStakingContract(options);
      } else if (proposalSubtype === 'redeploy-vesting-contract') {
        await reDeployVestingContract(options);
      } else if (proposalSubtype === 'redeploy-eosio-contract') {
        await reDeployEosioContract(options);
      } else if (proposalSubtype === 'redeploy-tonomy-contract') {
        await reDeployTonomyContract(options);
      } else if (proposalSubtype === 'setSettings') {
        await stakingSettings(options);
      } else printMsigHelp();
    } else if (proposalType === 'symbol') {
      if (proposalSubtype === 'migrate') {
        await symbolMigrate(options);
      } else if (proposalSubtype === 'migrate-app') {
        await migrateRebrandApps(options);
      } else printMsigHelp();
    } else {
      throw new Error(`Invalid msig proposal type ${proposalType}`);
    }
  } else if (args[0] === 'approve') {
    const proposalName = Name.from(args[1]);
    try {
      const transaction = await eosioMsigContract.approve(proposer, proposalName, signingAccount, undefined, signer);
      console.log('Transaction: ', JSON.stringify(transaction, null, 2));
      console.error('Transaction succeeded');
    } catch (e) {
      console.error('Error: ', JSON.stringify(e, null, 2));
      throw new Error('Transaction failed');
    }
  } else if (args[0] === 'exec') {
    const proposalName = Name.from(args[1]);
    try {
      const transaction = await eosioMsigContract.exec(proposer, proposalName, signingAccount, signer);
      console.log('Transaction: ', JSON.stringify(transaction, null, 2));
      console.error('Transaction succeeded');
    } catch (e) {
      console.error('Error: ', JSON.stringify(e, null, 2));
      throw new Error('Transaction failed');
    }
  } else {
    printCliHelp();
    throw new Error(`Invalid msig command ${args[0]}`);
  }
}
async function createProposal(proposer, proposalName, actions, privateKey, requested, dryRun) {
  const requestedPermissions = requested.map(actor => {
    if (typeof actor === 'string') {
      return {
        actor,
        permission: 'active'
      };
    }
    return {
      actor: actor.actor.toString(),
      permission: actor.permission.toString()
    };
  });
  console.log('Sending transaction', JSON.stringify({
    proposer,
    proposalName,
    requestedPermissions,
    actions,
    signer: privateKey.toPublic()
  }, null, 2));
  if (dryRun) {
    console.log('Dry run finished');
    // return a random hash to exit early
    return Checksum256.from('8e95684a0d281d0ed23ebf02f5888774dafb56d47174c4402122f05aef935bdd');
  }
  try {
    const {
      transaction,
      proposalHash
    } = await eosioMsigContract.propose(proposer, proposalName, requestedPermissions, actions, createSigner$1(privateKey));
    console.log('Transaction: ', JSON.stringify(transaction, null, 2));
    console.error('Transaction succeeded');
    console.log('Proposal name: ', proposalName.toString());
    console.log('You have 7 days to approve and execute the proposal.');
    return proposalHash;
  } catch (e) {
    var _e$error;
    if ((e == null || (_e$error = e.error) == null || (_e$error = _e$error.details[0]) == null ? void 0 : _e$error.message) === 'assertion failure with message: transaction authorization failed') {
      throw new Error('The transaction authorization requirements are not correct. Check the action authorizations are correct, and the "requested" permissions.');
    } else {
      console.error('Error: ', JSON.stringify(e, null, 2));
      throw new Error('Transaction failed');
    }
  }
}
async function executeProposal(proposer, proposalName, proposalHash, signingAccount) {
  if (!process.env.TONOMY_BOARD_PRIVATE_KEYS) throw new Error('TONOMY_BOARD_PRIVATE_KEYS not set');
  const tonomyGovKeys = JSON.parse(process.env.TONOMY_BOARD_PRIVATE_KEYS).keys;
  const tonomyGovSigners = tonomyGovKeys.map(key => createSigner$1(PrivateKey.from(key)));
  try {
    for (let i = 0; i < 2; i++) {
      await sleep(1000);
      await eosioMsigContract.approve(proposer, proposalName, governanceAccounts[i], proposalHash, tonomyGovSigners[i]);
    }
    console.log('Proposal approved succeeded');
    await eosioMsigContract.exec(proposer, proposalName, signingAccount != null ? signingAccount : proposer, tonomyGovSigners[0]);
    console.log('Proposal executed succeeded');
  } catch (e) {
    console.error('Error: ', JSON.stringify(e, null, 2));
    throw new Error('Transaction failed');
  }
}
function printMsigHelp() {
  console.log(`
        Usage:
            yarn run cli msig [commands]
            
            Commands:
                approve stakeacc
                cancel <proposalName>
                exec <proposalName>
                propose account create <proposalName>
                propose app create <proposalName>
                propose auth add-eosiocode <proposalName>
                propose auth create <proposalName>
                propose auth gov-migrate <proposalName>
                propose hypha accounts-create <proposalName>
                propose hypha add-permissions <proposalName>
                propose hypha contract-set <proposalName>
                propose producers add <proposalName>
                propose producers change <proposalName>
                propose producers remove <proposalName>
                propose producers update <proposalName>
                propose res-config-set <proposalName>
                propose set-chain-config <proposalName>
                propose staking account <proposalName>
                propose staking contract <proposalName>
                propose staking buyram <proposalName>
                propose staking deploy-staking-contract <proposalName>
                propose staking redeploy-vesting-contract <proposalName>
                propose staking redeploy-eosio-contract <proposalName>
                propose staking redeploy-tonomy-contract <proposalName>
                propose staking setSettings <proposalName>
                propose symbol migrate <proposalName>
                propose token transfer <proposalName>
                propose vesting bulk <proposalName>
                propose vesting migrate <proposalName>
                propose vesting migrate2 <proposalName>
                propose vesting migrate3 <proposalName>
                propose migrate-appsv2 <proposalName>
                propose ... --auto-execute
                propose ... --dry-run
        `);
}

const tonomyContract = TonomyContract.Instance;
async function apps(args) {
  if (args[0] === 'get') {
    const username = args[1];
    if (username.startsWith('@')) {
      console.log('Searching for username: ', username);
      const usernameInstance = TonomyUsername.fromUsername(username.split('@')[1], AccountType.PERSON, settings.config.accountSuffix);
      const {
        account_name: account
      } = await tonomyContract.getPerson(usernameInstance);
      console.log('Account name: ', account.toString());
    } else {
      console.log('Searching for account: ', username);
      const account = await getAccountInfo(Name.from(username));
      console.log('Account: ', JSON.stringify(account, null, 2));
    }
  } else {
    throw new Error(`Unknown command ${args[0]}`);
  }
}

const args = process.argv.slice(2);
setSettings(settings.config);
console.log('Using environment', settings.env);
async function main() {
  console.log('Arguments: ', args);
  if (args[0] === 'apps') {
    await apps$1(args.slice(1));
  } else if (args[0] === 'accounts') {
    await apps(args.slice(1));
  } else if (args[0] === 'authority') {
    await authority(args.slice(1));
  } else if (args[0] === 'bootstrap') {
    await bootstrap();
  } else if (args[0] === 'keys') {
    await keys(args.slice(1));
  } else if (args[0] === 'msig') {
    await msig(args.slice(1));
  } else if (args[0] === 'vesting') {
    await vesting(args.slice(1));
  } else if (args[0] === 'token') {
    if (args[1] === 'transfer') {
      await transfer(args.slice(1));
    } else if (args[1] === 'audit') {
      await audit();
    } else {
      printCliHelp();
    }
  } else {
    printCliHelp();
  }
}
function printCliHelp() {
  console.log(`
        Usage:
            yarn run cli [commands]
            
            Commands:
                accounts get <username>
                apps create <appName> <usernamePrefix> <description> <logoUrl> <origin> <publicKey>
                authority <publicKey/privateKey>
                authority <account1> <account2> ... <accountN>
                bootstrap
                keys create
                keys convert <publicKey/privateKey>
                msig [commands]
                vesting assign <sender> <username/accountName> <amount> <category>
                vesting audit
                vesting setsettings
                token transfer <from> <username/accountName> <amount> <memo>
                token audit
        `);
}
Promise.resolve(main()).catch(err => {
  console.error(err);
  process.exit(1);
});

export { printCliHelp };
//# sourceMappingURL=index.module.js.map
